// (c) 2018, SAS Institute Inc.

jQuery.sap.declare("sas.analytics.riskhp.datasource.DataSourceChangeManager");

jQuery.sap.require("sas.analytics.riskhp.datasource.DataSourceChanger");
jQuery.sap.require("sas.analytics.riskhp.models.DataSourceCacheManager");
jQuery.sap.require("sas.analytics.riskhp.models.DataSourceParams");
jQuery.sap.require("sas.analytics.riskhp.utils.CompatibilityCheck");
jQuery.sap.require("sas.analytics.riskhp.utils.HprAPI");
jQuery.sap.require("sas.analytics.riskhp.utils.HprConstants");
jQuery.sap.require("sas.analytics.riskhp.utils.HprUtil");


/**
 * @name sas.analytics.riskhp.datasource.DataSourceChangeManager
 */
sas.analytics.riskhp.datasource.DataSourceChangeManager = function () {
    // pseudo-imports
    var DataSourceChanger = sas.analytics.riskhp.datasource.DataSourceChanger;
    var DataSourceCacheManager = sas.analytics.riskhp.models.DataSourceCacheManager;
    var DataSourceParams = sas.analytics.riskhp.models.DataSourceParams;
    var CompatibilityCheck = sas.analytics.riskhp.utils.CompatibilityCheck;
    var HprAPI = sas.analytics.riskhp.utils.HprAPI;
    var HprConstants = sas.analytics.riskhp.utils.HprConstants;
    var HprUtil = sas.analytics.riskhp.utils.HprUtil;

    // statics
    var rb = HprUtil.getBundle();


    /**
     *
     * @param explorationAnalysis
     * @param oldDataSourceIds
     * @param launchControl
     * @param replaceSuccessCallback
     * @private
     */
    function _startReplaceDataSourceForExploration(explorationAnalysis, oldCubeName, oldDataSourceIds, bForOpenExploration, controller, launchControl, replaceSuccessCallback, replaceFailCallback) {
        // Load the cubes
        HprAPI.getCubes().done(
            function(cubes) {
                var chooseDatasource = function() {
                    // Open the ChooseDataSourceDialog with the list of cubes
                    var dialogModel = new sap.ui.model.json.JSONModel();
                    dialogModel.setProperty("/dataSourceItemsPath", "/dataSourceItems");
                    dialogModel.setProperty("/dataSourceItems", cubes);
                    dialogModel.setProperty("/dialogTitle", rb.getText('ChooseDataSourceDialog.replace.data.source.title.txt'));
                    dialogModel.setProperty("/dialogMode", HprConstants.HPR_CHOOSEDATASOURCE_DIALOG_REPLACE_DATA_SOURCE_MODE);
                    dialogModel.setProperty("/oldCubeName", oldCubeName);
                    dialogModel.setProperty("/openCallback", function(newDataSourceObject) {
                        _showDataSourceDifferencesConfirmationDialog(newDataSourceObject, explorationAnalysis, oldDataSourceIds, bForOpenExploration, controller, launchControl, replaceSuccessCallback, replaceFailCallback);
                    });
                    var chooseDataSourceDialog = sap.ui.jsfragment("sas.analytics.riskhp.fragments.ChooseDataSourceDialog", controller);
                    chooseDataSourceDialog.setModel(dialogModel);
                    chooseDataSourceDialog.open();
                };

                // Create an id that will be used by all Tabs so each one can be suppressed by the Global Settings
                var replaceDataSourceWarningDialogId = "replaceDataSourceWarningDialog";
                // Check the global settings to see if the dialog is being blocked from opening
                if (sas.hc.ui.core.settings.GlobalSettingsHelper.getSuppressedMessages().indexOf(replaceDataSourceWarningDialogId) === -1) {
                    var warning = sas.hc.m.MessageDialog.warning({
                        id: replaceDataSourceWarningDialogId,
                        text: rb.getText("HPRisk.ReplaceDataSource.Warning.txt"),
                        title: rb.getText("HPRisk.ReplaceDataSource.Warning.title.txt"),
                        includeDontShowCheckBox: true,
                        afterClose: function(evt) {
                            this.destroy();
                        }
                    });
                    warning.attachAfterClose(function(e) {
                        chooseDatasource();
                        warning.destroy();
                    });
                } else {
                    chooseDatasource();
                }

            }
        ).fail(
            function(jqxhr, textStatus, error) {
                HprUtil.displayDetailedErrorDialog(jqxhr);
            }
        );
    }

    /**
     * This method is called by the ChooseDataSourceDialog after the user selects to Replace a Data Source.  This method ends up displaying
     * a dialog which lists the differences between the currently selected DataSource and the one the user chose from ChooseDataSourceDialog.
     * Once the user sees the differences they have a choice to either Continue with the replace data source or Cancel.  Appropriate actions
     * for either choice are taken in this method.
     * @param newDataSourceObject An object returned from the ChooseDataSourceDialog with minimal properties used to find and load the desired DataSource object
     */
    function _showDataSourceDifferencesConfirmationDialog(newDataSourceObject, explorationAnalysis, oldDataSourceIds, bForOpenExploration, controller, launchControl, replaceSuccessCallback, replaceFailCallback) {
        // Set the Exploration view to busy because a lot may happen when we replace a Data Source
        launchControl.setBusy(true);

        // Create a DataSourceParams object that we can use to find details about it
        var newDataSourceParams = DataSourceParams.prototype.createRootParams(newDataSourceObject);

        var promise = HprAPI.getDataSourceDetails(newDataSourceParams);
        promise.done(function(newDataSourceDetails) {
            if (newDataSourceDetails && newDataSourceDetails.dataSources && newDataSourceDetails.dataSources.length > 0) {
                // TODO: Remove this snippet of code when 'portfolio' data source will be used
                // Remove the 'portfolio' data source because it is not used yet
                newDataSourceDetails.dataSources = _.filter(newDataSourceDetails.dataSources, function (dataSource) {
                    return dataSource.type === 'price' || dataSource.type === 'stress';
                });

                // Create references to handlers since they can be called in multiple places in the code below
                var successHandler = HprUtil.createSuccessHandler(launchControl, null);
                var failureHandler = HprUtil.createFailureHandler(launchControl, true, null);

                // For an open exploration we will perform a compatibility check between the new data source and the
                // one being replaced.
                if (bForOpenExploration === true) {
                    // get the previously (currently) selected data source
                    var previouslySelectedDataSource = DataSourceCacheManager.getSelectedDataSource();

                    // get the params object for the previously (currently) selected data source
                    var previouslySelectedDataSourceParam = DataSourceCacheManager.getSelectedDataSourceParam();

                    // Using the currently selected DataSourceParams object, find the type.  This is the type
                    // of DataSource we are going to try and replace the current one with
                    var previouslySelectedDataSourceType = previouslySelectedDataSourceParam.type;

                    // Create an Array of types for the new DataSource.  We are going to see if 'previouslySelectedDataSourceType' is in the list of types.
                    // Also, determine which type of DataSource we should try to replace with by using the only one in the Array (if that is the case), or use the
                    // previous DataSource type if it exists in the new DataSource types; We use the first one in the Array as a default fallback if the other cases fail
                    var newDataSourceTypes = _.pluck(newDataSourceDetails.dataSources, 'type');
                    var dataSourceTypeToUse = newDataSourceTypes.length === 1 ? newDataSourceTypes[0] : (_.contains(newDataSourceTypes, previouslySelectedDataSourceType) ? previouslySelectedDataSourceType : newDataSourceTypes[0]);

                    // Find matching DataSourceParam objects to the one we had selected; This allows us to find which ones we need to replace
                    var dataSourceParamsToRemove = DataSourceCacheManager.getMatchingDataSourceParamsUsingParams(previouslySelectedDataSourceParam);
                    // In order to remove them, we need to build an array of 'realDataSourceId' ids from 'dataSourceParamsToRemove'
                    var dataSourceParamsToRemoveIds = _.map(dataSourceParamsToRemove, function (dataSourceParam, key) {
                        return dataSourceParam.realDataSourceId;
                    });

                    // Since we don't know if the user really wants to replace data source with the new one yet, gather the data without going through the DataSourceCacheManager; This
                    // essentially cracks open the desired DataSource and gathers its data so we can compare the old DataSource with the one the user wants to replace with;
                    var promise = HprAPI.previewDataSource(newDataSourceDetails, dataSourceTypeToUse);
                    promise.done(
                        function (potentialDataSource) {
                            // Opened the new DataSource successfully, now we need to calculate the differences between the old DataSource and the new one so we can give the
                            // user an option to abort replace data source
                            var compatibilityDifferenceArray = CompatibilityCheck.computeDataSourceDiff(previouslySelectedDataSource, potentialDataSource);
                            if (compatibilityDifferenceArray && compatibilityDifferenceArray.length > 0) {
                                // Only open the ReplaceDataSourceDetailsDialog if there are differences to report
                                var dialogModel = new sap.ui.model.json.JSONModel();
                                dialogModel.setProperty("/differencesPath", "/differences");
                                dialogModel.setProperty("/differences", compatibilityDifferenceArray);
                                dialogModel.setProperty("/oldDataSourceLabel", previouslySelectedDataSource.getCubeDisplayLabel());
                                dialogModel.setProperty("/newDataSourceLabel", potentialDataSource.getCubeDisplayLabel());
                                var replaceDataSourceDetailsDialog = sap.ui.jsfragment("sas.analytics.riskhp.fragments.ReplaceDataSourceDetailsDialog", controller);
                                replaceDataSourceDetailsDialog.setModel(dialogModel);

                                // Once the user has made a selection in the ReplaceDataSourceDetailsDialog to Continue or Cancel, take appropriate action
                                replaceDataSourceDetailsDialog.attachAfterClose(function (evt) {
                                    if (this.buttonClicked && this.buttonClicked === this.CONTINUE_BUTTON) {
                                        // User chose to continue with replace data source
                                        _replaceDataSourceImpl(
                                            explorationAnalysis,
                                            newDataSourceDetails,
                                            dataSourceTypeToUse,
                                            dataSourceParamsToRemoveIds,
                                            bForOpenExploration,
                                            replaceSuccessCallback,
                                            replaceFailCallback,
                                            successHandler,
                                            failureHandler
                                        );
                                    }
                                    else {
                                        // User cancelled, we still need to DELETE the newly opened DataSource to help with grid resources
                                        $.when(
                                            HprAPI.releaseCubeResources(potentialDataSource.getCubeId())
                                        ).done(
                                            function () {
                                                successHandler();
                                            }
                                        ).fail(failureHandler);
                                    }
                                });

                                replaceDataSourceDetailsDialog.open();
                            }
                            else {
                                // Otherwise, if there are no differences, just do the replace data source
                                _replaceDataSourceImpl(explorationAnalysis, newDataSourceDetails, dataSourceTypeToUse, dataSourceParamsToRemoveIds, bForOpenExploration, replaceSuccessCallback, replaceFailCallback, successHandler, failureHandler);
                            }
                        }
                    );
                    promise.fail(failureHandler);
                }
                else {
                    // replace data source
                    _replaceDataSourceImpl(explorationAnalysis, newDataSourceDetails, dataSourceTypeToUse, oldDataSourceIds, bForOpenExploration, replaceSuccessCallback, replaceFailCallback, successHandler, failureHandler);
                }
            }
            else {
                launchControl.setBusy(false);

                // Display error dialog
                HprUtil.displayGenericErrorDialog(HprUtil.getBundle().getText('HPRisk.detailed.error.cube.information.failed.to.load.txt'));
            }
        });
    }

    /**
     * This method actually does the replacing of an old DataSource with a new DataSource. The new DataSource is loaded
     * and an event is fired so other components can react. The resources for the old DataSource are deleted so Grid
     * resources are happy.
     *
     * @param explorationAnalysis {String} the analysis JSON for the exploration
     * @param dataSourceDetails The cube data represented by a 'dataSourceDetails' object the user wants to load
     * @param dataSourceTypeToUse The type of 'dataSourceDetails' object that we need to use
     * @param dataSourceParamsToRemoveIds Array of DataSourceParam objects which need to be removed when replacing data sources
     * @param replaceSuccessCallback {Function} function to call after replace is complete
     * @param successHandler Function call on success
     * @param failureHandler Function to call on failure cases
     * @private
     */
    function _replaceDataSourceImpl(explorationAnalysis, dataSourceDetails, dataSourceTypeToUse, dataSourceParamsToRemoveIds, bForOpenExploration, replaceSuccessCallback, replaceFailCallback, successHandler, failureHandler) {
        // Since we are really going to replace the DataSource now, use the HprAPI class to load the cube data represented by 'dataSourceDetails' and go through
        // the DataSourceCacheManager this time so it is happy and in-sync
        var promise = HprAPI.getDataSource(dataSourceDetails, dataSourceTypeToUse);
        promise.done(function(dataSource) {
            // Since it is up to the components to decide whether they should replace data source, we need to pass all the possible DataSource IDs that a component can be replaced with
            var newDataSourceIds = _.pluck(dataSourceDetails.dataSources, 'id');

            // Method which is ran after updating the Exploration with replace data source changes
            var completionHandler = function(updatedExploration) {
                $.when(
                    // When the new data source is finished loading, release the old dataSourceParam objects from the mid-tier to free up memory
                    DataSourceCacheManager.releaseDataSourcesById(dataSourceParamsToRemoveIds)
                ).done(
                    function() {
                        // validate computed variables and display rules
                        _.each(dataSourceDetails.dataSources, function(dataSourceDetail) {
                            DataSourceCacheManager.getDataSource(dataSourceDetail.id).done(function(cubeData) {
                                var computedVarDeferred = sas.analytics.riskhp.utils.HprAPI._addComputedVariableArtifacts(cubeData, sas.analytics.riskhp.models.ComputedVariablesManager.getComputedVariables());
                                computedVarDeferred.done(function() {
                                    DisplayRulesManager.validateDisplayRules(cubeData);
                                });
                            });
                        });

                        replaceSuccessCallback(updatedExploration, dataSourceDetails);
                        successHandler(); // Finally done with everything, call the successHandler function
                    }
                ).fail(
                    function() {
                        replaceFailCallback();
                        failureHandler();
                    }
                );
            };

            // Here we update the exploration json to the data source change
            DataSourceChanger.updateExploration(true, bForOpenExploration, explorationAnalysis, dataSourceParamsToRemoveIds, newDataSourceIds, completionHandler);
        });
        promise.fail(failureHandler);
    }

    // public

    return {
        /**
         *
         *
         * @param explorationAnalysis {String} the exploration analysis JSON string
         * @param oldDataSourceIds {Array} the collection of data source IDs to be replaced
         * @param bForOpenExploration
         * @param controller
         * @param launchControl {Control} the control from which this was launched
         * @param replaceSuccessCallback {Function} function to be executed after successfully replacing the data source
         * @param replaceFailCallback
         */
        replaceDataSourceForExploration: function(explorationAnalysis, oldCubeName, oldDataSourceIds, bForOpenExploration, controller, launchControl, replaceSuccessCallback, replaceFailCallback) {
            _startReplaceDataSourceForExploration(explorationAnalysis, oldCubeName, oldDataSourceIds, bForOpenExploration, controller, launchControl, replaceSuccessCallback, replaceFailCallback);
        }
    };
}();