// (c) 2017, SAS Institute Inc.

jQuery.sap.declare("sas.analytics.riskhp.datasource.DataSourceChanger");

jQuery.sap.require("sas.analytics.riskhp.models.DataSourceCacheManager");
jQuery.sap.require("sas.analytics.riskhp.utils.ExplorationUtil");
jQuery.sap.require("sas.analytics.riskhp.utils.HprUtil");
jQuery.sap.require("sas.analytics.riskhp.models.CrosstabContext");
jQuery.sap.require("sas.analytics.riskhp.common.riskcommons.plot.HprChartModel");
jQuery.sap.require("sas.analytics.riskhp.portfolio.PortfolioDetailsModel");
jQuery.sap.require("sas.analytics.riskhp.sensitivities.RiskFactorSensitivitiesModel");

/**
 * @name sas.analytics.riskhp.datasource.DataSourceChanger
 */
sas.analytics.riskhp.datasource.DataSourceChanger = {
    DataSourceCacheManager: sas.analytics.riskhp.models.DataSourceCacheManager,
    ExplorationUtil: sas.analytics.riskhp.utils.ExplorationUtil,
    HprUtil: sas.analytics.riskhp.utils.HprUtil,
    CrosstabContext: sas.analytics.riskhp.models.CrosstabContext,
    HprChartModel: sas.analytics.riskhp.common.riskcommons.plot.HprChartModel,
    PortfolioDetailsModel: sas.analytics.riskhp.portfolio.PortfolioDetailsModel,
    RiskFactorSensitivitiesModel: sas.analytics.riskhp.sensitivities.RiskFactorSensitivitiesModel,
    TableViewerModel: sas.analytics.riskhp.common.riskcommons.table.TableViewerModel,

    PANEL_CONTENT_ID: "mainContentData",

    contentPaneTest: new RegExp( [
        HprConstants.HPR_CROSSTAB_COMPONENT_ID,
        HprConstants.HPR_PLOT_COMPONENT_ID,
        HprConstants.HPR_RISK_FACTOR_SENSITIVITIES_COMPONENT_ID,
        HprConstants.HPR_PORTFOLIO_DETAILS_COMPONENT_ID,
        HprConstants.HPR_SIMULATED_VALUES_COMPONENT_ID,
    ].join('|')),

    /**
     * Given an analysis/exploration object gives each panel an opportunity to respond if it
     * references any of the "dataSourceIds"
     * @param currentExploration - The analysis/exploration as it stands right now
     * @param dataSourceIds - The list of dataSourceIds
     * @returns {boolean} true | false - If any of the dataSourceIds are referenced in the currentExploration
     */
    areDataSourceIdsBeingReferenced: function(currentExploration, dataSourceIds) {
        var self = this;
        var reference = false;

        // Build a map of the JSON content for each panel in the exploration keyed by the panel's name attribute
        var panelContentMap = self.ExplorationUtil.buildPanelContentMap(currentExploration);

        // Setup our temporary model cache
        var modelCache = {};
        modelCache[HprConstants.HPR_CROSSTAB_PANEL] = new self.CrosstabContext("fakeId");
        modelCache[HprConstants.HPR_PLOT_PANEL] = new self.HprChartModel("fakeId");
        modelCache[HprConstants.HPR_PORTFOLIO_DETAILS_PANEL] = new self.PortfolioDetailsModel("fakeId");
        modelCache[HprConstants.HPR_RISK_FACTOR_SENSITIVITIES_PANEL] = new self.RiskFactorSensitivitiesModel("fakeId");
        modelCache[HprConstants.HPR_SIMULATED_VALUES_PANEL] = new self.TableViewerModel("fakeId");

        // Walk each item in the panelContentMap and give a model object of that type an opportunity to respond
        // with an answer if the data source ids being removed are referenced by that model
        $.each(_.values(panelContentMap), function(index, contentJSON) {
            // Set correct panel type; All Plots get type of Plot so HprChartModel can be called correctly
            var panelType = contentJSON.type;
            if (_.contains(_.values(HprConstants.PlotType), panelType)) {
                panelType = HprConstants.HPR_PLOT_PANEL;
            }
            var model = modelCache[panelType];
            var referenced = model.areDataSourceIdsBeingReferenced(contentJSON, dataSourceIds);
            if (referenced) {
                reference = true;
                return false; // Found at least one reference, break out of each loop
            }
        });

        return reference;
    },

    /**
     * Given an analysis/exploration object gives each panel an opportunity to provide updated json to reflect how it's
     * state should be modified based on whether we are replacing the data source or removing the data source
     * @param bReplaceDataSource - A flag to signify we are replacing the data source; 'false' indicates we are removing a data source
     * @param bReplaceForPanelsOnly - Indicates whether or not just the panels should be updated or the entire exploration
     * @param currentExploration - The analysis/exploration as it stands right now
     * @param oldDataSourceIds - The list of dataSourceIds that are being replaced or removed
     * @param newDataSourceIds - The list of dataSourceIds that the original list is being replaced with (should only be populated for replace data source)
     * @param completionHandler - Method to call when the updated exploration has been generated, the updated exploration JSON will be passed in
     */
    updateExploration: function(bReplaceDataSource, bReplaceForPanelsOnly, currentExploration, oldDataSourceIds, newDataSourceIds, completionHandler) {
        var self = this;

        var panelsUpdateDeferred = $.Deferred();
        var explorationItemsUpdateDeferred = $.Deferred();

        // clone of exploration to update
        var updatedExploration = jQuery.extend(true, {}, currentExploration);

        // Build a map of the JSON content for each panel in the exploration keyed by the panel's name attribute
        var panelContentMap = self.ExplorationUtil.buildPanelContentMap(currentExploration);

        // Setup our temporary model cache
        var modelCache = {};
        modelCache[HprConstants.HPR_CROSSTAB_PANEL] = new self.CrosstabContext("fakeId");
        modelCache[HprConstants.HPR_PLOT_PANEL] = new self.HprChartModel("fakeId");
        modelCache[HprConstants.HPR_PORTFOLIO_DETAILS_PANEL] = new self.PortfolioDetailsModel("fakeId");
        modelCache[HprConstants.HPR_RISK_FACTOR_SENSITIVITIES_PANEL] = new self.RiskFactorSensitivitiesModel("fakeId");
        modelCache[HprConstants.HPR_SIMULATED_VALUES_PANEL] = new sas.analytics.riskhp.common.riskcommons.table.TableViewerModel("fakeId");

        // Walk each item in the panelContentMap and give a model object of that type an opportunity to provide different
        // JSON after the correct model updates have been made.
        // For replacing a data source, each model must do the following:
        // - Provide a deferred which will be collected into the panelDeferredArray
        // - The deferred must return the JSON to represent the panel's state regardless of whether the panel should change because of the new data source
        // - The deferred must ALWAYS resolve successfully.
        // - The deferred may make subsequent/wrapped ajax calls but when it returns all processing should be complete.
        // - The deferred may not change state of the model or interact directly with the UI.
        // For removing a data source, each model must do the following:
        // - Provide a deferred which will be collected into the panelDeferredArray
        // - The deferred must return the JSON to represent the panel's state after the data source is removed
        // - The deferred must ALWAYS resolve successfully.
        var panelDeferredArray = [];
        _.each(_.values(panelContentMap), function(contentJSON) {
            // Set correct panel type; All Plots get type of Plot so HprChartModel can be called correctly
            var panelType = contentJSON.type;
            if (_.contains(_.values(HprConstants.PlotType), panelType)) {
                panelType = HprConstants.HPR_PLOT_PANEL;
            }
            var model = modelCache[panelType];

            // Call the correct model method based on whether we are calling this method for Replace Data Source or Remove Data Source
            var deferred = undefined;
            if (bReplaceDataSource) {
                deferred = model.replaceDataSource(contentJSON, oldDataSourceIds, newDataSourceIds);
            } else {
                deferred = model.removeDataSource(contentJSON, oldDataSourceIds);
            }
            if (deferred) {
                panelDeferredArray.push(deferred);
            }
        });

        // for now, removing a data source applies only to panels as it can only happen from within an open exploration
        var dataSourceParamsUpdateDeferred;
        var dataPaneUpdateDeferred;
        if (bReplaceDataSource === true && bReplaceForPanelsOnly !== true) {
            dataSourceParamsUpdateDeferred = self._replaceExplorationDataSourceParams(currentExploration, oldDataSourceIds, newDataSourceIds);
            dataPaneUpdateDeferred = self._replaceExplorationDataPaneDataSource(currentExploration, oldDataSourceIds, newDataSourceIds);
        }
        else {
            explorationItemsUpdateDeferred.resolve();
        }

        // Once all the deferreds in the panelDeferredArray are resolved update the panelContentMap with the updated JSON objects
        jQuery.when.apply(jQuery, panelDeferredArray).done(
            function() {
                // extract the content JSON for each panel
                _.each(arguments, function(result){
                    panelContentMap[result.name] = result;
                    console.log("Updating panel with id: " + result.name);
                });

                // Update the exploration/analysis object with the new panel content JSON that was just accumulated
                self._updateExplorationPanelContentJSON(updatedExploration, panelContentMap);

                panelsUpdateDeferred.resolve();
            }
        );

        if (dataSourceParamsUpdateDeferred/* && dataPaneUpdateDeferred*/) {
            jQuery.when(dataSourceParamsUpdateDeferred, dataPaneUpdateDeferred).done(
                function (updatedDataSourceParamsJSON, dataPaneUpdateItems) {
                    self._updateExplorationDataSourceParams(updatedExploration, updatedDataSourceParamsJSON);

                    // index 0 is the replacement data source ID, and item 1 is the updated data pane JSON
                    if (dataPaneUpdateItems) {
                        self._updateExplorationDataPane(updatedExploration, dataPaneUpdateItems[0], dataPaneUpdateItems[1]);
                    }

                    explorationItemsUpdateDeferred.resolve();
                }
            );
        }

        jQuery.when(panelsUpdateDeferred, explorationItemsUpdateDeferred).done(
            function() {
                if (completionHandler) {
                    completionHandler(updatedExploration);
                }
            }
        );
    },

    /**
     *
     * @param currentExploration
     * @param oldDataSourceIds
     * @param newDataSourceIds
     * @private
     */
    _replaceExplorationDataSourceParams: function(explorationAnalysis, oldDataSourceIds, newDataSourceIds) {
        var self = this;
        var deferred = $.Deferred();

        var dataSourceParams = this.ExplorationUtil.getDataSourceParams(explorationAnalysis);
        var oldDataSourceParamsMap = {};
        var updatedDataSourceParams = [];
        _.each(dataSourceParams, function(dataSourceParam) {
            if (_.contains(oldDataSourceIds, dataSourceParam.realDataSourceId) !== true) {
                updatedDataSourceParams.push(dataSourceParam);
            }
            else {
                oldDataSourceParamsMap[dataSourceParam.realDataSourceId] = dataSourceParam;
            }
        });

        // build a map of new ids to the ids they are replacing
        var paramsDeferredsArray = [];
        var idReplaceMap = {};
        _.each(newDataSourceIds, function(newDataSourceId) {
            var newDataSourceParam = self.DataSourceCacheManager.getDataSourceParam(newDataSourceId).toJSON();
            var idToReplace = self.HprUtil.getCompatibleDataSourceId(newDataSourceId, oldDataSourceIds);
            if (idToReplace) {
                idReplaceMap[newDataSourceId] = idToReplace;

                var oldDataSourceParams = oldDataSourceParamsMap[idToReplace];
                if (oldDataSourceParams.dataSourceTreeTableModels[oldDataSourceParams.realDataSourceId]) {
                    paramsDeferredsArray.push(self._updateOrderCaches(oldDataSourceParams.dataSourceTreeTableModels[oldDataSourceParams.realDataSourceId], newDataSourceId).done(
                        function (updatedOrderCache) {
                            newDataSourceParam.dataSourceTreeTableModels = {};
                            newDataSourceParam.dataSourceTreeTableModels[newDataSourceId] = updatedOrderCache;
                        }
                    ));
                }
            }
            updatedDataSourceParams.push(newDataSourceParam);
        });

        if (paramsDeferredsArray.length > 0) {
            jQuery.when.apply(jQuery, paramsDeferredsArray).done(
                function () {
                    deferred.resolve(updatedDataSourceParams);
                }
            );
        }
        else {
            deferred.resolve(updatedDataSourceParams);
        }

        return deferred.promise();
    },

    /**
     *
     * @param currentExploration
     * @param oldDataSourceIds
     * @param newDataSourceIds
     * @private
     */
    _replaceExplorationDataPaneDataSource: function(explorationAnalysis, oldDataSourceIds, newDataSourceIds) {
        var self = this;
        var deferred = $.Deferred();

        if (_.contains(oldDataSourceIds, explorationAnalysis.selectedDataSourceId) !== true) {
            deferred.resolve();
        }

        var dataPaneJSON = self.ExplorationUtil.getDataPane(explorationAnalysis);
        var replacementDataSourceId = self.HprUtil.getCompatibleDataSourceId(explorationAnalysis.selectedDataSourceId, newDataSourceIds);
        if (replacementDataSourceId) {
            //TODO handle horizon order cache
            self._updateOrderCaches(dataPaneJSON, replacementDataSourceId).done(
                function (updatedOrderCaches) {
                    deferred.resolve(replacementDataSourceId, updatedOrderCaches);
                }
            );
        }
        else {
            deferred.resolve(newDataSourceIds[0], null);
        }

        return deferred.promise();
    },

    /**
     *
     * @param oldOrderCaches
     * @param newDataSourceId
     * @private
     */
    _updateOrderCaches: function(oldOrderCaches, newDataSourceId) {
        var self = this;
        var deferred = $.Deferred();
        var internalDeferredArray = [];

        var __translateToArray = function(dataItemIndexMap) {
            var dataItemArray = [];
            _.each(Object.keys(dataItemIndexMap), function(dataItemId) {
                dataItemArray[dataItemIndexMap[dataItemId].index] = dataItemId;
            });
            return dataItemArray;
        };

        var __getUpdatedOrderMapForDataSource = function(oldDataItemArray, newDataSourceId) {
            var myDeferred = $.Deferred();

            self.DataSourceCacheManager.getDataSource(newDataSourceId).done(
                function(dataSource) {
                    var updatedDataItemArray = [];
                    _.each(oldDataItemArray, function(dataItemId) {
                        if (dataSource.getDataItemById(dataItemId)) {
                            updatedDataItemArray.push(dataItemId);
                        }
                    });

                    var updatedOrderMap = null;
                    if (updatedDataItemArray.length > 0) {
                        updatedOrderMap = {};
                        for (var i = 0; i < updatedDataItemArray.length; i++) {
                            updatedOrderMap[updatedDataItemArray[i]] = {index: i};
                        }
                    }
                    myDeferred.resolve(updatedOrderMap);
                }
            ).fail(
                function() {
                    myDeferred.resolve(null);
                }
            );

            return myDeferred.promise();
        };

        // updated order cache to be returned
        var updatedOrderCache = {
            horizonNodeChildrenCache: {},
            nodeOrderMap: {}
        };

        // update the horizon order cache
        if (!_.isEmpty(oldOrderCaches.horizonNodeChildrenCache)) {
            _.each(Object.keys(oldOrderCaches.horizonNodeChildrenCache), function (nodeOrderKey) {
                internalDeferredArray.push(__getUpdatedOrderMapForDataSource(__translateToArray(oldOrderCaches.horizonNodeChildrenCache[nodeOrderKey]), newDataSourceId).done(
                    function (updatedOrderMap) {
                        if (updatedOrderMap) {
                            updatedOrderCache.horizonNodeChildrenCache[nodeOrderKey] = updatedOrderMap;
                        }
                    }
                ));
            });
        }

        // update the node order map
        if (!_.isEmpty(oldOrderCaches.nodeOrderMap)) {
            _.each(Object.keys(oldOrderCaches.nodeOrderMap), function (nodeOrderKey) {
                internalDeferredArray.push(__getUpdatedOrderMapForDataSource(__translateToArray(oldOrderCaches.nodeOrderMap[nodeOrderKey].children), newDataSourceId).done(
                    function (updatedOrderMap) {
                        if (updatedOrderMap) {
                            updatedOrderCache.nodeOrderMap[nodeOrderKey] = {children: updatedOrderMap};
                        }
                    }
                ));
            });
        }

        jQuery.when.apply(jQuery, internalDeferredArray).done(
            function() {
                deferred.resolve(updatedOrderCache);
            }
        );

        return deferred.promise();
    },

    /**
     * Helper function to determine if the object passed in represents the json for a content pane
     * @param object
     * @returns {boolean}
     * @private
     */
    _isContentPaneObject: function(object) {
        if (!object) {
            return false;
        }
        var name = object.name;
        if (!name) {
            return false;
        }
        return this.contentPaneTest.test(name);
    },

    /**
     * Helper function to update the given exploration JSON with the JSON fragments held in the passed in contentMap.
     * It is expected the passed in contentMap contains all keys for all content on all pages of the exploration.
     * @param exploration
     * @param contentMap
     * @private
     */
    _updateExplorationPanelContentJSON: function(exploration, contentMap) {
        // Grab all the pages of the passed in exploration and walk them
        var pages = exploration.layout.pages;
        _.each(pages, function(page) {

            // Pluck the keys for the content panels on the current page and walk them;
            // Make sure the page has content and panelContentData in order to pull the keys
            var keys = (page.content && page.content.panelContentData) ? _.keys(page.content.panelContentData) : [];
            _.each(keys, function(key){

                // For each content key on the given page find the updated content in the passed in contentMap and replace the
                // content in the page's panelContentData map
                var newContentJSON = contentMap[key];
                if (newContentJSON) {
                    page.content.panelContentData[key] = newContentJSON;
                }
                // NOTE: If the content is not found in the contentMap then we have a problem.  Could/should we come up with
                // a generic zero state object that the RPVDelegate could use to go to zero state?  Need to ponder this. MCA
            })
        });
    },

    /**
     *
     * @param explorationAnalysis
     * @param updatedDataSourceParams
     * @private
     */
    _updateExplorationDataSourceParams: function(explorationAnalysis, updatedDataSourceParams) {
        explorationAnalysis.dataSourceParams = updatedDataSourceParams;
    },

    /**
     *
     * @param explorationAnalysis
     * @param newSelectedDataSourceID
     * @param updatedDataPaneJSON
     * @private
     */
    _updateExplorationDataPane: function(explorationAnalysis, newSelectedDataSourceID, updatedDataPaneJSON) {
        if (newSelectedDataSourceID) {
            explorationAnalysis.selectedDataSourceId = newSelectedDataSourceID;
        }
        if (updatedDataPaneJSON) {
            explorationAnalysis.layout.dataPane = updatedDataPaneJSON;
        }
    }
};