// (c) 2018, SAS Institute Inc.

jQuery.sap.require("sas.analytics.riskhp.models.DataSourceCacheManager");
jQuery.sap.require("sas.analytics.riskhp.models.dataitems.Hierarchy");
jQuery.sap.require("sas.analytics.riskhp.models.dataitems.Horizon");
jQuery.sap.require("sas.analytics.riskhp.models.dataitems.RelativeHorizon");
jQuery.sap.require("sas.analytics.riskhp.utils.HprUtil");
jQuery.sap.require("sas.analytics.riskhp.views.datapane.DataPaneModelManager");
jQuery.sap.require("sas.hc.m.Button");
jQuery.sap.require("sas.hc.m.Dialog");
jQuery.sap.require("sas.hc.ui.commons.formControl.Field");
jQuery.sap.require("sas.hc.ui.commons.formControl.FormControl");
jQuery.sap.require("sas.hc.ui.commons.formControl.Group");

/**
 * Data Item Rename dialog
 */
sap.ui.jsfragment("sas.analytics.riskhp.fragments.DataItemRenameDialog", {

    DataPaneModelManager: sas.analytics.riskhp.views.datapane.DataPaneModelManager,
    DataSourceCacheManager: sas.analytics.riskhp.models.DataSourceCacheManager,
    Hierarchy: sas.analytics.riskhp.models.dataitems.Hierarchy,
    HORIZON: sas.analytics.riskhp.models.dataitems.Horizon.HORIZON,
    RELATIVEHORIZON: sas.analytics.riskhp.models.dataitems.RelativeHorizon.RELATIVEHORIZON,
    HprUtil: sas.analytics.riskhp.utils.HprUtil,
    Button: sas.hc.m.Button,
    Dialog: sas.hc.m.Dialog,
    Field: sas.hc.ui.commons.formControl.Field,
    FormControl: sas.hc.ui.commons.formControl.FormControl,
    Group: sas.hc.ui.commons.formControl.Group,

    createContent: function(ctrl) {

        var self = this;

        // Store away some values for later use
        this.dataSource = null;
        this.selectedDataItem = null;
        this.existingLabels = [];

        this.labelField = new self.Field(ctrl.createId("-labelField"), {
            label: self.HprUtil.getBundle().getText("DataPane.data.item.rename.dialog.label.field.label"),
            type: self.Field.TEXT_TYPE,
            validator: {
                required: true,
                live: function(curVal) {
                    // Check to see if the current value is already an existing label for one of the other data items of the same type;
                    // If so, return an Error state with message
                    return {
                        valueState: _.contains(self.existingLabels, curVal) ? sap.ui.core.ValueState.Error : sap.ui.core.ValueState.None,
                        valueStateText: self.HprUtil.getBundle().getText("DataPane.properties.label.field.duplicate.label.txt")
                    }
                }
            }
        });

        this.formControl = new self.FormControl(ctrl.createId("-formControl"), {
            displayHeader: false,
            showGroupTitle: false,
            showResetToDefaults: false,
            rootGroup: new self.Group({
                groups: [
                    new self.Group({
                        fields: [this.labelField]
                    })
                ]
            })
        });

        var onOk = function() {
            // Check state of Label field and put focus back if in Error state
            if (!self.labelField.getValueValid()) {
                self.formControl.getControlByFieldId(self.labelField.getId()).focus();
                return;
            }

            // Label update is fine, set the label on the Data Item
            self.selectedDataItem.setLabel(self.labelField.getActiveValue());

            // For regular Horizon and Relative Horizon DataItems stored in the DataPane, we have to update the
            // label of the corresponding DataItem stored in the DataSource.  This is because the Crosstab uses
            // the AbsoluteHorizon DataItems and because for the DataPane we regenerate the RelativeHorizons when
            // we need them (which causes the objects to not match and binding to not work)
            if (self.selectedDataItem.getType() === self.RELATIVEHORIZON) {
                // Locate RelativeHorizon DataItem using id since DataPane and Crosstab use RelativeHorizon DataItems
                var dataSourceDataItem = self.dataSource.getDataItemById(self.selectedDataItem.getId());
                if (dataSourceDataItem) {
                    dataSourceDataItem.setLabel(self.labelField.getActiveValue());
                }
            } else if (self.selectedDataItem.getType() === self.HORIZON) {
                // Since Crosstab uses the AbsoluteHorizon DataItem, we need to locate it using horizonDate
                // and old label since it hasn't been updated yet
                var absHorizon = self.dataSource.findAbsoluteHorizon(self.selectedDataItem.horizonDate, self.labelField.getDefaultValue(), self.selectedDataItem.basecase);
                if (absHorizon) {
                    absHorizon.setLabel(self.labelField.getActiveValue());
                }
            }

            // Update the tree table model by refreshing it so new label will appear thanks to binding
            var dataSourceTreeTableModel = ctrl.getView().dataSourceTreeTable.getModel();
            dataSourceTreeTableModel.refresh();

            // Fire an event that signals a label for a Data Item has been updated and pass along some useful info
            sas.eventBus.publish(HprConstants.CHANNEL_EXPLORATION, HprConstants.EVENT_DATA_ITEM_LABEL_UPDATED, {
                dataSourceId: self.DataSourceCacheManager.getSelectedDataSourceId(),
                dataItem: self.selectedDataItem,
                oldLabel: self.labelField.getDefaultValue(),
                newLabel: self.labelField.getActiveValue()
            });

            // Update the dirty state of the Exploration view
            sas.eventBus.publish(HprConstants.CHANNEL_EXPLORATION, HprConstants.EVENT_UPDATE_DIRTY_STATE, {
                dirty: true
            });

            self.dialog.close();
        };

        var okButton = new self.Button(ctrl.createId("-okButton"), {
            text: self.HprUtil.getBundle().getText("hprisk.okButton.txt"),
            press: onOk
        });

        var onCancel = function() {
            self.dialog.close();
        };

        var cancelButton = new self.Button(ctrl.createId("-cancelButton"), {
            text: self.HprUtil.getBundle().getText("hprisk.cancelButton.txt"),
            press: onCancel
        });

        this.dialog = new self.Dialog(ctrl.createId("-renameDialog"), {
            title: self.HprUtil.getBundle().getText("DataPane.data.item.rename.dialog.title"),
            content: [this.formControl],
            leftButton: okButton,
            defaultButton: okButton,
            rightButton: cancelButton,
            afterClose: function(evt) {
                this.destroy();
            }
        });

        this.dialog.attachBeforeOpen(function() {
            var model = self.dialog.getModel();
            var selectedDataItem = model.getProperty("/selectedDataItem");

            // Get the current DataSource from the DataPaneModelManager
            var deferred = self.DataSourceCacheManager.getDataSource(self.DataPaneModelManager.getSelectedDataSourceId());
            deferred.done(function(dataSource) {
                // Retrieve the selected data item type and turn Forced Hierarchies back into regular Hierarchies so we can search for labels
                var selectedDataItemType = selectedDataItem.getType();
                if (selectedDataItemType === self.Hierarchy.FORECEDHIERARCHY) { selectedDataItemType = self.Hierarchy.HIERARCHY; }

                // Gather all the labels for the other data items of the same type but do not include the selected one
                var dataItems = dataSource.getDataItemsForType(selectedDataItemType);
                self.existingLabels = _.without(_.pluck(dataItems, "label"), selectedDataItem.getLabel());

                // Keep track of the selected Data Source
                self.dataSource = dataSource;

                // Keep track of the selected Data item
                self.selectedDataItem = selectedDataItem;

                // Fill in the Label field's value and default value (default value seems to help with ESC key and resetting value)
                self.labelField.setValue(selectedDataItem.getLabel());
                self.labelField.setDefaultValue(selectedDataItem.getLabel());

                // Set the width of the dialog
                self.dialog.setContentWidth("20%");

                // Set the height of the dialog
                self.dialog.setContentHeight("30%");
            });
        });

        return this.dialog;
    }
});