// (c) 2017, SAS Institute Inc.

jQuery.sap.declare("sas.analytics.riskhp.models.DataSource");

jQuery.sap.require("sas.analytics.riskhp.models.DataSourceUtil");
jQuery.sap.require("sas.analytics.riskhp.models.dataitems.Benchvar");
jQuery.sap.require("sas.analytics.riskhp.models.dataitems.CashFlowLeg");
jQuery.sap.require("sas.analytics.riskhp.models.dataitems.Classvar");
jQuery.sap.require("sas.analytics.riskhp.models.dataitems.Compvar");
jQuery.sap.require("sas.analytics.riskhp.models.dataitems.Hierarchy");
jQuery.sap.require("sas.analytics.riskhp.models.dataitems.Horizon");
jQuery.sap.require("sas.analytics.riskhp.models.dataitems.RelativeHorizon");
jQuery.sap.require("sas.analytics.riskhp.models.dataitems.AbsoluteHorizon");
jQuery.sap.require("sas.analytics.riskhp.models.dataitems.Outvar");
jQuery.sap.require("sas.analytics.riskhp.models.dataitems.RiskCube");
jQuery.sap.require("sas.analytics.riskhp.models.dataitems.Statistic");
jQuery.sap.require("sas.analytics.riskhp.models.dataitems.ResultSet");
jQuery.sap.require("sas.analytics.riskhp.models.dataitems.Scenario");
jQuery.sap.require("sas.analytics.riskhp.models.options.Option");
jQuery.sap.require("sas.analytics.riskhp.models.restrictions.Restriction");
jQuery.sap.require("sas.analytics.riskhp.utils.HprConstants");

/**
 * @typedef {Object} CapabilitiesObject - Holds capabilities of the cube
 * @property {boolean} densitySupported - Is density chart supported
 * @property {boolean} distributionSupported - Is distribution chart supported
 * @property {boolean} sensitivitiesSupported - Are sensitivities supported
 * @property {boolean} detailsSupported
 * @property {boolean} vectorsSupported
 * @property {boolean} rfimSupported - Are risk factor information measures supported
 */
// noinspection JSClosureCompilerSyntax
/**
 * @name sas.analytics.riskhp.models.DataSource
 * @extends sap.ui.base.Object
 */
sap.ui.base.Object.extend("sas.analytics.riskhp.models.DataSource", {

    constructor : function(json) {
        var self = this;

        this.json = json;
        this.masterDataItemArray = []; // Big list of dataItems.  ALL dataItems.
        this.classVariable = [];
        this.statistic = [];
        this.horizon = [];
        this.relativeHorizon = [];
        this.absoluteHorizon = [];
        this.outputVariable = [];
        this.computedVariable = [];
        this.benchmarkVariable = [];
        this.riskCube = [];
        this.hierarchy = [];
        this.cashFlowLeg = [];
        this.resultSet = [];
        this.scenarios = [];
        this.dataSource = json.dataSource; // TODO: Used now, create DataSource model object class?
		this.defaults = json.defaults;
        this.capabilities = {};
        this.restrictions = self._parseCubeRestrictionsData(json).restrictions;
        this.optionsObject = self._parseCubeOptionsData(json);


        ////////////////
        // Data Items //
        ////////////////

        var dataItemsFromJSON = json.dataItems;

        // class variables
        _.each(dataItemsFromJSON.classVariables, function(data) {
            var value = new sas.analytics.riskhp.models.dataitems.Classvar(data);
            self.classVariable.push(value);
            self.masterDataItemArray.push(value);
        });

        // risk cubes
        _.each(dataItemsFromJSON.riskCubes, function(data) {
            var riskCube = new sas.analytics.riskhp.models.dataitems.RiskCube(data);
            self.riskCube.push(riskCube);
            self.masterDataItemArray.push(riskCube);

            // Add a reference to the horizons in a big list of horizons.
            if (self.isStressCube() === true) {
                _.each(data.resultSets, function(resultSetJSON) {
                    var resultSet = new sas.analytics.riskhp.models.dataitems.ResultSet(resultSetJSON);
                    resultSet.riskCubeId = riskCube.getId();
                    if (resultSet.isDefault()) {
                        self.defaultResultSet = resultSet;
                    }
                    self.resultSet.push(resultSet);
                    self.masterDataItemArray.push(resultSet);

                    if (resultSet.horizon !== undefined) {
                        resultSet.horizon.scenarioId = undefined;
                        resultSet.horizon.riskCubeId = undefined;
                        resultSet.horizon.resultSetId = resultSet.getId();

                        self.masterDataItemArray.push(resultSet.horizon);

                        if (resultSet.isDefault() && resultSet.horizon.isBasecase()) {
                            self.horizon.push(resultSet.horizon);
                        }
                    }
                    else {
                        resultSet.horizon = undefined;
                    }

                    // Save off the subtending scenario dataItems and their subtending horizon dataItems
                    // to the big list of DataItems.
                    _.each(resultSet.scenarios, function(scenario) {
                        scenario.resultSetId = resultSet.getId();
                        scenario.setActivated(true); // Technically activated since it was baked into the cube

                        // Do not add scenario to '.scenarios' master list because we don't want it show up on the DataPane
                        self.masterDataItemArray.push(scenario);

                        _.each(scenario.horizons, function(horizon) {
                            horizon.scenarioId = scenario.getId();
                            horizon.resultSetId = undefined;
                            horizon.riskCubeId = undefined;

                            // Do not add horizon to '.horizon' master list because we don't want it show up on the DataPane
                            self.masterDataItemArray.push(horizon);
                        });
                    });
                });
            }
            else {
                _.each(riskCube.horizons, function (horizon) {
                    horizon.resultSetId = undefined;
                    horizon.scenarioId = undefined;
                    horizon.riskCubeId = riskCube.getId();
                    self.horizon.push(horizon);
                    self.masterDataItemArray.push(horizon);
                });
            }
        });
        if (self.riskCube.length > 0) {
            self.capabilities = self.riskCube[0].capabilities;
        }

        // hierarchies
        _.each(dataItemsFromJSON.hierarchies, function(data) {
            var value = new sas.analytics.riskhp.models.dataitems.Hierarchy(data);
            self.hierarchy.push(value);
            self.masterDataItemArray.push(value);
            var cvList = [];
            _.each(data.classVariables, function(cvID) {
                var cv = undefined;
                cv = self.getDataItemById(cvID);
                cvList.push(cv);
            });
            value.classVariables = cvList;
        });

        // output variables
        _.each(dataItemsFromJSON.outputVariables, function(data) {
            var value = new sas.analytics.riskhp.models.dataitems.Outvar(data);
            self.outputVariable.push(value);
            self.masterDataItemArray.push(value);
        });

        // benchmark variables
        _.each(dataItemsFromJSON.benchmarkVariables, function(data) {
            var value = new sas.analytics.riskhp.models.dataitems.Benchvar(data);
            self.benchmarkVariable.push(value);
            self.masterDataItemArray.push(value);
        });

        // computed variables
        _.each(dataItemsFromJSON.computedVariables, function(data) {
            var value = new sas.analytics.riskhp.models.dataitems.Compvar(data);
            self.computedVariable.push(value);
            self.masterDataItemArray.push(value);
        });

        // stats
        _.each(dataItemsFromJSON.statistics, function(data) {
            var value = new sas.analytics.riskhp.models.dataitems.Statistic(data);
            self.statistic.push(value);
            self.masterDataItemArray.push(value);
        });

        // cash flow legs
        _.each(dataItemsFromJSON.cashFlowLegs, function(data) {
            var value = new sas.analytics.riskhp.models.dataitems.CashFlowLeg(data);
            self.cashFlowLeg.push(value);
            self.masterDataItemArray.push(value);
        });

        // process instrument variables (the artist previously known as portfolio columns)
        self.instrumentVariables = dataItemsFromJSON.instrumentVariables;
        self.instrumentTypes = dataItemsFromJSON.instrumentTypes;
        self.requiredInstrumentVariables = dataItemsFromJSON.requiredInstrumentVariables;


        //////////////////////////////////
        // Create our UI-only dataItems //
        //////////////////////////////////

        // Take 'horizons' and create RelativeHorizon objects but don't apply label changes in this case
        self._generateRelativeHorizonDataItems(false);

        // Take 'horizons' and create AbsoluteHorizon objects
        self._generateAbsoluteHorizonDataItems();
    }
});

//*****************************************************************************************************
//*****************************************************************************************************
// General functions
//*****************************************************************************************************
//*****************************************************************************************************

sas.analytics.riskhp.models.DataSource.prototype.getId = function() {
    return this.dataSource.id;
};

/**
 * Returns the name of the data source.
 * @return {String}
 */
sas.analytics.riskhp.models.DataSource.prototype.getName = function() {
    return this.dataSource.info.label ? this.dataSource.info.label : this.dataSource.info.name;
};

/**
 * Returns the date the cube was created.
 * @return {String}
 */
sas.analytics.riskhp.models.DataSource.prototype.getCreatedDate = function() {
    return this.dataSource.info.sasCreationDate;
};

/**
 * Gather the list of properties to compare between two DataSources when a Compatibility Check takes place (i.e. "classVariable", "dataSource", etc.).
 * @returns [String] Returns an Array of Strings which represent the properties which should be compared by two DataSources
 */
sas.analytics.riskhp.models.DataSource.prototype.getValidPropertiesToCompare = function() {
    // This is all the properties (captured as a String representation) for this DataSource object captured in an Array
    var allKeys = _.keys(this); // (i.e. "classVariable", "capabilities", etc.)

    // This is an array of properties (more specifically, string representations of them) which should NOT be included in the Compatibility Check between two DataSources
    var keysToExclude = ["absoluteHorizon", "defaults", "json", "masterDataItemArray", "optionsObject", "relativeHorizon"];

    // The difference between the two Arrays are the properties we want to check between two DataSources during the Compatibility Check
    return _.difference(allKeys, keysToExclude);
};

/**
 * Uses the 'dataSource.type' property to determine if this cube represents a stress cube.
 * @returns {boolean}
 */
sas.analytics.riskhp.models.DataSource.prototype.isStressCube = function() {
    return this.dataSource.type === HprConstants.HPR_CUBE_INFORMATION_TYPE_STRESS;
};

/**
 * Uses the 'capabilities' property to determine if this cube supports stress testing
 * @returns {boolean}
 */
sas.analytics.riskhp.models.DataSource.prototype.isStressSupported = function() {
    return this.capabilities.stressSupported;
};

/**
 * Returns the 'dataSource.info.joinType' property so the type of cube can be determined
 * @return {String}
 */
sas.analytics.riskhp.models.DataSource.prototype.getJoinType = function() {
    return this.dataSource.info.joinType;
};

/**
 * Uses the 'dataSource.info.joinType' property to determine if this cube is a Side by Side
 * @returns {boolean}
 */
sas.analytics.riskhp.models.DataSource.prototype.isSideBySide = function() {
    return this.dataSource.info.joinType === HprConstants.HPR_CUBE_INFORMATION_JOINTYPE_SIDEBYSIDE;
};

/**
 * Uses the 'dataSource.info.joinType' property to determine if this cube is an Aggregate Join
 * @returns {boolean}
 */
sas.analytics.riskhp.models.DataSource.prototype.isAggregateJoin = function() {
    return this.dataSource.info.joinType === HprConstants.HPR_CUBE_INFORMATION_JOINTYPE_AGGREGATE;
};

//TODO: Talk to Bill as to whether this is a good idea
/**
 * If the DataSource only has one RiskCube this method returns an empty array so UI components
 * will render correctly.  Otherwise, this method returns the full array.
 * @returns Array The array of valid Risk Cubes for this DataSource or an empty array
 */
sas.analytics.riskhp.models.DataSource.prototype.getValidRiskCubes = function() {

	// If there is only one risk cube, then cube is an aggregate join, price cube, or stress cube and
	// Cubes with only one risk cube do not need to display a risk cube under that category
    // For Price (including Aggregate Joins) and Stress cubes, there is only one riskCube dataItem.
	// This should not be displayed on the DataPane, but will be needed by the Crosstab.
	// For Price Comparison Join cubes, there are multiple riskCube dataItems. Each should be displayed
	// in the DataPane under the RiskCubes category and all will be needed by the Crosstab.

    if (this.riskCube.length === 1) {
        return [];
    }
	return this.riskCube;
};

/**
 * Returns the set of all risk cubes for this data source. This is needed for some cases (see
 * PortfolioDetailsModel.replaceDataSource) since the getDataItemsForType returns an empty array
 * for risk cubes when the data source is not a side-by-side.
 * @return
 */
sas.analytics.riskhp.models.DataSource.prototype.getAllRiskCubes = function() {
    return this.riskCube;   //TODO should this be cloned?
};

/**
 * Get the id of the DataSource
 * @returns String The id of this DataSource
 */
sas.analytics.riskhp.models.DataSource.prototype.getCubeId = function() {
    return this.dataSource.id;
};

/**
 * Get the label of the DataSource.  For a displayable label in the UI which contains the DataSource 'type' property, see the "getCubeDisplayLabel" method.
 * @returns String The label of this DataSource
 */
sas.analytics.riskhp.models.DataSource.prototype.getCubeLabel = function() {
    return this.dataSource.label;
};

/**
 * Get the displayable label of the DataSource.  This will usually be the 'label' property from the mid-tier (if it exists) concatenated
 * with the 'type' property.  If the 'label' property does not exist, then the 'name' property is used.  The resulting displayable label
 * will be something like "AutoPrice (price)".
 * @returns String The displayable label of this DataSource
 */
sas.analytics.riskhp.models.DataSource.prototype.getCubeDisplayLabel = function() {
    // Since we also have cases where the DataSource displayable label is needed in code that does not use a DataSource object/class, just rely on the generic function in HprUtil here
    return sas.analytics.riskhp.utils.HprUtil.getCubeDisplayLabel(this.dataSource.label, this.dataSource.name, this.dataSource.type);
};

/**
 * Get the displayable path of the DataSource.  This will usually be the DataSource 'workGroup' folder path appended with the 'relPath' folder path.  If
 * the 'relPath' does not exist it is just the 'workGroup' folder path.  NOTE: The application folder is not contained in this cube path as it was deemed
 * not useful.
 * @returns {String}
 */
sas.analytics.riskhp.models.DataSource.prototype.getCubePath = function() {
    // Since we also have cases where the DataSource cube path is needed in code that does not use a DataSource object/class, just rely on the generic function in HprUtil here
    return sas.analytics.riskhp.utils.HprUtil.getCubePath(this.dataSource.workGroup, this.dataSource.relPath);
};

/**
 * Returns the default filters saved in the cube.
 * @return {Array}
 */
sas.analytics.riskhp.models.DataSource.prototype.getDefaultFilters = function() {
    return this.defaults.filters ? this.defaults.filters : [];
};

/**
 * Returns the JSON representation of this data source suitable for use in a query as the
 * value of the "dataSource" parameter
 * @returns {{id: *, application: *, workGroup: *, relPath: *, type: *}}
 */
sas.analytics.riskhp.models.DataSource.prototype.queryJSON = function() {
    return {
        id: this.dataSource.id,
        application: this.dataSource.application,
        workGroup: this.dataSource.workGroup,
        relPath: this.dataSource.relPath,
        type: this.dataSource.type
    };
};


//*****************************************************************************************************
//*****************************************************************************************************
// Instrument variable functions
//*****************************************************************************************************
//*****************************************************************************************************

sas.analytics.riskhp.models.DataSource.prototype.getInstrumentVariables = function() {
    return this.instrumentVariables;
};

sas.analytics.riskhp.models.DataSource.prototype.getInstrumentTypes = function() {
    return this.instrumentTypes;
};

sas.analytics.riskhp.models.DataSource.prototype.getRequiredInstrumentVariables = function() {
    return this.requiredInstrumentVariables;
};

sas.analytics.riskhp.models.DataSource.prototype.getDefaultInstrumentVariables = function() {
    return this.defaults.instrumentVariables;
};


//*****************************************************************************************************
//*****************************************************************************************************
// Horizon functions
//*****************************************************************************************************
//*****************************************************************************************************

/**
 * Returns the horizons saved in the cube.
 * @return {Array}
 */
sas.analytics.riskhp.models.DataSource.prototype.getHorizons = function() {
	return $.extend(true, [], this.horizon);
};

/**
 * Takes a list of Horizons and generates RelativeHorizon data items.
 * @param applyChangedLabels {Boolean} Apply the label changes stored in the DataSourceParams
 * object to the new RelativeHorizon data items
 * @returns {Array}
 */
sas.analytics.riskhp.models.DataSource.prototype._generateRelativeHorizonDataItems = function(applyChangedLabels) {
    var self = this;

    // clean up master data items list
    self.masterDataItemArray = _.reject(self.masterDataItemArray, function(dataItem) {
        return _.findIndex(self.relativeHorizon, function(relativeHorizon) {
            return dataItem.getId() === relativeHorizon.getId();
        }) !== -1;
    });

    // build a set of all horizons
    var allHorizons = $.extend(true, [], self.horizon);
    _.each(self.resultSet, function(rs) {
        if (rs.horizon !== undefined){
            allHorizons.push(rs.horizon);
        }
        _.each(rs.scenarios, function(scenario) {
            _.each(scenario.horizons, function(horizon) {
                allHorizons.push(horizon);
            });
        });
    });

    // replace the set of relative horizons with the set built from allHorizons
    self.relativeHorizon = sas.analytics.riskhp.models.DataSourceUtil.generateRelativeHorizonDataItems(allHorizons, applyChangedLabels, self.getId());
    self.masterDataItemArray = self.masterDataItemArray.concat(self.relativeHorizon);
};

/**
 * Takes a list of Horizons and generates AbsoluteHorizon data items.
 * @returns {Array}
 */
sas.analytics.riskhp.models.DataSource.prototype._generateAbsoluteHorizonDataItems = function() {
    var self = this;

    // clean up master data items list
    self.masterDataItemArray = _.reject(self.masterDataItemArray, function(dataItem) {
        return _.findIndex(self.absoluteHorizon, function(absoluteHorizon) {
            return dataItem.getId() === absoluteHorizon.getId();
        }) !== -1;
    });

    // build a set of all horizons
    var allHorizons = $.extend(true, [], self.horizon);
    _.each(self.resultSet, function(rs) {
        if (rs.horizon !== undefined){
            allHorizons.push(rs.horizon);
        }
        _.each(rs.scenarios, function(scenario) {
            _.each(scenario.horizons, function(horizon) {
                allHorizons.push(horizon);
            });
        });
    });

    // replace the set of relative horizons with the set built from allHorizons
    self.absoluteHorizon = sas.analytics.riskhp.models.DataSourceUtil.generateAbsoluteHorizonDataItems(allHorizons);
    self.masterDataItemArray = self.masterDataItemArray.concat(self.absoluteHorizon);
};

/**
 * Looks through the relative horizons and returns the first match
 * @param interval
 * @param horizonNumber
 */
sas.analytics.riskhp.models.DataSource.prototype.findRelativeHorizon = function(interval, horizonNumber, basecase) {
    if (basecase === undefined)
        basecase = false;
    return _.findWhere(this.relativeHorizon, {interval: interval, horizonNumber: horizonNumber, basecase: basecase});
};

/**
 * Looks through the absolute horizons and returns the first match
 * @param horizonDate
 * @param horizonLabel
 */
sas.analytics.riskhp.models.DataSource.prototype.findAbsoluteHorizon = function(horizonDate, horizonLabel, basecase) {
    if (basecase === undefined)
        basecase = false;
    if (horizonLabel && horizonLabel !== undefined) {
        return _.findWhere(this.absoluteHorizon, {horizonDate: horizonDate, label: horizonLabel, basecase: basecase});
    }
    return _.findWhere(this.absoluteHorizon, {horizonDate: horizonDate, basecase: basecase});
};

/**
 * Finds the relative horizon related to the passed in horizonId.  May return undefined if the horizonId is not specified
 * or the horizon could not be found, or the relative horizon could not be found
 * @param horizonId
 * @returns {undefined | DataItem}
 */
sas.analytics.riskhp.models.DataSource.prototype.getRelatedRelativeHorizon = function(horizonId) {
    if (!horizonId) return undefined;

    var horizon = this.getDataItemById(horizonId);
    if (horizon) {
        return this.findRelativeHorizon(horizon.interval, horizon.horizonNumber, horizon.basecase);
    }
    return undefined;
};

/**
 * Determines if a given Data Item represents a 0 Day Horizon
 * @param dataItem
 * @returns {boolean} true | false
 */
sas.analytics.riskhp.models.DataSource.prototype.isDataItem0DayHorizon = function(dataItem) {
    if (!dataItem) return false;

    if (dataItem && (dataItem.getType() !== sas.analytics.riskhp.models.dataitems.Horizon.HORIZON && dataItem.getType() !== sas.analytics.riskhp.models.dataitems.RelativeHorizon.RELATIVEHORIZON)) {
        return false;
    }

    return dataItem.getHorizonNumber() === 0 && dataItem.getInterval().toLowerCase() === "day";
};

/**
 * Determines if a given Data Item represents the BaseCase Horizon
 * @param dataItem
 * @returns {boolean} true | false
 */
sas.analytics.riskhp.models.DataSource.prototype.isDataItemBaseCaseHorizon = function(dataItem) {
    if (!dataItem) return false;

    if (_.isFunction(dataItem.isBasecase)) {
        return dataItem.isBasecase();
    }
};

// // Determines if a given horizon is associated with a scenario.
// //FIXME Something to think about: adding scenarioId to horizon and resultSetId to scenario.
// //FIXME This would make look ups fast, not like this
// sas.analytics.riskhp.models.DataSource.prototype.isHorizonAssociatedWithScenario = function(horizon) {
//
//     _.each(this.resultSet, function(rsDataItem){
//         _.each(rsDataItem.scenarios, function(scenDataItem) {
//             _.each(scenDataItem.horizons, function(horizonDataItem) {
//                 if (horizonDataItem.getId() === horizon.getId()) {
//                     return true;
//                 }
//             })
//         })
//     });
//     return false;
// };
//

//*****************************************************************************************************
//*****************************************************************************************************
// DataItem functions
//*****************************************************************************************************
//*****************************************************************************************************

/**
 * Given one of the dataitem types returns an array of those representing what this cube has
 * @param dataItemType
 * @returns {Array}
 */
sas.analytics.riskhp.models.DataSource.prototype.getDataItemsForType = function(dataItemType) {
	switch(dataItemType) {
        case sas.analytics.riskhp.models.dataitems.RiskCube.RISKCUBE:
            return this.getValidRiskCubes();
        case sas.analytics.riskhp.models.dataitems.Classvar.CLASSVAR:
            return this.classVariable;
        case sas.analytics.riskhp.models.dataitems.Hierarchy.HIERARCHY:
            return this.hierarchy;
        case sas.analytics.riskhp.models.dataitems.Horizon.HORIZON:
            return this.horizon;
        case sas.analytics.riskhp.models.dataitems.RelativeHorizon.RELATIVEHORIZON:
            return this.relativeHorizon;
        case sas.analytics.riskhp.models.dataitems.AbsoluteHorizon.ABSOLUTEHORIZON:
            return this.absoluteHorizon;
        case sas.analytics.riskhp.models.dataitems.Outvar.OUTVAR:
            return this.outputVariable;
        case sas.analytics.riskhp.models.dataitems.Compvar.COMPVAR:
            return this.computedVariable;
        case sas.analytics.riskhp.models.dataitems.Benchvar.BENCHVAR:
            return this.benchmarkVariable;
        case sas.analytics.riskhp.models.dataitems.Statistic.STATISTIC:
            return this.statistic;
        case sas.analytics.riskhp.models.dataitems.ResultSet.RESULTSET:
            return this.resultSet;
        case sas.analytics.riskhp.models.dataitems.Scenario.SCENARIO:
            return this.scenarios;
        default:
            return [];
	}
};

/**
 * Given a name and a type returns a dataItem object
 * @param name
 * @param type
 * @returns sas.analytics.riskhp.models.dataitems.BaseDataItem
 */
sas.analytics.riskhp.models.DataSource.prototype.getDataItemByNameAndType = function(name, type) {
    var dataItems = [];
    switch(type) {
        case sas.analytics.riskhp.models.dataitems.Classvar.CLASSVAR:
            dataItems = this._findDataItemByName([name], this.classVariable);
            break;
        case sas.analytics.riskhp.models.dataitems.RiskCube.RISKCUBE:
            dataItems = this._findDataItemByName([name], this.riskCube);
            break;
        case sas.analytics.riskhp.models.dataitems.Horizon.HORIZON:
            if (this.dataSource.type === HprConstants.HPR_CUBE_INFORMATION_TYPE_STRESS) {
                _.each(this.resultSet, function(aResultSet, index) {
                    _.each(aResultSet.scenarios, function(aScenario, index) {
                        _.each(aScenario.horizons, function(aHorizon, index) {
                            if (aHorizon.name === name) {
                                dataItems.push(aHorizon);
                            }
                        });
                    });
                });
            }
            _.each(this.riskCube, function(aRiskCube, index) {
                _.each(aRiskCube.horizons, function(aHorizon, index) {
                    if (aHorizon.name === name) {
                        dataItems.push(aHorizon);
                    }
                });
            });

            break;
        case sas.analytics.riskhp.models.dataitems.RelativeHorizon.RELATIVEHORIZON:
            dataItems = this._findDataItemByName([name], this.relativeHorizon);
            break;
        case sas.analytics.riskhp.models.dataitems.AbsoluteHorizon.ABSOLUTEHORIZON:
            dataItems = this._findDataItemByName([name], this.absoluteHorizon);
            break;
        case sas.analytics.riskhp.models.dataitems.Outvar.OUTVAR:
            dataItems = this._findDataItemByName([name], this.outputVariable);
            break;
        case sas.analytics.riskhp.models.dataitems.Compvar.COMPVAR:
            dataItems = this._findDataItemByName([name], this.computedVariable);
            break;
        case sas.analytics.riskhp.models.dataitems.Benchvar.BENCHVAR:
            dataItems = this._findDataItemByName([name], this.benchmarkVariable);
            break;
        case sas.analytics.riskhp.models.dataitems.Statistic.STATISTIC:
            dataItems = this._findDataItemByName([name], this.statistic);
            break;
        case sas.analytics.riskhp.models.dataitems.Hierarchy.HIERARCHY:
            dataItems = this._findDataItemByName([name], this.hierarchy);
            break;
        default:
            break;
    }

    // Only return the first one found.
    if (dataItems.length > 0) {
        return dataItems[0];
    } else {
        return undefined;
    }
};

/**
 * Looks through the list of 'names' and for each 'name' the method returns the first data item from 'items' that matches the name property.
 * @param names
 * @param items
 * @returns {*}
 * @private
 */
sas.analytics.riskhp.models.DataSource.prototype._findDataItemByName = function(names, items) {
    return _.map(names, function(name) {
        return _.findWhere(items, {name: name});
    });
};

/**
 * Given an id returns a dataItem object
 * @param id
 * @returns sas.analytics.riskhp.models.dataitems.BaseDataItem
 */
sas.analytics.riskhp.models.DataSource.prototype.getDataItemById = function(id) {
    return _.findWhere(this.masterDataItemArray, {id: id});
};

/**
 * Given a list of ids, returns an Array of dataItem objects corresponding to each id
 * @param ids
 * @returns {Array} An Array of sas.analytics.riskhp.models.dataitems.BaseDataItem objects
 */
sas.analytics.riskhp.models.DataSource.prototype.getDataItemsByIds = function(ids) {
    var self = this;
    var dataItems = [];

    _.each(ids, function(id) {
        dataItems.push(self.getDataItemById(id));
    });

    return dataItems;
};

/**
 * Returns a JSONModel object containing the list of data items this data source knows about.  This is for the
 * debug component
 * @returns {JSONModel}
 */
sas.analytics.riskhp.models.DataSource.prototype.getDataItemJSONModel = function() {
    var model = new sap.ui.model.json.JSONModel();
    model.setProperty("/dataItems",this.masterDataItemArray);
    return model;
};

sas.analytics.riskhp.models.DataSource.prototype.getDataItemHTML = function() {
    var returnString = "<table>";
    returnString += "<tr>";
    returnString += "<th>Id</th>";
    returnString += "<th>Type</th>";
    returnString += "<th>Sub-type</th>";
    returnString += "<th>Name</th>";
    returnString += "<th>Label</th>";
    returnString += "<th>Tooltip</th>";
    returnString += "<th>Format</th>";
    returnString += "<th>Data Type</th>";
    returnString += "</tr>";

    var sanitizeValue = function(value) {
        if (value === null || value === undefined) {
            return '';
        }
        return value;
    };
    _.each(this.masterDataItemArray, function(dataItem) {
        returnString += "<tr>";
        returnString += "<td>" + sanitizeValue(dataItem.id) + "</td>";
        returnString += "<td>" + sanitizeValue(dataItem.type) + "</td>";
        returnString += "<td>" + sanitizeValue(dataItem.subType) + "</td>";
        returnString += "<td>" + sanitizeValue(dataItem.name) + "</td>";
        returnString += "<td>" + sanitizeValue(dataItem.label) + "</td>";
        returnString += "<td>" + sanitizeValue(dataItem.tooltip) + "</td>";
        returnString += "<td>" + sanitizeValue(dataItem.format) + "</td>";
        returnString += "<td>" + sanitizeValue(dataItem.dataType) + "</td>";
        returnString += "</tr>";
    });
    return returnString;
};

/**
 * Find a Data Item in this DataSource by narrowing down the data items by 'type' and searching on the 'label'
 * property for the data item.  NOTE: 'label' and 'type' are required.
 *
 * @param label (Required) The label of the Data Item to look for
 * @param type (Required) The type of Data Item to look for (i.e. sas.analytics.riskhp.models.dataitems.Classvar.CLASSVAR, etc.)
 * @returns {undefined | Object} Returns the Data Item itself or undefined if a Data Item can not be located.  undefined
 * is also returned if 'label' or 'type' is missing.
 */
sas.analytics.riskhp.models.DataSource.prototype.getDataItemByLabel = function(label, type) {
    if (!label || !type) {
        return undefined;
    }

    var self = this;
    // Gather all the Data Items by the desired 'type'
    var dataItemsByType = _.propertyOf(self)(type);

    // Using the 'label', find the corresponding Data Item
    if (dataItemsByType) {
        var dataItem = _.findWhere(dataItemsByType, {label: label});
        if (dataItem) {
            return dataItem;
        }
    }
    return undefined;
};

/**
 * Restores the tuple dehydrated data items to full data items.
 * @param dehydratedTuplesArrays
 * @returns {Array}
 */
sas.analytics.riskhp.models.DataSource.prototype.hydrateTuples = function(dehydratedTuplesArrays) {
    var self = this;
    var hydratedTuplesArray = [];
    _.each(dehydratedTuplesArrays, function(dehydratedTuplesArray){
        var hydratedTuples = [];
        _.each(dehydratedTuplesArray, function(dehydratedTuple){
            var dataItem = self.getDataItemById(dehydratedTuple.id);
            if (dataItem !== null) {
                hydratedTuples.push(dataItem);
            }
        });
        hydratedTuplesArray.push(hydratedTuples);
    });
    return hydratedTuplesArray;
};

/**
 * Reduces the tuple data items to simple objects used for persistence.
 * @param hydratedTuplesArrays
 * @returns {Array}
 */
sas.analytics.riskhp.models.DataSource.prototype.dehydrateTuples = function(hydratedTuplesArrays) {
    var dehydratedTuplesArray = [];
    _.each(hydratedTuplesArrays, function(hydratedTuplesArray){
        var dehydratedTuples = [];
        _.each(hydratedTuplesArray, function(hydratedTuple){
            dehydratedTuples.push({id:hydratedTuple.getId(), type: hydratedTuple.getType()});
        });
        dehydratedTuplesArray.push(dehydratedTuples);
    });
    return dehydratedTuplesArray;
};

//*****************************************************************************************************
//*****************************************************************************************************
// Statistic functions
//*****************************************************************************************************
//*****************************************************************************************************

/**
 * Determines if the Data Item is a Row Statistic
 * @param statDataItem
 * @returns {boolean}
 */
sas.analytics.riskhp.models.DataSource.prototype.isRowStatistic=function(statDataItem){
    if (statDataItem && statDataItem.getType() === sas.analytics.riskhp.models.dataitems.Statistic.STATISTIC) {
        if (statDataItem.getStatisticType() === sas.analytics.riskhp.models.dataitems.Statistic.typeRow) {
            return true;
        }
    }
    return false;
};


//*****************************************************************************************************
//*****************************************************************************************************
// Computed Variable functions
//*****************************************************************************************************
//*****************************************************************************************************
sas.analytics.riskhp.models.DataSource.prototype.addComputedVariable = function(data) {
    var self = this;

    var computedvar = new sas.analytics.riskhp.models.dataitems.Compvar(data);

    self.computedVariable.push(computedvar); // Add to '.computedVariable' master list so they will show up in DataPane
    self.masterDataItemArray.push(computedvar);
};

sas.analytics.riskhp.models.DataSource.prototype.updateComputedVariable = function(data) {
    var self = this;

    // crosstab keeps references to the dataItems so can't replace them in the dataSource model, need to update their fields.
    // masterDataItemArray has reference to the same dataItem object so it is updated by reference.
    var computedVar = _.find(self.computedVariable, function(cv) {
        return cv.id === data.id;
    });

    computedVar.label = data.label;
    computedVar.setValid(data.error === 0);
};


/**
 * Removes all the computed variables associated with this DataSource.
 * @returns {boolean} true if at least one computed variable was removed from this object
 */
sas.analytics.riskhp.models.DataSource.prototype.removeAllComputedVariables = function() {
    var self = this;
    var removedAnItem = false;
    _.each(this.computedVariables,function(compVar){
        var itemRemoved = self.removeComputedVariableWithId(compVar.getId());
        if (itemRemoved) {
            removedAnItem = true;
        }
    });
    return removedAnItem;
};

/**
 * Removes the computed variable with the specified id from the list of computed variables this DateSource is tracking.  No-op if the compVarId
 * is not being tracked by this object
 * @param compVarId The computed variable id to remove
 * @returns {boolean} true if a computed variable with the given id was removed.
 */
sas.analytics.riskhp.models.DataSource.prototype.removeComputedVariableWithId = function(compVarId) {
    var rejectedACompVar = false;

    // get a reference to the computed variable we are removing
    var compVarToRemove = this.getDataItemById(compVarId);

    // no-op if the computed variable is not part of this data source
    if (!compVarToRemove) {
        return rejectedACompVar;
    }

    // remove the computed variable from the computedVariables collection
    this.computedVariable = _.reject(this.computedVariable, function(compVar){
        var found = compVarId === compVar.getId();
        if (found) {
            rejectedACompVar = true;
        }
        return found;
    });

    // remove the computed variable from the master data item array
    this.masterDataItemArray = _.reject(this.masterDataItemArray, function(item){
        return compVarId === item.getId();
    });

    return rejectedACompVar;
};

//*****************************************************************************************************
//*****************************************************************************************************
// Scenario functions
//*****************************************************************************************************
//*****************************************************************************************************

/**
 * Adds a scenario to the cube
 * @param scenario
 */
sas.analytics.riskhp.models.DataSource.prototype.addScenario = function(scenario) {
    var self = this;

    // if the scenario is already on the data source, this is a no-op
    var dsScenario = this.getDataItemById(scenario.getId());
    if (dsScenario) {
        //TODO no-op, or replace?
        return;
    }

    scenario.resultSetId = self.defaultResultSet.getId();
    self.defaultResultSet.scenarios.push(scenario);

    self.scenarios.push(scenario); // Add to '.scenarios' master list so they will show up in DataPane
    self.masterDataItemArray.push(scenario);

    _.each(scenario.horizons, function(horizon) {
        horizon.scenarioId = scenario.getId();
        horizon.riskCubeId = null;
        // Do not add horizon to '.horizon' master list because we don't want it to show up in the DataPane
        self.masterDataItemArray.push(horizon);
    });

    // Keep the crosstab happy with newly created relative and absolute horizons for these
    // new scenario horizons.
    self._generateRelativeHorizonDataItems(true); // Apply label changes on new Relative Horizons
    self._generateAbsoluteHorizonDataItems();
};

/**
 * Removes all the scenarios associated with this DataSource.
 * @returns {boolean} true if at least one scenario was removed from this object
 */
sas.analytics.riskhp.models.DataSource.prototype.removeAllScenarios = function() {
    var self = this;
    var removedAnItem = false;
    _.each(this.scenarios,function(scenario){
        var itemRemoved = self.removeScenarioWithId(scenario.getId());
        if (itemRemoved) {
            removedAnItem = true;
        }
    });
    return removedAnItem;
};

/**
 * Removes the scenario with the specified id from the list of scenarios this DateSource is tracking.  No-op if the scenarioId
 * is not being tracked by this object
 * @param scenarioId The scenario id to remove
 * @returns {boolean} true if a scenario with the given id was removed.
 */
sas.analytics.riskhp.models.DataSource.prototype.removeScenarioWithId = function(scenarioId) {
    var rejectedAScenario = false;

    // get a reference to the scenario we are removing
    var scenarioToRemove = this.getDataItemById(scenarioId);

    // no-op if the scenario is not part of this data source
    if (!scenarioToRemove) {
        return rejectedAScenario;
    }

    // remove the scenario from the scenarios collection
    this.scenarios = _.reject(this.scenarios, function(scenario){
        var found = scenarioId === scenario.getId();
        if (found) {
            rejectedAScenario = true;
        }
        return found;
    });

    // remove the scenario from the default result set
    this.defaultResultSet.scenarios = _.reject(this.scenarios, function(scenario) {
        return scenario.getId() === scenarioId;
    });

    // remove the scenario from the master data item array
    this.masterDataItemArray = _.reject(this.masterDataItemArray, function(item){
        return scenarioId === item.getId();
    });

    // clean up the scenario horizons from the master data item array
    this.masterDataItemArray = _.reject(this.masterDataItemArray, function(item) {
        return _.findIndex(scenarioToRemove.horizons, function(horizon) {return horizon.getId() === item.getId();}) !== -1;
    });

    return rejectedAScenario;
};

/**
 * Replaces the appropriate scenario in this data source with the specified scenario. No-op
 * if the data source is not tracking a scenario with the id of the specified scenario.
 *
 * @param updatedScenario the updated version of the scenario to replace in this data source
 */
sas.analytics.riskhp.models.DataSource.prototype.replaceScenario = function(updatedScenario) {
    var self = this;

    // get a reference to the scenario we are replacing
    var scenarioToReplace = this.getDataItemById(updatedScenario.getId());

    // no-op if the scenario does not exist
    if (!scenarioToReplace) {
        return;
    }

    // determine the index of the scenario
    var index = _.findIndex(self.scenarios, function(scenario) {
        return scenario.getId() === updatedScenario.getId();
    });

    // remove the scenario
    self.removeScenarioWithId(updatedScenario.getId());

    // insert the updated scenario at the previous index
    self.scenarios.splice(index, 0, updatedScenario);
    updatedScenario.resultSetId = self.defaultResultSet.getId();
    self.defaultResultSet.scenarios.push(updatedScenario);
    self.masterDataItemArray.push(updatedScenario);

    // add the horizons from the updated scenario
    _.each(updatedScenario.horizons, function(horizon) {
        horizon.scenarioId = updatedScenario.getId();
        horizon.riskCubeId = undefined;

        // Do not add horizon to '.horizon' master list because we don't want it to show up in the DataPane
        self.masterDataItemArray.push(horizon);
    });

    // Keep the crosstab happy with newly created relative and absolute horizons for these
    // new scenario horizons.
    self._generateRelativeHorizonDataItems(true); // Apply label changes on new Relative Horizons
    self._generateAbsoluteHorizonDataItems();
};

/**
 * Given a repoId returns a scenario dataItem object
 * @param id
 * @returns sas.analytics.riskhp.models.dataitems.BaseDataItem
 */
sas.analytics.riskhp.models.DataSource.prototype.getScenarioDataItemByRepoId = function(repoId) {
    return _.findWhere(this.scenarios, {repoId: repoId});
};


//*****************************************************************************************************
//*****************************************************************************************************
// Option functions
//*****************************************************************************************************
//*****************************************************************************************************

/**
 * Get the options for this DataSource
 * @returns {Object} Object representing the options for this DataSource}
 */
sas.analytics.riskhp.models.DataSource.prototype.getOptions = function() {
    // Make a copy so we don't change the original Data Source Options
    return sas.analytics.riskhp.utils.HprUtil.hprClone(this.optionsObject);
};

/**
 * Parses the cube options data and creates an object representing the cube options data
 * which will be used in the UI.
 *
 * @param cubeOptionsData
 * @returns
 * @private
 */
sas.analytics.riskhp.models.DataSource.prototype._parseCubeOptionsData = function(cubeOptionsData) {
    var alpha = {};
    var pctldef = {};
    var ndensity = {};
    var kernel = {};
    var nrfim = {};
    var nrows = {};
    var ndetails = {};
    var mktcache = {};
    var poscache = {};
    var hierarchy = {};
    _.each(cubeOptionsData.dataSource.options, function(data) {
        var value = new sas.analytics.riskhp.models.options.Option(data);
        switch(data.name) {
            case HprConstants.HPR_OPTIONS_PANE_ALPHA_NAME:
                sas.analytics.riskhp.utils.HprUtil._deepCopy(value, alpha);
                break;
            case HprConstants.HPR_OPTIONS_PANE_PCTLDEF_NAME:
                sas.analytics.riskhp.utils.HprUtil._deepCopy(value, pctldef);
                break;
            case HprConstants.HPR_OPTIONS_PANE_NDENSITY_NAME:
                sas.analytics.riskhp.utils.HprUtil._deepCopy(value, ndensity);
                break;
            case HprConstants.HPR_OPTIONS_PANE_KERNEL_NAME:
                sas.analytics.riskhp.utils.HprUtil._deepCopy(value, kernel);
                break;
            case HprConstants.HPR_OPTIONS_PANE_NRFIM_NAME:
                sas.analytics.riskhp.utils.HprUtil._deepCopy(value, nrfim);
                break;
            case HprConstants.HPR_OPTIONS_PANE_NROWS_NAME:
                sas.analytics.riskhp.utils.HprUtil._deepCopy(value, nrows);
                break;
            case HprConstants.HPR_OPTIONS_PANE_NDETAILS_NAME:
                sas.analytics.riskhp.utils.HprUtil._deepCopy(value, ndetails);
                break;
            case HprConstants.HPR_OPTIONS_PANE_MKTCACHE_NAME:
                sas.analytics.riskhp.utils.HprUtil._deepCopy(value, mktcache);
                break;
            case HprConstants.HPR_OPTIONS_PANE_POSCACHE_NAME:
                sas.analytics.riskhp.utils.HprUtil._deepCopy(value, poscache);
                break;
            case HprConstants.HPR_OPTIONS_PANE_HIERARCHY_NAME:
                sas.analytics.riskhp.utils.HprUtil._deepCopy(value, hierarchy);
                break;
            default:
                console.log("Tried to parse the option '" + value.name + "' that does not exist.");
                break;
        }
    });

    return {
        "alpha" : alpha,
        "pctldef" : pctldef,
        "ndensity" : ndensity,
        "kernel" : kernel,
        "nrfim" : nrfim,
        "nrows" : nrows,
        "ndetails" : ndetails,
        "mktcache" : mktcache,
        "poscache" : poscache,
        "hierarchy" : hierarchy
    };
};

//*****************************************************************************************************
//*****************************************************************************************************
// Restriction functions
//*****************************************************************************************************
//*****************************************************************************************************

/**
 * Get the restrictions for this DataSource
 * @returns {Object} Object with a 'restrictions' property whose value is an Array of restrictions for this DataSource
 */
sas.analytics.riskhp.models.DataSource.prototype.getRestrictions = function() {
    return this.restrictions;
};

/**
 * Check the array of 'dataItems' are compatible with each other and therefore truly make a valid column.
 * 'dataItems' is required to be an array of at least 3 objects with the order of Result Set, Scenario,
 * Risk Cube (only required for Side by Side cubes), Horizon, Output Variable, Statistic.  Only one
 * type of each Data Item should exist in the 'dataItems' array.
 *
 * @param dataItems Array An array of Data Item objects which should be checked for compatibility with each other
 * @returns {boolean}
 */
sas.analytics.riskhp.models.DataSource.prototype.areDataItemsCompatible = function(dataItems) {
    // 'tri-state boolean' (yes, that is a contradiction in terms)--true means pass on explicit
    // rule, null means we didn't fail (wildcard), and false means we failed the check
    var compatible = null;

    // Other than Computed and Benchmark Variables, DataItems should have at least 2 items.
    //
    // Valid combinations are:
    // 1) Risk Cube, Statistic
    // 2) Risk Cube, Horizon, Output Variable, Statistic (price)
    // 3) Risk Cube, Result Set, Horizon, Ouput Variable, Statistic (stress)
    // 4) Risk Cube, Result Set, Scenario, Horizon, Output Variable, Statistic (stress)
    if (dataItems) {
        if (dataItems.length >= 2) {
            // Use Underscore to chain together methods which will make sure there is only one dataitem of each type
            var validNumberOfDataItems = _.chain(dataItems)
                .countBy(function (dataItem) {
                    return dataItem.getType();
                }) // Count the dataitems by their type
                .every(function (dataItem) {
                    return dataItem === 1;
                }) // Make sure each dataitem type has only 1
                .value(); // Will return true or false

            if (!validNumberOfDataItems) {
                return false; // Some dataitem type had the wrong count (!1) so do not continue
            }

            // Use the cube restrictions if they exist
            var restrictions = this.getRestrictions();
            if (restrictions) {
                // Group the dataitems array by their type (returns an Object)
                var groupDataItemsByType = _.groupBy(dataItems, function (dataItem) {
                    return dataItem.getType();
                });

                // Group restrictions by their type (returns an Object)
                var groupRestrictionsByType = _.groupBy(restrictions, function (restriction) {
                    return restriction.getType();
                });

                // parse out the data items in the tuple
                var riskCube = (groupDataItemsByType.riskCube && groupDataItemsByType.riskCube.length > 0) ? groupDataItemsByType.riskCube[0] : null;
                var resultSet = (groupDataItemsByType.resultSet && groupDataItemsByType.resultSet.length > 0) ? groupDataItemsByType.resultSet[0] : null;
                var scenario = (groupDataItemsByType.scenario && groupDataItemsByType.scenario.length > 0) ? groupDataItemsByType.scenario[0] : null;
                var horizon = (groupDataItemsByType.horizon && groupDataItemsByType.horizon.length > 0) ? groupDataItemsByType.horizon[0] : null;
                var outputVariable = (groupDataItemsByType.outputVariable && groupDataItemsByType.outputVariable.length > 0) ? groupDataItemsByType.outputVariable[0] : null;
                var statistic = (groupDataItemsByType.statistic && groupDataItemsByType.statistic.length > 0) ? groupDataItemsByType.statistic[0] : null;

                // if we have some combination of risk cube, result set, scenario, and/or horizon, make sure they all reference each other properly
                if (riskCube) {
                    // stress scenario tuple w/ mtmOnly output variable
                    if (resultSet && scenario && horizon && outputVariable) {
                        // S1442821 - Scenario Horizons should not display MTM Only Output Variables
                        if (outputVariable.isMtmOnly()) {
                            return false;
                        }
                    }
                    // stress scenario tuple
                    else if (resultSet && scenario && horizon) {
                        if (resultSet.riskCubeId !== riskCube.id || scenario.resultSetId !== resultSet.id || horizon.scenarioId !== scenario.id) {
                            return false;
                        }
                    }
                    // stress result set mtm tuple
                    else if (resultSet && horizon) {
                        if (resultSet.riskCubeId !== riskCube.id || horizon.resultSetId !== resultSet.id) {
                            return false;
                        }
                    }
                    // price basecase tuple
                    else if (horizon) {
                        if (horizon.riskCubeId !== riskCube.id) {
                            return false;
                        }
                    }
                    // row statistic tuple
                    else if (statistic) {
                        if (this._areRestrictionRulesCompatible(sas.analytics.riskhp.models.dataitems.RiskCube.RISKCUBE, groupDataItemsByType, groupRestrictionsByType) !== false) {
                            return this._areRestrictionRulesCompatible(sas.analytics.riskhp.models.dataitems.Statistic.STATISTIC, groupDataItemsByType, groupRestrictionsByType) !== false;
                        }
                        else {
                            return false;
                        }
                    }
                    else {
                        return false;
                    }
                }

                // If things are still compatible, check the Horizon Restrictions against the Horizon dataitem passed in
                if (horizon) {
                    compatible = this._areRestrictionRulesCompatible(sas.analytics.riskhp.models.dataitems.Horizon.HORIZON, groupDataItemsByType, groupRestrictionsByType);
                }

                // If we passed horizon on a wildcard, check the Output Variable Restrictions against the Output Variable dataitem passed in
                if (compatible === null && outputVariable) {
                    compatible = this._areRestrictionRulesCompatible(sas.analytics.riskhp.models.dataitems.Outvar.OUTVAR, groupDataItemsByType, groupRestrictionsByType);
                }

                // If things are still compatible, check the Statistic Restrictions against the Statistic dataitem passed in
                if (compatible !== false && statistic) {
                    compatible = this._areRestrictionRulesCompatible(sas.analytics.riskhp.models.dataitems.Statistic.STATISTIC, groupDataItemsByType, groupRestrictionsByType);
                }
            }
            else {
                // No Restrictions checks were found (this is an error and should never happen--fail the check).
                compatible = false;
            }
        }
        else if (dataItems.length === 1) {
            // Computed Variables and Benchmark Variables will be represented by a tuple with one dataItem;
            var singleDataItem = dataItems[0];
            if (singleDataItem.getType() === sas.analytics.riskhp.models.dataitems.Benchvar.BENCHVAR ||
                (singleDataItem.getType() === sas.analytics.riskhp.models.dataitems.Compvar.COMPVAR && singleDataItem.isValid())) {
                compatible = true;
            }
            else {
                compatible = false;
            }
        }
    }
    else {
        return false;
    }

    // null means we didn't fail (pass), otherwise we had hit on a non-wildcard rule with a pass or fail
    return compatible === null ? true : compatible;
};

/**
 * Determine if the rules for a certain 'type' of Restriction are compatible. Return
 * values are as follows:
 *
 * null - we did not fail the data item, but we didn't pass on an explicit rule either
 * true - we passed the data item on an explicit rule
 * false - we failed the data item
 *
 * @param type String which represents type of Data Item or Restriction to use
 * @param groupDataItemsByType Object of Data Items grouped by type
 * @param groupRestrictionsByType Object of Restrictions grouped by type
 * @returns {boolean} null | true | false
 * @private
 */
sas.analytics.riskhp.models.DataSource.prototype._areRestrictionRulesCompatible = function(type, groupDataItemsByType, groupRestrictionsByType) {
    // If there are Restrictions of the type passed in, then we need to process their Restriction Rules
    if (groupRestrictionsByType[type] && groupRestrictionsByType[type].length > 0) {
        if (groupDataItemsByType[type] && groupDataItemsByType[type].length > 0 ) {
            // Gather the dataitem information that we need
            var dataItemId = groupDataItemsByType[type][0].getId();
            var dataItemType = groupDataItemsByType[type][0].getType();

            // Loop through the restrictions based on the type passed in until we find a match
            for (var i=0; i < groupRestrictionsByType[type].length; i++) {
                var restrictionId = groupRestrictionsByType[type][i].getId();
                var restrictionType = groupRestrictionsByType[type][i].getType();

                // If the id & type of the dataitem match the id & type of the restriction... use it!
                // Or, if the restriction is a wildcard ("") and the type of the dataitem matches the
                // type of the restriction... use it!
                if (dataItemType === restrictionType && (restrictionId === "" || restrictionId === dataItemId)) {
                    if (groupRestrictionsByType[type][i].processRulesForRestriction(groupDataItemsByType) === true) {
                        return restrictionId === "" ? null : true;
                    }
                    else {
                        return false;
                    }
                }
            }
        }
    }

    // if no restrictions for a given type or item, then assume it is good
    return true;
};

/**
 * Parses the cube restrictions data and creates an object representing the cube restrictions data
 * which will be used in the UI.
 *
 * @param cubeRestrictionsData
 * @returns Object Returns an Object with a 'restrictions' property whose value is an Array of 'sas.analytics.riskhp.models.restrictions.Restriction' objects
 * @private
 */
sas.analytics.riskhp.models.DataSource.prototype._parseCubeRestrictionsData = function(cubeRestrictionsData) {
    var restrictionArray = [];
    var restrictions = cubeRestrictionsData.restrictions;
    _.each(restrictions, function(data) {
        var value = new sas.analytics.riskhp.models.restrictions.Restriction(data);
        restrictionArray.push(value);
    });

    return {
        "restrictions": restrictionArray
    };
};

/**
 * Finds the valid columns that can be displayed in the Crosstab based on the Restrictions and Rules for the Data Item.
 * Any number of valid columns can be returned with -1 returning all valid columns generated.  Any column using the
 * BaseCase Horizon can also be excluded by setting the 'noBaseCase' parameter to true.  If the 'filterValidCols' parameter
 * is true, the number of valid columns returned is filtered down by the Risk Cubes and Output Variables currently
 * displayed in the Crosstab.
 *
 * @param dataItem Data Item to find valid columns for
 * @param numCols Number of valid columns to return (-1 will return all valid columns; 0 will return 1 valid column)
 * @param filterValidCols Performs extra steps to filter the number of valid columns returned by Risk Cubes and Output
 * Variables currently displayed in the Crosstab
 * @param noBaseCase Flag to signify not to use BaseCase Horizon (true means don't use BaseCase Horizon).  Option is ignored
 * if the dataItem given is the BaseCase Horizon.
 * @param crosstabContext Context of the Crosstab to find valid columns for
 * @return [Array] Returns an Array of valid columns; Each valid column is an Array of Data Items
 */
sas.analytics.riskhp.models.DataSource.prototype.findValidColumnsForDataItem = function(dataItem, numCols, filterValidCols, noBaseCase, crosstabContext) {
    var self = this;
    var RELATIVEHORIZON = sas.analytics.riskhp.models.dataitems.RelativeHorizon.RELATIVEHORIZON;
    var RISKCUBE = sas.analytics.riskhp.models.dataitems.RiskCube.RISKCUBE;
    var OUTVAR = sas.analytics.riskhp.models.dataitems.Outvar.OUTVAR;
    var validColumns = [];

    // Check to see if Data Item
    if (!dataItem || (dataItem && (!_.isFunction(dataItem.getId) || !_.isFunction(dataItem.getType)))) return validColumns;

    // Check to see if we have a Crosstab context
    if (!crosstabContext) return validColumns;

    // Want to always return at least one column
    if (numCols === 0) numCols = 1;

    // Ignore the noBaseCase parameter if the Data Item given is a BaseCase Horizon
    if (this.isDataItemBaseCaseHorizon(dataItem)) {
        noBaseCase = undefined;
    }

    // Convert RelativeHorizon Data Item to Base Horizon
    if (dataItem.getType() === RELATIVEHORIZON) {
        var searchKey = {horizonNumber: dataItem.horizonNumber, interval: dataItem.interval};

        // Try and narrow down the searchKey by using the first Risk Cube id that is present in the Crosstab
        var riskCubesCurrentlyInCrosstab = _.filter(crosstabContext.getTuplesAsOneArray(), function(dataItem) { return dataItem.getType() === RISKCUBE; });
        if (riskCubesCurrentlyInCrosstab.length > 0) {
            var searchKeyWithRiskCube = {};
            searchKeyWithRiskCube = Object.assign({}, searchKey);
            searchKeyWithRiskCube.riskCubeId = riskCubesCurrentlyInCrosstab[0].getId(); // Not sure what else we can do here other than take the first one
            dataItem = _.findWhere(this.horizon, searchKeyWithRiskCube);

            // S1432569
            // Well, the horizon may not be associated with the risk cube already displayed.
            // So here were will do a less restrictive search for the horizon and use that if found.
            if (dataItem === undefined) {
                dataItem = _.findWhere(this.horizon, searchKey);
            }
        }
        else {
            dataItem = _.findWhere(this.horizon, searchKey);
        }
    }

    if (dataItem) {
        // Use the cube restrictions if they exist
        var restrictions = this.getRestrictions();
        if (restrictions) {
            // Group restrictions by their type (returns an Object)
            var groupRestrictionsByType = _.groupBy(restrictions, function (restriction) {
                return restriction.getType();
            });

            // Only add the columns if there are some
            var potentialValidColumns = this._findRestrictionToUseToFindValidColumns(dataItem, groupRestrictionsByType, noBaseCase);
            if (potentialValidColumns && potentialValidColumns.length > 0) {
                validColumns = validColumns.concat(potentialValidColumns);
            }
        }

        // Determine the number of valid columns to return based on the 'numCols' argument and number of valid columns
        var columnsToReturnLength = (numCols === -1) ? validColumns.length : numCols;

        // S1422854 & S1425758: If we should filter the valid columns to return, do some magic and try to find the best columns to return; In some
        // cases, more than one column will be returned in order to try and match the number of Output Variables that are displayed in the Crosstab
        if (filterValidCols) {
            // First, we are going to try and filter the generated valid columns by Risk Cube(s); So, see if the dataItem is a Risk Cube first and then try the 'riskCubeId' property
            var dataItemRiskCubeId = (dataItem && dataItem.getType() === RISKCUBE) ? [dataItem.getId()] : (dataItem && dataItem.riskCubeId) ? [dataItem.riskCubeId] : []; // Make it an Array

            // Next, create an Array of Risk Cubes currently displayed in the Crosstab; Otherwise just use an empty array which will not hurt anything
            var dataItemsByTypeCurrentlyInCrosstab = _.groupBy(crosstabContext.getTuplesAsOneArray(), "type");
            var riskCubeIdsInCrosstab = _.pluck(dataItemsByTypeCurrentlyInCrosstab[RISKCUBE], "id");

            // Capture the id of the first Risk Cube in this DataSource in case it is not currently shown in the Crosstab
            var firstRiskCubeInDataSource = [self.riskCube[0].id];
            // Make an Array of unique Risk Cube Id's, should be no duplicates
            var combinedRiskCubeIds = _.union(dataItemRiskCubeId, riskCubeIdsInCrosstab, firstRiskCubeInDataSource);
            var filterColumnsByRiskCubeId = [];

            // Perform a while loop until we find valid columns filtered by the desired Risk Cube Id or we run out of Risk Cube id's to filter on
            while (filterColumnsByRiskCubeId.length === 0 && combinedRiskCubeIds.length > 0) {
                var riskCubeIdToUse = combinedRiskCubeIds.shift(); // Grab the first id and shorten the original 'combinedRiskCubeIds' array
                filterColumnsByRiskCubeId = _.filter(validColumns, function (validColumn) {
                    return _.some(validColumn, {id: riskCubeIdToUse}); // Returns true if any of the values pass
                });
            }

            // Now, see if we can filter the valid columns even more by using Output Variables
            var combinedFilteredRiskCubeOutputVariableArray = [];
            var outputVariableIdsInCrosstab = _.pluck(dataItemsByTypeCurrentlyInCrosstab[OUTVAR], "id");
            var outVarDataItemsLength = outputVariableIdsInCrosstab.length;
            if (outVarDataItemsLength > 0) {
                // If we have Output Variables to work with, continue to filter down the valid columns we are already working with (i.e. 'filterColumnsByRiskCubeId') by looking
                // for columns with the Output Variables in the Crosstab
                var filterColumnsByOutputVariablesInCrosstab = _.filter(filterColumnsByRiskCubeId, function (column) {
                    return _.some(column, function (dataItem) {
                        // Returns true if the Output Variable is in the column or the Data Item is the PL Output Variable;
                        // PL is special because we are going to substitute it for VALUE in some cases
                        return outputVariableIdsInCrosstab.indexOf(dataItem.getId()) > -1 || dataItem.getName() === sas.analytics.riskhp.models.dataitems.Outvar.PL;
                    });
                });

                // Group the columns by the id of the Output Variables
                var groupColumnsByOutputVariableIds = _.groupBy(filterColumnsByOutputVariablesInCrosstab, function (columns) {
                    return _.find(columns, function (column) {
                        return column.getType() === OUTVAR;
                    }).id;
                });

                // Since we are going to try and substitute the PL column for the VALUE column, we need to keep track of lots of things
                var valueOutputVariable = self.getDataItemByNameAndType(sas.analytics.riskhp.models.dataitems.Outvar.VALUE, OUTVAR);
                var valueOutputVariableId = valueOutputVariable ? valueOutputVariable.getId() : undefined;
                var valueOutputVariableColumnIndex = -1; // Index of column where the VALUE Output Variable is located
                var valueOutputVariableColumnArray = [];
                var plOutputVariable = self.getDataItemByNameAndType(sas.analytics.riskhp.models.dataitems.Outvar.PL, OUTVAR);
                var plOutputVariableId = plOutputVariable ? plOutputVariable.getId() : undefined;
                var plOutputVariableColumnArray = [];

                // Helper function to determine the column index which contains the Output Variable id
                var outVarColumnIndex = function (oCrosstabContext, ovId) {
                    return _.findIndex(oCrosstabContext.getTuples(), function (tuple) {
                        return _.findWhere(tuple, {id: ovId});
                    });
                };

                // Create an Object which stores data about each Output Variable we are working with; The data we are most
                // interested in is the index (i.e. column index) and column array (includes all data items which make up the column)
                var outputVariableColumnObject = {};
                $.each(groupColumnsByOutputVariableIds, function (ovId, ovCols) {
                    var ovSubObject = {};
                    ovSubObject.id = ovId;
                    ovSubObject.colIndex = outVarColumnIndex(crosstabContext, ovId);

                    // There can be multiple columns which contain the Output Variable but we only need the first one, so grab it
                    ovSubObject.colArray = [].concat(_.first(ovCols));

                    // If we come across PL or VALUE, store some things off for those for later use since we will be substituting those columns
                    if (ovId === plOutputVariableId) {
                        plOutputVariableColumnArray = [].concat(_.first(ovCols));
                    } else if (ovId === valueOutputVariableId) {
                        valueOutputVariableColumnIndex = ovSubObject.colIndex; // Keep track of the column where VALUE OutVar is located
                        valueOutputVariableColumnArray = [].concat(_.first(ovCols));
                    }

                    // Store the newly created sub object
                    outputVariableColumnObject[ovId] = ovSubObject;
                });

                // Sort the Output Variable column object by the 'colIndex' property which gives us the correct order the columns should go in
                var outputVariableColumnObjectSortedByIndex = _.sortBy(outputVariableColumnObject, 'colIndex');
                $.each(outputVariableColumnObjectSortedByIndex, function (index, obj) {
                    // Loop over the sorted column object and add all Output Variables to an Array (except PL and VALUE) at the correct index
                    if (obj.id !== valueOutputVariableId && obj.id !== plOutputVariableId) {
                        combinedFilteredRiskCubeOutputVariableArray.splice(obj.colIndex, 0, obj.colArray);
                    }
                });

                // Determine if we can substitute the PL column where the VALUE column should go and do it
                if (plOutputVariableColumnArray.length > 0) {
                    if (valueOutputVariableColumnIndex > -1) {
                        combinedFilteredRiskCubeOutputVariableArray.splice(valueOutputVariableColumnIndex, 0, plOutputVariableColumnArray);
                    }
                }

                // Always push VALUE column as last object in case we need to display it
                if (valueOutputVariableColumnArray.length > 0) {
                    combinedFilteredRiskCubeOutputVariableArray.push(valueOutputVariableColumnArray);
                }
            }

            // Create an Array which will be used to return the filtered valid columns
            var finalFilteredColumns = [];

            // Here we determine if we should return more than one valid column to the calling method based on the number of Output Variables displayed in the Crosstab
            var finalFilteredColumnsLength = outVarDataItemsLength > 0 ? outVarDataItemsLength : columnsToReturnLength;

            // If the columns were filtered by Risk Cubes and Output Variables, use them; Otherwise, just use the ones filtered by Risk Cubes
            if (combinedFilteredRiskCubeOutputVariableArray.length > 0) {
                finalFilteredColumns = [].concat(combinedFilteredRiskCubeOutputVariableArray);
            } else {
                finalFilteredColumns = [].concat(filterColumnsByRiskCubeId);
            }

            return finalFilteredColumns.slice(0, finalFilteredColumnsLength);
        }

        return validColumns.slice(0, columnsToReturnLength);
    }
    return validColumns;
};

/**
 * Returns all valid columns for the 'dataItem' passed in based on its type.  Potential valid columns for the 'dataItem' are
 * constructed based on known properties and Restrictions and then filled out by searching through the Restriction Rules for
 * valid data items that fit those rules.
 *
 * @param dataItem Data Item to find valid columns for
 * @param groupRestrictionsByType The Restrictions for this DataSource grouped by the 'type' property
 * @param noBaseCase Flag to signify not to use BaseCase Horizon (true means don't use BaseCase Horizon).  Option is ignored
 * if the dataItem given is the BaseCase Horizon.
 * @returns {Array} Returns an Array of valid columns
 * @private
 */
sas.analytics.riskhp.models.DataSource.prototype._findRestrictionToUseToFindValidColumns = function(dataItem, groupRestrictionsByType, noBaseCase) {
    var self = this;
    var validColumns = [];

    // Gather the dataitem information that we need
    var dataItemType = dataItem.getType();

    // Create some variables for the types of Data Items we might need to use and other things
    var RISKCUBE = sas.analytics.riskhp.models.dataitems.RiskCube.RISKCUBE;
    var RESULTSET = sas.analytics.riskhp.models.dataitems.ResultSet.RESULTSET;
    var SCENARIO = sas.analytics.riskhp.models.dataitems.Scenario.SCENARIO;
    var HORIZON = sas.analytics.riskhp.models.dataitems.Horizon.HORIZON;
    var OUTVAR = sas.analytics.riskhp.models.dataitems.Outvar.OUTVAR;
    var STATISTIC = sas.analytics.riskhp.models.dataitems.Statistic.STATISTIC;
    var COMPVAR = sas.analytics.riskhp.models.dataitems.Compvar.COMPVAR;
    var BENCHVAR = sas.analytics.riskhp.models.dataitems.Benchvar.BENCHVAR;
    var DATASOURCEUTIL = sas.analytics.riskhp.models.DataSourceUtil;

    var potentialColumns = []; // Array of potential valid columns
    var potentialColumnTemplate = []; // Array of Data Items representing a valid column to be placed inside potentialColumns

    // Depending on the type of the Data Item, construct some potential valid columns.  For this method, potential valid columns do not include the Output Variables
    // and Statistics.  Those data items are added via the Restriction.useRestrictionToFindValidColumns() method.  So a couple of examples of a potential valid column
    // produced by this method are: [Risk Cube], [Risk Cube, Horizon], [Risk Cube, Result Set, Horizon], etc.
    //
    // For reference, when the "useRestrictionToFindValidColumns()" method finishes, full valid columns consist of:
    // 1) Risk Cube, Statistic
    // 2) Risk Cube, Horizon, Output Variable, Statistic (price)
    // 3) Risk Cube, Result Set, Horizon, Output Variable, Statistic (stress)
    // 4) Risk Cube, Result Set, Scenario, Horizon, Output Variable, Statistic (stress)
    switch (dataItemType) {
        // The following commented out Data Items don't need special processing
        // case COMPVAR:
        // case BENCHVAR:
        case RISKCUBE:
            if (!self.isStressCube()) {
                // For Price cubes only; Not sure this will get called for a Stress cube since a Risk Cube doesn't show up in Data Pane for them;
                // Start the potential column by adding the Risk Cube;
                potentialColumnTemplate.push(dataItem);
                potentialColumns.push($.extend(true, [], potentialColumnTemplate)); // Risk Cube by itself is a valid potential column (i.e. Row Statistics are added later)

                // Try to add valid columns for the Horizons; Weed out BaseCase Horizons; Don't use BaseCase Horizons if 'noBaseCase' is set to true
                // (value of 'noBaseCase' can be 'undefined' if the dataItem is BaseCase Horizon)
                $.each(dataItem.horizons, function(index, horizon) {
                    if (noBaseCase && self.isDataItemBaseCaseHorizon(horizon)) {
                        return true; // Skip this Horizon
                    } else {
                        potentialColumnTemplate.push(horizon);
                        potentialColumns.push($.extend(true, [], potentialColumnTemplate));
                        potentialColumnTemplate.pop(); // Need to pop off Horizon, so next Horizon can be added on potential column
                    }
                });
            }
            break;
        case RESULTSET:
            // Trying to add Result Set, first item in potential column is Risk Cube the Result Set belongs too
            potentialColumnTemplate.push(self.getDataItemById(dataItem.riskCubeId));
            potentialColumnTemplate.push(dataItem);

            // See if the Result Set Horizon (which is singular) can be added as a valid column
            if (dataItem.horizon) {
                // Weed out BaseCase Horizons; Don't use BaseCase Horizons if 'noBaseCase' is set to true (value of 'noBaseCase' can be 'undefined' if the dataItem is BaseCase Horizon)
                if (noBaseCase && self.isDataItemBaseCaseHorizon(dataItem.horizon)) {
                    // If this is the DefaultResultSet, always add the Horizon regardless of noBaseCase parameter
                    if (dataItem.isDefault()) {
                        potentialColumnTemplate.push(dataItem.horizon);
                        potentialColumns.push($.extend(true, [], potentialColumnTemplate));
                    }
                } else {
                    potentialColumnTemplate.push(dataItem.horizon);
                    potentialColumns.push($.extend(true, [], potentialColumnTemplate));
                    potentialColumnTemplate.pop(); // Need to pop off Horizon, so next Horizon can be added on potential column
                }
            }

            // Process the Scenarios and their Horizons to try and build potential columns
            if (dataItem.scenarios && dataItem.scenarios.length > 0) {
                // Loop through the Scenarios
                _.each(dataItem.scenarios, function(scenario) {
                    // Add Scenario on potential column
                    potentialColumnTemplate.push(scenario);

                    // Loop through the Horizons of the Scenario now; Weed out BaseCase Horizons; Don't use BaseCase Horizons if 'noBaseCase' is set
                    // to true (value of 'noBaseCase' can be 'undefined' if the dataItem is BaseCase Horizon)
                    $.each(scenario.horizons, function(index, horizon) {
                        if (noBaseCase && self.isDataItemBaseCaseHorizon(horizon)) {
                            return true; // Skip this Horizon
                        } else {
                            potentialColumnTemplate.push(horizon);
                            potentialColumns.push($.extend(true, [], potentialColumnTemplate));
                            potentialColumnTemplate.pop(); // Pop Horizon so we can add next one
                        }
                    });

                    potentialColumnTemplate.pop(); // Pop Scenario so we can add next one on potential column
                });
            }
            break;
        case SCENARIO:
            // Trying to add Scenario, first item in potential column is Risk Cube, use the resultSetId to find it and add all 3 to potential column
            var resultSetDataItem = self.getDataItemById(dataItem.resultSetId);
            potentialColumnTemplate.push(self.getDataItemById(resultSetDataItem.riskCubeId));
            potentialColumnTemplate.push(resultSetDataItem);
            potentialColumnTemplate.push(dataItem);

            // Loop through the Horizons of the Scenario now; Weed out BaseCase Horizons; Don't use BaseCase Horizons if 'noBaseCase' is set
            // to true (value of 'noBaseCase' can be 'undefined' if the dataItem is BaseCase Horizon)
            $.each(dataItem.horizons, function(index, horizon) {
                if (noBaseCase && self.isDataItemBaseCaseHorizon(horizon)) {
                    return true; // Skip this Horizon
                } else {
                    potentialColumnTemplate.push(horizon);
                    potentialColumns.push($.extend(true, [], potentialColumnTemplate));
                    potentialColumnTemplate.pop(); // Pop Horizon so we can add next one on potential column
                }
            });
            break;
        case HORIZON:
            // Trying to add a Horizon, one of these values should be set: resultSetId; riskCubeId; scenarioId.  Locate other data items by using the appropriate
            // properties to build a potential column
            if (dataItem.riskCubeId) {
                potentialColumnTemplate.push(self.getDataItemById(dataItem.riskCubeId));
            } else if (dataItem.resultSetId) {
                var resultSetDataItem = self.getDataItemById(dataItem.resultSetId);
                potentialColumnTemplate.push(self.getDataItemById(resultSetDataItem.riskCubeId));
                potentialColumnTemplate.push(resultSetDataItem);
            } else if (dataItem.scenarioId) {
                var scenarioDataItem = self.getDataItemById(dataItem.scenarioId);
                var resultSetDataItem = self.getDataItemById(scenarioDataItem.resultSetId);
                potentialColumnTemplate.push(self.getDataItemById(resultSetDataItem.riskCubeId));
                potentialColumnTemplate.push(resultSetDataItem);
                potentialColumnTemplate.push(scenarioDataItem);
            }

            // Add Horizon now, but weed out BaseCase Horizons; Don't use BaseCase Horizons if 'noBaseCase' is set to true (value of 'noBaseCase' can be 'undefined' if the dataItem is BaseCase Horizon)
            if (noBaseCase && self.isDataItemBaseCaseHorizon(dataItem)) {
                // No-op, skip this Horizon and don't add a potential column
            } else {
                potentialColumnTemplate.push(dataItem);
                potentialColumns.push($.extend(true, [], potentialColumnTemplate));
            }
            break;
        case OUTVAR:
            // Build potential columns for the Output Variable, the Statistics for the potential column will be added via the "useRestrictionToFindValidColumns" method call below
            potentialColumns = [].concat(DATASOURCEUTIL.generatePotentialColumnsForOutputVariable(self, dataItem, groupRestrictionsByType, noBaseCase));
            break;
        case STATISTIC:
            // We have all the information we need to build a valid column, so no further processing necessary for Statistics; Just set validColumns
            validColumns = [].concat(DATASOURCEUTIL.generateValidColumnsForStatistic(self, dataItem, groupRestrictionsByType, noBaseCase));
            break;
        default:
            break;
    }

    if (dataItemType === COMPVAR || dataItemType === BENCHVAR) {
        // Special type of column with no restrictions in DataSource (i.e. Computed Variable, Benchmark Variables); Just add them to validColumns as-is
        validColumns.push([dataItem]);
    } else {
        // NOTE: Big assumptions here based on format of current Restrictions and Restriction Rules; Skip this section if we are trying to find valid columns
        // for Statistics (we have all the information we need for those data items)
        if (potentialColumns.length > 0 && dataItemType !== STATISTIC) {
            var riskCubeId = potentialColumns[0][0].getId(); // All potential columns generated above should have the same Risk Cube as first item

            // We use the Risk Cube Restrictions here because they have rules for Row Statistics; Also, the "useRestrictionToFindValidColumns"
            // method lives in the Restriction.js class and we needed a way to access it
            if (groupRestrictionsByType[RISKCUBE] && groupRestrictionsByType[RISKCUBE].length > 0) {
                // Loop through the restrictions based on the Risk Cube type until we find a match
                for (var i = 0; i < groupRestrictionsByType[RISKCUBE].length; i++) {
                    var restrictionId = groupRestrictionsByType[RISKCUBE][i].getId();
                    var restrictionType = groupRestrictionsByType[RISKCUBE][i].getType();
                    // If the id & type of the Risk Cube match the id & type of the restriction... use it!
                    // Or, if the restriction is a wildcard ("") and the type of the Risk Cube matches the type of the restriction... use it!
                    if ((riskCubeId === restrictionId && RISKCUBE === restrictionType) || (restrictionId === "" && RISKCUBE === restrictionType)) {
                        // Found a match, use that Restriction to find the valid columns
                        validColumns = groupRestrictionsByType[RISKCUBE][i].useRestrictionToFindValidColumns(self, dataItem, potentialColumns, groupRestrictionsByType);
                        break;
                    }
                }
            }
        }
    }

    return validColumns;
};