/*
 * (c) 2017, SAS Institute Inc.
 */

jQuery.sap.declare("sas.analytics.riskhp.models.DataSourceCacheManager");

jQuery.sap.require("sas.analytics.riskhp.models.DataSourceParams");
jQuery.sap.require("sas.analytics.riskhp.utils.HprUtil");
jQuery.sap.require("sas.analytics.riskhp.utils.HprAPI");
jQuery.sap.require("sas.analytics.riskhp.utils.HprConstants");
jQuery.sap.require("sas.analytics.riskhp.utils.CubeKeepAlive");

/**
 * @name sas.analytics.riskhp.models.DataSourceCacheManager
 */
sas.analytics.riskhp.models.DataSourceCacheManager = (function () {
    "use strict";
    var DATATYPE_OPTIONS = 'options';
    var DATATYPE_RESTRICTIONS = 'restrictions';
    var DATATYPE_DATASOURCE = 'dataSource';

    var listenerMap = {};
    var dataSourceMap = {};
    var dataSourceParamsMap = {};
    var selectedDataSourceId = null;

    function _updateDataSource(newDataSource){
        // TODO: Merge object with currently held object or replace based on current state
        var currentDataSource = newDataSource;
        return currentDataSource;
    }

    /**
     * The main code to return the data source from the cache or start the process of loading it.  This method ALWAYS
     * returns a Deferred.  If the requested data is loaded in the cache the returned Deferred will be a resolved Deferred
     * with the requested data.  Also, note, right now the cached data is cached as a JSON model, but (again for right now)
     * we return the underlying data object.
     *
     * @param dataType The type of data to return
     * @param dataSourceId The id of the data source to load or return
     * @returns {*} A deferred object
     * @private
     */
    function _getData(dataType, dataSourceId) {
        // Create the uber Deferred that this function will ALWAYS return
        var deferred = $.Deferred();

        // If there is no data source id then we fail
        if (!dataSourceId || (dataSourceId && _.isEmpty(dataSourceId))) {
            deferred.reject();
            return deferred;
        }

        // Retrieve the current object in the cache, it will be either a Deferred (presumably in flight), the real
        // loaded DataSource object wrapped in a JSONModel, or null
        var deferredOrDataSource = dataSourceMap[dataSourceId];

        // If something was retrieved from the cache move into this block to determine what to do next
        if (deferredOrDataSource) {

            // If the object we retrieved from the cache has a .done function on it then we know it is a Deferred
            if (_.isFunction(deferredOrDataSource.done)){

                // Assign the retrieved object to a variable that makes the code read better
                var cachedDeferred = deferredOrDataSource;

                // Check which type of data has been asked for (options, restrictions or DataSource), construct a
                // function to pull out the right object and attach it to the .done handler of the retrieved Deferred. If
                // the type of data being asked for is options or restrictions we need to embed a new Deferred in the
                // done handler function so the caller will get the proper data
                if (dataType === DATATYPE_OPTIONS) {
                    cachedDeferred.done( function() {

                        // Get the DataSource from the cache, at this point we know it has been loaded
                        var dataSource = dataSourceMap[dataSourceId];

                        // get the options off the underlying DataSource object
                        deferred.resolve(dataSource.oData.getOptions());
                    });

                    // Make sure we propagate failures up
                    cachedDeferred.fail( function(jqxhr, textStatus, error) {
                        deferred.reject(jqxhr, textStatus, error);
                    });
                }
                else if (dataType === DATATYPE_RESTRICTIONS) {
                    cachedDeferred.done( function() {

                        // Get the DataSource from the cache, at this point we know it has been loaded
                        var dataSource = dataSourceMap[dataSourceId];

                        // get the restrictions off the underlying DataSource object
                        deferred.resolve(dataSource.oData.getRestrictions());
                    });

                    // Make sure we propagate failures up
                    cachedDeferred.fail( function(jqxhr, textStatus, error) {
                        deferred.reject(jqxhr, textStatus, error);
                    });
                }
                else {
                    // Since the default behavior of the cached Deferred is to notify attached .done handlers when the
                    // DataSource object has been loaded into the cache we can just return the cached
                    // Deferred directly and there is no need embed our Deferred
                    deferred = cachedDeferred;
                }
            }
            else {
                // Case where we got the actual DataSource object wrapped as a JSONModel.  In this case
                // we resolve the Deferred we are going to pass to the caller with the correct object they asked for.

                // Assign the retrieved object to a variable that makes the code read better
                var dataSource = deferredOrDataSource;

                // Resolve the Deferred object we will return with the data requested
                if (dataType === DATATYPE_OPTIONS) {
                    deferred.resolve(dataSource.oData.getOptions());
                }
                else if (dataType === DATATYPE_RESTRICTIONS) {
                    deferred.resolve(dataSource.oData.getRestrictions());
                }
                else {
                    deferred.resolve(dataSource.oData);
                }
            }
        }
        else {
            // case were nothing was in the cache, which means the DataSource has never been loaded or
            // attempted to be loaded

            // First put the deferred into the cache, so subsequent callers will get the deferred
            dataSourceMap[dataSourceId] = deferred;

            // Grab the associated DataSourceParams object
            var dataSourceParams = dataSourceParamsMap[dataSourceId];

            // Call the method to open the data source and capture it's deferred object
            var promise = sas.analytics.riskhp.utils.HprAPI.openDataSource(dataSourceParams);

            // Attach a .done handler to be notified when the DataSource has been loaded
            promise.done(function(data) {

                //console.log(JSON.stringify(data));
                // Run the retrieved data through our updateDataSource method to apply changes and wrap the result in
                // a JSONModel.  Not sure _updateDataSource applies here, more investigation required
                var cubeDataModel = new sap.ui.model.json.JSONModel(_updateDataSource(data));

                // Finally replace the Deferred in the cache with the DataSource we just loaded and
                // wrapped in a JSONModel
                dataSourceMap[dataSourceId] = cubeDataModel;
                _updateCubeKeepAlive();

                // Since the Data Source is now open, apply the Data Item that were stored back to the cube data
                dataSourceParams.applyDataItemModifications(data);

                // Resolve the Deferred we will return passing the DataSource object
                // TODO: DSCM should we return it here?
                deferred.resolve(cubeDataModel.oData);
            });

            // Because the intent is to hide the captured promise from the openDataSource method call we must attach a
            // failure handler as well so the fail case of the promise gets propagated to the caller.
            promise.fail(function (jqxhr, textStatus, error) {
                // Since the load of the data source failed, we must remove it from the relevant maps since it is not valid
                delete dataSourceMap[dataSourceId];
                delete dataSourceParamsMap[dataSourceId];

                deferred.reject(jqxhr, textStatus, error);
            });
        }
        return deferred;
    }

    /**
     * Update the cube keep alive with updated cube information.
     */
    function _updateCubeKeepAlive() {
        var self = this;
        var cubeKeepAliveInfoArray = [];
        var loadedDataSourceIds = DataSourceCacheManager.getLoadedDataSourceIds();
        _.each(loadedDataSourceIds, function (dataSourceId) {
            // If this is a deferred, we're still waiting for the midtier to
            // return the cube data, so we don't want to attempt to add it to
            // the keepalive list yet.
            if (_.isFunction(dataSourceMap[dataSourceId].done)) {
                return;
            }

            // otherwise, add the data source to the keepalive list
            var ds = dataSourceMap[dataSourceId].oData.dataSource;
            cubeKeepAliveInfoArray.push({
                id: ds.id,
                name: ds.label,
                application: ds.application,
                workGroup: ds.workGroup,
                relPath: ds.relPath
                //lastModifiedTimeStamp: ds.lastModifiedTimeStamp
            })
        });
        CubeKeepAlive.update(cubeKeepAliveInfoArray);
    }


    return {

        /**
         * Resets the cache manager to it's default state.  This should be called each time the exploration view is opened
         */
        reset: function() {
            listenerMap = {};
            dataSourceMap = {};
            _updateCubeKeepAlive();
            dataSourceParamsMap = {};
            selectedDataSourceId = null;
        },

        /**
         * Returns the id of the currently selected data source
         * @returns {string}
         */
        getSelectedDataSourceId: function() {
            return selectedDataSourceId;
        },

        /**
         * Set the id of the currently selected data source. This is intended to be called from the data pane when adding
         * a data source via the dropdown box, or probably when changing the data source
         * @param id
         */
        setSelectedDataSourceId: function(id) {
            selectedDataSourceId = id;
        },

        /**
         * Returns the DataSourceParam object representing the currently selected data source
         * @returns {sas.analytics.riskhp.models.DataSourceParams}
         */
        getSelectedDataSourceParam: function() {
            return dataSourceParamsMap[selectedDataSourceId];
        },

        /**
         * Convenience method to return the data source which is currently selected
         * @returns {DataSource} Will return null if the data source for the currentlySelectedDataSourceId is not loaded
         */
        getSelectedDataSource: function() {
            return this.getLoadedDataSource(this.getSelectedDataSourceId());
        },

        /**
         * Adds a new DataSourceParam to the data source params list this cache manager knows about.  For a data source to
         * be loaded, it's corresponding DataSourceParam object must have already been added.
         * @param dataSourceParam DataSourceParam
         */
        addDataSourceParam: function(dataSourceParam) {
            dataSourceParamsMap[dataSourceParam.realDataSourceId] = dataSourceParam;
        },

        /**
         * Given a data source id returns the associated DataSourceParam object
         * @param dataSourceId
         * @returns {sas.analytics.riskhp.models.DataSourceParams} Will return null if the corresponding DataSourceParam object has not been added to the cache
         */
        getDataSourceParam: function(dataSourceId) {
            return dataSourceParamsMap[dataSourceId];
        },

        /**
         * Returns the list of DataSourceParam objects this cache knows about
         * @returns {[sas.analytics.riskhp.models.DataSourceParams]}
         */
        getDataSourceParams: function() {
            return _.values(dataSourceParamsMap);
        },

        /**
         * Finds all the DataSourceParams objects which match some properties in dataSourceDetails.
         * The properties compared are: abstractDataSourceId, application, workGroup, and relPath.  NOTE: abstractDataSourceId
         * is compared against dataSourceDetails.id because that property does not exist in dataSourceDetails
         * @param dataSourceDetails
         * @returns {*}
         */
        getMatchingDataSourceParamsUsingDetails: function(dataSourceDetails) {
            var allParams = _.values(dataSourceParamsMap);

            // Get the dataSource params that match the root data source
            return _.filter(allParams, function(dataSourceParam){
                return dataSourceParam.abstractDataSourceId === dataSourceDetails.id &&
                    dataSourceParam.application === dataSourceDetails.application &&
                    dataSourceParam.workGroup === dataSourceDetails.workGroup &&
                    dataSourceParam.relPath === dataSourceDetails.relPath;
            });
        },

        /**
         * Finds all the DataSourceParams objects which match some properties in another dataSourceParams object.
         * The properties compared are: abstractDataSourceId, application, workGroup, and relPath.
         * @param dataSourceParams
         * @returns {*}
         */
        getMatchingDataSourceParamsUsingParams: function(dataSourceParams) {
            var allParams = _.values(dataSourceParamsMap);

            // Get the dataSource params that match the data source params
            return _.filter(allParams, function(dataSourceParam){
                return dataSourceParam.abstractDataSourceId === dataSourceParams.abstractDataSourceId &&
                    dataSourceParam.application === dataSourceParams.application &&
                    dataSourceParam.workGroup === dataSourceParams.workGroup &&
                    dataSourceParam.relPath === dataSourceParams.relPath;
            });
        },

        /**
         * Returns the number of unique DataSourceParams currently stored in this DataSourceCacheManager.  Unique in this case means that the
         * DataSourceParams with matching "abstractDataSourceId", "application", "workGroup", and "relPath" properties are considered the same.
         * Notice the "type" property is not considered in this case.  Basically this method counts the "price" and/or "stress" version of the
         * DataSourceParams as the same one.
         * @returns {number} 0 if no DataSourceParams objects exist or the number of unique DataSourceParams objects
         */
        getCountOfUniqueDataSourceParams: function() {
            var allParams = _.values(dataSourceParamsMap);

            // Group all the DataSourceParams objects together this cache knows about by building a unique key; The "type" property is excluded
            var groupDataSourceParamsByUniqueId = _.groupBy(allParams, function(dataSourceParam) {
                return dataSourceParam.abstractDataSourceId + '|' + dataSourceParam.application + '|' + dataSourceParam.workGroup + '|' + dataSourceParam.relPath;
            });

            // Return 0 if no DataSourceParams exist or the actual length of unique DataSourceParams
            return _.isEmpty(groupDataSourceParamsByUniqueId) ? 0 : _.values(groupDataSourceParamsByUniqueId).length;
        },

        /**
         * Update the dataSourceParamsMap object for this cache manager by ultimately looping through the 'dataSourceDetails.dataSources'
         * property and determining if we need to add new DataSourceParam objects or remove some.
         * @param dataSourceDetails
         */
        updateDataSourceParams: function(dataSourceDetails) {
            if (dataSourceDetails && dataSourceDetails.dataSources) {
                var updatedDataSourceParams = [];
                var dataSourceParamsToAdd = [];

                // Get the dataSource params that match the root data source
                var matchedDataSourceParams = this.getMatchingDataSourceParamsUsingDetails(dataSourceDetails);

                // Loop through the 'dataSources' array and see if each dataSource already exists, or if
                // we need to create a new DataSourceParams object
                _.each(dataSourceDetails.dataSources, function(dataSource) {
                    // Try to find the DSP in our matched DSP array
                    var foundDSP = _.find(matchedDataSourceParams, function (dataSourceParam) {
                        return dataSource.id === dataSourceParam.realDataSourceId;
                    });

                    // If the DSP was found, add it to the updated array
                    if (foundDSP) {
                        // TODO: add any updates here
                        updatedDataSourceParams.push(foundDSP);
                    }
                    else {
                        // Otherwise, create a new DataSourceParams object by using the appropriate properties from the
                        // 'dataSourceDetails' and 'dataSource' objects
                        var newDataSourceParams = new sas.analytics.riskhp.models.DataSourceParams();
                        newDataSourceParams.abstractDataSourceId = dataSourceDetails.id;
                        newDataSourceParams.application = dataSourceDetails.application;
                        newDataSourceParams.workGroup = dataSourceDetails.workGroup;
                        newDataSourceParams.relPath = dataSourceDetails.relPath;
                        newDataSourceParams.type = dataSource.type;
                        newDataSourceParams.realDataSourceId = dataSource.id;
                        newDataSourceParams.label = sas.analytics.riskhp.utils.HprUtil.getCubeDisplayLabel(dataSource.label, dataSourceDetails.name, dataSource.type);
                        //newDataSourceParams.name = dataSourceDetails.name;
                        //newDataSourceParams.description = dataSourceDetails.description;
                        //newDataSourceParams.lastModifiedTimeStamp = dataSourceDetails.lastModifiedTimeStamp;

                        // Add the new DataSourceParams object to the add array
                        dataSourceParamsToAdd.push(newDataSourceParams);
                    }
                });

                // Figure out which DataSourceParams need to be removed by comparing the 'realDataSourceId' on the
                // matched DataSourceParams with the 'realDataSourceId' on the updated DataSourceParams.  The _.reject
                // method returns the values without the elements that the truth test (predicate) passes.  So the
                // DataSourceParams objects that were matched but not updated will be removed
                var dataSourceParamsToRemove = _.reject(matchedDataSourceParams, function(matchedDataSourceParam){
                    var foundDSP = _.find(updatedDataSourceParams, function (updatedDataSourceParam) {
                        return matchedDataSourceParam.realDataSourceId === updatedDataSourceParam.realDataSourceId;
                    });
                    return foundDSP !== null;
                });

                // At this point we have a list of the DataSourceParams we updated, a list of the DataSourceParams to
                // add, and a list of the DataSourceParams to remove. Nothing to do with the DataSourceParams that were updated.

                // Remove the DataSourceParams that need to be removed
                _.each(dataSourceParamsToRemove, function(dataSourceParam) {
                    delete dataSourceParamsMap[dataSourceParam.realDataSourceId];
                });

                // Add the DataSourceParams that need to be added
                _.each(dataSourceParamsToAdd, function(dataSourceParam) {
                    dataSourceParamsMap[dataSourceParam.realDataSourceId] = dataSourceParam;
                });
            }
        },

        /**
         * Returns a list of data source ids representing the data sources that have been loaded and therefore are cached
         * by the cache manager
         * @returns {[string]}
         */
        getLoadedDataSourceIds: function() {
            return _.keys(dataSourceMap);
        },

        /**
         * Method to release (from the midtier) all the data sources this cache manager has loaded.
         * @returns {Deferred} The deferred object returned from this method will resolve when all cached data sources
         * have been successfully released by the midtier. Right now this does not handle the fail case (mostly because
         * we don't know what to do)
         */
        releaseAllLoadedDataSources: function() {
            var loadedDataSourceIds = this.getLoadedDataSourceIds();
            var deferredArray = [];
            _.each(loadedDataSourceIds, function(dataSourceId) {
                var deferred = sas.analytics.riskhp.utils.HprAPI.releaseCubeResources(dataSourceId);
                deferred.done(function(){
                    delete dataSourceMap[dataSourceId];
                    delete dataSourceParamsMap[dataSourceId];
                });
                // TODO: What about fail case
                deferredArray.push(deferred);
            });
            // Pass an array to the when function - https://stackoverflow.com/questions/5627284/pass-in-an-array-of-deferreds-to-when
            var combinedDeferred = $.Deferred();
            $.when.apply(null, deferredArray).done(function() {
                _updateCubeKeepAlive();
                combinedDeferred.resolve();
            });
            // TODO: What about fail case
            return combinedDeferred;
        },

        /**
         * Method that tries to release (from the midtier) all the data sources listed in 'ids'.  If the data source is loaded, it
         * is released by the midtier and deleted from the appropriate maps of this cache.  If the data source is not loaded, it
         * is just deleted from the appropriate maps of this cache.
         * @param ids Array of data source ids which are to be released by the midtier
         * @returns {Deferred} The deferred object returned from this method will resolve when all cached data sources
         * have been successfully released by the midtier. Right now this does not handle the fail case (mostly because
         * we don't know what to do)
         */
        releaseDataSourcesById: function(ids) {
            if (ids && ids.length === 0) return; // Do nothing if we don't have ids

            var loadedDataSourceIds = _.intersection(DataSourceCacheManager.getLoadedDataSourceIds(), ids); // Finds the id's of loaded Data Sources that are passed in via 'ids' parameter
            var notLoadedDataSourceIds = _.difference(ids, DataSourceCacheManager.getLoadedDataSourceIds()); // Finds the id's of Data Sources that are not loaded

            var deferredArray = [];
            _.each(loadedDataSourceIds, function(dataSourceId) {
                var deferred = sas.analytics.riskhp.utils.HprAPI.releaseCubeResources(dataSourceId);
                deferred.done(function(){
                    delete dataSourceMap[dataSourceId];
                    delete dataSourceParamsMap[dataSourceId];
                });
                // TODO: What about fail case
                deferredArray.push(deferred);
            });
            // Pass an array to the when function - https://stackoverflow.com/questions/5627284/pass-in-an-array-of-deferreds-to-when
            var combinedDeferred = $.Deferred();
            $.when.apply(null, deferredArray).done(function() {

                // Still have to delete the Data Sources that were passed in that are not currently loaded
                _.each(notLoadedDataSourceIds, function(dataSourceId) {
                    delete dataSourceMap[dataSourceId];
                    delete dataSourceParamsMap[dataSourceId];
                });
                _updateCubeKeepAlive();
                combinedDeferred.resolve();
            });
            // TODO: What about fail case
            return combinedDeferred;
        },

        /**
         * Given a data source id returns a Deferred object (which may already be resolved) which when resolved will return
         * the associated DataSource object.  Consumers should attach a function(cube){} method to the done handler of this
         * deferred object to retrieve the associated data source object (DataSource)
         * @param dataSourceId
         * @returns {Deferred} This deferred may already be resolved, in which case the consumer's done handler will get
         * called immediately
         */
        getDataSource: function(dataSourceId) {
            return _getData(DATATYPE_DATASOURCE, dataSourceId);
        },

        /**
         * A temporary method to by-pass the default behaviour of the getDataSource(id) method so as to make it easier to
         * transition to this new design.  This method returns the cached data source associated with the passed in
         * data source id, or null if it isn't in the cache.  Consumers should move away from this method and transition
         * to the preferred getDataSource method.  Note, this method internally uses the getDataSource(id) method and
         * hides the returned deferred from that method
         * @param dataSourceId
         * @returns {DataSource} Will return null if the data source associated with passed in id has not been loaded yet. If null
         * is returned, a warning message will be logged to the console to indicate why
         * @deprecated Use getDataSource
         */
        getLoadedDataSource: function(dataSourceId) {
            if (!dataSourceId) {
                console.warn("getLoadedData called without dataSourceParamsId");
            }
            var retrievedDataSource = null;
            var deferred = this.getDataSource(dataSourceId);
            deferred.done(function(cubeData){
                retrievedDataSource = cubeData;
            });
            if (!retrievedDataSource) {
                console.warn("retrievedDataSource could not be found with dataSourceId: "+dataSourceId);
            }
            return retrievedDataSource;
        },

        /**
         * Returns a string representation of the current state of the DataSourceCacheManager.  This is intended to be
         * used for debugging purposes
         */
        getState: function() {
            var result = "Selected DataSource: " + selectedDataSourceId + "\n";

            result = result + "DataSourceParams: \n";
            var keys = _.keys(dataSourceParamsMap);
            if (keys && keys.length > 0) {
                _.each(keys, function (key){
                    result = result + "    " + key + "\n"
                });
            }
            else {
                result = result + "    None\n";
            }
            result = result + "Cached DataSources:\n";

            keys = _.keys(dataSourceMap);
            if (keys && keys.length > 0) {
                _.each(keys, function (key) {
                    var value = dataSourceMap[key];
                    var valueString = "DataSource object";
                    if (_.isFunction(value.done)) {
                        valueString = "Deferred object";
                    }
                    result = result + "    " + key + ": " + valueString + "\n";
                });
            }
            else {
                result = result + "    None\n";
            }

            result = result +"Listeners:\n";
            keys = _.keys(listenerMap);
            if (keys && keys.length > 0) {
                _.each(_.keys(listenerMap), function (key){
                    result = result + "    " + key + "\n"
                });
            }
            else {
                result = result + "    None\n";
            }
            return result;
        },

        /**
         * Still in work
         * @param dataSourceId
         * @param dataSourceType
         * @returns {string}
         */
        register: function(dataSourceId, dataSourceType) {
            var cacheId = _generateCacheId(dataSourceId, dataSourceType);
            var listeners = listenerMap[cacheId];
            if (!listeners) {
                listeners = [];
                listenerMap[cacheId] = listeners;
            }
            var listenerToken = "" + new Date().getTime();
            listeners.push(listenerToken);
            return listenerToken;
        },

        /**
         * Still in work
         * @param dataSourceId
         * @param dataSourceType
         * @param listenerToken
         */
        deregister: function(dataSourceId, dataSourceType, listenerToken) {
            var cacheId = _generateCacheId(dataSourceId, dataSourceType);
            var listeners = listenerMap[cacheId];
            if (listeners) {
                var newListeners = _.without(listeners, [listenerToken]);
                if (newListeners.length === 0) {
                    delete listenerMap[cacheId];
                    delete dataSourceMap[dataSourceId];
                    delete dataSourceParamsMap[dataSourceId];
                    _updateCubeKeepAlive();
                }
                else {
                    listenerMap[cacheId] = newListeners;
                }
            }
        }
    };
}());

DataSourceCacheManager = sas.analytics.riskhp.models.DataSourceCacheManager;