// (c) 2017, SAS Institute Inc.

/**
 * @name sas.analytics.riskhp.models.DataSourceParams
 */
sap.ui.base.Object.extend("sas.analytics.riskhp.models.DataSourceParams", {
    constructor: function () {
        this.dataItemModifications = {};
        this.attachedScenarioIds = [];
        this.dataSourceTreeTableModels = {};
    }
});

/**
 * Creates a "RootParam" DataSourceParams object which is an object used to fetch more detailed information about
 * a DataSource.  A "RootParam" object consists of these properties: id, application, workGroup, and relPath.
 * @param dataSourceData
 */
sas.analytics.riskhp.models.DataSourceParams.prototype.createRootParams = function(dataSourceData) {
    var dataSourceParam = new sas.analytics.riskhp.models.DataSourceParams();
    dataSourceParam.abstractDataSourceId = dataSourceData.id;
    dataSourceParam.application = dataSourceData.application;
    dataSourceParam.workGroup = dataSourceData.workGroup;
    dataSourceParam.relPath = dataSourceData.relPath;
    //dataSourceParam.type = dataSourceData.type;
    return dataSourceParam;
};

/**
 * Creates a new DataSourceParams object with properties from all 3 parameters.
 * @param rootParams
 * @param detailsData
 * @param dataSourceData
 */
sas.analytics.riskhp.models.DataSourceParams.prototype.fromRootParams = function(rootParams, detailsData, dataSourceData) {
    var dataSourceParam = new sas.analytics.riskhp.models.DataSourceParams();
    dataSourceParam.abstractDataSourceId = rootParams.id;
    dataSourceParam.application = rootParams.application;
    dataSourceParam.workGroup = rootParams.workGroup;
    dataSourceParam.relPath = rootParams.relPath;
    dataSourceParam.type = dataSourceData.type;
    dataSourceParam.realDataSourceId = dataSourceData.id;
    dataSourceParam.name = detailsData.name;
    dataSourceParam.label = sas.analytics.riskhp.utils.HprUtil.getCubeDisplayLabel(detailsData.label, detailsData.name, dataSourceData.type);
    dataSourceParam.description = detailsData.description;
    dataSourceParam.lastModifiedTimeStamp = detailsData.lastModifiedTimeStamp;
    return dataSourceParam;
};

/**
 * Creates a new DataSourceParams object based on JSON being read in.  This method is
 * useful when creating DataSourceParam objects from a saved Exploration.
 * @param json
 */
sas.analytics.riskhp.models.DataSourceParams.prototype.fromJSON = function(json) {
    var dataSourceParam = new sas.analytics.riskhp.models.DataSourceParams();
    dataSourceParam.abstractDataSourceId = json.abstractDataSourceId;
    dataSourceParam.application = json.application;
    dataSourceParam.workGroup = json.workGroup;
    dataSourceParam.relPath = json.relPath;
    dataSourceParam.type = json.type;
    dataSourceParam.realDataSourceId = json.realDataSourceId;
    dataSourceParam.label = json.label;
    dataSourceParam.attachedScenarioIds = json.attachedScenarioIds;
    dataSourceParam.dataItemModifications = json.dataItemModifications;
    dataSourceParam.dataSourceTreeTableModels = json.dataSourceTreeTableModels;
    // dataSourceParam.name = json.name;
    // dataSourceParam.lastModifiedTimeStamp = json.lastModifiedTimeStamp;
    // dataSourceParam.description = json.description;
    return dataSourceParam;
};

sas.analytics.riskhp.models.DataSourceParams.prototype.toJSON = function() {
    return {
        abstractDataSourceId: this.abstractDataSourceId,
        application: this.application,
        workGroup: this.workGroup,
        relPath: this.relPath,
        type: this.type,
        realDataSourceId: this.realDataSourceId,
        label: this.label,
        attachedScenarioIds: this.attachedScenarioIds,
        dataItemModifications: this.dataItemModifications,
        dataSourceTreeTableModels: this.dataSourceTreeTableModels,  // contains the nodeOrderMap and horizonNodeChildrenCache
    };
};

/**
 * Returns an encoded string of properties and their values which are used to help retrieve and
 * load the DataSource using a REST call.  NOTE: The properties used to build the
 * query are: application, workGroup, relPath, & type
 */
sas.analytics.riskhp.models.DataSourceParams.prototype.getQuery = function() {
    // Empty string is required as first value so _.reduce function works properly;
    // Only build the query string using these 5 keys/properties
    var keys = ["", "application", "workGroup", "relPath", "type"];
    var self = this;
    var result = _.reduce(keys, function(memo, key) {
        // If this DataSourceParams object has the 'key' stored, use it
        if (self[key]) {
            // This builds up the query string by concatenating the previous string
            // with the new key and encoded value
            return memo + key + '=' + encodeURIComponent(self[key]) + '&';
        }
        return memo;
    });
    return result.substring(0, result.length-1); // Will remove the extra '&' on the end
};

//*****************************************************************************************************
// DataSource TreeTable model functions
//*****************************************************************************************************
/**
 * Add a DataSource TreeModel to this DataSourceParams so it can be saved and recalled.
 * @param dataSourceId
 * @param dataSourceTreeTableModelData
 */
sas.analytics.riskhp.models.DataSourceParams.prototype.addDataSourceTreeTableModelObject = function(dataSourceId, dataSourceTreeTableModelData) {
    this.dataSourceTreeTableModels[dataSourceId] = dataSourceTreeTableModelData;
};

/**
 * Delete a DataSource TreeModel from this DataSourceParams.
 * @param dataSourceId
 */
sas.analytics.riskhp.models.DataSourceParams.prototype.deleteDataSourceTreeTableModelObject = function(dataSourceId) {
    if ((dataSourceId in this.dataSourceTreeTableModels)) {
        delete this.dataSourceTreeTableModels[dataSourceId];
    }
};

/**
 * Retrieve a specific DataSource TreeModel
 * @param dataSourceId
 * @returns {DataSourceTreeModel object | undefined} Will return undefined if a DataSource TreeModel can not be found
 */
sas.analytics.riskhp.models.DataSourceParams.prototype.getDataSourceTreeTableModelObject = function(dataSourceId) {
    if (dataSourceId in this.dataSourceTreeTableModels) {
        return this.dataSourceTreeTableModels[dataSourceId];
    }
    return undefined;
};

//*****************************************************************************************************
// Data Item modification functions
//*****************************************************************************************************
/**
 * This method is useful when opening a Saved Exploration because it re-applies modifications to Data Item objects.  Currently the method re-applies
 * label changes and count property updates.
 * @param dataSource The DataSource which contains the Data Item who should be modified
 */
sas.analytics.riskhp.models.DataSourceParams.prototype.applyDataItemModifications = function(dataSource) {
    this._applyChangedLabelsToDataSource(dataSource);
    this._applyChangedCountsToDataSource(dataSource);
};

/**
 * Reset the Object which holds the list of changed Data Item labels to an empty Object
 */
sas.analytics.riskhp.models.DataSourceParams.prototype.resetDataItemChangedLabelObjects = function() {
    if (this.dataItemModifications.dataItemChangedLabelObjects) this.dataItemModifications.dataItemChangedLabelObjects = {};
};

/**
 * Returns the hashmap of Objects representing the Data Items who had label changes.
 *
 * @returns {{}} Returns an Object representing the Data Items with label changes or an empty Object if none exist
 */
sas.analytics.riskhp.models.DataSourceParams.prototype.getDataItemChangedLabelObjects = function() {
    return this.dataItemModifications.dataItemChangedLabelObjects ? this.dataItemModifications.dataItemChangedLabelObjects : {};
};

/**
 * Add an Object to the 'dataItemModifications.dataItemChangedLabelObjects' hashmap which holds information about a Data Item label update.
 *
 * @param data Object with information about a Data Item label update
 */
sas.analytics.riskhp.models.DataSourceParams.prototype.addDataItemChangedLabelObject = function(data) {
    var self = this;

    // Based on the properties for the data passed in, manipulate it before adding the object to the dataItemChangedLabelObjects hashmap
    if (data && data.dataItem) {
        var dataItemRef = data.dataItem;
        var dataItemRefId = dataItemRef.getId();
        var dataItemRefType = dataItemRef.getType();
        var uniqueId = dataItemRefId + "_" + dataItemRefType;
        var labelObject = {
            dataSourceId: data.dataSourceId,
            dataItemId: dataItemRefId,
            dataItemType: dataItemRefType,
            newLabel: data.newLabel,
            oldLabel: data.oldLabel
        };

        // If not present, create the required parent object; Add or update correct property with new 'labelObject'
        if (!self.dataItemModifications.dataItemChangedLabelObjects) self.dataItemModifications.dataItemChangedLabelObjects = {};
        self.dataItemModifications.dataItemChangedLabelObjects[uniqueId] = labelObject;

        // For Horizon DataItems, we need to make sure we also store an object representing the label change
        // for the AbsoluteHorizon DataItem, this way the Crosstab will be happy
        if (dataItemRefType === sas.analytics.riskhp.models.dataitems.Horizon.HORIZON) {
            // Retrieve the DataSource that was used for the label update so we can get to the AbsoluteHorizon DataItem
            var deferred = DataSourceCacheManager.getDataSource(data.dataSourceId);
            deferred.done(function(dataSource) {
                // NOTE: Have to use 'data.newLabel' here since label has already been updated on DataItem
                var absHorizon = dataSource.findAbsoluteHorizon(dataItemRef.horizonDate, data.newLabel, dataItemRef.basecase);
                if (absHorizon) {
                    // Create a unique lebel change object for AbsoluteHorizon now
                    dataItemRefType = sas.analytics.riskhp.models.dataitems.AbsoluteHorizon.ABSOLUTEHORIZON;
                    uniqueId = absHorizon.getId() + "_" + dataItemRefType;
                    labelObject = {
                        dataSourceId: data.dataSourceId,
                        dataItemId: absHorizon.getId(),
                        dataItemType: dataItemRefType,
                        newLabel: data.newLabel,
                        oldLabel: data.oldLabel
                    };

                    self.dataItemModifications.dataItemChangedLabelObjects[uniqueId] = labelObject;
                }
            });
        }
    }
};

/**
 * Loop through this DataSourceParams 'dataItemModifications.dataItemChangedLabelObjects' hashmap and apply the label changes to the corresponding Data Item's stored
 * in the cube data for this DataSourceParams object.
 * @param dataSource The DataSource to apply the label changes to
 */
sas.analytics.riskhp.models.DataSourceParams.prototype._applyChangedLabelsToDataSource = function(dataSource) {
    var self = this;
    var changedLabelObjects = self.getDataItemChangedLabelObjects();
    _.each(changedLabelObjects, function(object) {
        // Find the Data Item by 'id' in the cubeData represented by this DataSourceParams object
        var dataItem = dataSource.getDataItemById(object.dataItemId);
        if (dataItem) {
            dataItem.setLabel(object.newLabel); // Update label
        }
    });
};

/**
 * Returns the hashmap of Objects representing the Data Items who had count and updateCount property changes.
 *
 * @returns {{}} Returns an Object representing the Data Items with count property changes or an empty Object if none exist
 */
sas.analytics.riskhp.models.DataSourceParams.prototype.getDataItemChangedCountObjects = function() {
    return this.dataItemModifications.dataItemChangedCountObjects ? this.dataItemModifications.dataItemChangedCountObjects : {};
};

/**
 * Add an Object to the 'dataItemModifications.dataItemChangedCountObjects' hashmap which holds information about a Data Item count property update.
 *
 * @param data Object with information about a Data Item count property update
 */
sas.analytics.riskhp.models.DataSourceParams.prototype.addDataItemChangedCountObject = function(data) {
    var self = this;

    // Based on the properties for the data passed in, manipulate it before adding the object to the dataItemChangedCountObjects hashmap
    if (data && data.dataItem) {
        var dataItemRef = data.dataItem;
        var uniqueId = dataItemRef.getId() + "_" + dataItemRef.getType();
        var countObject = {
            dataSourceId: data.dataSourceId,
            dataItemId: dataItemRef.getId(),
            dataItemType: dataItemRef.getType(),
            count: dataItemRef.getCount(),
            updateCount: dataItemRef.getUpdateCount()
        };

        // If not present, create the required parent object; Add or update correct property with new 'countObject'
        if (!self.dataItemModifications.dataItemChangedCountObjects) self.dataItemModifications.dataItemChangedCountObjects = {};
        self.dataItemModifications.dataItemChangedCountObjects[uniqueId] = countObject;
    }
};

/**
 * Loop through this DataSourceParams 'dataItemModifications.dataItemChangedCountObjects' hashmap and apply the count property changes to the corresponding Data Item's stored
 * in the cube data for this DataSourceParams object.
 * @param dataSource The DataSource to apply the count property changes to
 */
sas.analytics.riskhp.models.DataSourceParams.prototype._applyChangedCountsToDataSource = function(dataSource) {
    var self = this;
    var changedCountObjects = self.getDataItemChangedCountObjects();
    _.each(changedCountObjects, function(object) {
        // Find the Data Item by 'id' in the cubeData represented by this DataSourceParams object
        var dataItem = dataSource.getDataItemById(object.dataItemId);
        if (dataItem) {
            dataItem.setCount(object.count); // Update 'count' property
            dataItem.setUpdateCount(object.updateCount); // Update the 'updateCount' property
        }
    });
};

//*****************************************************************************************************
// Scenario functions
//*****************************************************************************************************
/**
 * Adds the specified scenario id to the list of scenarioId's this DataSourceParam is tracking. Duplicates will be ignored
 * @param scenarioId
 */
sas.analytics.riskhp.models.DataSourceParams.prototype.addScenarioId = function(scenarioId) {
    var foundId = _.find(this.attachedScenarioIds, function(id) {
        return id === scenarioId;
    });

    if (!foundId) {
        this.attachedScenarioIds.push(scenarioId);
    }
};

/**
 * Removes the specified scenarioId from the list of scenarioId's this DataSourceParam is tracking. No-op if the specified
 * scenarioId is not being tracked.
 * @param scenarioId
 */
sas.analytics.riskhp.models.DataSourceParams.prototype.removeScenarioId = function(scenarioId) {
    this.attachedScenarioIds = _.reject(this.attachedScenarioIds, function(id){
        return scenarioId === id;
    });
};

/**
 * Removes all the scenario ids this DataSourceParam is tracking
 */
sas.analytics.riskhp.models.DataSourceParams.prototype.removeAllScenarioIds = function() {
    this.attachedScenarioIds = [];
};
