// (c) 2017, SAS Institute Inc.


/**
 * @name sas.analytics.riskhp.models.DataSourceUtil
 */
sas.analytics.riskhp.models.DataSourceUtil = {
    /**
     * Takes a list of Horizons and generates RelativeHorizon data items, including the BaseCase Horizons.
     * @param horizons {Array} List of Horizon data items to used to generate new Relative Horizons
     * @param applyChangedLabels {Boolean} Whether the changed labels stored in the DataSourceParams object (found by using dataSourceId)
     * should be applied to the new RelativeHorizon data items
     * @param dataSourceId {String} id of the DataSource to look up and find the corresponding DataSourceParams object for
     * @returns {Array}
     */
    generateRelativeHorizonDataItems: function(horizons, applyChangedLabels, dataSourceId) {
        var relativeHorizons = [];

        // reject non-basecase horizons without a positive horizon number or valid interval
        var candidateHorizons = _.reject(horizons, function(horizon) {
            return !horizon.isBasecase() && (!horizon.interval || horizon.horizonNumber < 0);
        });

        // Group the horizons by their horizonNumber property
        var groupHorizonsByHorizonNumber = _.groupBy(candidateHorizons, function(horizon) {
            return horizon.horizonNumber;
        });

        // Create a RelativeHorizon data item which represents each horizonNumber group; This also includes BaseCase Horizons
        // because we need to display a RelativeHorizon data item in the Data pane for it
        _.each(groupHorizonsByHorizonNumber, function(horizonGroup, horizonNumber) {
            var groupHorizonsByInterval = _.groupBy(horizonGroup, function(horizon) {
                return horizon.interval;
            });
            _.each(groupHorizonsByInterval, function(horizonIntervalGroup, horizonInterval) {
                var horizon0 = horizonIntervalGroup[0]; // Grab the first Horizon to create the Relative Horizon from because it has all the info that we need
                var horizonIntervalGroupIds = _.pluck(horizonIntervalGroup, "id"); // Grab all the Horizon IDs which this Relative Horizon will represent
                relativeHorizons.push(new sas.analytics.riskhp.models.dataitems.RelativeHorizon(horizon0, horizonIntervalGroupIds));
            });
        });

        // Since we are generating new Relative Horizon data items, in some cases when this method is called we need
        // to re-apply labels that have been changed for the Data Source; Changed label modifications are stored in
        // the corresponding DataSourceParams object
        if (applyChangedLabels && dataSourceId) {
            var dsParams = sas.analytics.riskhp.models.DataSourceCacheManager.getDataSourceParam(dataSourceId);
            var changedLabelObjects = dsParams.getDataItemChangedLabelObjects();
            // Loop over the changed label objects in the DataSourceParams object and see if any match up to the new Relative Horizons we are generating
            _.each(changedLabelObjects, function(object) {
                var relHorizon = _.find(relativeHorizons, function(relativeHorizon) {
                    return relativeHorizon.getId() === object.dataItemId;
                });
                if (relHorizon) {
                    relHorizon.setLabel(object.newLabel);
                }
            });
        }

        return relativeHorizons;
    },

    /**
     * Takes a list of Horizons and generates AbsoluteHorizon data items.
     * @param horizons
     * @returns {Array}
     */
    generateAbsoluteHorizonDataItems: function(horizons) {
        var absoluteHorizons = [];

        var basecaseGroupedHorizons = _.groupBy(horizons, function(horizon) {
            return horizon.basecase;
        });

        var dateGroupedHorizons = _.groupBy(basecaseGroupedHorizons["false"], function(horizon) {
            return horizon.horizonDate;
        });

        _.each(dateGroupedHorizons, function(dateGroupedHorizon, date) {
            var dateAndLabelGroupedHorizons = _.groupBy(dateGroupedHorizon, function(horizon) {
                return horizon.label;
            });
            _.each(dateAndLabelGroupedHorizons, function(dateAndLabelGroupedHorizon, horizonLabel){
                var horizon0 = dateAndLabelGroupedHorizon[0]; // Grab the first one.  The others have the same info we need.
                if (horizon0.horizonDate !== "1959-12-31") { //FIXME "1959-12-31" is a hack for now to determine no absolute data
                    absoluteHorizons.push(new sas.analytics.riskhp.models.dataitems.AbsoluteHorizon(horizon0));
                }
            });
        });

        var basecaseGroupedHorizons = _.groupBy(basecaseGroupedHorizons["true"], function(horizon) {
            return horizon.horizonDate;
        });

        _.each(basecaseGroupedHorizons, function(basecaseGroupedHorizon, date) {
            var basecaseAndLabelGroupedHorizons = _.groupBy(basecaseGroupedHorizon, function(horizon) {
                return horizon.label;
            });
            _.each(basecaseAndLabelGroupedHorizons, function(basecaseAndLabelGroupedHorizon, horizonLabel){
                var horizon0 = basecaseAndLabelGroupedHorizon[0]; // Grab the first one.  The others have the same info we need.
                if (horizon0.horizonDate !== "1959-12-31") { //FIXME "1959-12-31" is a hack for now to determine no absolute data
                    absoluteHorizons.push(new sas.analytics.riskhp.models.dataitems.AbsoluteHorizon(horizon0));
                }
            });
        });

        return absoluteHorizons;
    },

    /**
     * Returns the the data source corresponding to the specified type.
     *
     * @param dataSourceDetails
     * @param dataSourceType
     */
    getDataSourceForType: function(dataSourceDetails, dataSourceType) {
        var dataSource = _.find(dataSourceDetails.dataSources, function(dataSource) {
            return dataSource.type === dataSourceType;
        });

        return dataSource;
    },

    /**
     * Returns the ID of the data source corresponding to the specified type.
     *
     * @param dataSourceDetails
     * @param dataSourceType
     */
    getDataSourceIDForType: function(dataSourceDetails, dataSourceType) {
        var dataSource = this.getDataSourceForType(dataSourceDetails, dataSourceType);
        if (!dataSource) {
            return null;
        }
        return dataSource.id;
    },

    /**
     * Find the potentially valid columns for an Output Variable data item.  No columns are returned if the data item
     * passed in is not an Output Variable.  The method uses the Restrictions from the DataSource to locate valid Horizons
     * for the Output Variable and then starts to build the potentially valid columns from those.
     *
     * @param dataSource The DataSource which contains the data item
     * @param dataItem The Output Variable data item to find columns for
     * @param groupRestrictionsByType The Restrictions for a DataSource grouped by the 'type' property
     * @param noBaseCase Flag to signify not to use BaseCase Horizon (true means don't use BaseCase Horizon)
     * @returns {Array} Returns an Array of potentially valid columns for an Output Variable data item
     */
    generatePotentialColumnsForOutputVariable: function(dataSource, dataItem, groupRestrictionsByType, noBaseCase) {
        var self = this;
        var potentialColumns = []; // Array of potential valid columns
        var potentialColumnTemplate = []; // Array of Data Items representing a valid column to be placed inside potentialColumns

        // Create some variables for the types of Data Items we might need to use
        var OUTVAR = sas.analytics.riskhp.models.dataitems.Outvar.OUTVAR;

        // If there is not a Data Item or if it is not an Output Variable, return empty array of potential columns
        if (!dataItem || (dataItem && dataItem.getType() != OUTVAR)) { return potentialColumns; }

        // Gather all the Horizons for the Data Source (i.e. Regular, Result Set, Scenarios, etc.); We need to find the Risk Cubes,
        // Result Sets, Scenarios, etc. so we can start to build valid columns
        var validHorizons = [].concat(self._findValidHorizonsForDataItem(dataSource, dataItem, groupRestrictionsByType));

        // We finally have a list of valid Horizons for the Output Variable; Loop through them
        // and build potentially valid columns
        $.each(validHorizons, function(index, validHorizon) {
            if (noBaseCase && dataSource.isDataItemBaseCaseHorizon(validHorizon)) {
                // If noBaseCase is specified, weed out the Base Case Horizon
                return true;
            } else {
                // Create a column based on the properties gathered from the Horizon
                if (validHorizon.riskCubeId) {
                    potentialColumnTemplate.push(dataSource.getDataItemById(validHorizon.riskCubeId));
                } else if (validHorizon.resultSetId) {
                    var resultSetDataItem1 = dataSource.getDataItemById(validHorizon.resultSetId);
                    potentialColumnTemplate.push(dataSource.getDataItemById(resultSetDataItem1.riskCubeId));
                    potentialColumnTemplate.push(resultSetDataItem1);
                } else if (validHorizon.scenarioId) {
                    var scenarioDataItem = dataSource.getDataItemById(validHorizon.scenarioId);
                    var resultSetDataItem2 = dataSource.getDataItemById(scenarioDataItem.resultSetId);
                    potentialColumnTemplate.push(dataSource.getDataItemById(resultSetDataItem2.riskCubeId));
                    potentialColumnTemplate.push(resultSetDataItem2);
                    potentialColumnTemplate.push(scenarioDataItem);
                }

                potentialColumnTemplate.push(validHorizon); // Push Horizon data item
                potentialColumnTemplate.push(dataItem); // Push Output Variable data item
                potentialColumns.push($.extend(true, [], potentialColumnTemplate));
                potentialColumnTemplate = []; // Reset to empty array after every Horizon since we are looping
            }
        });

        return potentialColumns;
    },

    /**
     * Find the valid columns for a Statistic data item.  No columns are returned if the data item
     * passed in is not a Statistic.  The method uses the Restrictions from the DataSource to locate valid Horizons
     * for the Statistic and then starts to build the valid columns from those.
     *
     * @param dataSource The DataSource which contains the data item
     * @param dataItem The Statistic data item to find columns for
     * @param groupRestrictionsByType The Restrictions for a DataSource grouped by the 'type' property
     * @param noBaseCase Flag to signify not to use BaseCase Horizon (true means don't use BaseCase Horizon)
     * @returns {Array} Returns an Array of valid columns for a Statistic data item
     */
    generateValidColumnsForStatistic: function(dataSource, dataItem, groupRestrictionsByType, noBaseCase) {
        var self = this;
        var validColumns = []; // Array of valid columns
        var potentialColumnTemplate = []; // Array of Data Items representing a valid column to be placed inside validColumns

        // Create some variables for the types of Data Items we might need to use
        var RISKCUBE = sas.analytics.riskhp.models.dataitems.RiskCube.RISKCUBE;
        var STATISTIC = sas.analytics.riskhp.models.dataitems.Statistic.STATISTIC;

        // If there is not a Data Item or if it is not a Statistic, return empty array of columns
        if (!dataItem || (dataItem && dataItem.getType() != STATISTIC)) { return validColumns; }

        // Create some valid columns based on the type of Statistic
        if (dataSource.isRowStatistic(dataItem)) {
            // For Row Statistics, loop through the Risk Cubes for the DataSource and use the Restrictions to see if this Row Statistic is valid for each Risk Cube
            $.each(dataSource.riskCube, function(index, riskCube) {
                if (groupRestrictionsByType[RISKCUBE] && groupRestrictionsByType[RISKCUBE].length > 0) {
                    var riskCubeRestrictionTypeToUse = null;

                    // First, we need to locate a Risk Cube Restriction
                    for (var j = 0; j < groupRestrictionsByType[RISKCUBE].length; j++) {
                        var restrictionId = groupRestrictionsByType[RISKCUBE][j].getId();
                        var restrictionType = groupRestrictionsByType[RISKCUBE][j].getType();
                        if ((riskCube.getId() === restrictionId && RISKCUBE === restrictionType) || (restrictionId === "" && RISKCUBE === restrictionType)) {
                            riskCubeRestrictionTypeToUse = groupRestrictionsByType[RISKCUBE][j];
                            break;
                        }
                    }

                    // Use the Risk Cube Restriction to look at the Statistic Rules for the Row Statistic
                    if (riskCubeRestrictionTypeToUse) {
                        var riskCubeRestrictionRulesByType = _.groupBy(riskCubeRestrictionTypeToUse.rules, function(rule) { return rule.type; });

                        // See if there are any Statistic Restriction Rules for this Risk Cube Restriction; It may have valid/invalid Row Statistics listed
                        if (riskCubeRestrictionRulesByType[STATISTIC] && riskCubeRestrictionRulesByType[STATISTIC].length > 0) {
                            // Try to locate a Statistic rule by using the Statistic data item id
                            var statisticRiskCubeRule = _.findWhere(riskCubeRestrictionRulesByType[STATISTIC], {id: dataItem.getId()});
                            if (!statisticRiskCubeRule) {
                                // Could be undefined if not found, use wildcard Statistic rule instead
                                statisticRiskCubeRule = _.findWhere(riskCubeRestrictionRulesByType[STATISTIC], {id: ""});
                            }

                            if (statisticRiskCubeRule && statisticRiskCubeRule.include === true) {
                                // Valid Row Statistic according to Risk Cube Restriction
                                potentialColumnTemplate.push(riskCube);
                                potentialColumnTemplate.push(dataItem); // Push Statistic data item
                                validColumns.push($.extend(true, [], potentialColumnTemplate));
                                potentialColumnTemplate = []; // Reset to empty array after every Risk Cube since we are looping
                            }
                        }
                    }
                } else {
                    // If there are no Risk Cube Restrictions, assume the Row Statistic is valid for the Risk Cube
                    potentialColumnTemplate.push(riskCube);
                    potentialColumnTemplate.push(dataItem); // Push Statistic data item
                    validColumns.push($.extend(true, [], potentialColumnTemplate));
                    potentialColumnTemplate = []; // Reset to empty array after every Risk Cube since we are looping
                }
            });
        } else {
            // Gather all the Horizons for the Data Source (i.e. Regular, Result Set, Scenarios, etc.); We need to find the Risk Cubes,
            // Result Sets, Scenarios, etc. so we can start to build valid columns
            var validHorizons = [].concat(self._findValidHorizonsForDataItem(dataSource, dataItem, groupRestrictionsByType));

            // We finally have a list of valid Horizons for the Statistic; Loop through them and build potentially valid columns
            $.each(validHorizons, function(index, validHorizon) {
                if (noBaseCase && dataSource.isDataItemBaseCaseHorizon(validHorizon)) {
                    // If noBaseCase is specified, weed out the Base Case Horizon
                    return true;
                } else {
                    // Create a column based on the properties gathered from the Horizon
                    if (validHorizon.riskCubeId) {
                        potentialColumnTemplate.push(dataSource.getDataItemById(validHorizon.riskCubeId));
                    } else if (validHorizon.resultSetId) {
                        var resultSetDataItem1 = dataSource.getDataItemById(validHorizon.resultSetId);
                        potentialColumnTemplate.push(dataSource.getDataItemById(resultSetDataItem1.riskCubeId));
                        potentialColumnTemplate.push(resultSetDataItem1);
                    } else if (validHorizon.scenarioId) {
                        var scenarioDataItem = dataSource.getDataItemById(validHorizon.scenarioId);
                        var resultSetDataItem2 = dataSource.getDataItemById(scenarioDataItem.resultSetId);
                        potentialColumnTemplate.push(dataSource.getDataItemById(resultSetDataItem2.riskCubeId));
                        potentialColumnTemplate.push(resultSetDataItem2);
                        potentialColumnTemplate.push(scenarioDataItem);
                    }

                    potentialColumnTemplate.push(validHorizon); // Push Horizon data item

                    // Loop through the Output Variables in the Data Source to see if we can make a valid column using Restriction check
                    $.each(dataSource.outputVariable, function(index, outputVariable) {
                        potentialColumnTemplate.push(outputVariable); // Push Output Variable data item
                        potentialColumnTemplate.push(dataItem); // Push Statistic data item

                        // Run potential column (array of data items) through the Restriction check to see if it is valid; If so, use it!
                        if (dataSource.areDataItemsCompatible(potentialColumnTemplate)) {
                            validColumns.push($.extend(true, [], potentialColumnTemplate));
                        }

                        // Since we are looping through Output Variables and we added an Output Variable and Statistic to the potential
                        // column template, we need to pop both so we can reuse that column template
                        potentialColumnTemplate.pop();
                        potentialColumnTemplate.pop();
                    });

                    potentialColumnTemplate = []; // Reset to empty array after every Horizon since we are looping
                }
            });
        }

        return validColumns;
    },

    /**
     * Find the valid Horizons for a data item.  The method uses the Restrictions from the DataSource to locate the valid Horizons.
     *
     * @param dataSource The DataSource which contains the data item
     * @param dataItem The data item to find Horizons for
     * @param groupRestrictionsByType The Restrictions for a DataSource grouped by the 'type' property
     * @returns {Array} Returns an Array of valid Horizon data items for a data item
     * @private
     */
    _findValidHorizonsForDataItem: function(dataSource, dataItem, groupRestrictionsByType) {
        var potentialValidHorizons = [];

        // If there is not a Data Item, return empty array of potential Horizons
        if (!dataItem) { return potentialValidHorizons; }

        var dataItemId = dataItem.getId();
        var dataItemType = dataItem.getType();
        var HORIZON = sas.analytics.riskhp.models.dataitems.Horizon.HORIZON;

        // Gather all the Horizons for the Data Source (i.e. Regular, Result Set, Scenarios, etc.); We need to find the Risk Cubes,
        // Result Sets, Scenarios, etc. so we can start to build valid columns
        var allHorizons = _.filter(dataSource.masterDataItemArray, function(dataItem) { return dataItem.getType() === HORIZON });
        var checkedHorizonRestrictions = false;

        // Check the Horizon Restrictions against the 'dataItem' to find potentially valid Horizons for our columns
        if (groupRestrictionsByType[HORIZON] && groupRestrictionsByType[HORIZON].length > 0) {
            checkedHorizonRestrictions = true;

            // Loop through all the Horizons and try to find valid ones
            _.each(allHorizons, function(horizon) {
                var horizonRestrictionTypeToUse = null;

                // First, we need to locate a Horizon Restriction
                for (var j = 0; j < groupRestrictionsByType[HORIZON].length; j++) {
                    var restrictionId = groupRestrictionsByType[HORIZON][j].getId();
                    var restrictionType = groupRestrictionsByType[HORIZON][j].getType();
                    if ((horizon.getId() === restrictionId && HORIZON === restrictionType) || (restrictionId === "" && HORIZON === restrictionType)) {
                        horizonRestrictionTypeToUse = groupRestrictionsByType[HORIZON][j];
                        break;
                    }
                }

                if (!horizonRestrictionTypeToUse) return potentialValidHorizons; // Couldn't find a Horizon Restriction, return empty array

                var horizonRestrictionRulesByType = _.groupBy(horizonRestrictionTypeToUse.rules, function(rule) { return rule.type; });

                // Try to locate a specific Restriction Rule by using the data item id
                var horizonDataItemTypeRule = _.findWhere(horizonRestrictionRulesByType[dataItemType], {id: dataItemId});
                if (!horizonDataItemTypeRule) {
                    // Could be undefined if not found, use wildcard Restriction Rule instead
                    horizonDataItemTypeRule = _.findWhere(horizonRestrictionRulesByType[dataItemType], {id: ""});
                }
                if (!horizonDataItemTypeRule || (horizonDataItemTypeRule && horizonDataItemTypeRule.include === true)) {
                    // Assume Horizon is potentially valid
                    potentialValidHorizons.push(horizon);
                }
            });
        }

        // No Horizon Restrictions present, so just assume all Horizons are potentially valid
        if (!checkedHorizonRestrictions && potentialValidHorizons.length === 0) {
            potentialValidHorizons = [].concat(allHorizons);
        }

        // If there are no potentially valid Horizons, we can't create a valid column so return empty array
        if (potentialValidHorizons.length === 0) {
            return potentialValidHorizons;
        }

        // Now use the Restrictions based on the Data Item type to try and find a list of invalid Horizons
        var invalidHorizonIds = [];
        if (groupRestrictionsByType[dataItemType] && groupRestrictionsByType[dataItemType].length > 0) {
            var restrictionTypeToUse = null;

            // First, we need to locate a Restriction to use
            for (var j = 0; j < groupRestrictionsByType[dataItemType].length; j++) {
                var restrictionId = groupRestrictionsByType[dataItemType][j].getId();
                var restrictionType = groupRestrictionsByType[dataItemType][j].getType();
                if ((dataItemId === restrictionId && dataItemType === restrictionType) || (restrictionId === "" && dataItemType === restrictionType)) {
                    restrictionTypeToUse = groupRestrictionsByType[dataItemType][j];
                    break;
                }
            }

            if (restrictionTypeToUse) {
                var restrictionRulesByType = _.groupBy(restrictionTypeToUse.rules, function(rule) { return rule.type; });

                // See if there are any Horizon Restriction Rules for this Restriction; It may have invalid Horizons listed
                if (restrictionRulesByType[HORIZON] && restrictionRulesByType[HORIZON].length > 0) {
                    _.each(potentialValidHorizons, function(horizon) {
                        // Try to locate a Horizon rule by using the Horizon data item id
                        var horizonRule = _.findWhere(restrictionRulesByType[HORIZON], {id: horizon.getId()});
                        if (!horizonRule) {
                            // Could be undefined if not found, use wildcard Horizon rule instead
                            horizonRule = _.findWhere(restrictionRulesByType[HORIZON], {id: ""});
                            if (horizonRule && horizonRule.include === false) {
                                // Bad Horizon according to Restriction
                                invalidHorizonIds.push(horizon.getId());
                            }
                        }
                    });
                }
            }
        }

        // Use the list of invalid Horizons to create a master list of valid Horizons
        var validHorizons = [];
        if (invalidHorizonIds.length > 0) {
            validHorizons = _.filter(potentialValidHorizons, function(potentialValidHorizon) {
                return !_.contains(invalidHorizonIds, potentialValidHorizon.getId());
            });
        } else {
            validHorizons = [].concat(potentialValidHorizons);
        }

        return validHorizons;
    }
};