// (c) 2018, SAS Institute Inc.

jQuery.sap.declare("sas.analytics.riskhp.models.DisplayRulesManager");

jQuery.sap.require("sas.analytics.riskhp.utils.HprUtil");
jQuery.sap.require("sas.analytics.riskhp.utils.HprAPI");
jQuery.sap.require("sas.analytics.riskhp.utils.HprConstants");
jQuery.sap.require("sas.analytics.riskhp.models.DisplayRule");

/**
 * @name sas.analytics.riskhp.models.DisplayRulesManager
 */
sas.analytics.riskhp.models.DisplayRulesManager = (function () {
    "use strict";
    var displayRules = [];
    var crosstabDisplayRulesMap = {}; // map of crosstab id to displayRule id and whether the rule is enabled for the crosstab
    var dataSourceDisplayRulesMap = {}; // map of dataSource id to displayRule id and whether the rule is valid for the dataSource

    var HprUtil = sas.analytics.riskhp.utils.HprUtil;

    var crosstabDeleted = function(channelId, eventId, data) {
        delete crosstabDisplayRulesMap[data.id];
    };

    sas.eventBus.subscribe(HprConstants.CHANNEL_EXPLORATION, HprConstants.EVENT_CROSSTAB_DELETED, crosstabDeleted, this);


    return {

        //*****************************************************************************************************
        // Display Rules functions
        //*****************************************************************************************************

        /**
         * Adds the specified displayRule. Duplicates will be ignored
         * @param displayRule
         */
        addDisplayRule: function(displayRule, crosstabId) {
            var self = this;
            var foundId = _.find(displayRules, function(dr) {
                return displayRule.id === dr.id;
            });

            if (!foundId) {
                displayRules.push(displayRule);
            }

            this.addDisplayRuleToDataSourceMap(displayRule);
            this.addDisplayRuleToCrosstabMap(crosstabId, displayRule);

            sas.eventBus.publish(HprConstants.CHANNEL_EXPLORATION, HprConstants.EVENT_DISPLAY_RULES_CHANGED, {
                crosstabIds: [crosstabId],
                tuples: [self.getAppliesToTuple(displayRule)]
            });

            // Update the dirty state of the Exploration view
            sas.eventBus.publish(HprConstants.CHANNEL_EXPLORATION, HprConstants.EVENT_UPDATE_DIRTY_STATE, {
                dirty: true
            });
        },

        getCrosstabsWithRuleActive: function(displayRuleId) {
            var crosstabIds = [];
            _.each(Object.keys(crosstabDisplayRulesMap), function(crosstabId){
                var crosstabDisplayRule = _.find(crosstabDisplayRulesMap[crosstabId], function(dr){
                    return dr.ruleId === displayRuleId;
                });
                if (crosstabDisplayRule && crosstabDisplayRule.active) {
                    crosstabIds.push(crosstabId);
                }
            });
            return crosstabIds;
        },

        deleteDisplayRule: function(displayRuleId) {
            var self = this;
            var crosstabsWithRuleActive = this.getCrosstabsWithRuleActive(displayRuleId);

            var origAppliesToTuple = self.getAppliesToTuple(self.getDisplayRuleById(displayRuleId));

            // remove display rule from the list of display rules in the exploration
            displayRules = _.reject(displayRules, function(displayRule){
                return displayRule.id === displayRuleId;
            });

            // remove the reference to the display rule from crosstabs
            _.each(Object.keys(crosstabDisplayRulesMap), function(crosstabId) {
                var crosstabDisplayRule = _.find(crosstabDisplayRulesMap[crosstabId], function (dr) {
                    return dr.ruleId === displayRuleId;
                });
                if (crosstabDisplayRule) {
                    crosstabDisplayRulesMap[crosstabId] = _.reject(crosstabDisplayRulesMap[crosstabId], function (dr) {
                        return dr.ruleId === displayRuleId;
                    });
                }
            });
            // remove the reference to the display rule from dataSources
            _.each(Object.keys(dataSourceDisplayRulesMap), function(dataSourceId) {
                var dataSourceDisplayRule = _.find(dataSourceDisplayRulesMap[dataSourceId], function (dr) {
                    return dr.ruleId === displayRuleId;
                });
                if (dataSourceDisplayRule) {
                    dataSourceDisplayRulesMap[dataSourceId] = _.reject(dataSourceDisplayRulesMap[dataSourceId], function (dr) {
                        return dr.ruleId === displayRuleId;
                    });
                }
            });
            sas.eventBus.publish(HprConstants.CHANNEL_EXPLORATION, HprConstants.EVENT_DISPLAY_RULES_CHANGED, {
                crosstabIds: crosstabsWithRuleActive,
                tuples: [origAppliesToTuple]
            });

            // Update the dirty state of the Exploration view
            sas.eventBus.publish(HprConstants.CHANNEL_EXPLORATION, HprConstants.EVENT_UPDATE_DIRTY_STATE, {
                dirty: true
            });
        },

        updateDisplayRule: function(displayRule) {
            var self = this;
            var index = _.findIndex(displayRules, function(dr){
                return dr.id === displayRule.id;
            });
            if (index != -1) {
                var origAppliesToTuple = self.getAppliesToTuple(displayRules[index]);
                var newAppliesToTuple = self.getAppliesToTuple(displayRule);
                displayRules[index] = displayRule;
                self.updateDisplayRuleInDataSourceMap(displayRule);

                var crosstabsWithRuleActive = this.getCrosstabsWithRuleActive(displayRule.id);
                sas.eventBus.publish(HprConstants.CHANNEL_EXPLORATION, HprConstants.EVENT_DISPLAY_RULES_CHANGED, {
                    crosstabIds: crosstabsWithRuleActive,
                    tuples: [origAppliesToTuple, newAppliesToTuple] // could optimize to only pass both if different
                });
            }
        },

        moveDisplayRule: function(up, displayRuleId, crosstabId) {
            var self = this;
            var rules = crosstabDisplayRulesMap[crosstabId];
            var index = _.findIndex(rules, function (rule) {
                return rule.ruleId === displayRuleId;
            });

            if (up) {
                if (index > 0) {
                    // swap item in rules list with prev item
                    var prev = rules[index - 1];
                    rules[index - 1] = rules[index];
                    rules[index] = prev;

                    var displayRule = self.getDisplayRuleById(displayRuleId);
                    sas.eventBus.publish(HprConstants.CHANNEL_EXPLORATION, HprConstants.EVENT_DISPLAY_RULES_CHANGED, {
                        crosstabIds: [crosstabId],
                        tuples: [self.getAppliesToTuple(displayRule)]
                    });

                    // Update the dirty state of the Exploration view
                    sas.eventBus.publish(HprConstants.CHANNEL_EXPLORATION, HprConstants.EVENT_UPDATE_DIRTY_STATE, {
                        dirty: true
                    });
                }
             } else { // move down
                if (index < (rules.length - 1)) {
                    // swap item in rules list with next item
                    var next = rules[index + 1];
                    rules[index + 1] = rules[index];
                    rules[index] = next;

                    var displayRule = self.getDisplayRuleById(displayRuleId);
                    sas.eventBus.publish(HprConstants.CHANNEL_EXPLORATION, HprConstants.EVENT_DISPLAY_RULES_CHANGED, {
                        crosstabIds: [crosstabId],
                        tuples: [self.getAppliesToTuple(displayRule)]
                    });

                    // Update the dirty state of the Exploration view
                    sas.eventBus.publish(HprConstants.CHANNEL_EXPLORATION, HprConstants.EVENT_UPDATE_DIRTY_STATE, {
                        dirty: true
                    });
                }
            }
        },

        toggleDisplayRule: function(displayRuleId, crosstabId) {
            var self = this;
            var rules = crosstabDisplayRulesMap[crosstabId];
            var index = _.findIndex(rules, function (rule) {
                return rule.ruleId === displayRuleId;
            });
            var crosstabRule = _.find(crosstabDisplayRulesMap[crosstabId], function(rule) {
                return rule.ruleId === displayRuleId;
            });
            if (crosstabRule) {
                crosstabRule.active = !crosstabRule.active;
            }
            crosstabDisplayRulesMap[crosstabId][index] = crosstabRule;
            var displayRule = self.getDisplayRuleById(displayRuleId);
            sas.eventBus.publish(HprConstants.CHANNEL_EXPLORATION, HprConstants.EVENT_DISPLAY_RULES_CHANGED, {
                crosstabIds: [crosstabId],
                tuples: [self.getAppliesToTuple(displayRule)]
            });

            // Update the dirty state of the Exploration view
            sas.eventBus.publish(HprConstants.CHANNEL_EXPLORATION, HprConstants.EVENT_UPDATE_DIRTY_STATE, {
                dirty: true
            });
        },

        /**
         * Returns the display rules
         */
        getDisplayRules: function() {
            return displayRules;
        },

        setDisplayRules: function(rules) {
            displayRules = rules;
        },


        getDisplayRuleById: function(ruleId) {
            return _.find(displayRules, function(rule) {
                return rule.id === ruleId;
            })
        },

        // Note: does not check for appliesTo Row since this is not yet supported
        getAppliesToTuple: function(displayRule) {
            return (displayRule.appliesTo.type === sas.analytics.riskhp.models.DisplayRule.APPLIES_TO_TYPE_COL ? displayRule.appliesTo.column : displayRule.referenceColumn);
        },

        // returns true if tuples are the same
        tuplesMatch: function (t1, t2) {
            return _.every(t1, function(item) {
                return _.find(t2, function(t2Item) {
                    return t2Item.id === item.id;
                })
            });
        },

        // returns active display rules that are apllied to the tuple for the crosstab
        getActiveDisplayRulesForCrosstabTuple: function(crosstabId, tuple) {
            var self = this;
            var displayRules = self.getDisplayRulesForCrosstab(crosstabId);
            // remove any display rules for which the appliesTo tuple does not match tuple
            return _.reject(displayRules, function(rule){
                var displayRule = self.getDisplayRuleById(rule.ruleId);
                return !rule.active || !self.tuplesMatch(self.getAppliesToTuple(displayRule), tuple);
            });
        },

        getDisplayRulesForCrosstab: function(crosstabId) {
            if (!crosstabDisplayRulesMap[crosstabId]) {
                var displayRulesForCrosstab = [];
                _.each(displayRules, function(rule) {
                    displayRulesForCrosstab.push({
                        ruleId: rule.id,
                        active: false
                    });
                });
                crosstabDisplayRulesMap[crosstabId] = displayRulesForCrosstab;
            }
            return HprUtil.hprClone(crosstabDisplayRulesMap[crosstabId]);
        },

        setDisplayRulesForCrosstab: function(crosstabId, rules) {
            crosstabDisplayRulesMap[crosstabId] = HprUtil.hprClone(rules);

            sas.eventBus.publish(HprConstants.CHANNEL_EXPLORATION, HprConstants.EVENT_DISPLAY_RULES_CHANGED, {
                crosstabIds: [crosstabId]
            });
        },


        addDisplayRuleToCrosstabMap: function(crosstabId, rule) {
            _.each(Object.keys(crosstabDisplayRulesMap), function(crosstabMapId) {
                var value = {
                    ruleId: rule.id,
                    active: crosstabMapId === crosstabId
                };
                if (crosstabMapId === crosstabId) {
                    // add to the top of the rules since its active
                    crosstabDisplayRulesMap[crosstabMapId].unshift(value)
                } else {
                    crosstabDisplayRulesMap[crosstabMapId].push(value)
                }
            });
        },

        addDisplayRuleToDataSourceMap: function(rule) {
            var self = this;
            var loadedDataSourceIds = DataSourceCacheManager.getLoadedDataSourceIds();

            _.each(Object.keys(dataSourceDisplayRulesMap), function (dataSourceMapId) {
                // when a dataSources is loaded, the displayRules for it are revalidated, so only need to check the
                // already-loaded dataSources, so the deferred from getDataSource should synchronous
                if (_.find(loadedDataSourceIds, function(loadedId) {
                    return loadedId === dataSourceMapId;
                })) {
                    DataSourceCacheManager.getDataSource(dataSourceMapId).done(function (dataSource) {
                        var isValid = self.isValidDisplayRule(dataSource, rule);
                        var value = {
                            ruleId: rule.id,
                            valid: isValid
                        };
                        dataSourceDisplayRulesMap[dataSourceMapId].push(value);
                    });
                } else {
                    dataSourceDisplayRulesMap[dataSourceMapId].push({
                        ruleId: rule.id,
                        valid: false
                    });
                }
            });
        },

        /**
         * revalidate the display rule across all open dataSources
         * @param rule
         */
        updateDisplayRuleInDataSourceMap: function(rule) {
            var self = this;

            var loadedDataSourceIds = DataSourceCacheManager.getLoadedDataSourceIds();
            _.each(Object.keys(dataSourceDisplayRulesMap), function (dataSourceMapId) {
                // when a dataSources is loaded, the displayRules for it are revalidated, so only need to check the
                // already-loaded dataSources, so the deferred from getDataSource should synchronous
                if (_.find(loadedDataSourceIds, function(loadedId) {
                        return loadedId === dataSourceMapId;
                    })) {
                    DataSourceCacheManager.getDataSource(dataSourceMapId).done(function (dataSource) {
                        var isValid = self.isValidDisplayRule(dataSource, rule);
                        var dataSourceRule = _.find(dataSourceDisplayRulesMap[dataSourceMapId], function(dataSourceRule){
                            return dataSourceRule.ruleId === rule.id;
                        });
                        if (dataSourceRule) {
                            dataSourceRule.valid = isValid;
                        }
                    });
                } else {
                    dataSourceDisplayRulesMap[dataSourceMapId].push({
                        ruleId: rule.id,
                        valid: false
                    });
                }
            });
        },

        reset: function() {
            displayRules = [];
            crosstabDisplayRulesMap = {};
            dataSourceDisplayRulesMap = {};
        },

        toJSON: function() {
            var rulesJSON = [];
            _.each(displayRules, function(rule) {
                rulesJSON.push(rule.toJSON());
            });

            return {
                rules: rulesJSON,
                crosstabDisplayRulesMap: crosstabDisplayRulesMap
            };
        },

        fromJSON: function(json) {
            if (json) {
                var rules = [];
                _.each(json.rules, function (rule) {
                    rules.push(sas.analytics.riskhp.models.DisplayRule.prototype.fromJSON(rule));
                });
                this.setDisplayRules(rules);

                crosstabDisplayRulesMap = HprUtil.hprClone(json.crosstabDisplayRulesMap);
            }
        },

        /**
         * returns true if the rule has been marked valid for the dataSource
         * @param dataSourceId
         * @param ruleId
         * @returns {boolean}
         */
        isDisplayRuleValidForDataSource: function(dataSourceId, ruleId) {
            var rules = dataSourceDisplayRulesMap[dataSourceId];
            if (rules) {
                var ruleData = _.find(rules, function(rule){
                    return rule.ruleId === ruleId;
                });
                if (ruleData) {
                    return ruleData.valid;
                }
            }
            return false;
        },


        isValidColumn: function(cubeData, column) {
            var self = this;
            var isValid = true;

            _.each(column, function (dataItem) {
                var item =cubeData.getDataItemById(dataItem.id);
                if (item) {
                    if (item.type === sas.analytics.riskhp.models.dataitems.Scenario.SCENARIO) {
                        if (!item.isActivated()) {
                            isValid = false;
                        }
                    } else if (item.type === sas.analytics.riskhp.models.dataitems.Compvar.COMPVAR) {
                        if (!item.isValid()) {
                            isValid = false;
                        }
                    }
                } else {
                    isValid = false;
                }
            });

            return isValid;
        },

        isValidDisplayRule: function(cubeData, rule) {
            var self = this;
            var valid = self.isValidColumn(cubeData, rule.referenceColumn);

            if (valid &&
                sas.analytics.riskhp.views.displayrules.DisplayRulesUtil.requiresDataItem1(rule.operator) &&
                rule.dataItem1.type === sas.analytics.riskhp.models.DisplayRule.DATA_ITEM_TYPE_COL) {
                valid = self.isValidColumn(cubeData, rule.dataItem1.column);
            }
            if (valid &&
                sas.analytics.riskhp.views.displayrules.DisplayRulesUtil.requiresDataItem2(rule.operator) &&
                rule.dataItem2.type === sas.analytics.riskhp.models.DisplayRule.DATA_ITEM_TYPE_COL) {
                valid = self.isValidColumn(cubeData, rule.dataItem2.column);
            }
            if (valid &&
                rule.appliesTo.type === sas.analytics.riskhp.models.DisplayRule.APPLIES_TO_TYPE_COL) {
                valid = self.isValidColumn(cubeData, rule.appliesTo.column);
            }
            return valid;
        },

        validateDisplayRules: function(cubeData) {
            var self = this;
            dataSourceDisplayRulesMap[cubeData.dataSource.id] = [];
            _.each(displayRules, function(rule) {
                var isValid = self.isValidDisplayRule(cubeData, rule);
                dataSourceDisplayRulesMap[cubeData.dataSource.id].push({
                    ruleId: rule.id,
                    valid: isValid
                })
            });

            // send bus message to inform everyone that display rules have been updated for the dataSource
            sas.eventBus.publish(HprConstants.CHANNEL_EXPLORATION, HprConstants.EVENT_DISPLAY_RULES_VALIDATED, {
                dataSourceId: cubeData.dataSource.id
            });
        }

    };
}());

DisplayRulesManager = sas.analytics.riskhp.models.DisplayRulesManager;