// (c) 2018, SAS Institute Inc.

jQuery.sap.declare("sas.analytics.riskhp.models.cache.CacheManager");

jQuery.sap.require("sas.analytics.riskhp.models.cache.CacheKey");
jQuery.sap.require("sas.analytics.riskhp.models.cache.ResultsCache");

/**
 * @name sas.analytics.riskhp.models.cache.CacheManager
 */
sas.analytics.riskhp.models.cache.CacheManager = (function () {
    "use strict";

    var _cacheMap = {}; // {cacheId: {cache: cache, clients: [clientID, ... ]}, ..., cacheIdX: {}}
    var _clientMap = {}; // {clientId: cacheId, ..., clientIdN: cacheId}

    // Keep track of outstanding cache action requests.  We will use this
    // to prevent multiple midtier queries by allowing a request to the midtier
    // to complete and populate the cache before another request is serviced.
    var _deferredCacheRequests = [];

    // Add cache to cache map
    function _addCache(cacheKey, clientId) {
        var deferred = $.Deferred();

        var cacheId = cacheKey.hashCode();
        var dataSourceId = cacheKey.getDataSourceId();

        // Make sure we have the data source loaded
        var deferredDataSource = DataSourceCacheManager.getDataSource(dataSourceId);
        deferredDataSource.done(function(cube) {
            // Create the cache and set the cube data model
            var cache = new sas.analytics.riskhp.models.cache.ResultsCache(cacheKey);
            cache.setCubeDataModel(cube);

            // Add clientId if given
            var clients = [];
            if (clientId) {
                clients.push(clientId);
            }

            // Add entry into cache map
            _cacheMap[cacheId] = {cache: cache, clients: clients};

            _clientMap[clientId] = cacheId;

            deferred.resolve(true);
            })
            .fail(function (jqxhr, textStatus, error) {
                // FIXME what to do here?
                //jQuery.sap.log.error('Request Failed. Response text: ' + jqxhr.responseText);
                deferred.reject(jqxhr);
            });
        return deferred;
    };

    // Remove cache from cache map
    function _removeCache(cacheId) {
        delete _cacheMap[cacheId];
    };

    // Add client to cache entry
    function _addClientToCache(cacheKey, clientId) {
        var deferred = $.Deferred();
        var cacheId = cacheKey.hashCode();
        var cacheEntry = _cacheMap[cacheId];
        if (cacheEntry) {
            var clients = cacheEntry.clients;
            var found = _.contains(clients, clientId);
            // If not already in client list, then add.  NO DUPS
            if (!found) {
                clients.push(clientId);
            }
            _clientMap[clientId] = cacheId;
            return deferred.resolve(true);
        }
        else {
            return _addCache(cacheKey, clientId);
        }
    };

    // Remove client from the cache entry
    function _removeClientFromCache(cacheId, clientId) {
        var cacheEntry = _cacheMap[cacheId];
        if (cacheEntry) {
            var clients = cacheEntry.clients;
            cacheEntry.clients = _.without(clients, clientId);
            // If no clients after this remove, remove the cache
            if (cacheEntry.clients.length === 0) {
                _removeCache(cacheId);
            }
        }
        delete _clientMap[clientId];
    };

    // Add deferred to pending list.
    function _addDeferredCacheRequests (deferred) {
        _deferredCacheRequests.push(deferred);
    };

    // Remove deferred from pending list.
    function _removeDeferredCacheRequests (deferred) {
        var index = _.indexOf(_deferredCacheRequests, deferred);
        _deferredCacheRequests.splice(index, 1);
    };


    return {

        /**
         * Register a client to use a cache.  If the cache does not exist it will be created
         * and the client id is stored to keep track of cache use.
         *
         * @param clientId
         * @param cacheKey  CacheKey
         * @return deferred
         */
        register: function (cacheKey, clientId) {
            var self = this;

            var cacheId = cacheKey.hashCode();

            // Does clientId already have entry for cache?
            var foundCacheId = _clientMap[clientId];
            if (foundCacheId) {
                // Does the existing cache entry for this client match this new requested cache?
                if (cacheId === foundCacheId) {
                    // Ensure cache has reference to this client
                    return _addClientToCache(cacheKey, clientId);
                }
                else {
                    // A different cache
                    _removeClientFromCache(foundCacheId, clientId);
                    return _addClientToCache(cacheKey, clientId);
                }
            }
            else {
                var cacheEntry = _cacheMap[cacheId];

                // If request cache exists, then add this client to map
                if (cacheEntry) {
                    return _addClientToCache(cacheKey, clientId);
                }
                // Cache miss, so create a new cache
                else {
                    return _addCache(cacheKey, clientId);
                }
            }
        },

        deregister: function (cacheKey, clientId) {
            var self = this;

            var cacheId = cacheKey.hashCode();

            _removeClientFromCache(cacheId, clientId);
        },

        /**
         * This function will return an array of data based on the expansion.
         * What data, is defined by the tuple(s).
         *
         * Note: this function will get the data from the tree, if present.  If expansion
         * not found or page not found, then the data is pulled from midtier/server.  The
         * tree is updated with this new data prior to being returned to caller.
         *
         * @param cacheKey = CacheKey object
         * @param clientId = clientId string
         * @param oExpansion = { key: [ccValue], startRow: startRow, maxRows: maxRows } or []
         * @param tuples = [ [dataItem, ... dataItem], ..., [dataItem, ... dataItem] ]
         * @returns deferred     results = [{expansion: oExpansion, data: rowsOfData},...]
         *                    rowsOfData = [{value: ccValue, data: [x,y,z]}, ... ]
         */
        getData : function (cacheKey, clientId, expansionKey, tuples) {
            var self = this;
            var deferred = $.Deferred();

            // var _removeDeferredCacheRequests = function(deferred) {
            //     var index = _.indexOf(_deferredCacheRequests, deferred);
            //     _deferredCacheRequests.splice(index, 1);
            // };

            // Check client against any old cacheKey with this (potentially) new cacheKey
            self.register(cacheKey, clientId).done(function() {
                var cacheId = cacheKey.hashCode();
                var cacheEntry = _cacheMap[cacheId];

                // If request cache exists
                if (cacheEntry) {
                    var cache = _cacheMap[cacheId].cache;

                    // If there are outstanding deferred requests then wait until done.
                    if (_deferredCacheRequests.length > 0) {
                        // https://stackoverflow.com/questions/5627284/pass-in-an-array-of-deferreds-to-when
                        $.when.apply($, _deferredCacheRequests).done(function() {
                            _addDeferredCacheRequests(deferred);
                            cache.getData(expansionKey, tuples).done(function (results) {
                                _removeDeferredCacheRequests(deferred);
                                deferred.resolve(results)
                            }).fail(function (jqxhr) {
                                _removeDeferredCacheRequests(deferred);
                                deferred.reject(jqxhr);
                            });
                       })
                    }
                    // No deferreds so go right on and request data from cache.
                    else {
                        _addDeferredCacheRequests(deferred);
                        cache.getData(expansionKey, tuples).done(function (results) {
                            _removeDeferredCacheRequests(deferred);
                            deferred.resolve(results)
                        }).fail(function (jqxhr) {
                            _removeDeferredCacheRequests(deferred);
                            deferred.reject(jqxhr);
                        });
                    }
                }
                else {
                    // Should never get here if register did it's job
                    deferred.reject(false); //FIXME with something more informative
                }
            }).fail(function(jqxhr) {
                deferred.reject(jqxhr);
            });
            return deferred;
        },

        /**
         * Ask the cache to refresh the data contained within it.
         *
         * @param clientId
         * @param cacheKey  CacheKey
         * @param force  force refresh of given cache, even if client is not a client of given cache
         * @return deferred
         */
        refresh: function (cacheKey, clientId, force) {
            var self = this;
            var deferred = $.Deferred();
            var rejectMsg = "";
            var cacheId = cacheKey.hashCode();
            if (force === undefined)
                force = false;

            // Does clientId have entry for cache?
            var foundCacheId = _clientMap[clientId];
            if (foundCacheId || force) {
                // Does the existing cache entry for this client match this new requested cache?
                if (cacheId === foundCacheId || force) {
                    var cacheEntry = _cacheMap[cacheId];
                    if (cacheEntry) {
                        var clients = cacheEntry.clients;
                        var found = _.contains(clients, clientId);
                        // If clientId found in client list of this cache, then perform refresh
                        if (found || force) {
                            //return cacheEntry.cache.refresh();

                            // If there are outstanding deferred requests then wait until done.
                            if (_deferredCacheRequests.length > 0) {
                                // https://stackoverflow.com/questions/5627284/pass-in-an-array-of-deferreds-to-when
                                $.when.apply($, _deferredCacheRequests).done(function() {
                                    _addDeferredCacheRequests(deferred);
                                    cacheEntry.cache.refresh().done(function () {
                                        _removeDeferredCacheRequests(deferred);
                                        deferred.resolve()
                                    }).fail(function (jqxhr) {
                                        _removeDeferredCacheRequests();
                                        deferred.reject(jqxhr);
                                    });
                                })
                            }
                            // No deferreds so go right on and request data from cache.
                            else {
                                _addDeferredCacheRequests(deferred);
                                cacheEntry.cache.refresh().done(function () {
                                    _removeDeferredCacheRequests(deferred);
                                    deferred.resolve()
                                }).fail(function (jqxhr) {
                                    _removeDeferredCacheRequests(deferred);
                                    deferred.reject(jqxhr);
                                });
                            }
                            return deferred;
                        }
                        else {
                            rejectMsg = "cache not used by client";
                        }
                    }
                    else {
                        rejectMsg = "cache not recognized";
                    }
                }
                else {
                    rejectMsg = "client registered with different cache";
                }
            }
            else {
                rejectMsg = "client not recognized";
            }
            return deferred.reject(rejectMsg);
        },

        /**
         * Ask the cache(s) associated to a given datasource to refresh the data contained within them.
         *
         * @param dataSourceId
         * @return deferred
         */
        refreshRelatedCaches: function (dataSourceId) {
            var self = this;
            var refreshDeferred = $.Deferred();

            // Get list of cacheIds associated with a dataSource
            var getAssociatedCacheIds = function() {
                var cacheIds = Object.keys(_cacheMap);
                var ids = [];
                _.each(cacheIds, function (cacheId) {
                    var cacheEntry = _cacheMap[cacheId];
                    if (cacheEntry && cacheEntry.cache.getDataSourceId() === dataSourceId) {
                        ids.push(cacheId);
                    }
                });
                return ids;
            };

            // This function is recursive.
            var refreshAssociatedCache = function(idList) {
                var deferred = $.Deferred();
                if (idList.length < 1) {
                    return deferred.resolve();
                }
                var cacheEntry = _cacheMap[idList.pop()];
                cacheEntry.cache.refresh().done(function () {
                    refreshAssociatedCache(idList).done(function () {
                        deferred.resolve();
                    })
                }).fail(function (jqxhr) {
                    deferred.resolve(); // Ignore.  (ok?)
                });
                return deferred;
            };

            // If there are outstanding deferred cache requests then wait until done.
            if (_deferredCacheRequests.length > 0) {
                // https://stackoverflow.com/questions/5627284/pass-in-an-array-of-deferreds-to-when
                $.when.apply($, _deferredCacheRequests).done(function() {
                    _addDeferredCacheRequests(refreshDeferred);
                    refreshAssociatedCache(getAssociatedCacheIds()).done(function () {
                        _removeDeferredCacheRequests(refreshDeferred);
                        refreshDeferred.resolve();
                    });
                })
            }
            // No outstanding deferred cache requests so go right on and do cache action.
            else {
                _addDeferredCacheRequests(refreshDeferred);
                refreshAssociatedCache(getAssociatedCacheIds()).done(function () {
                    _removeDeferredCacheRequests(refreshDeferred);
                    refreshDeferred.resolve();
                });
            }
            return refreshDeferred;
        },

        /**
         * Invalidate the cache(s) associated to a given datasource.  This will force the caches to
         * refresh the data contained within them next time data is requested.
         *
         * @param dataSourceId
         * @return deferred
         */
        invalidate: function (dataSourceId) {
            var self = this;
            var cacheIds = Object.keys(_cacheMap);
            var ids = [];
            _.each(cacheIds, function (cacheId) {
                var cacheEntry = _cacheMap[cacheId];
                if (cacheEntry && cacheEntry.cache.getDataSourceId() === dataSourceId) {
                    cacheEntry.cache.setRefreshNeeded(true);
                }
            });
        },

    }
}());

CacheManager = sas.analytics.riskhp.models.cache.CacheManager;