// (c) 2017, SAS Institute Inc.
jQuery.sap.declare("sas.analytics.riskhp.utils.ExplorationUtil");

jQuery.sap.require("sas.analytics.riskhp.models.DataSourceCacheManager");
jQuery.sap.require("sas.analytics.riskhp.models.HPRiskModelManager");
jQuery.sap.require("sas.analytics.riskhp.utils.HprConstants");
jQuery.sap.require("sas.analytics.riskhp.utils.HprUtil");

/**
 * @name sas.analytics.riskhp.util.ExplorationUtil
 */
sas.analytics.riskhp.utils.ExplorationUtil = {
    PANEL_CONTENT_ID: "mainContentData",

    contentPaneTest: new RegExp( [
        HprConstants.HPR_CROSSTAB_COMPONENT_ID,
        HprConstants.HPR_PLOT_COMPONENT_ID,
        HprConstants.HPR_RISK_FACTOR_SENSITIVITIES_COMPONENT_ID,
        HprConstants.HPR_PORTFOLIO_DETAILS_COMPONENT_ID
    ].join('|')),

    /**
     * Helper function to determine if the object passed in represents the json for a content pane
     * @param object
     * @returns {boolean}
     * @private
     */
    _isContentPaneObject: function(object) {
        if (!object) {
            return false;
        }
        var name = object.name;
        if (!name) {
            return false;
        }
        return this.contentPaneTest.test(name);
    },

    /**
     * Returns an array of objects representing the set of crosstabs in the exploration associated with stress
     * data sources. The objects in the returned array contain properties for crosstab ID (id), crosstab name
     * (name), the name of the page the crosstab is on (pageName), and the data source name (dataSourceName).
     *
     * @param explorationView
     * @returns {Array}
     */
    getStressCrosstabs: function(explorationView) {
        var self = this;

        var pageCrosstabMaps = this.buildPagePanelContentMaps(explorationView, sas.analytics.riskhp.utils.HprConstants.HPR_CROSSTAB_PANEL);

        var stressCrosstabs = [];
        _.each(pageCrosstabMaps, function (pageCrosstabMap) {
            _.each(Object.keys(pageCrosstabMap.panelMap), function (crosstabId) {
                var crosstab = pageCrosstabMap.panelMap[crosstabId];
                var dataSourceParam = sas.analytics.riskhp.models.DataSourceCacheManager.getDataSourceParam(crosstab.dataSourceId);
                if (dataSourceParam.type === sas.analytics.riskhp.utils.HprConstants.HPR_CUBE_INFORMATION_TYPE_STRESS) {
                    stressCrosstabs.push({
                        id: crosstab.name,
                        name: crosstab.options.name,
                        pageId: pageCrosstabMap.id,
                        pageName: pageCrosstabMap.name,
                        dataSourceId: dataSourceParam.realDataSourceId,
                        dataSourceName: dataSourceParam.label // NOTE: This is already the correct display label
                    });
                }
            });
        });
        return stressCrosstabs;
    },


    getCrosstabsUsingDataSource: function(explorationView, dataSource) {
        var self = this;

        var pageCrosstabMaps = this.buildPagePanelContentMaps(explorationView, sas.analytics.riskhp.utils.HprConstants.HPR_CROSSTAB_PANEL);

        var sameDataSourceCrosstabs = [];
        _.each(pageCrosstabMaps, function (pageCrosstabMap) {
            _.each(Object.keys(pageCrosstabMap.panelMap), function (crosstabId) {
                var crosstab = pageCrosstabMap.panelMap[crosstabId];
                if (crosstab.dataSourceId === dataSource.id) {
                    sameDataSourceCrosstabs.push({
                        id: crosstab.name,
                        pageId: pageCrosstabMap.id,
                        pageName: pageCrosstabMap.name
                    });
                }
            });
        });

        return sameDataSourceCrosstabs;
    },

    /**
     * Returns an array of objects representing the set of crosstabs in the exploration associated with the same
     * data source. The objects in the returned array contain properties for crosstab ID (id), the id of
     * the page the crosstab is on (pageId), and the name of the page the crosstab is on (pageName).
     *
     * @param explorationView
     * @returns {Array}
     */
    getCrosstabsUsingSameDataSource: function(explorationView) {
        var self = this;

        var pageCrosstabMaps = this.buildPagePanelContentMaps(explorationView, sas.analytics.riskhp.utils.HprConstants.HPR_CROSSTAB_PANEL);

        var sameDataSourceCrosstabs = [];
        _.each(pageCrosstabMaps, function (pageCrosstabMap) {
                _.each(Object.keys(pageCrosstabMap.panelMap), function (crosstabId) {
                    var crosstab = pageCrosstabMap.panelMap[crosstabId];
                    var selectedDataSourceId = sas.analytics.riskhp.models.DataSourceCacheManager.getSelectedDataSourceId();
                    if (crosstab.dataSourceId === selectedDataSourceId) {
                        sameDataSourceCrosstabs.push({
                            id: crosstab.name,
                            pageId: pageCrosstabMap.id,
                            pageName: pageCrosstabMap.name
                        });
                    }
                });
        });
        return sameDataSourceCrosstabs;
    },

    getAllCrosstabs: function(explorationView) {
        var self = this;

        var pageCrosstabMaps = this.buildPagePanelContentMaps(explorationView, sas.analytics.riskhp.utils.HprConstants.HPR_CROSSTAB_PANEL);

        var crosstabs = [];
        _.each(pageCrosstabMaps, function (pageCrosstabMap) {
            _.each(Object.keys(pageCrosstabMap.panelMap), function (crosstabId) {
                var crosstab = pageCrosstabMap.panelMap[crosstabId];

                    crosstabs.push({
                        id: crosstab.name,
                        pageId: pageCrosstabMap.id,
                        pageName: pageCrosstabMap.name
                    });

            });
        });
        return crosstabs;
    },

    /**
     * Build an array of objects corresponding to pages which contain a panel map. The panel map is an object containing
     * the panel metadata keyed by panel ID. If no type is specified, all page panels are included in the map, otherwise
     * only panels of that type are included.
     *
     * @param explorationView
     * @param type
     * @returns {Array}
     */
    buildPagePanelContentMaps: function(explorationView, type) {
        var self = this;

        var analysisObject = sas.analytics.riskhp.models.HPRiskModelManager.generateExplorationAnalysisJSON(explorationView);

        var pagePanelContentMaps = [];
        // Loop through each page looking for the page content
        _.each(analysisObject.layout.pages, function(page) {
            // Make sure page and page.content exists (they don't for ZeroState pages)
            if (page && page.content) {
                var panelMapObject = page.content.panelContentData;
                // If there is a 'type' specified, just use those types of controls in the panel content
                if (type) {
                    var panelContentData = page.content.panelContentData;
                    _.each(panelContentData, function(panelData) {
                        if (panelData.type !== type) {
                            // Since it is not needed, just delete the control from the original object
                            delete panelMapObject[panelData.name];
                        }
                    });
                }

                var pagePanelContentMap = {
                    id: page.id,
                    name: page.name,
                    panelMap: panelMapObject
                };

                pagePanelContentMaps.push(pagePanelContentMap);
            }
        });

        return pagePanelContentMaps;
    },

    /**
     * Takes an exploration object and retrieves a hash of all the content metadata for all panels across all pages
     * @param exploration
     * @returns {{}}
     */
    buildPanelContentMap: function(exploration) {
        var panelContentMap = {};

        var pages = exploration.layout.pages;
        _.each(pages, function(page) {
            // Make sure the page has content and panelContentData in order to pull the keys
            var keys = (page.content && page.content.panelContentData) ? _.keys(page.content.panelContentData) : [];
            _.each(keys, function(key) {
                panelContentMap[key] = sas.analytics.riskhp.utils.HprUtil.hprClone(page.content.panelContentData[key]);
            });
        });

        return panelContentMap;
    },

    /**
     * Returns the data source param JSON stored in the exploration analysis.
     * @param explorationAnalysis
     */
    getDataSourceParams: function(explorationAnalysis) {
        if (!explorationAnalysis) {
            return null;
        }
        return explorationAnalysis.dataSourceParams;
    },

    /**
     * Returns the Data Pane JSON stored in the exploration analysis.
     * @param explorationAnalysis
     */
    getDataPane: function(explorationAnalysis) {
        if (!explorationAnalysis) {
            return null;
        }
        return explorationAnalysis.layout.dataPane;
    },

    /**
     * Generates the data sources part of the exploration from the analysis object data source
     * params. Data sources passed on the exploration are a representation of the cubes used in
     * the exploration and are used for filtering out explorations from the welcome screen if
     * they contain cubes the user is not allowed to see.
     *
     * @param analysisObject
     * @returns {*}
     */
    generateDataSourceCubesFromAnalysisObject: function(analysisObject) {
        var cubeMap = {};

        _.each(analysisObject.dataSourceParams, function(dataSourceParam) {
            // we are reducing to the cube set, so we only need one entry for a cube having multiple data sources (i.e. price/stress)
            if (cubeMap[dataSourceParam.abstractDataSourceId]) {
                return;
            }

            // add the cube data for the data source to the map
            cubeMap[dataSourceParam.abstractDataSourceId] = {
                name: dataSourceParam.abstractDataSourceId,
                application: dataSourceParam.application,
                workGroup: dataSourceParam.workGroup,
                relPath: dataSourceParam.relPath
            };
        });

        return _.values(cubeMap);
    }
};