// (c) 2015, SAS Institute Inc.

jQuery.sap.declare("sas.analytics.riskhp.utils.HprAPI");

jQuery.sap.require("sas.analytics.riskhp.models.DataSource");
jQuery.sap.require("sas.analytics.riskhp.models.DataSourceCacheManager");
jQuery.sap.require("sas.analytics.riskhp.models.DataSourceParams");
jQuery.sap.require('sas.analytics.riskhp.models.HPRiskModelManager');
jQuery.sap.require('sas.analytics.riskhp.models.DisplayRulesManager');
jQuery.sap.require("sas.analytics.riskhp.utils.HprConstants");
jQuery.sap.require("sas.analytics.riskhp.utils.HprUtil");
jQuery.sap.require("sas.analytics.riskhp.views.computedvariableeditor.ComputedVariableUtil");
jQuery.sap.require('sas.analytics.riskhp.models.ComputedVariablesManager');


/**
 * Collection of methods to interface with the midtier
 *
 * Design Guidelines:
 * - If you are going to make an AJAX call to a midtier endpoint and the result of that call
 * is to populate or update some model then you should handle the model update in the AJAX promise's done
 * method and wrap the AJAX call's promise in a Deferred object and return the Deferred's promise object.
 * - If a method is going to return a Deferred's promise, document it as returning a "restricted" promise.
 * That way the caller knows they cannot resolve/reject the return as it will be handled internally in the method,
 * but they can tack on their own done/fail callbacks as needed.
 * - If you are going to make an AJAX call to the midtier and return the results without modification, consider
 * if you can just send the jqxhr object from the $.ajax call since it implements the Promise interface anyway.
 *
 * @name sas.analytics.riskhp.utils.HprAPI
 */
sas.analytics.riskhp.utils.HprAPI = {

    CompVarUtil: sas.analytics.riskhp.views.computedvariableeditor.ComputedVariableUtil,


/**
     * @memberOf HprAPI
     */
    GUID_PARAM_STRING : "guid=" + new Date().getTime(),

    _userName : '',
    _rootLinks: [],

    //*****************************************************************************************************
    //*****************************************************************************************************
    // Exploration functions
    //*****************************************************************************************************
    //*****************************************************************************************************
    /**
     * Returns a list of the explorations.
     *
     * COMMAND - GET …/SASHighPerformanceRisk/rest/explorations
     * @returns {*} - a restricted promise
     */
    getExplorations: function() {
        var api = this;
        var deferred = $.Deferred();
        var link = this.findRequiredLink(deferred, api.getRootLinks(), {rel:'explorations'});
        if (link) {
            // TODO: Could this just return the ajax promise object?
			$.getJSON(link.href, function (data) {
                deferred.resolve(data);
            }).fail(function(jqxhr) {
                jQuery.sap.log.error('Request Failed. Response text: ' + jqxhr.responseText);
                deferred.reject(jqxhr);
            });
        }

        return deferred.promise();
    },

    /**
     * Load a new exploration by retrieving the correct Data Source to use based on the 'type' selected.
     *
     * @param dataSourceDetails The dataSourceDetails object (which consists of a DataSourceParams object + multiple types of supported dataSources) needed to load the correct DataSource
     * @param selectedDataSourceType The selected type of DataSource to open
     * @returns {*} - a restricted promise
     */
    loadNewExploration: function(dataSourceDetails, selectedDataSourceType) {
        // For a new Exploration, we already have the DataSource details and a 'type' of DataSource selected.
        // So all we have to do is get the DataSource from the DataSourceCacheManager
        return sas.analytics.riskhp.utils.HprAPI.getDataSource(dataSourceDetails, selectedDataSourceType);
    },

    /**
     * Load a saved exploration using the following logic.
     *
     * When we go to open a saved RE, that is really just fetching it from Metadata and getting the saved off JSON.
     * The UI will then take this and fire off the actual open of the specified cube or data source. The JSON represents the crosstab’s query
     * and there will be some for plots as well.  After opening the cube, this is all processed to then bring the UI back up to its originally saved state.
     *
     * @param explorationId
     * @param bLoadExplorationOnly
     * @returns {*} - a restricted promise
     */
    loadSavedExploration: function(explorationId, bLoadExplorationOnly) {
        // In some cases the explorationId might contain '/Transformation', strip that part off and just use the metadata id
        if (explorationId.indexOf('/') > -1) {
            explorationId = explorationId.split('/')[0];
        }

        return sas.analytics.riskhp.utils.HprAPI.loadExploration(explorationId, bLoadExplorationOnly);
    },

    /**
     * Loads the saved Exploration corresponding to 'explorationId'.  This includes recalling the
     * saved exploration data and loading the correct Cube.  Returns the exploration content (JSON in our case).
     *
     * COMMAND - GET …/SASHighPerformanceRisk/rest/explorations/{explorationId}
     * @param explorationId id of the saved Exploration which should be loaded
     * @param bLoadExplorationOnly
     * @returns {*} - a restricted promise
     */
    loadExploration: function(explorationId, bLoadExplorationOnly) {
        var deferred = $.Deferred();

        var get = $.ajax({
            headers: {
                "Accept": "application/vnd.sas.risk.exploration+json;charset=utf-8"
            },
            contentType: "application/json;charset=utf-8",
            method: "GET",
            url: HprConstants.HPR_BASE_EXPLORATIONS_URL + '/' + explorationId,
            dataType: "json"
        });
        get.done(function (data) {
            // resolve just the exploration data if the exploration only flag is set
            if (bLoadExplorationOnly === true) {
                deferred.resolve(data);
            }
            else {
                // After loading the Exploration, parse the data to create an Exploration object
                var explorationData = sas.analytics.riskhp.utils.HprAPI._parseExplorationData(data, explorationId);

                // Gather the selected DataSourceParam object which holds the data required to load the Cube
                var dataSourceParams = DataSourceCacheManager.getSelectedDataSourceParam();

                // Open the cube associated with the Exploration and resolve the deferred
                var promise = sas.analytics.riskhp.utils.HprAPI.loadDataSource(dataSourceParams);
                promise.done(function () {
                    deferred.resolve(explorationData);
                });
                // Because we wrapped the promise we must add a failure handler as well so the fail case of the
                // promise gets propagated to the deferred's promise returned by this method.
                promise.fail(function (jqxhr, textStatus, error) {
                    deferred.reject(jqxhr, textStatus, error);
                });
            }
        });
        get.fail(function (jqxhr, textStatus, error) {
            deferred.reject(jqxhr, textStatus, error);
        });

        return deferred.promise();
    },

    /**
     * Create a new Risk Exploration
     *
     * COMMAND - POST …/SASHighPerformanceRisk/rest/explorations/create
     * @param explorationObject
     * @returns {*}
     */
    createExploration: function(explorationObject) {
        var deferred = $.Deferred();

        // Midtier expects analysis JSON to be a string
        explorationObject.analysis = JSON.stringify(explorationObject.analysis);

        // jQuery cannot encode the Content-Type for parms so we have to stringify the entire data payload
        var stringifiedData = JSON.stringify(explorationObject);

        // TODO: Determine if this should return a restricted promise
        var promise = jQuery.ajax({
            headers: {
                "Accept": "application/vnd.sas.risk.exploration.summary+json;charset=utf-8",
                "Content-Type": "application/vnd.sas.risk.exploration.parms+json;charset=utf-8"
            },
            method: "POST",
            url: HprConstants.HPR_BASE_EXPLORATIONS_URL + "/create",
            data: stringifiedData,
            dataType: "json"
        });

        promise.done(function (data) {
            deferred.resolve(data);
        });
        promise.fail(function (jqxhr) { //textStatus, error
            jQuery.sap.log.error("Request Failed. Response text: " + jqxhr.responseText);
            deferred.reject(jqxhr);
        });

        return deferred;
    },

    /**
     * Delete a list of Risk Explorations
     *
     * COMMAND - POST …/SASHighPerformanceRisk/rest/explorations/delete
     * @param explorationObjects An array of Exploration objects to delete; Object contains 'id' and 'name' properties for the Explorations
     * @returns {*}
     */
    deleteExplorations: function(explorationObjects) {
        var deferred = $.Deferred();

        // Chain together some methods to build up an array of Exploration Ids for the mid-tier
        var explorationIds = _.chain(explorationObjects)
            .pluck("id")
            .map(function(expId) { return {id: expId}; })
            .value();
        var stringifiedData = JSON.stringify(explorationIds);

        $.ajax({
            headers: {
                "Content-Type": "application/json",
                "Content-Item": "application/vnd.sas.risk.exploration.parms+json;charset=utf-8",
                "Accept": "application/json;charset=utf-8",
                "Accept-Item": "application/vnd.sas.risk.exploration+json;charset=utf-8"
            },
            method: "POST",
            url: HprConstants.HPR_BASE_EXPLORATIONS_URL + "/delete",
            data: stringifiedData,
            dataType: "json"
        }).done(function(deletedExplorationsObject) {
            // Returns an Object with some data about the successfully deleted Explorations;
            // Populate with empty Object if server doesn't send anything back
            if (!deletedExplorationsObject) {
                deletedExplorationsObject = {};
                deletedExplorationsObject.items = [];
            }

            // If the number of deleted Explorations is not the same as the number passed in, notify the user
            if (deletedExplorationsObject.items.length < explorationIds.length) {
                // Get pass/fail counts
                var failCount = explorationIds.length - deletedExplorationsObject.items.length;
                var expCount = explorationIds.length;

                // Get the ids of the Explorations that failed to delete
                var failedExplorationIds = _.difference(explorationIds, deletedExplorationsObject.items);

                // Error messaging
                var dialogText = null;
                var detailText = null;
                if (failCount === 1) {
                    dialogText = sas.analytics.riskhp.utils.HprUtil.getBundle().getText("hprisk.delete.exploration.error.single.fmt", [explorationObjects[0].name]);
                } else {
                    dialogText = sas.analytics.riskhp.utils.HprUtil.getBundle().getText("hprisk.delete.exploration.error.multi.fmt", [failCount, expCount]);

                    // Build the detail text for display in the error dialog (shows the explorations that failed to be deleted)
                    var explorationListText = "";
                    _.each(failedExplorationIds, function(explorationId) {
                        if (explorationListText.length > 0) {
                            explorationListText += "\n";
                        }
                        // Find the specific Exploration in the original Exploration objects passed in and pull the name for display
                        explorationListText += "\t" + _.findWhere(explorationObjects, {id: explorationId.id}).name;
                    });
                    detailText = sas.analytics.riskhp.utils.HprUtil.getBundle().getText("hprisk.delete.exploration.error.detail.txt") + "\n" + explorationListText;
                }

                // Show the error
                sas.hc.m.MessageDialog.error({
                    text: dialogText,
                    title: sas.analytics.riskhp.utils.HprUtil.getBundle().getText("error.window.title"),
                    detailText: detailText ? detailText : null
                });
            }

            deferred.resolve(); // Don't return anything in deferred
        }).fail(function (jqxhr) { //textStatus, error
            jQuery.sap.log.error("Request Failed. Response text: " + jqxhr.responseText);
            deferred.reject(jqxhr);
        });

        return deferred;
    },

    /**
     * Update a Risk Exploration
     *
     * COMMAND - POST /rest/explorations/update
     * @param explorationId - The id of the exploration to be updated
     * @param explorationObject - The object/data representing the Exploration
     * @returns {*}
     */
    updateExploration: function(explorationId, explorationObject) {
        var deferred = $.Deferred();

        // Midtier expects analysis JSON to be a string
        explorationObject.analysis = JSON.stringify(explorationObject.analysis);

        // jQuery cannot encode the Content-Type for parms so we have to stringify the entire data payload
        var stringifiedData = JSON.stringify(explorationObject);

        // TODO: Determine if this should return a restricted promise
        $.ajax({
            headers: {
                "Accept": "application/vnd.sas.risk.exploration.summary+json;charset=utf-8",
                "Content-Type": "application/vnd.sas.risk.exploration.parms+json;charset=utf-8"
            },
            method: "POST",
            url: HprConstants.HPR_BASE_EXPLORATIONS_URL + '/update',
            data: stringifiedData,
            dataType: "json"
        }).done(function () {
            deferred.resolve();
        }).fail(function (jqxhr) { // textStatus, error
            jQuery.sap.log.error("Request Failed. Response text: " + jqxhr.responseText);
            deferred.reject(jqxhr);
        });

        return deferred;
    },

    /**
     * Parses the data and creates an exploration object which will be used in the UI.
     *
     * @param explorationData
     * @param explorationId
     * @returns exploration object
     * @private
     */
    _parseExplorationData: function(explorationData, explorationId) {
        return sas.analytics.riskhp.models.HPRiskModelManager.createExplorationObject(explorationData, explorationId);
    },

    //*****************************************************************************************************
    //*****************************************************************************************************
    // Cube functions
    //*****************************************************************************************************
    //*****************************************************************************************************
    /**
     * Returns a list of the cubes
     *
     * COMMAND - GET …/SASHighPerformanceRisk/rest/cubes
     * @returns {*}
     */
    getCubes: function() {
        var api = this;
        var deferred = $.Deferred();
        var link = this.findRequiredLink(deferred, api.getRootLinks(), {rel:'cubes'});
        if (link) {
            // TODO: Could this just return the ajax promise object?
            $.getJSON(link.href, function (data) {
                deferred.resolve(data.items);
            }).fail(function(jqxhr) {
                jQuery.sap.log.error('Request Failed. Response text: ' + jqxhr.responseText);
                deferred.reject(jqxhr);
            });
        }

        return deferred.promise();
    },

    /**
     * Get the detailed information for a DataSource.  One of the most important parts of information returned is the array of 'dataSources'.  They symbolize the different
     * ways that a particular DataSource can be opened.
     *
     * COMMAND - GET …/SASHighPerformanceRisk/rest/cubes/{dataSourceId}?application=...&workGroup=...&relPath=...
     * @param rootDataSourceParams An Object which holds the required parameters (application, workGroup, and relPath) which should be encoded
     * @returns {*} - a restricted promise
     */
    getDataSourceDetails: function(rootDataSourceParams) {
        var deferred = $.Deferred();

        var get = $.ajax({
            headers: {
                "Accept": "application/vnd.sas.risk.cube.summary+json;charset=utf-8"
            },
            contentType: "application/json;charset=utf-8",
            method: "GET",
            url: HprConstants.HPR_BASE_CUBES_URL + '/' + encodeURIComponent(rootDataSourceParams.abstractDataSourceId) + '?' + rootDataSourceParams.getQuery() + "&" + this.GUID_PARAM_STRING,
            dataType: "json"
        });
        get.done(function (dataSourceDetails) {
            //FIXME (temporary) remove the data source for portfolio until it is supported
            dataSourceDetails.dataSources =_.filter(dataSourceDetails.dataSources, function(dataSource) {
                return dataSource.type === HprConstants.HPR_CUBE_INFORMATION_TYPE_PRICE || dataSource.type === HprConstants.HPR_CUBE_INFORMATION_TYPE_STRESS;
            });

            deferred.resolve(dataSourceDetails);
        });
        get.fail(function (jqxhr, textStatus, error) {
            deferred.reject(jqxhr, textStatus, error);
        });

        return deferred.promise();
    },

    /**
     * Load the data for a new data source based on the 'type' selected without going through the DataSourceCacheManager.
     * This is primarily used to open the DataSource and look at its data for comparison purposes.
     *
     * @param dataSourceDetails The dataSourceDetails object (which consists of a DataSourceParams object + multiple types of supported dataSources) needed to load the new DataSource
     * @param selectedDataSourceType The selected type of the new DataSource
     * @returns {*} - a restricted promise
     */
    previewDataSource: function(dataSourceDetails, selectedDataSourceType) {
        var deferred = $.Deferred();

        // If there are no dataSourceDetails or dataSources, then we need to bail
        if (!dataSourceDetails || (dataSourceDetails && !dataSourceDetails.dataSources) || !selectedDataSourceType) {
            deferred.reject();
            return deferred.promise();
        }

        // Group the dataSources for this dataSourceDetails by their type
        var groupDetailsByType = _.groupBy(dataSourceDetails.dataSources, function(dataSource) {
            return dataSource.type;
        });
        // Get the correct DataSource by using the selected type and create a new DataSourceParams object which is needed to
        // actually load the data for the DataSource
        var correctDataSource = groupDetailsByType[selectedDataSourceType][0]; // Should only be one of each type here
        if (correctDataSource) {
            var newDataSourceParams = new sas.analytics.riskhp.models.DataSourceParams();
            newDataSourceParams.abstractDataSourceId = dataSourceDetails.id;
            newDataSourceParams.application = dataSourceDetails.application;
            newDataSourceParams.workGroup = dataSourceDetails.workGroup;
            newDataSourceParams.relPath = dataSourceDetails.relPath;
            newDataSourceParams.type = correctDataSource.type;
            newDataSourceParams.realDataSourceId = correctDataSource.id;
            newDataSourceParams.label = sas.analytics.riskhp.utils.HprUtil.getCubeDisplayLabel(correctDataSource.label, dataSourceDetails.name, correctDataSource.type);

            // Call the method to open the data source and capture it's deferred object
            var promise = sas.analytics.riskhp.utils.HprAPI.openDataSource(newDataSourceParams);

            // Attach a .done handler to be notified when the DataSource has been loaded
            promise.done(function(data) {
                // data is what we want to return here, don't need a JSONModel
                deferred.resolve(data);
            });

            // Create fail case which gets propagated to the caller
            promise.fail(function (jqxhr, textStatus, error) {
                deferred.reject(jqxhr, textStatus, error);
            });
        } else {
            deferred.reject();
        }

        return deferred.promise();
    },

    /**
     * Load a new data source based on the 'type' selected.  Take advantage of the DataSourceCacheManager so the new
     * DataSource is loaded and tracked via the cache manager code.
     *
     * @param dataSourceDetails The dataSourceDetails object (which consists of a DataSourceParams object + multiple types
     *                          of supported dataSources) needed to load the new DataSource
     * @param selectedDataSourceType The selected type of the new DataSource
     * @returns {*} - a restricted promise
     */
    getDataSource: function(dataSourceDetails, selectedDataSourceType) {
        var deferred = $.Deferred();

        // Since we already have the DataSource details and a 'type' of DataSource selected, all we have to do is get
        // the DataSource from the DataSourceCache
        var promise = sas.analytics.riskhp.utils.HprAPI._getDataSource(dataSourceDetails, false, selectedDataSourceType);
        promise.done(function(dataSource) {
            deferred.resolve(dataSource);
        });
        promise.fail(function(jqxhr, textStatus, error) {
            deferred.reject(jqxhr, textStatus, error);
        });

        return deferred.promise();
    },

    /**
     * Load the contents of a DataSource by performing a couple of steps.  The first step of the process is to retrieve (GET) the detailed information for the DataSource
     * represented by the 'rootDataSourceParams.abstractDataSourceId' property and using required parameters (i.e. application, workGroup, and relPath).
     * 'rootDataSourceParams.abstractDataSourceId' for this method should be the simple 'id' of the DataSource (i.e. HPRKitchenSink_AggJoin_HLKJR164).  The real 'id'
     * is retrieved in the results of the GET call.  The second step of the process is to take some of the valid info ('id' and 'type') for the DataSource and call
     * 'getDataSource' on the DataSourceCacheManager to actually open the DataSource to retrieve the contents using a 'POST' call.
     *
     * 1st COMMAND - GET …/SASHighPerformanceRisk/rest/cubes/{dataSourceId}?application=...&workGroup=...&relPath=...
     * 2nd COMMAND - POST …/SASHighPerformanceRisk/rest/cubes/{dataSourceId}?application=...&workGroup=...&relPath=...&type=...
     * @param rootDataSourceParams An Object which holds the required parameters (application, workGroup, and relPath) which should be encoded
     * @returns {*} - a restricted promise
     */
    loadDataSource: function(rootDataSourceParams) {
        var deferred = $.Deferred();

        var get = $.ajax({
            headers: {
                "Accept": "application/vnd.sas.risk.cube.summary+json;charset=utf-8"
            },
            contentType: "application/json;charset=utf-8",
            method: "GET",
            url: HprConstants.HPR_BASE_CUBES_URL + '/' + encodeURIComponent(rootDataSourceParams.abstractDataSourceId) + '?' + rootDataSourceParams.getQuery() + "&" + this.GUID_PARAM_STRING,
            dataType: "json"
        });
        get.done(function (dataSourceDetails) {
            // Check to see if we have any 'dataSources' returned.  Those are the available data sources for this DataSource (i.e. price, stress, portfolio)
            if (dataSourceDetails && dataSourceDetails.dataSources) {

                // TODO: Remove this snippet of code when 'portfolio' data source will be used
                // Remove the 'portfolio' data source because it is not used yet
                dataSourceDetails.dataSources = _.filter(dataSourceDetails.dataSources, function(dataSource) {
                    return dataSource.type === 'price' || dataSource.type === 'stress';
                });

                // Get the DataSource from the DataSourceCache
                var promise = sas.analytics.riskhp.utils.HprAPI._getDataSource(dataSourceDetails, true, null);
                promise.done(function(dataSource) {
                    deferred.resolve(dataSource);
                });
                promise.fail(function(jqxhr, textStatus, error) {
                    deferred.reject(jqxhr, textStatus, error);
                });
            } else {
                // TODO: What to do here?  Can this happen?
            }
        });
        get.fail(function (jqxhr, textStatus, error) {
            deferred.reject(jqxhr, textStatus, error);
        });

        return deferred.promise();
    },

    /**
     * This method runs through logic to determine which DataSource the DataSourceCacheManager should try and open.  A restricted promise is returned once
     * the DataSource has been opened.
     *
     * @param dataSourceDetails The dataSourceDetails object (which consists of a DataSourceParams object + multiple types of supported dataSources) needed to load the correct DataSource
     * @param isSavedExploration Whether we are working with a saved Exploration
     * @param selectedDataSourceType The selected type of DataSource to open
     * @returns {*} - a restricted promise
     * @private
     */
    _getDataSource: function(dataSourceDetails, isSavedExploration, selectedDataSourceType) {
        var deferred = $.Deferred();

        // Have the DataSourceCacheManager update the DataSourceParams objects it tracks
        DataSourceCacheManager.updateDataSourceParams(dataSourceDetails);

        // In some cases we will need to set the 'selectedDataSoourceId' on the DataSourceCacheManager
        var shouldSetSelectedDataSourceId = false;
        if (isSavedExploration) {
            // If we are working with a saved exploration, only set the selectedDataSourceId if there
            // is not a selectedDataSourceParam object
            var selectedDSP = DataSourceCacheManager.getSelectedDataSourceParam();
            if (!selectedDSP) {
                shouldSetSelectedDataSourceId = true;
            }
        }
        else {
            // Always do it for new explorations
            shouldSetSelectedDataSourceId = true;
        }

        // If we need too, determine which DataSource we should use
        if (shouldSetSelectedDataSourceId) {
            // Gather all the DataSourceParams object we can work with
            var availableDataSourceParams = DataSourceCacheManager.getMatchingDataSourceParamsUsingDetails(dataSourceDetails);
            var dataSourceParamToUse = availableDataSourceParams[0]; // Default to the first one
            if (isSavedExploration) {
                // If we have more than one possibility for a saved exploration, determine if we can use the Price version first, then Stress,
                // otherwise fall back to the default
                if (availableDataSourceParams.length > 1) {
                    // Group the dataSources by the 'type' property
                    var groupDataSourcesByType = _.groupBy(availableDataSourceParams, function (dataSourceParam) {
                        return dataSourceParam.type;
                    });

                    // If there is a Price dataSource, use it; Otherwise, use the Stress; If neither of those types exist, just fall
                    // back to the default first DataSource
                    if (groupDataSourcesByType.price) {
                        dataSourceParamToUse = groupDataSourcesByType.price[0];
                    } else if (groupDataSourcesByType.stress) {
                        dataSourceParamToUse = groupDataSourcesByType.stress[0];
                    }
                }
            } else {
                // For a new exploration, if the selected type was passed in, find the DataSourceParam object which matches it and use it
                if (selectedDataSourceType !== null) {
                    var newExplorationDataSourceParamToUse = _.find(availableDataSourceParams, function(availableDataSourceParam) {
                        return availableDataSourceParam.type === selectedDataSourceType;
                    }); // Can be undefined
                    if (newExplorationDataSourceParamToUse) {
                        dataSourceParamToUse = newExplorationDataSourceParamToUse;
                    }
                }
            }
            DataSourceCacheManager.setSelectedDataSourceId(dataSourceParamToUse.realDataSourceId);
        }

        // Retrieve the DataSource id which we should load, and do it.  Handle the returned object properly.
        var dataSourceIdToLoad = DataSourceCacheManager.getSelectedDataSourceId();
        var deferredOrModel = DataSourceCacheManager.getDataSource(dataSourceIdToLoad);
        if (_.isFunction(deferredOrModel.done)) {
            // We have a deferred here
            deferredOrModel.done(function(dataSource) {
                deferred.resolve(dataSource);
            });
            deferredOrModel.fail(function (jqxhr, textStatus, error) {
                deferred.reject(jqxhr, textStatus, error);
            });
        }
        else {
            // deferredOrModel is the cubeData's JSON Model here
            // TODO: What if fail here? Do we ever get here?
            deferred.resolve(deferredOrModel);  // We don't believe this ever gets called
        }

        return deferred.promise();
    },

    /**
     * Open the specified cube based on the 'id' passed in and the required parameters (i.e. application, workGroup, relPath, & type)
     * COMMAND - POST …/SASHighPerformanceRisk/rest/cubes/{dataSourceId}?application=...&workGroup=...&relPath=...&type=...
     *
     * <p><b>Ajax Properties:</b>
     * <ul>
     *     <li> restricted deferred </li>
     *     <li> resolve(cubeData)</li>
     *     <li> rejects</li>
     *     <li> calls {@link _addStressCubeArtifacts} when opening a stress cube</li>
     *     <li> calls {@link _addComputedVariableArtifacts} regardless of the type of cube</li>
     *  </ul>
     * @param dataSourceParams The DataSourceParams object we want to open
     * @returns {Deferred} The returned Deferred is restricted
     */
    openDataSource: function(dataSourceParams) {
        var deferred = $.Deferred();

        // without params, the data source cannot be opened
        if (!dataSourceParams) {
            deferred.reject();
            return deferred;
        }

        var stringifiedData = JSON.stringify({}); // Requires empty object
        var post = $.ajax({
            headers: {
                "Accept": "application/vnd.sas.risk.cube+json;charset=utf-8",
                "Content-Type": "application/json;charset=utf-8"
            },
            contentType: "application/json;charset=utf-8",
            method: "POST",
            url: HprConstants.HPR_BASE_CUBES_URL + '/' + encodeURIComponent(dataSourceParams.realDataSourceId) +
                '?' + dataSourceParams.getQuery() +
                '&' + this.GUID_PARAM_STRING +
                '&keepAlive=' + !HprConstants.ENABLE_DEBUG,
            data: stringifiedData,
            dataType: "json"
        });
        post.done(function (data) {
            // After loading the cube, parse the required data and resolve the deferred object
            var cubeData = new sas.analytics.riskhp.models.DataSource(data);


            if (cubeData.isStressCube()) {
                var promise = sas.analytics.riskhp.utils.HprAPI._addStressCubeArtifacts(cubeData,dataSourceParams.attachedScenarioIds);
                promise.always(function () {

                    // Create exploration computed variables on the server and add them to the datasource AFTER scenarios
                    // have been added and activated on the server since scenarios can be used in creating computed vars
                    var computedVarDeferred = sas.analytics.riskhp.utils.HprAPI._addComputedVariableArtifacts(cubeData, sas.analytics.riskhp.models.ComputedVariablesManager.getComputedVariables());
                    computedVarDeferred.done(function() {
                        DisplayRulesManager.validateDisplayRules(cubeData);

                        deferred.resolve(cubeData);
                    });
                    // No fail case for _addComputedVariableArtifacts as it will always resolve successfully
                });
                // No fail case for _addStressCubeArtifacts as it will always resolve successfully
            }
            else {
                var computedVarDeferred = sas.analytics.riskhp.utils.HprAPI._addComputedVariableArtifacts(cubeData, sas.analytics.riskhp.models.ComputedVariablesManager.getComputedVariables());
                computedVarDeferred.done(function() {
                    DisplayRulesManager.validateDisplayRules(cubeData);

                    deferred.resolve(cubeData);
                });
                // No fail case for _addComputedVariableArtifacts as it will always resolve successfully
            }
        });
        post.fail(function (jqxhr, textStatus, error) {
            jQuery.sap.log.error("Request Failed. Response text: " + jqxhr.responseText);
            deferred.reject(jqxhr, textStatus, error);
        });

        return deferred.promise();
    },


    /**
     * Update the cube with some values.  Eventually this REST call will be used to
     * access specific values for a class variable or portfolio column to construct a filter.
     *
     * COMMAND - POST …/SASHighPerformanceRisk/rest/cubes/{dataSourceId}/values
     * @param dataSourceId The id of the cube
     * @param arrayToStringify The array with required objects to stringify.
     * @param dataFilter
     * @returns {*}
     */
    updateCubeValues: function(dataSourceId, arrayToStringify, dataFilter) {
        var deferred = $.Deferred();
        // TODO: Determine if this should return a restricted promise
        var queryURL = HprConstants.HPR_BASE_CUBES_URL + '/' + encodeURIComponent(dataSourceId) + "/values" + "?" + this.GUID_PARAM_STRING;
        var stringifiedData = JSON.stringify(arrayToStringify);

        $.ajax({
            headers: {
                "Accept": "application/vnd.sas.collection+json;charset=utf-8",
                "Accept-Item": "application/vnd.sas.risk.values+json;charset=utf-8",
                "Content-Type": "application/json",
                "Content-Item": "application/vnd.sas.risk.values.parms+json;charset=utf-8"
            },
            contentType: "application/json;charset=utf-8",
            method: "POST",
            url: queryURL,
            data: stringifiedData,
            dataType: "json",
            dataFilter: dataFilter
        })
        .done(function (data) {
            deferred.resolve(data);
        })
        .fail(function (jqxhr) { // textStatus, error
            jQuery.sap.log.error('Request Failed. Response text: ' + jqxhr.responseText);
            deferred.reject(jqxhr);
        });
        return deferred;
    },

    /**
    * Called to notify the midtier/server it can remove the cache for
    * the values referenced by the specified class variable. This should
    * only be called for class variables that are high-cardinality (the
    * total value count exceeds the specified page size).
    * @param dataSourceId
    * @param classVariableId
    * @returns {*}
    */
    destroyCubeValuesCache: function(dataSourceId, classVariableId) {
       // var deferred = $.Deferred();

        var data = [{
            version: 1,
            type: "classVariable",
            id: classVariableId,
            clearCache: true
        }];

        return this.updateCubeValues(dataSourceId, data);
    },

    /**
     * Release the specified cube so resources on the Grid will free up.
     *
     * @param dataSourceId
     * COMMAND - DELETE .../SASHighPerformanceRisk/rest/cubes/{dataSourceId}
     */
    releaseCubeResources: function(dataSourceId) {
        var deferred = $.Deferred();
        // TODO: Determine if this should return a restricted promise
        // TODO: Could this just return the ajax promise object?
        var queryURL = HprConstants.HPR_BASE_CUBES_URL + '/' + encodeURIComponent(dataSourceId) + "?" + this.GUID_PARAM_STRING;

        $.ajax({
                method: "DELETE",
                url: queryURL
            })
            .done(function (data) {
                deferred.resolve(data);
            })
            .fail(function (jqxhr) { // textStatus, error
                jQuery.sap.log.error('Request Failed. Response text: ' + jqxhr.responseText);
                deferred.reject(jqxhr);
            });
        return deferred;
    },

    /**
     * Retrieve the SAS Log for the specified Data Source.
     *
     * COMMAND - GET …/SASHighPerformanceRisk/rest/cubes/{dataSourceId}/log
     * @param dataSourceId
     * @returns {*}
     */
    getSASLog: function(dataSourceId) {
        var deferred = $.Deferred();
        var queryURL = HprConstants.HPR_BASE_CUBES_URL + "/" + encodeURIComponent(dataSourceId) + "/log?" + this.GUID_PARAM_STRING;

        $.ajax({
            method: "GET",
            url: queryURL
        })
        .done(function (logData) {
            deferred.resolve(logData);
        })
        .fail(function (jqxhr) { // textStatus, error
            jQuery.sap.log.error("Request Failed. Response text: " + jqxhr.responseText);
            deferred.reject(jqxhr);
        });

        return deferred;
    },

    //*****************************************************************************************************
    //*****************************************************************************************************
    // MidTier keep alive functions
    //*****************************************************************************************************
    //*****************************************************************************************************
    cubeKeepAlive: function (cubeList) {
        //var self = this;
        var deferred = $.Deferred();

        var keepAliveURL = HprConstants.HPR_CUBE_KEEP_ALIVE_URL + "?" + this.GUID_PARAM_STRING;

        var stringifiedData = JSON.stringify(cubeList);

        // NOTE: dataType must be set to "text" because no data is being returned. dataType = "json" throws a
        // json parsing error.
        // https://stackoverflow.com/questions/37057711/unexpected-end-of-json-input-from-an-ajax-call
        var ajaxRequest = {
            headers: {
                "Content-Type": "application/json",
                "Content-Item": "application/vnd.sas.risk.cube.summary+json;charset=utf-8"
            },
            contentType: "application/json",
            method: "POST",
            url: keepAliveURL,
            data: stringifiedData,
            dataType: "text"
        };

        // Get the logger for this object and output stuff
        var logger = sas.log.getLogger("sas.analytics.riskhp.utils.HprAPI");
        logger.debug('HprAPI.cubeKeepAlive ' + keepAliveURL);
        logger.trace('HprAPI.cubeKeepAlive ' + JSON.stringify(ajaxRequest));

        $.ajax(ajaxRequest)
            .done(function (data, textStatus, jqxhr) {
                logger.debug('HprAPI.cubeKeepAlive DONE');
                deferred.resolve({status: jqxhr.status});
            })
            .fail(function (jqxhr, textStatus) { // error
                jQuery.sap.log.error("HprAPI.cubeKeepAlive  Request Failed. " + textStatus + " Response text: " + jqxhr.responseText);
                deferred.reject(jqxhr);
            });
        return deferred;
    },

    // FIXME this may/will be needed later.
    //*****************************************************************************************************
    //*****************************************************************************************************
    // Cube Polling functions
    //*****************************************************************************************************
    //*****************************************************************************************************
    // pollCubes: function (cubeList) {
    //     var self = this;
    //     var deferred = $.Deferred();
    //
    //     var queryURL = HprConstants.HPR_BASE_CUBES_URL;
    //
    //     var stringifiedData = JSON.stringify(cubeList);
    //
    //     var ajaxRequest = {
    //         headers: {
    //             "Accept": "application/vnd.sas.collection+json;charset=utf-8",
    //             "Accept-Item": "application/vnd.sas.risk.cube.summary+json;charset=utf-8",
    //             "Content-Type": "application/json",
    //             "Content-Item": "application/vnd.sas.risk.cube.summary+json;charset=utf-8"
    //         },
    //         contentType: "application/json",
    //         method: "POST",
    //         url: queryURL,
    //         data: stringifiedData,
    //         dataType: "json"
    //     };
    //
    //     // Get the logger for this object and output stuff
    //     var logger = sas.log.getLogger("sas.analytics.riskhp.utils.HprAPI");
    //     logger.debug('HprAPI.pollCubes ' + queryURL);
    //     logger.trace('HprAPI.pollCubes ' + JSON.stringify(ajaxRequest));
    //
    //     $.ajax(ajaxRequest)
    //         .done(function (data, textStatus, jqxhr) {
    //             if (data === undefined) {
    //                 // If no data returned, the create an object to carry the status code to caller
    //                 data = {};
    //             }
    //             data.status = jqxhr.status;
    //             logger.debug('HprAPI.pollCubes DONE');
    //             logger.trace('HprAPI.pollCubes DONE ' + JSON.stringify(data));
    //             deferred.resolve(data);
    //         })
    //         .fail(function (jqxhr, textStatus, error) {
    //             jQuery.sap.log.error('HprAPI.pollCubes  Request Failed. Response text: ' + jqxhr.responseText);
    //             deferred.reject(jqxhr);
    //         });
    //     return deferred;
    // },

    //*****************************************************************************************************
    //*****************************************************************************************************
    // Export functions
    //*****************************************************************************************************
    //*****************************************************************************************************
    exportPanel: function (exportData) {
        //var self = this;
        var deferred = $.Deferred();
        var HPRConstants = sas.analytics.riskhp.utils.HprConstants;
        var exportURL = HPRConstants.HPR_EXPORT_URL + "?" + this.GUID_PARAM_STRING;
        var stringifiedData = JSON.stringify([exportData]);

        var ajaxRequest = {
            headers: {
                "Accept": "application/json;charset=utf-8",
                "Content-Type": "application/json;charset=utf-8",
                "Content-Item": "application/vnd.sas.risk.query.parms+json;charset=utf-8"
            },
            contentType: "application/json;charset=utf-8",
            method: "POST",
            url: exportURL,
            data: stringifiedData,
            dataType: "text"
        };

        // Get the logger for this object and output stuff
        var logger = sas.log.getLogger("sas.analytics.riskhp.utils.HprAPI");
        logger.debug('HprAPI.exportPanel ' + exportURL);
        logger.trace('HprAPI.exportPanel ' + JSON.stringify(ajaxRequest));

        $.ajax(ajaxRequest)
            .done(function (url) { // textStatus, jqxhr
                var __handleDownloadFinished = function() {
                    // debug
                    logger.debug('HprAPI.exportPanel DONE');
                    logger.trace('HprAPI.exportPanel DONE ' + url);

                    deferred.resolve(url);
                };

                // Trigger the download. Both techniques below are used in order to avoid triggering
                // the unload event, which due to the handler set in HPRShell.controller.js, would
                // result in the release (close) of all opened data sources. See S1431155 for more
                // information.
                if (window.navigator.msSaveBlob) {
                    // For Microsoft browsers (even though Edge supports the download attribute
                    // of the anchor tag, it is not worth trying to distinguish it from IE), create
                    // an XMLHttpRequest to download the binary data into a Blob object. Then, use
                    // the msSaveBlob function to prompt the user to save the binary content as a
                    // file.
                    var xhr = new XMLHttpRequest();
                    xhr.open("GET", url);
                    xhr.responseType = "blob";
                    xhr.onload = function() {
                        // attempt to capture the file name returned by the server
                        var filename = null;
                        var filenameHeader = xhr.getResponseHeader(HPRConstants.HPR_EXPORT_FILENAME_HEADER_NAME);
                        if (filenameHeader) {
                            var filenameKeyIndex = filenameHeader.indexOf(HPRConstants.HPR_EXPORT_FILENAME_HEADER_KEY);
                            if (filenameKeyIndex !== -1) {
                                filename = filenameHeader.substring(filenameKeyIndex + HPRConstants.HPR_EXPORT_FILENAME_HEADER_KEY.length, filenameHeader.length);
                            }
                        }

                        // prompt the user to save or open the file
                        window.navigator.msSaveBlob(xhr.response, filename ? filename : "HPRisk.xlsx");

                        // perform final tasks
                        __handleDownloadFinished();
                    };
                    xhr.send();
                }
                else {
                    // For non-Microsoft browsers, build a link with a download attribute, which
                    // tells the browser that this is a download instead of a redirect. Note: We
                    // must append the link to the document body for FF to work (Chrome does not
                    // require this).
                    var link = document.createElement("a");
                    link.href = window.location.protocol + "//" + window.location.hostname + ":" + window.location.port + url;
                    link.download = "";
                    document.body.appendChild(link);    // for FF
                    link.click();
                    document.body.removeChild(link);    // for FF

                    // perform final tasks
                    __handleDownloadFinished();
                }
            })
            .fail(function (jqxhr) { // textStatus, error
                jQuery.sap.log.error('Request Failed. Response text: ' + jqxhr.responseText);
                deferred.reject(jqxhr);
            });
        return deferred;
    },

    //*****************************************************************************************************
    //*****************************************************************************************************
    // Query functions
    //*****************************************************************************************************
    //*****************************************************************************************************
    query: function (hprReportRequest) {
        //var self = this;
        var deferred = $.Deferred();
        // TODO: Determine if this should return a restricted promise
        // TODO: Could this just return the ajax promise object?

        var dataSourceRef = hprReportRequest.dataSources;
        if (!dataSourceRef) {
            dataSourceRef = hprReportRequest.dataSource;
        }
        var cubeID = dataSourceRef.id;
        var queryURL = HprConstants.HPR_BASE_CUBES_URL + '/' + encodeURIComponent(cubeID) + "/query?" + this.GUID_PARAM_STRING;

        var stringifiedData = JSON.stringify(hprReportRequest);

        var ajaxRequest = {
            headers: {
                "Accept": "application/vnd.sas.risk.query+json;charset=utf-8",
                "Content-Type": "application/vnd.sas.risk.query.parms+json;charset=utf-8" // FIXME see DA for name change
            },
            contentType: "application/json;charset=utf-8",
            method: "POST",
            url: queryURL,
            data: stringifiedData,
            dataType: "json"
        };

        // Get the logger for this object and output stuff
        var logger = sas.log.getLogger("sas.analytics.riskhp.utils.HprAPI");
        logger.debug('HprAPI.query ' + queryURL);
        logger.trace('HprAPI.query ' + JSON.stringify(ajaxRequest));

        $.ajax(ajaxRequest)
            .done(function (data, textStatus, jqxhr) {
                if (data === undefined) {
                    // If no data returned, the create an object to carry the status code to caller
                    data = {};
                }
                data.status = jqxhr.status;
                logger.debug('HprAPI.query DONE');
                logger.trace('HprAPI.query DONE ' + JSON.stringify(data));
                deferred.resolve(data);
            })
            .fail(function (jqxhr) { // textStatus, error
                jQuery.sap.log.error('Request Failed. Response text: ' + jqxhr.responseText);
                deferred.reject(jqxhr);
            });
        return deferred;
    },

    // This will eventually become the new query function that is used with the cache.
    queryExpansion: function (hprReportRequest) {
        //var self = this;
        var deferred = $.Deferred();

        var dataSourceRef = hprReportRequest.dataSources;
        if (!dataSourceRef) {
            dataSourceRef = hprReportRequest.dataSource;
        }
        var cubeID = dataSourceRef.id;
        var queryURL = HprConstants.HPR_BASE_CUBES_URL + '/' + encodeURIComponent(cubeID) + "/query?" + this.GUID_PARAM_STRING;

        //FIXME TEMP   hprReportRequest must use "crosstab2" as type for now.  Applegate 1/25/18
        var stringifiedData = JSON.stringify(hprReportRequest);

        var ajaxRequest = {
            headers: {
                "Accept": "application/vnd.sas.risk.query+json;charset=utf-8",
                "Content-Type": "application/vnd.sas.risk.query.parms+json;charset=utf-8"
            },
            contentType: "application/json;charset=utf-8",
            method: "POST",
            url: queryURL,
            data: stringifiedData,
            dataType: "json"
        };

        // Get the logger for this object and output stuff
        var logger = sas.log.getLogger("sas.analytics.riskhp.utils.HprAPI");
        logger.debug('HprAPI.queryExpansion ' + queryURL);
        logger.trace('HprAPI.queryExpansion ' + JSON.stringify(ajaxRequest));

        $.ajax(ajaxRequest)
            .done(function (data, textStatus, jqxhr) { // Is this the right order of parameters?
                if (data === undefined) {
                    // If no data returned, the create an object to carry the status code to caller
                    data = {};
                }
                data.status = jqxhr.status;
                logger.debug('HprAPI.queryExpansion DONE');
                logger.trace('HprAPI.queryExpansion DONE ' + JSON.stringify(data));
                deferred.resolve(data);
            })
            .fail(function (jqxhr) { // textStatus, error
                jQuery.sap.log.error('Request Failed. Response text: ' + jqxhr.responseText);
                deferred.reject(jqxhr);
            });
        return deferred;
    },

    /**
     * Is this deprecated? If so consider removing.
     * @param cubeId
     * @returns {void|*}
     * @deprecated
     */
    portfolioMetadata: function (cubeId) {
        //var self = this;
        var deferred = $.Deferred();
        //TODO: Determine if this should return a restricted promise

        var queryURL = HprConstants.HPR_BASE_CUBES_URL + '/' + encodeURIComponent(cubeId) + "?fields=dataItems.portfolioColumns&" + this.GUID_PARAM_STRING;

        $.ajax({
            headers: {
                "Accept": "application/vnd.sas.risk.cube+json;charset=utf-8",
                "Content-Type": "application/vnd.sas.risk.cube+json;charset=utf-8"
            },
            contentType: "application/json;charset=utf-8",
            method: "GET",
            url: queryURL,
            dataType: "json"
        }).done(
            function (data) {
                deferred.resolve(data.dataItems);
            }
        ).fail(
            function (jqxhr) { // textStatus, error
                jQuery.sap.log.error('Request Failed. Response text: ' + jqxhr.responseText);
                deferred.reject(jqxhr);
            }
        );
        return deferred;
    },

    //*****************************************************************************************************
    //*****************************************************************************************************
    // Scenario functions
    //*****************************************************************************************************
    //*****************************************************************************************************
    getScenarioList: function(cube) {
        //var self = this;
        var cubeId = cube.dataSource.id;
        var deferred = $.Deferred();
        // TODO: Determine if this should return a restricted promise

        var queryURL = HprConstants.HPR_BASE_CUBES_URL + '/' + encodeURIComponent(cubeId) + "/scenarios?" + this.GUID_PARAM_STRING;

        $.ajax({
            headers: {
                "Accept": "application/vnd.sas.risk.cube+json;charset=utf-8"
            },
            contentType: "application/json;charset=utf-8",
            method: "GET",
            url: queryURL,
            dataType: "json"
        }).done(
            function (data) {
                //data has the following properties - links, dataItems, version, count

                // generate a map of scenarios (needed for populating baseline scenario information)
                var scenariosMap = {};
                _.each(data.dataItems.scenarios, function(scenario) {
                    scenariosMap[scenario.id] = scenario;
                });

                // process scenarios
                var scenarios = [];
                _.each(data.dataItems.scenarios, function(scenario) {
                    var baselineScenario = scenariosMap[scenario.baselineId];
                    if (baselineScenario) {
                        scenario.baselineName = baselineScenario.name;
                        scenario.baselineVersion = baselineScenario.scenarioVersion;
                    }
                    scenarios.push(new sas.analytics.riskhp.models.dataitems.Scenario(scenario));
                });

                deferred.resolve(scenarios);
            }
        ).fail(
            function (jqxhr) { // textStatus, error
                jQuery.sap.log.error('Request Failed. Response text: ' + jqxhr.responseText);
                deferred.reject(jqxhr);
            }
        );
        return deferred;
    },

    /**
     * Load the scenario objects for the given scenarioIds in the cube with the given dataSourceId
     * Note: There is no guarantee the scenarios returned from this call are all activated.  The code
     * does it's best to activate them but there may be server side issues that preclude this from happening.
     * The impact of this is consumers need to be able to deal with non-activated scenarios in some fashion
     * specific to their use case.
     * <p><b>Ajax Properties:</b>
     * <ul>
     *     <li> restricted deferred </li>
     *     <li> resolve(Scenario[])</li>
     *     <li> rejects</li>
     *     <li> calls {@link _activateScenarios}</li>
     *  </ul>
     *
     * @param {string} dataSourceId
     * @param {string[]} scenarioIds
     * @param {boolean} usingRepoId
     * @return {Deferred} The returned deferred is restricted
     */
    loadDetailedScenarios: function(dataSourceId, scenarioIds, usingRepoId) {
        var deferred = $.Deferred(),
            self =this;

        if (usingRepoId === undefined) {
            usingRepoId = false;
        }

        if (scenarioIds === undefined || scenarioIds.length === 0){
            deferred.resolve([]);
            return deferred;
        }

        // Build a string to represent the scenarios to load (suitable for the query)
        var idString = _.reduce(scenarioIds,function(memo,value){
            return memo + "%7C" + value; // %7C is a url encoded pipe ('|')
        });

        var queryURL = HprConstants.HPR_BASE_CUBES_URL + '/' +
            encodeURIComponent(dataSourceId) + "/scenarios?" + (usingRepoId ? "repoId=" : "id=") + idString + "&" + self.GUID_PARAM_STRING + "&activate=true";

        $.ajax({
            headers: {
                "Accept": "application/vnd.sas.risk.cube+json;charset=utf-8",
                "Content-Type": "application/vnd.sas.risk.cube+json;charset=utf-8"
            },
            contentType: "application/json;charset=utf-8",
            method: "GET",
            url: queryURL,
            dataType: "json"
        }).done(
            function (data) {
                // process the scenarios and track how many were activated
                var activatedScenarioCount = 0;
                var scenarios = [];
                _.each(data.dataItems.scenarios, function(scenario) {
                    scenarios.push(new sas.analytics.riskhp.models.dataitems.Scenario(scenario));
                    if (scenario.activated === true) {
                        activatedScenarioCount++;
                    }
                });

                // If any scenarios were activated, we need to force the caches to requery
                if (activatedScenarioCount > 0) {
                    CacheManager.invalidate(dataSourceId);
                }

                deferred.resolve(scenarios);
            }
        ).fail(
            function (jqxhr) { // textStatus, error
                jQuery.sap.log.error('Request Failed. Response text: ' + jqxhr.responseText);
                deferred.reject(jqxhr);
            }
        );
        return deferred.promise();
    },

    /**
     * Activates the scenarios with the given ids for the given cube id
     * <p><b>Ajax Properties:</b>
     * <ul>
     *     <li> restricted deferred </li>
     *     <li> resolves</li>
     *     <li> never rejects</li>
     *     <li> calls nothing</li>
     * </ul>
     *
     * @param {string} dataSourceId - The id of the cube to load the scenarios for
     * @param {Scenario[]} scenarios - if list is null or empty will return a resolved deferred.
     * @private
     * @returns {Deferred} The returned deferred is restricted
     */
    _activateScenarios: function(dataSourceId, scenarios) {
        var self = this;
        var deferred = $.Deferred();

        // no scenarios, then this is a no-op
        if (!scenarios || scenarios.length === 0){
            deferred.resolve();
            return deferred;
        }
        var scenarioIds = _.map(scenarios, function(scenario){
            return scenario.getId();
        });
        var promise = $.ajax(
            {
                headers: {
                    "Accept": "application/json;charset=utf-8"
                },
                method: "POST",
                url: sas.analytics.riskhp.utils.HprConstants.HPR_BASE_CUBES_URL +
                    '/' + encodeURIComponent(dataSourceId) + "/scenarios?id=" +
                    self._buildScenarioIDString(scenarioIds) +
                    "&" + this.GUID_PARAM_STRING
            }
        );
        promise.done(
            function (activatedScenarioIds) {
                // populate with empty array if server doesn't send anything back
                if (!activatedScenarioIds) {
                    activatedScenarioIds = [];
                }

                // mark the scenarios that were activated as activated
                var scenario = null;
                _.each(activatedScenarioIds, function(scenarioId) {
                    scenario = _.find(scenarios, function(scenario){
                        return scenario.getId() === scenarioId;
                    });
                    if (scenario && _.isFunction(scenario.setActivated)) {
                        scenario.setActivated(true);
                    }
                });
                deferred.resolve();
            }
        );
        promise.fail(
            function (jqxhr) { // textStatus, error
                // Failure of this call is not terminal so log the error and press on
                jQuery.sap.log.error('Request Failed. Response text: ' + jqxhr.responseText);
                deferred.resolve();
            }
        ); // end fail handler

        return deferred.promise();
    },

    // Left in for historical reasons until we know we got scenario activation right
    // /**
    //  * Activates the scenarios with the given ids for the given cube id
    //  * @param {string} cubeId
    //  * @param {string[]} scenarioIds - if list is null or empty will return a resolved deferred.
    //  * @returns {*} A deferred object
    //  */
    // activateScenarios: function(cubeId, scenarioIds) {
    //     var self = this;
    //     var HprUtil = sas.analytics.riskhp.utils.HprUtil;
    //     var deferred = $.Deferred();
    //
    //     // no scenarios, then this is a no-op
    //     if (!scenarioIds || scenarioIds.length === 0){
    //         deferred.resolve([]);
    //         return deferred;
    //     }
    //
    //     $.ajax(
    //         {
    //             headers: {
    //                 "Accept": "application/json;charset=utf-8"
    //             },
    //             method: "POST",
    //             url: sas.analytics.riskhp.utils.HprConstants.HPR_BASE_CUBES_URL + '/' + encodeURIComponent(cubeId) + "/scenarios?id=" + self._buildScenarioIDString(scenarioIds) + "&" + this.GUID_PARAM_STRING
    //         }
    //     ).done(
    //         function (activatedScenarioIds) {
    //             // populate with empty array if server doesn't send anything back
    //             if (!activatedScenarioIds) {
    //                 activatedScenarioIds = [];
    //             }
    //
    //             sas.analytics.riskhp.models.DataSourceCacheManager.getDataSource(cubeId).done(
    //                 function(dataSource) {
    //                     // if the number of activated scenarios is not the same as the number passed in, notify the user
    //                     if (activatedScenarioIds.length < scenarioIds.length) {
    //                         // get pass/fail counts
    //                         var failCount = scenarioIds.length - activatedScenarioIds.length;
    //                         var scenCount = scenarioIds.length;
    //
    //                         // get the ids of the scenarios that failed activation
    //                         var failedScenarioIds = _.difference(scenarioIds, activatedScenarioIds);
    //
    //                         // error messaging
    //                         var dialogText = null;
    //                         var detailText = null;
    //                         if (failCount === 1) {
    //                             dialogText = HprUtil.getBundle().getText("ActivateScenario.error.single.fmt", [dataSource.getDataItemById(failedScenarioIds[0]).getLabel(), dataSource.getCubeDisplayLabel()]);
    //                         }
    //                         else {
    //                             dialogText = HprUtil.getBundle().getText("ActivateScenario.error.multi.fmt", [failCount, scenCount, dataSource.getCubeDisplayLabel()]);
    //
    //                             // build the detail text for display in the error dialog (shows the scenarios that failed activation)
    //                             var scenarioListText = "";
    //                             _.each(failedScenarioIds, function(scenarioId) {
    //                                 if (scenarioListText.length > 0) {
    //                                     scenarioListText += "\n";
    //                                 }
    //                                 scenarioListText += "\t" + dataSource.getDataItemById(scenarioId).getLabel();
    //                             });
    //                             detailText = HprUtil.getBundle().getText("ActivateScenario.error.detail.txt") + "\n" + scenarioListText;
    //                         }
    //
    //                         // show the error
    //                         sas.hc.m.MessageDialog.error({
    //                             text: dialogText,
    //                             title: HprUtil.getBundle().getText("error.window.title"),
    //                             detailText: detailText ? detailText : null
    //                         });
    //                     }
    //
    //                     // mark the scenario data items as activated
    //                     var scenario = null;
    //                     _.each(activatedScenarioIds, function(scenarioId) {
    //                         scenario = dataSource.getDataItemById(scenarioId);
    //                         if (scenario && _.isFunction(scenario.setActivated)) {
    //                             scenario.setActivated(true);
    //                         }
    //                     });
    //
    //                     deferred.resolve(activatedScenarioIds);
    //                 }
    //             ).fail(
    //                 function (jqxhr, textStatus, error) {
    //                     deferred.reject(jqxhr, textStatus, error);
    //                 }
    //             );
    //         }
    //     ).fail(
    //         function (jqxhr, textStatus, error) {
    //             jQuery.sap.log.error('Request Failed. Response text: ' + jqxhr.responseText);
    //             deferred.reject(jqxhr, textStatus, error);
    //         }
    //     );
    //
    //     return deferred.promise();
    // },

    deactivateScenarios: function(cubeId, scenarioIds) {
        var self = this;
        var deferred = $.Deferred();

        $.ajax(
            {
                headers: {},
                method: "DELETE",
                url: sas.analytics.riskhp.utils.HprConstants.HPR_BASE_CUBES_URL + '/' + encodeURIComponent(cubeId) + "/scenarios?id=" + self._buildScenarioIDString(scenarioIds) + "&" + self.GUID_PARAM_STRING
            }
        ).done(
            function () {
                // try to mark the scenario data items as deactivated (they may have already been removed from the data source)
                sas.analytics.riskhp.models.DataSourceCacheManager.getDataSource(cubeId).done(
                    function(dataSource) {
                        var scenario = null;
                        _.each(scenarioIds, function (scenarioId) {
                            scenario = dataSource.getDataItemById(scenarioId);
                            if (scenario && _.isFunction(scenario.setActivated)) {
                                scenario.setActivated(false);
                            }
                        });
                    }
                );
                deferred.resolve();
            }
        ).fail(
            function (jqxhr, textStatus, error) {
                jQuery.sap.log.error('Request Failed. Response text: ' + jqxhr.responseText);
                deferred.reject(jqxhr, textStatus, error);
            }
        );

        return deferred.promise();
    },

    /**
     * Helper method to build the string representing scenario ids for inclusing in an API call
     * @param scenarioIds
     * @returns {*}
     * @private
     * @memberOf HprAPI
     */
    _buildScenarioIDString: function(scenarioIds) {
        return _.reduce(scenarioIds, function(memo, value){
            return memo + "%7C" + value; // %7C is a url encoded pipe ('|')
        });
    },

    /**
     * Adds the pieces that need to be loaded for stress cubes to be useable
     * <p><b>Ajax Properties:</b>
     * <ul>
     *     <li> restricted deferred </li>
     *     <li> resolve(cubeData)</li>
     *     <li> never rejects</li>
     *     <li> calls {@link loadDetailedScenarios}</li>
     *  </ul>
     * @param cubeData
     * @param {string[]} attachedScenarioIds
     * @returns {Deferred} The returned Deferred is restricted
     * @private
     */
    _addStressCubeArtifacts: function(cubeData, attachedScenarioIds) {
        var deferred = $.Deferred();

        var promise = this.loadDetailedScenarios(cubeData.getId(),attachedScenarioIds);

        promise.done(function(scenarios) {
            // Add the scenarios to the cube object and resolve
            _.each(scenarios, jQuery.proxy(cubeData.addScenario,cubeData));
            deferred.resolve(cubeData);
        });

        promise.fail(
            function (jqxhr) { // textStatus, error
                // Failure of this call is not terminal so log the error and press on
                jQuery.sap.log.error('Request Failed. Response text: ' + jqxhr.responseText);
                deferred.resolve(cubeData);
            }
        ); // end fail handler

        return deferred.promise();
    },

    //*****************************************************************************************************
    //*****************************************************************************************************
    // Risk Factor Sensitivities functions
    //*****************************************************************************************************
    //*****************************************************************************************************

    /**
     * recursively retrieve all leaf node risk factors from the riskFactorItem
     * @param riskFactorItem - risk factor
     * @returns {void|*} array of risk factors
     */
    getRiskFactorMembers: function(riskFactorItem) {
        var self = this;
        var deferred = $.Deferred();
        var link = this.findRequiredLink(deferred, riskFactorItem.links, {rel:'members'});
        if (link) {
            var threshold = HprConstants.HPR_RISK_FACTOR_SENSITIVITIES_RISK_FACTOR_THRESHOLD;
            var limit = Math.min(threshold, 10000); // 10000 is max page size for pageable model
            var semaphore = {
                deferred: deferred,
                threshold: threshold, // max total to retrieve
                start: 0,
                limit: limit,   // number to request
                rfs: [] // retrieved risk factors
            };
            self._getRiskFactors(link, semaphore);
        }

        return deferred;
    },

    /**
     * returns all risk factors according to semaphore state
     * @returns {void|*}
     * @memberOf HprAPI
     */
    _getRiskFactors: function(link, semaphore) {
        var self = this;

        $.getJSON(link.href + "&start=" + semaphore.start + "&limit=" + semaphore.limit, function(data) {
            semaphore.count = data.count;
            semaphore.rfs = semaphore.rfs.concat(data.items);
            semaphore.start = semaphore.rfs.length;
            var min = Math.min(semaphore.count, semaphore.threshold);
            if (min > semaphore.rfs.length) {
                // don't ask for more than the threshold
                if (semaphore.count > semaphore.threshold && // limiting by threshold
                    semaphore.limit > semaphore.threshold - semaphore.rfs.length) {
                    semaphore.limit = semaphore.threshold - semaphore.rfs.length;
                }
                self._getRiskFactors(link, semaphore);
            } else {
                semaphore.deferred.resolve({
                    total: semaphore.count,
                    threshold: semaphore.threshold,
                    riskFactors: semaphore.rfs
                });
            }
        }).fail(function(jqxhr) {
            jQuery.sap.log.error('getRiskFactors request failed. Response text: ' + jqxhr.responseText);
            semaphore.deferred.reject(jqxhr);
        });
    },

    /**
     * Retrieves all risk factors from the server, up to the threshold
     * @returns {void|*} an object containing count (the number of risk factors on server), threshold (max number of
     * risk factors to retrieve), and riskFactors array
     */
    getAllRiskFactors: function(thresholdVal) {
        var self = this;
        var deferred = $.Deferred();

        var threshold = thresholdVal && !isNaN(thresholdVal) ? thresholdVal : HprConstants.HPR_RISK_FACTOR_SENSITIVITIES_RISK_FACTOR_THRESHOLD;
        var limit = Math.min(threshold, 10000); // 10000 is max page size for pageable model

        var links = sas.analytics.riskhp.views.datapane.DataPaneModelManager.getSelectedDataSource().json.links;
        var link = _.find(links, {rel: "riskFactors"});
        if (link) {
            var semaphore = {
                deferred: deferred,
                threshold: threshold, // max total to retrieve
                start: 0,
                limit: limit,   // number to request
                rfs: [] // retrieved risk factors
            };
            self._getRiskFactors(link, semaphore);
        }

        return deferred;
    },

    /**
     * returns definitions of supplied riskFactors that are in the selected datasource
     * @param dataSource
     * @param variables - array of risk factor names
     * @returns {void|*}
     */
    getVariableDefinitions: function(dataSource, variables) {
        var deferred = $.Deferred();
        var links = dataSource.json.links;
        var link = _.find(links, {rel: "variableDefinitions"});
        if (link) {
            var stringifiedData = JSON.stringify(variables);
            $.ajax({
                headers: {
                    "Content-Type": link.type,
                    "Accept": link.responseType,
                    "Accept-Item": link.responseItemType
                },
                contentType: "application/json;charset=utf-8",
                method: "POST",
                url: link.href,
                data: stringifiedData,
                dataType: "json"
            }).done(function (data) {
                deferred.resolve(data);
            }).fail(function (jqxhr) { // textStatus, error
                jQuery.sap.log.error('Request Failed. Response text: ' + jqxhr.responseText);
                deferred.reject(jqxhr);
            });
        }
        return deferred;
    },


    //*****************************************************************************************************
    //*****************************************************************************************************
    // Username functions
    //*****************************************************************************************************
    //*****************************************************************************************************
	loadUserName: function() {
		console.log('loadUserName');
		var api = this;
		var deferred = $.Deferred();
		$.getJSON('rest/commons/apps/userdetails', function(user) {
			api.setUserName(user.displayName);
			deferred.resolve(user);
			// NOTE:we can grab the roles here
		});
		return deferred;
	},

	setUserName: function(userName) {
		this._userName = userName;
	},

	getUserName: function() {
		var oUserInfo = sas.hc.ui.core.util.UserInfo.instance;

		// I don't know what object property to use. Assuming name.
		return (oUserInfo !== undefined) ? oUserInfo.name : this._userName;
	},

    //*****************************************************************************************************
    //*****************************************************************************************************
    // Computed Variables functions
    //*****************************************************************************************************
    //*****************************************************************************************************


    /**
     * Validates the loaded computed varaibles
     * <p><b>Ajax Properties:</b>
     * <ul>
     *     <li> restricted deferred </li>
     *     <li> resolves(data)</li>
     *     <li> rejects</li>
     *     <li> calls nothing</li>
     *  </ul>
     *
     * @param cubeDataModel
     * @param request
     * @returns {Deferred} The returned deferred is restricted
     */
    validateComputedVariable: function(cubeDataModel, request) {
        var deferred = $.Deferred();

        var link = _.find(cubeDataModel.json.links, {rel: "validateComputedVariables"});
        if (link) {
            var stringifiedData = JSON.stringify(request);
            $.ajax({
                headers: {
                    "Content-Type": link.type,
                    "Content-Item": link.itemType,
                    "Accept": link.responseType,
                    "Accept-Item": link.responseItemType
                },
                method: link.method,
                url: link.href,
                data: stringifiedData,
                dataType: "json"
            }).done(function (data) {
                deferred.resolve(data);
            }).fail(function (jqxhr) { // textStatus, error
                jQuery.sap.log.error('Request Failed. Response text: ' + jqxhr.responseText);
                jQuery.sap.log.error('hprReportRequest: ' + stringifiedData);
                deferred.reject(jqxhr);
            });
        }
        return deferred.promise();
    },

    /**
     * Updates a computed variable
     * <p><b>Ajax Properties:</b>
     * <ul>
     *     <li> restricted deferred </li>
     *     <li> resolves(data)</li>
     *     <li> rejects</li>
     *     <li> calls nothing</li>
     *  </ul>
     *
     * @param cubeDataModel
     * @param request
     * @returns {Deferred} The returned deferred is restricted
     */
    updateComputedVariable: function(cubeDataModel, request) {
        var deferred = $.Deferred();

        var link = _.find(cubeDataModel.json.links, {rel: "updateComputedVariables"});
        if (link) {
            var stringifiedData = JSON.stringify(request);
            $.ajax({
                headers: {
                    "Content-Type": link.type,
                    // "Content-Item": link.itemType,
                    "Accept": link.responseType
                    // "Accept-Item": link.responseItemType
                },
                method: link.method,
                url: link.href,
                data: stringifiedData,
                dataType: "json"
            }).done(function (data) {
                deferred.resolve(data);
            }).fail(function (jqxhr) { // textStatus, error
                jQuery.sap.log.error('Request Failed. Response text: ' + jqxhr.responseText);
                jQuery.sap.log.error('hprReportRequest: ' + stringifiedData);
                deferred.reject(jqxhr);
            });
        }
        return deferred.promise();
    },

    /**
     * Creates a computed variable
     * <p><b>Ajax Properties:</b>
     * <ul>
     *     <li> restricted deferred </li>
     *     <li> resolves(data)</li>
     *     <li> rejects</li>
     *     <li> calls nothing</li>
     *  </ul>
     *
     * @param cubeDataModel
     * @param request
     * @returns {Deferred} The returned deferred is restricted
     */
    createComputedVariable: function(cubeDataModel, request) {
        var deferred = $.Deferred();

        var link = _.find(cubeDataModel.json.links, {rel: "createComputedVariables"});
        if (link) {
            var stringifiedData = JSON.stringify(request);
            $.ajax({
                headers: {
                    "Content-Type": link.type,
                    "Content-Item": link.itemType,
                    "Accept": link.responseType,
                    "Accept-Item": link.responseItemType
                },
                method: link.method,
                url: link.href,
                data: stringifiedData,
                dataType: "json"
            }).done(function (data) {
                deferred.resolve(data);
            }).fail(function (jqxhr) { // textStatus, error
                jQuery.sap.log.error('Request Failed. Response text: ' + jqxhr.responseText);
                jQuery.sap.log.error('hprReportRequest: ' + stringifiedData);
                deferred.reject(jqxhr);
            });
        }
        return deferred.promise();
    },

    deleteComputedVariable: function(cubeDataModel, request) {
        var deferred = $.Deferred();

        var link = _.find(cubeDataModel.json.links, {rel: "deleteComputedVariables"});
        if (link) {
            var stringifiedData = JSON.stringify(request);
            $.ajax({
                headers: {
                    "Content-Type": link.type,
                    "Content-Item": link.itemType,
                    "Accept": link.responseType,
                    "Accept-Item": link.responseItemType
                },
                method: link.method,
                url: link.href,
                data: stringifiedData,
                dataType: "json"
            }).done(function (data) {
                deferred.resolve(data);
            }).fail(function (jqxhr) { // textStatus, error
                jQuery.sap.log.error('Request Failed. Response text: ' + jqxhr.responseText);
                jQuery.sap.log.error('hprReportRequest: ' + stringifiedData);
                deferred.reject(jqxhr);
            });
        }
        return deferred;
    },


    /**
     * Adds the computed variables to the DataSource
     * <p><b>Ajax Properties:</b>
     * <ul>
     *     <li> restricted deferred </li>
     *     <li> resolves</li>
     *     <li> never rejects</li>
     *     <li> calls {@link validateComputedVariable}</li>
     *     <li> calls {@link createComputedVariable}</li>
     *  </ul>
     *
     * @param cubeData
     * @param computedVariables
     * @returns {Deferred} The returned deferred is restricted
     * @private
     */
    _addComputedVariableArtifacts: function(cubeData, computedVariables) {
        var self = this;
        var deferred = $.Deferred();

        if (computedVariables && computedVariables.length > 0) {
            var clonedCompVars = sas.analytics.riskhp.utils.HprUtil.hprClone(computedVariables);

            // sort clonedCompVars according to items.  Computed variables may be created or updated to be based on
            // other computed variables, so they need to be created on the server in order.
            clonedCompVars.sort(function(a, b){
                var aAncestors = [];  // array of names of computed variables used in the creation of computed variable a
                var bAncestors = [];  // array of names of computed variables used in the creation of computed variable b

                self.CompVarUtil.getComputedVariableAncestors(aAncestors, a, clonedCompVars, cubeData);
                self.CompVarUtil.getComputedVariableAncestors(bAncestors, b, clonedCompVars, cubeData);

                if (_.contains(aAncestors, b.name)) {
                    return 1;
                } else if (_.contains(bAncestors, a.name)) {
                    return -1;
                }
                return 0;
            });

            var requests = [];
            _.each(clonedCompVars, function(compVarData) {
                // ignore exploration computed variables if the cube already has a computed variable of the same name
                if (!_.find(cubeData.computedVariable, function(cubeCV) {
                        return cubeCV.name === compVarData.name;
                    })) {

                    var isValid = true;
                    _.each(compVarData.items, function(column) {
                        _.each(column.dataItems, function (dataItem) {
                            var item = cubeData.getDataItemById(dataItem.id);

                            if (item) {
                                if (item.type === sas.analytics.riskhp.models.dataitems.Scenario.SCENARIO) {
                                    if (!item.isActivated()) {
                                        isValid = false;
                                    }
                                } else if (item.type === sas.analytics.riskhp.models.dataitems.Compvar.COMPVAR) {
                                    if (!item.isValid()) {
                                        isValid = false;
                                    }
                                }
                            } else {
                                // it could be a computed variable that has not yet been created on the datasource
                                if (dataItem.type === sas.analytics.riskhp.models.dataitems.Compvar.COMPVAR) {
                                    // we could check that the missing computed variable is in the array of requests, but
                                    // this should always be the case
                                } else {
                                    isValid = false;
                                }
                            }
                        });
                    });

                    if (isValid) {
                        requests.push(sas.analytics.riskhp.views.computedvariableeditor.ComputedVariableUtil.buildComputedVarRequest(compVarData, cubeData));
                    } else {
                        // add the invalid computed variable to the data source so it can show up in the data pane and can be edited/corrected
                        var invalidCompVarData = {
                            name: compVarData.name,
                            label: compVarData.label,
                            id: compVarData.serverId, // id provided by the server when the computed variable was originally validated/created
                            error: 1,
                            exploration: true,
                            type: sas.analytics.riskhp.models.dataitems.Compvar.COMPVAR
                        };
                        cubeData.addComputedVariable(invalidCompVarData);
                    }
                } else {
                    console.log("Ignoring exploration computed variable " + compVarData.name + " since there is a cube computed variable of the same name")
                }
            });

            if (requests.length > 0) {
                sas.analytics.riskhp.utils.HprAPI.createComputedVariable(cubeData, requests).done(function (createdCVs) {
                    _.each(createdCVs.items, function (createdCV) {
                        // If the server failed to create the CV it will have a non-zero error field and will therefore be added as invalid
                        cubeData.addComputedVariable(createdCV);
                    });
                    deferred.resolve();
                }).fail(function (jqxhr) { // textStatus, error
                    jQuery.sap.log.error('Request Failed. Response text: ' + jqxhr.responseText);
                    deferred.resolve();
                });
            } else { // none to create, just resolve original deferred
                deferred.resolve();
            }
        } else {
            deferred.resolve();
        }

        return deferred.promise();
    },


    //*****************************************************************************************************
    //*****************************************************************************************************
    // REST Link functions
    //*****************************************************************************************************
    //*****************************************************************************************************
	loadRootLinks: function() {
		var api = this;
		var deferred = $.Deferred();
		$.getJSON('rest', function(data) {
			api._rootLinks = data.links;
			deferred.resolve(data.links);
		});
		return deferred;
	},

    /**
     * Returns the root level links owned by this singleton object
     * @returns {Array}
     * @memberOf HprAPI
     */
	getRootLinks: function() {
		return this._rootLinks;
	},

	findOptionalLink: function(links, properties) {
		return _.findWhere(links, properties);
	},

	findRequiredLink: function(deferred, links, properties) {
		var link = this.findOptionalLink(links, properties);
		if (!link) {
			var msg = sas.analytics.riskhp.utils.HprUtil.getBundle().getText('error.missingLink.fmt', JSON.stringify(properties));
			deferred.reject({
				responseText: msg,
				responseJSON: {
					message: msg,
					errorCode: 400
				}
			});
		}

		return link;
	},

    /**
     * Encodes the parameters passed in so they can be passed on a request
     *
     * @param cubeParms An Object containing parameters to encode
     * @returns {string} A concatenated string of encoded parameters
     * @deprecated This doesn't appear to be used, recommend removal
     * @private
     */
    _encodeArgsAsQuery: function(cubeParms) {
        var conj = '', str = '';
        for (var key in cubeParms) {
            if (cubeParms.hasOwnProperty(key) && key !== "attachedScenarioIds") {
                str += conj + key + '=' + encodeURIComponent(cubeParms[key]);
                conj = '&';
            }
        }

        return str;
	}
};