// (c) 2017, SAS Institute Inc.

jQuery.sap.declare("sas.analytics.riskhp.utils.HprApplicationSwitcherController");

jQuery.sap.require("sap.ui.model.json.JSONModel");

jQuery.sap.require("sas.analytics.riskhp.models.DataSourceCacheManager");
jQuery.sap.require("sas.analytics.riskhp.models.DataSourceUtil");
jQuery.sap.require("sas.analytics.riskhp.utils.ExplorationUtil");
jQuery.sap.require("sas.analytics.riskhp.utils.HprAPI");
jQuery.sap.require("sas.analytics.riskhp.utils.HprConstants");
jQuery.sap.require("sas.analytics.riskhp.utils.HPRiskPanelsViewDelegate");
jQuery.sap.require("sas.analytics.riskhp.utils.HprUtil");
jQuery.sap.require("sas.analytics.riskhp.utils.ViewManager");
jQuery.sap.require("sas.analytics.riskhp.views.datapane.DataPaneUtil");


/**
 * @name sas.analytics.riskhp.utils.HprApplicationSwitcherController
 */
sas.analytics.riskhp.utils.HprApplicationSwitcherController = function() {
    // pseudo-imports (static constants)
    var DataSourceCacheManager = sas.analytics.riskhp.models.DataSourceCacheManager;
    var DataSourceUtil = sas.analytics.riskhp.models.DataSourceUtil;
    var ExplorationUtil = sas.analytics.riskhp.utils.ExplorationUtil;
    var HprAPI = sas.analytics.riskhp.utils.HprAPI;
    var HprConstants = sas.analytics.riskhp.utils.HprConstants;
    var HPRiskPanelsViewDelegate = sas.analytics.riskhp.utils.HPRiskPanelsViewDelegate;
    var HprUtil = sas.analytics.riskhp.utils.HprUtil;
    var ViewManager = sas.analytics.riskhp.utils.ViewManager;
    var DataPaneUtil = sas.analytics.riskhp.views.datapane.DataPaneUtil;

    // application constants
    var SCENARIO_MANAGER_APP = "RiskScenarioManager";

    // message type constants
    var CREATE_SCENARIO_MSG = "CreateScenario";
    var OPEN_SCENARIO_MSG = "OpenScenario";
    var SCENARIOS_UPDATED_MSG = "ScenariosUpdated";
    var RUN_SCENARIO_MSG = "RunScenario";
    var OPEN_MIP_CUBE = "App.RunInstance";

    // message data constants
    var SCENARIO_UPDATE_ACTION = "UPDATE";
    var SCENARIO_DELETE_ACTION = "DELETE";

    // internal static vars
    var scenarioManagerURL;
    var rb = HprUtil.getBundle();
    var vaHubURL;


    // private message handling functions

    function _handleAppShowMessage(type, data) {
        // confirm that it is HP Risk to be displayed
        if (data.appId === sas.config.getAppId()) {
            //TODO do something?
        }
    }

    function _handleScenariosUpdatedMessage(type, messageDataItems) {
        _.each(messageDataItems, function(messageDataItem) {
            if (messageDataItem.action === SCENARIO_UPDATE_ACTION) {
                _handleScenarioUpdated(messageDataItem.scenarioId);
            }
            else if (messageDataItem.action === SCENARIO_DELETE_ACTION) {
                _handleScenarioDeleted(messageDataItem.scenarioId);
            }
        });
    }

    /**
     * Runs the scenario specified in the data. Running the scenario is adding it to the
     * crosstab and displaying the calculations for it.
     *
     * The expected behavior for the following conditions is as follows:
     *
     * If no exploration is open, open a new exploration using the data source associated
     * with the scenario, and run it on the existing crosstab.
     *
     * If an exploration is open with no stress data sources and/or crosstabs open, or the
     * data source associated with the scenario is not open, then open the data source
     * associated with the scenario, create a page containing a crosstab pointing to that
     * data source, and run the scenario on the resulting crosstab.
     *
     * If an exploration is open, and the data source associated with the scenario is already
     * open, provide the user the option to select an existing crosstab or create a new crosstab
     * for the scenario data source.
     */
    function _handleRunScenario(type, data) {
        // create the data source params object so we can open the cube
        var dataSourceParams = new sas.analytics.riskhp.models.DataSourceParams.prototype.createRootParams({
            id: data.dataSourceId,
            application: data.dataSourceParms.application,
            workGroup: data.dataSourceParms.workGroup,
            relPath: data.dataSourceParms.relPath
        });

        var __getDataSourceAndScenario = function(dataSourceId, dataSourceName, scenarioData) {
            var deferred = $.Deferred();
            DataSourceCacheManager.getDataSource(dataSourceId).done(
                function(stressDataSource) {
                    // try to get the scenario from the data source
                    // S1484568 scenarioData.scenarioId is really repoId internal to HPRisk UI and is
                    // NOT id!
                    var dsScenario = stressDataSource.getScenarioDataItemByRepoId(scenarioData.scenarioId);
                    if (dsScenario) {
                        // resolve the data source and the already added scenario
                        deferred.resolve(stressDataSource, dsScenario);
                    }
                    else {
                        // if the scenario is not already in the data source, load it and add it.
                        // NOTE: usingRepoId flag is true because scenarioData.scenarioId is repoId
                        HprAPI.loadDetailedScenarios(stressDataSource.getId(), [scenarioData.scenarioId], true).done(
                            function (scenarios) {
                                // add the scenario to the data source
                                stressDataSource.addScenario(scenarios[0]);
                                DataSourceCacheManager.getDataSourceParam(stressDataSource.getId()).addScenarioId(scenarios[0].getId());

                                // resolve the data source and the added scenario
                                deferred.resolve(stressDataSource, stressDataSource.getDataItemById(scenarios[0].getId()));
                            }
                        ).fail(
                            function (jqxhr) {
                                HprUtil.displayGenericErrorDialog(rb.getText("RunScenario.LoadScenario.error.fmt", [scenarioData.scenarioName]));
                                deferred.reject();
                            }
                        );
                    }
                }
            ).fail(
                function() {
                    HprUtil.displayGenericErrorDialog(rb.getText("DataPane.data.source.load.error.fmt", [dataSourceName]));
                    deferred.reject();
                }
            );
            return deferred.promise();
        };

        var __updateAndSelectDataSource = function(stressDataSource, scenario, dataPaneController) {
            // update the data pane to have the data source as active/selected
            dataPaneController.setDataPaneDataSource(stressDataSource, true);

            // update the scenario nodes
            DataPaneUtil.updateScenarioNodes(stressDataSource, dataPaneController);
        };

        var __runScenarioOnNewPage = function(activeView, scenarioData, dataSourceDetails, stressDataSourceDetails, dataPaneController) {
            DataSourceCacheManager.updateDataSourceParams(dataSourceDetails);

            __getDataSourceAndScenario(stressDataSourceDetails.id, stressDataSourceDetails.label, scenarioData).done(
                function(stressDataSource, scenario) {
                    // add scenario to the data source and select it on the data pane
                    __updateAndSelectDataSource(stressDataSource, scenario, dataPaneController);

                    // open the new page with a crosstab
                    activeView.getController().addPageToExploration(true);

                    // run the scenario
                    _getActiveCrosstab().runScenario(scenario);
                }
            ).always(
                function() {
                    activeView.setBusy(false);
                }
            );
        };

        // open the cube
        HprAPI.getDataSourceDetails(dataSourceParams).done(
            function(dataSourceDetails) {
                if (dataSourceDetails && dataSourceDetails.dataSources && dataSourceDetails.dataSources.length > 0) {
                    // get the active view
                    var activeView = ViewManager.getActiveView();

                    // set the active view to busy
                    if (activeView) {
                        activeView.setBusyIndicatorDelay(0);
                        activeView.setBusy(true);
                    }

                    var stressDataSourceDetails = DataSourceUtil.getDataSourceForType(dataSourceDetails, HprConstants.HPR_CUBE_INFORMATION_TYPE_STRESS);

                    // If an exploration is already open, present the user with options for running the scenario
                    // on a new or existing crosstab. Otherwise, just create a new exploration with the scenario's
                    // data source and run the scenario on the resulting crosstab.
                    if (activeView && activeView.getId() === HprConstants.HPR_EXPLORATION_VIEW_ID) {
                        // get the data pane controller
                        var dataPaneController = activeView.getDataPane().getDataPaneView().getController();

                        //var loadedDataSourceIds = DataSourceCacheManager.getLoadedDataSourceIds();
                        var stressCrosstabs = ExplorationUtil.getStressCrosstabs(activeView);

                        // if the data source is loaded, then let the user choose an existing crosstab or a new one
                        if (/*_.indexOf(loadedDataSourceIds, stressDataSourceDetails.id) !== -1 && */stressCrosstabs.length > 0) {
                            var dialog = sap.ui.jsfragment("sas.analytics.riskhp.fragments.RunScenarioCrosstabSelectorDialog", activeView.getController());
                            dialog.setModel(new sap.ui.model.json.JSONModel(stressCrosstabs));
                            dialog.attachRunScenarioHandler(
                                function(crosstabProperties) {
                                    // If crosstab properties is undefined, the user opted for a new crosstab and we'll
                                    // run the scenario on it. Otherwise, run the scenario on the selected crosstab.
                                    if (crosstabProperties) {
                                        // set the page of the selected crosstab as the selected/active page
                                        activeView.setActivePage(crosstabProperties.pageId);

                                        __getDataSourceAndScenario(crosstabProperties.dataSourceId, crosstabProperties.dataSourceName, data).done(
                                            function(stressDataSource, scenario) {
                                                // add scenario to the data source and select it on the data pane
                                                __updateAndSelectDataSource(stressDataSource, scenario, dataPaneController);

                                                // get the crosstab control
                                                var crosstab = sap.ui.getCore().byId(crosstabProperties.id);

                                                // run the scenario on the crosstab
                                                crosstab.runScenario(scenario);
                                            }
                                        ).always(
                                            function() {
                                                activeView.setBusy(false);
                                            }
                                        );
                                    }
                                    else {
                                        __runScenarioOnNewPage(activeView, data, dataSourceDetails, stressDataSourceDetails, dataPaneController);
                                    }
                                }
                            );
                            dialog.attachCancelHandler(function() {activeView.setBusy(false);});
                            dialog.open();
                        }
                        else {
                            __runScenarioOnNewPage(activeView, data, dataSourceDetails, stressDataSourceDetails, dataPaneController);
                        }
                    }
                    else {
                        _openNewExploration(dataSourceDetails, HprConstants.HPR_CUBE_INFORMATION_TYPE_STRESS, activeView, data);
                    }
                } else {
                    // get the active view
                    var activeView = ViewManager.getActiveView();
                    if (activeView) {
                        activeView.setBusy(false);
                    }
                    HprUtil.displayGenericErrorDialog(rb.getText("HPRisk.detailed.error.cube.information.failed.to.load.txt"));
                }
            }
        ).fail(
            function(jqxhr, textStatus, error) {
                // get the active view
                var activeView = ViewManager.getActiveView();
                if (activeView) {
                    activeView.setBusy(false);
                }
                HprUtil.displayGenericErrorDialog(rb.getText("DataPane.data.source.load.error.fmt", [dataSourceParams.abstractDataSourceId]));
            }
        );
    }

    function _viewOrEditScenario(scenario, dataSourceParams, readOnly) {
        sas.applicationSwitcher.gotoApp(SCENARIO_MANAGER_APP, scenarioManagerURL, true);
        sas.applicationSwitcher.sendMessageToApp(SCENARIO_MANAGER_APP, OPEN_SCENARIO_MSG, {
            scenarioWorkGroup: dataSourceParams.workGroup,
            scenarioName: scenario.getName(),
            scenarioVersion: scenario.getScenarioVersion(),
            dataSourceId: dataSourceParams.abstractDataSourceId,
            dataSourceParms: {
                workGroup: dataSourceParams.workGroup,
                application: dataSourceParams.application,
                relPath: dataSourceParams.relPath
            },
            readOnly: readOnly
        });
    }

    /**
     * Reloads the specified scenario and reactivates it if it was previously activated for each
     * data source to which it was added. An event is then broadcast so affected controls can react
     * to the changed scenario.
     *
     * @param scenarioId
     * @private
     */
    function _handleScenarioUpdated(scenarioId) {
        var dsCache = sas.analytics.riskhp.models.DataSourceCacheManager;
        var loadedDataSourceIds = dsCache.getLoadedDataSourceIds();

        _.each(loadedDataSourceIds, function(dataSourceId) {
            dsCache.getDataSource(dataSourceId).done(
                function(dataSource) {
                    // get the original data source
                    // S1484568 scenarioData.scenarioId is really repoId internal to HPRisk UI and is
                    // NOT id!
                    var originalScenario = dataSource.getScenarioDataItemByRepoId(scenarioId);

                    // if the scenario is not loaded in the data source, bail
                    if (!originalScenario) {
                        return;
                    }

                    // Load the scenario and update the data source
                    HprAPI.loadDetailedScenarios(dataSource.getId(), [scenarioId], true).done(
                        function(scenarios) {
                            var updatedScenario = scenarios[0];
                            var dsParams = dsCache.getDataSourceParam(dataSourceId);

                            // figure out the horizon difference
                            var existingHorizonIds = _.pluck(originalScenario.horizons, "id");
                            var updatedHorizonIds = _.pluck(updatedScenario.horizons, "id");
                            var deletedHorizonIds = _.filter(existingHorizonIds, function(existingHorizonId) {
                                return _.findIndex(updatedHorizonIds, function(updatedHorizonId){return updatedHorizonId === existingHorizonId;}) === -1;
                            });
                            var addedHorizonIds = _.filter(updatedHorizonIds, function(updatedHorizonId) {
                                return _.findIndex(existingHorizonIds, function(existingHorizonId) {return existingHorizonId === updatedHorizonId;}) === -1;
                            });

                            // update the name of the data source if the label has changed
                            var changedLabelObjects = dsParams.getDataItemChangedLabelObjects();
                            var changedScenarioLabel = _.find(changedLabelObjects, function(changedLabelObject) {
                                return updatedScenario.getId() === changedLabelObject.dataItemId;
                            });
                            if (changedScenarioLabel) {
                                updatedScenario.setLabel(changedScenarioLabel.newLabel);
                            }

                            // replace the udpated scenario in the data source
                            dataSource.replaceScenario(scenarios[0]);

                            // notify the interested parties the scenario was updated
                            sas.eventBus.publish(HprConstants.CHANNEL_EXPLORATION, HprConstants.EVENT_SCENARIO_UPDATED, {
                                dataSourceId: dataSource.getId(),
                                scenarioId: updatedScenario.getId(),
                                addedHorizonIds: addedHorizonIds,
                                deletedHorizonIds: deletedHorizonIds
                            });
                        }
                    ).fail(
                        function(jqxhr) {
                            //TODO log this?
                        }
                    );
                }
            );
        });
    }

    function _handleScenarioDeleted(scenarioId) {
        var dsCache = sas.analytics.riskhp.models.DataSourceCacheManager;
        var loadedDataSourceIds = dsCache.getLoadedDataSourceIds();

        _.each(loadedDataSourceIds, function(dataSourceId) {
            dsCache.getDataSource(dataSourceId).done(
                function (dataSource) {
                    // get the original data source
                    // S1484568 scenarioData.scenarioId is really repoId internal to HPRisk UI and is
                    // NOT id!
                    var originalScenario = dataSource.getScenarioDataItemByRepoId(scenarioId);

                    // if the scenario is not loaded in the data source, bail
                    if (!originalScenario) {
                        return;
                    };

                    // De-activate the scenario if it is active
                    if (originalScenario.isActivated() === true) {
                        HprAPI.deactivateScenarios(dataSource.getId(), [originalScenario.getId()]);
                    };

                    // scenario is in the data source, so remove it from there and from the data source params
                    dataSource.removeScenarioWithId(originalScenario.getId());
                    dsCache.getDataSourceParam(dataSource.getId()).removeScenarioId(originalScenario.getId());

                    // notify the interested parties the scenario was deleted
                    sas.eventBus.publish(HprConstants.CHANNEL_EXPLORATION, HprConstants.EVENT_SCENARIO_DELETED, {dataSourceId: dataSource.getId(), scenarioIds: [originalScenario.getId()]});
                }
            );
        });
    }


    // MIP handling

    function _handleOpenMIPCube(type, messageData) {
        var self = this;
        var rb = HprUtil.getBundle();

        // create the data source params object so we can open the cube
        var dataSourceParams = new sas.analytics.riskhp.models.DataSourceParams.prototype.createRootParams({
            id: messageData.cubeName,
            application: messageData.application,
            workGroup: messageData.workGroup,
            relPath: messageData.relPath
        });

        // open the cube
        HprAPI.getDataSourceDetails(dataSourceParams).done(
            function(dataSourceDetails) {
                if (dataSourceDetails && dataSourceDetails.dataSources && dataSourceDetails.dataSources.length > 0) {
                    // get the active view
                    var activeView = ViewManager.getActiveView();

                    // if the cube has a price data source, open it, otherwise open the stress data source (MIP cube should always have stress)
                    var dataSourceToOpen = DataSourceUtil.getDataSourceForType(dataSourceDetails, HprConstants.HPR_CUBE_INFORMATION_TYPE_PRICE);
                    if (!dataSourceToOpen) {
                        dataSourceToOpen = DataSourceUtil.getDataSourceForType(dataSourceDetails, HprConstants.HPR_CUBE_INFORMATION_TYPE_STRESS);
                    }

                    // If an exploration is already open, we need to allow the user to save any unsaved changes before
                    // creating the new exploration with the MIP cube. Otherwise, just create a new exploration with the
                    // MIP cube.
                    if (activeView && activeView.getId() === HprConstants.HPR_EXPLORATION_VIEW_ID) {
                        if (activeView.getController().isExplorationDirty()) {
                            var saveButton = new sas.hc.m.Button({
                                text: rb.getText("HPRisk.OpenMIPCube.UnsavedExploration.Save.action.txt"),
                                press: function (evt) {
                                    // save the exploration and open a new one
                                    activeView.getController().onSaveExploration().done(
                                        function () {
                                            _openNewExploration(dataSourceDetails, dataSourceToOpen.type, activeView);
                                        }
                                    );
                                }
                            });
                            var discardButton = new sas.hc.m.Button({
                                text: rb.getText("HPRisk.OpenMIPCube.UnsavedExploration.NoSave.action.txt"),
                                press: function (evt) {
                                    // open the new exploration
                                    _openNewExploration(dataSourceDetails, dataSourceToOpen.type, activeView);
                                }
                            });
                            new sas.hc.m.MessageDialog.warning({
                                title: rb.getText("HPRisk.OpenMIPCube.UnsavedExploration.title.txt"),
                                text: rb.getText("HPRisk.OpenMIPCube.UnsavedExploration.Message.fmt", [dataSourceToOpen.label]),
                                buttonActions: [
                                    saveButton,
                                    discardButton,
                                    sas.hc.m.MessageDialogAction.Cancel
                                ],
                                defaultButton: saveButton
                            });
                        }
                    }
                    else {
                        _openNewExploration(dataSourceDetails, dataSourceToOpen.type, activeView);
                    }
                } else {
                    HprUtil.displayGenericErrorDialog(rb.getText(HPRisk.detailed.error.cube.information.failed.to.load.txt));
                }
            }
        ).fail(
            function(jqxhr, textStatus, error) {
                HprUtil.displayGenericErrorDialog(rb.getText("DataPane.data.source.load.error.fmt", [dataSourceParams.abstractDataSourceId]));
            }
        );
    }


    // internal utils

    function _openNewExploration(dataSourceDetails, dataSourceType, activeView, scenarioData) {
        // open an exploration with the specified name
        var _openNewExplorationFunction = function (explorationName) {
            sas.analytics.riskhp.utils.ViewManager.loadNewExploration(
                explorationName,
                dataSourceDetails,
                dataSourceType
            ).done(
                function(stressDataSource) {
                    activeView = ViewManager.getActiveView(); // active view is now the new exploration

                    if (scenarioData && scenarioData.scenarioId) {
                        HprAPI.loadDetailedScenarios(stressDataSource.dataSource.id, [scenarioData.scenarioId]).done(
                            function(scenarios) {
                                // add the scenario to the data source
                                var scenario = scenarios[0];
                                stressDataSource.addScenario(scenario);
                                DataSourceCacheManager.getDataSourceParam(stressDataSource.getId()).addScenarioId(scenario.getId());
                                DataPaneUtil.updateScenarioNodes(stressDataSource, activeView.getDataPane().getDataPaneView().getController());

                                // run the scenario
                                _getActiveCrosstab().runScenario(scenario);
                            }
                        ).fail(
                            function(jqxhr) {
                                HprUtil.displayGenericErrorDialog(rb.getText("RunScenario.LoadScenario.error.fmt", [scenarioData.scenarioName]));
                            }
                        ).always(
                            function() {
                                if (activeView) {
                                    activeView.setBusy(false);
                                }
                            }
                        );
                    }
                    else {
                        if (activeView) {
                            activeView.setBusy(false);
                        }
                    }
                }
            ).fail(
                function (jqxhr, textStatus, error) {
                    HprUtil.displayGenericErrorDialog(rb.getText("DataPane.data.source.load.error.fmt", [dataSourceDetails.label]));
                    if (activeView) {
                        activeView.setBusy(false);
                    }
                }
            );
        };

        // set the active view to busy while we create the exploration
        if (activeView) {
            activeView.setBusy(true);
        }

        // get the explorations so we can attempt to set a reasonably valid name for it
        HprAPI.getExplorations().done(
            function (data) {
                _openNewExplorationFunction(HprUtil.getUniqueNewExplorationName(data.items));
            }
        ).fail(
            function (jqxhr, textStatus, error) {
                // attempt a reasonable default if we couldn't get the explorations for some crazy reason
                _openNewExplorationFunction(HprUtil.getUniqueNewExplorationName());
            }
        );
    }

    /**
     * Returns the first crosstab from the active page.
     * @returns {sas.analytics.riskhp.common.riskcommons.crosstab.HprCrosstab}
     * @private
     */
    function _getActiveCrosstab() {
        return sap.ui.getCore().byId(HprUtil.getSelectedRiskPanelsView().getDelegate().crosstabIds[0]);
    }


    return {
        /**
         * Subscribes to application switcher messages that HPR will
         * respond to.
         */
        initialize: function () {
            // appswitcher messages
            sas.applicationSwitcher.subscribe(sas.applicationSwitcher.APP_SHOW_MESSAGETYPE, _handleAppShowMessage);

            // scenario manager messages
            sas.applicationSwitcher.subscribe(SCENARIOS_UPDATED_MSG, _handleScenariosUpdatedMessage);
            sas.applicationSwitcher.subscribe(RUN_SCENARIO_MSG, _handleRunScenario);

            // mip messages
            sas.applicationSwitcher.subscribe(OPEN_MIP_CUBE, _handleOpenMIPCube);

            // get URLs for supported apps
            scenarioManagerURL = sas.applicationSwitcher.getApplicationUrl(SCENARIO_MANAGER_APP, SCENARIO_MANAGER_APP);
            vaHubURL = sas.applicationSwitcher.getApplicationUrl(HprConstants.VAHUB_APPLICATION_ID, HprConstants.VAHUB_APPLICATION_ID);
        },


        // public functions

        getVAHubURL: function() {
            return vaHubURL;
        },

        getScenarioManagerAppId: function() {
            return SCENARIO_MANAGER_APP;
        },

        isScenarioManagerAvailable: function() {
            return !!scenarioManagerURL;
        },

        createScenario: function(dataSourceId, dataSourceParams) {
            // should never get here, but just in case
            if (!scenarioManagerURL) {
                return;
            }

            // open scenario manager and tell it we want to create a scenario
            sas.applicationSwitcher.gotoApp(SCENARIO_MANAGER_APP, scenarioManagerURL, true);
            sas.applicationSwitcher.sendMessageToApp(SCENARIO_MANAGER_APP, CREATE_SCENARIO_MSG, {
                scenarioWorkGroup: dataSourceParams.workGroup,
                dataSourceId: dataSourceParams.abstractDataSourceId,
                dataSourceParms: {
                    workGroup: dataSourceParams.workGroup,
                    application: dataSourceParams.application,
                    relPath: dataSourceParams.relPath
                }
            });
        },

        viewScenario: function(scenario, dataSourceParams) {
            // should never get here, but just in case
            if (!scenarioManagerURL) {
                return;
            }

            _viewOrEditScenario(scenario, dataSourceParams, true);
        },

        editScenario: function(scenario, dataSourceParams) {
            // should never get here, but just in case
            if (!scenarioManagerURL) {
                return;
            }

            _viewOrEditScenario(scenario, dataSourceParams);
        }
    };
}();