// (c) 2015, SAS Institute Inc.

jQuery.sap.declare("sas.analytics.riskhp.utils.HprUtil");

jQuery.sap.require("sap.ui.core.format.DateFormat");
jQuery.sap.require("sap.ui.core.routing.Router");
jQuery.sap.require("sas.analytics.riskhp.models.dataitems.Horizon");
jQuery.sap.require("sas.analytics.riskhp.models.dataitems.Outvar");
jQuery.sap.require('sas.analytics.riskhp.models.dataitems.RelativeHorizon');
jQuery.sap.require("sas.analytics.riskhp.models.dataitems.ResultSet");
jQuery.sap.require("sas.analytics.riskhp.models.dataitems.RiskCube");
jQuery.sap.require("sas.analytics.riskhp.models.dataitems.Scenario");
jQuery.sap.require("sas.analytics.riskhp.utils.HprConstants");
jQuery.sap.require("sas.hc.m.MessageDialog");
jQuery.sap.require("sas.hc.ui.services.navigation.ContentSelector");
jQuery.sap.require("sas.hc.ui.services.navigation.classes.NavigatorToolBar");
jQuery.sap.require("sas.hc.ui.services.navigation.config.FilterItem");
jQuery.sap.require("sas.hc.ui.services.navigation.config.FilteredItemsConfig");
jQuery.sap.require("sas.hc.ui.services.navigation.config.ModelConfig");
jQuery.sap.require("sas.hc.ui.services.navigation.config.ServiceConfig");
jQuery.sap.require("sas.hc.ui.services.navigation.types.ObjectTypeConstants");
jQuery.sap.require("sas.risk.common.utils.NavChecker");
jQuery.sap.require("sas.ltjs.sasicu.util.SASFormatterUtil");

/**
 * @name sas.analytics.riskhp.utils.HprUtil
 */
sas.analytics.riskhp.utils.HprUtil = {
    _pageCountMap: {},

    //*****************************************************************************************************
    //*****************************************************************************************************
    // General functions
    //*****************************************************************************************************
    //*****************************************************************************************************
    /**
     * Return the HPRisk resource bundle
     * @returns {jQuery.sap.util.ResourceBundle}
     */
    getBundle: function() {
        return sap.ui.getCore().getLibraryResourceBundle("sas.analytics.riskhp.nls");
    },

    /**
     * Create a NavChecker and start if needed.  NavChecker detects navigation requests for a page and warns/prompts
     * a user before proceeding if the page has unsaved changes.
     * @param {boolean} start Whether or not start immediately
     * @param {sap.ui.core.mvc.View} view The View that the NavChecker needs to destroy when exit
     * @return {sas.risk.common.utils.NavChecker} the navChecker
     */
    createNavChecker: function (start, view) {
        var router = sap.ui.core.routing.Router.getRouter(sas.analytics.riskhp.utils.HprConstants.HPR_ROUTER_ID);
        var navChecker = new sas.risk.common.utils.NavChecker(
            sas.config.getAppId(),
            sas.analytics.riskhp.utils.HprUtil.getBundle().getText('ExplorationView.unsaved.changes.txt'),
            router
        );
        if (start) {
            if (view && view.destroy) {
                var callback = $.proxy(view.destroy, view);
            }
            navChecker.start(callback);
        }
        return navChecker;
    },

    /**
     * Selects all the text within an input field.
     * @param {sas.hc.m.Input} input input control
     * @return {void}
     */
    selectAllInputText: function(input) {
        input.selectText(0, input.getValue().length);
    },

    /**
     * A "poor-man" clone of a json object that avoids the jQuery version so we get "true" deep copies of json objects
     * @param json
     * @return A deep copy of the passed in JSON.
     */
    hprClone: function(json) {
        return JSON.parse(JSON.stringify(json));
    },

    //*****************************************************************************************************
    //*****************************************************************************************************
    // String functions
    //*****************************************************************************************************
    //*****************************************************************************************************
    /**
     * Return true if 'string' ends with 'token'.
     * @param string The string to search through
     * @param token The token to find at the end of the string
     * @returns {boolean} Whether the string ends with the specified token
     */
    stringEndsWith: function(string, token) {
        if($.type(string) === "string" && $.type(token) === "string")
            return string.indexOf(token, string.length - token.length) !== -1;
        return false;
    },

    //*****************************************************************************************************
    //*****************************************************************************************************
    // Date format functions
    //*****************************************************************************************************
    //*****************************************************************************************************
    /**
     * Given a value, create a Date object
     * @param value Value (Date, Number, String) which will be turned into a Date object
     * @returns {Date} A Date representation of the value.
     */
    valueToDate: function(value) {
        var dateValue = null;
        if($.type(value) === "date") {
            dateValue = value;
        } else if($.type(value) === "number") {
            dateValue = new Date(value);
        } else if(!isNaN(Date.parse(value))) {
            dateValue = new Date(value);
        }

        return dateValue;
    },

    /**
     * Convert the given date (either Date or String) to a String label that displays the date and time.
     * @param value Value to convert to String representation of a Date
     * @param formatOptions Format options for the Date
     * @returns {String} String which displays the data and time according to the format.
     */
    formatDateTime: function(value, formatOptions) {
        var dateValue = sas.analytics.riskhp.utils.HprUtil.valueToDate(value);
        if(dateValue) {
            var fmt = sap.ui.core.format.DateFormat.getDateTimeInstance(formatOptions);
            return fmt.format(dateValue);
        }

        return "";
    },

    /**
     * Takes the given ISO Date string and converts it to an equivalent localized JavaScript Date.  May return null if the
     * string could not be converted.
     * @param isoDateString
     * @returns {Date|null}
     */
    convertISODateToLocalJSDate: function(isoDateString) {
        var dateValue = sas.analytics.riskhp.utils.HprUtil.valueToDate(isoDateString);
        if(!dateValue) {
            return null;
        }

        // Have to account for locale, which is specific for each date due to DST considerations

        // Get the time offset in minutes
        var offset = dateValue.getTimezoneOffset();

        // Figure out how many hours we are off
        var hours = Math.floor(Math.abs(offset)/60);

        // If minutes are greater than zero then the date is ahead of UTC so we subtract
        var offsetSign = "+";
        if (offset > 0) {
            offsetSign = "-";
        }

        // Make sure we generate a two digit hour string
        var hoursString = "";
        if (hours < 10) {
            hoursString = "0"+hours;
        }
        else {
            hoursString += hours;
        }

        // Calculate remaining minutes and make sure we create a two digit minute string
        var minutes = offset % 60;
        var minutesString = "";
        if (minutes < 10) {
            minutesString = "0"+minutes;
        }
        else {
            minutesString += minutes;
        }

        // Create string representation of the date with locale so when we run it through the formatter it will be
        // midnight in the user's locale and the resultant date String will not be off by a day.
        var localizedDate = isoDateString + "T00:00"+offsetSign+hoursString+":"+minutesString;
        return new Date(localizedDate);
    },

    /**
     * Convert the given date (either Date or String) to a String label that displays the date only.
     * @param isoDateString an ISO representation of a date (i.e. YYYY-MM-DD)
     * @returns {String} String suitable for display of a date or blank if no date was passed in.
     */
    formatISODateString: function(isoDateString) {
        var localizedJSDate = sas.analytics.riskhp.utils.HprUtil.convertISODateToLocalJSDate(isoDateString);
        if(!localizedJSDate) {
            return "";
        }
        return sas.ltjs.sasicu.util.SASFormatterUtil.format(localizedJSDate, 'NLDATE10');
    },

    //*****************************************************************************************************
    //*****************************************************************************************************
    // Select & List functions
    //*****************************************************************************************************
    //*****************************************************************************************************
    /**
     * Adds a array of objects to a field as list items.
     * @param {field} field form field with addItem() function
     * @param {array} array array of objects
     * @param {string} textPropertyName object text property name
     * @param {string} [keyPropertyName] optional object key property name
     * @return {void}
     */
    addListItems: function(field, array, textPropertyName, keyPropertyName) {
        for(var i = 0; i < array.length; i++) {
            var item = array[i];
            if(keyPropertyName) {
                field.addItem(new sas.hc.ui.core.Item({
                    text: item[textPropertyName],
                    key: item[keyPropertyName]
                }));
            } else {
                field.addItem(new sas.hc.ui.core.Item({
                    text: item[textPropertyName]
                }));
            }
        }
    },

    /**
     * Copy all properties from obj1 into obj2 so that obj2
     * is a deep copy of obj1.  Any properties that previously
     * existed in obj2 but not obj1 will be deleted.
     * @param {object} obj1 object to copy properties from
     * @param {object} obj2 object to copy all obj1 properties too
     * @return {object} returns modified obj2 object
     */
    _deepCopy: function(obj1, obj2) {
        if(!obj1 || !obj2) return;

        // delete all existing properties in obj2
        for(prop in obj2) {
            if(obj2.hasOwnProperty(prop)) {
                delete obj2[prop];
            }
        }

        $.extend(true, obj2, obj1);
    },

    /**
     * Simple sleep function for testing
     * @param {number} ms - time to wait in milliseconds
     */
    sleep: function( ms ){
        var now = new Date().getTime();
        while(new Date().getTime() < now + ms){ /* do nothing */ }
    },

    /**
     * Search the given list and return the item that matches the
     * given value.  If property is specified (e.g. id), it will return
     * the item where that property matches the given value.  If no match
     * is found, returns null.
     * @param {array} list array items to search
     * @param {string | number} value value to match property value too
     * @param {string} property property of array item to try and match
     * @param {boolean} ignoreCaseForValue whether we should match case for value
     * @return {object} returns null or matching item
     */
    findObject: function(list, value, property, ignoreCaseForValue) {
        var rval = null;
        if($.isArray(list)) {
            $.each(list, function(index, item) {
                if(property) {
                    if(item.hasOwnProperty(property)) {
                        var propertyValue = item[property];
                        if (ignoreCaseForValue) {
                            if (String(propertyValue).toLowerCase() === String(value).toLowerCase()) {
                                rval = item;
                                return false; // stop iterating
                            }
                        } else {
                            if (propertyValue === value) {
                                rval = item;
                                return false; // stop iterating
                            }
                        }
                    }
                } else if(item === value) {
                    rval = item;
                    return false; // stop iterating
                }
            });
        }
        return rval;
    },

    //*****************************************************************************************************
    //*****************************************************************************************************
    // Exploration functions
    //*****************************************************************************************************
    //*****************************************************************************************************
    /**
     * Returns a unique name for a new exploration using a template of 'Risk Exploration [n]' where [n] is
     * an integer, i.e. 'Risk Exploration 1'.
     * @param explorations
     */
    getUniqueNewExplorationName: function(explorations) {
        var self = this;
        var count = 0;

        var _getNextDefaultExplorationName = function() {
            count++;
            return self.getBundle().getText("ChooseDataSourceDialog.defaultNewExplorationName.fmt", count);
        };

        var uniqueName = _getNextDefaultExplorationName();
        if (explorations) {
            // Loop through the Explorations until we don't have a match on the default name that we would like to use
            while (self.findObject(explorations, uniqueName, "name", true)) {
                uniqueName = _getNextDefaultExplorationName();
            }
        }

        return uniqueName;
    },

    /**
     * Opens the ChooseDataSourceDialog in the New Exploration mode which will gather information for creating a new exploration
     * @param cubes
     * @param explorations
     * @param releaseAllLoadedDataSources
     * @param controller
     */
    openNewExplorationDialog: function(cubes, explorations, releaseAllLoadedDataSources, controller) {
        // Open the ChooseDataSourceDialog with the list of cubes and explorations
        var dialogModel = new sap.ui.model.json.JSONModel();
        dialogModel.setProperty("/dataSourceItemsPath", "/dataSourceItems");
        dialogModel.setProperty("/dataSourceItems", cubes);
        dialogModel.setProperty("/explorations", explorations);
        dialogModel.setProperty("/releaseAllLoadedDataSources", releaseAllLoadedDataSources);
        dialogModel.setProperty("/dialogTitle", sas.analytics.riskhp.utils.HprUtil.getBundle().getText('ChooseDataSourceDialog.new.exploration.title.txt'));
        dialogModel.setProperty("/dialogMode", HprConstants.HPR_CHOOSEDATASOURCE_DIALOG_NEW_EXPLORATION_MODE);
        var chooseDataSourceDialog = sap.ui.jsfragment("sas.analytics.riskhp.fragments.ChooseDataSourceDialog", controller);
        chooseDataSourceDialog.setModel(dialogModel);
        chooseDataSourceDialog.open();
    },

    /**
     * Handler for when the "New Exploration" action is initiated.  "this" is assumed to be a controller which can be
     * accomplished by using the JQuery proxy API and attaching this method to an Event handler for a MenuItem.
     * A dialog fragment (which needs a controller) will be opened.
     */
    handleNewExploration: function() {
        if (!(this instanceof sap.ui.core.mvc.Controller)) {
            return;
        }
        var mainController = this; // Main controller for this method

        // Function to perform when we need to create a new Exploration
        var onNewExploration = function(controllerToUse) {

            // Disable the New button on the Welcome screen to prevent fast double-click errors while cubes are loading
            var welcomeViewNewButton = undefined;
            var welcomeView = sas.analytics.riskhp.utils.ViewManager.getWelcomeView();
            if (welcomeView) {
                // Welcome view will only be initialized if Exploration is dirty so account for that
                welcomeViewNewButton = welcomeView.getNewButton();
                welcomeViewNewButton.setEnabled(false);
            }

            $.when (
                // Load the cubes
                sas.analytics.riskhp.utils.HprAPI.getCubes(),
                // Load the explorations
                sas.analytics.riskhp.utils.HprAPI.getExplorations()
            ).done(function(cubes, explorations) {
                // Pass the controller we need to use, mainly because we get the associated view from it and set the busy state on it
                sas.analytics.riskhp.utils.HprUtil.openNewExplorationDialog(cubes, explorations.items, true, controllerToUse);
            }).fail(function() {
                sas.analytics.riskhp.utils.HprUtil.displayGenericErrorDialog();
            }).always(function() {
                if (welcomeViewNewButton) {
                    welcomeViewNewButton.setEnabled(true); // Re-enable New button after everything is done running
                }
            });
        };

        // Would love to use "sas.risk.common.utils.NavChecker" here like in other places of the UI, but it only supports navigating
        // to another view using the router.  In this case, we need to stay on the current view and reset.  Just use the MessageDialog
        // which the NavChecker uses behind the scenes.
        if (_.isFunction(mainController.isExplorationDirty)) {
            // If the Exploration is dirty, put up a MessageDialog to see if the user wants to navigate away
            if (mainController.isExplorationDirty()) {
                sas.analytics.riskhp.utils.HprUtil._displayDirtyExplorationMessageDialog(onNewExploration);
            } else {
                onNewExploration(mainController);
            }
        }
    },

    /**
     *
     *
     * @param explorationID
     * @param explorationName
     */
    handleCopyExploration: function(explorationID, explorationName, controller) {
        var deferred = $.Deferred();
        var newExplorationName = this.getBundle().getText("SaveExplorationDialog.saveAsName.fmt", [explorationName]);

        // S1422616: The ContentSelector Save dialog has Breadcrumb links with an href of '#'.  This causes problems for our
        // Router and NavChecker code in that '#' triggers the Router code into thinking it is a hash and view change.  We have
        // to stop the NavChecker and Router before the ContentSelector dialog appears so we do not change views when the
        // Breadcrumb links are clicked.  The NavChecker is re-started in the ContentSelector.closeHandler code below.
        controller.navChecker.stop();
        sas.analytics.riskhp.utils.ViewManager.getRouter().stop();

        var __copyExplorationCallback = function(result, dataDescriptor) {
            // S1422616: When the ContentSelector is closed, restart the NavChecker (which ultimately ends up restarting the Router) so
            // we can safely navigate to views and properly receive NavChecker discard dialogs
            var view = controller.getView();
            var navCheckerCallback = $.proxy(view.destroy, view);
            controller.navChecker.start(navCheckerCallback);

            // User clicked 'Save'
            if ( result.detail === sas.hc.ui.services.navigation.ContentSelector.CloseDetail.COMMIT ) {
                // Create an Exploration object which will be saved with the necessary information
                var explorationObject = {
                    id: explorationID,  // tells the midtier this is a copy
                    version: 1,
                    name: result.saveItemName,
                    folderEntityKey: dataDescriptor.getId(result.containerItem),
                    folderPath: dataDescriptor.getAbsolutePath(result.containerItem),
                };

                sas.analytics.riskhp.utils.HprAPI.createExploration(explorationObject).done(
                    function(data) {
                        deferred.resolve(data);
                    }
                ).fail(
                    function(jqxhr, textStatus, error) {
                        deferred.reject(jqxhr, textStatus, error);
                    }
                );
            }
            else {
                deferred.resolve();
            }
        };

        this.showContentSelector(
            controller.createId("-contentSelectorSaveDlg"),
            sas.hc.ui.services.navigation.ContentSelector.Mode.SAVE,
            __copyExplorationCallback,
            newExplorationName
        );

        return deferred.promise();
    },

    /**
     * Handler for when the "Find Exploration" action is initiated.  "this" is assumed to be a controller which can be
     * accomplished by using the JQuery proxy API and attaching this method to an Event handler for a MenuItem.  The
     * ContentSelector Open dialog is presented to the user where the can go find an Exploration in metadata to open.
     */
    handleFindExploration: function() {
        // Display a ContentSelector Open Dialog
        if (!(this instanceof sap.ui.core.mvc.Controller)) {
            return;
        }
        var mainController = this; // Main controller for this method

        // Function to perform when we need to open an Exploration
        var onOpenExploration = function(controllerToUse) {
            // S1422616: The ContentSelector Save dialog has Breadcrumb links with an href of '#'.  This causes problems for our
            // Router and NavChecker code in that '#' triggers the Router code into thinking it is a hash and view change.  We have
            // to stop the NavChecker and Router before the ContentSelector dialog appears so we do not change views when the
            // Breadcrumb links are clicked.  The NavChecker is re-started in the ContentSelector.closeHandler code below.
            controllerToUse.navChecker.stop();
            sas.analytics.riskhp.utils.ViewManager.getRouter().stop();

            sas.analytics.riskhp.utils.HprUtil.showContentSelector(
                controllerToUse.createId("-contentSelectorOpenDlg"),
                sas.hc.ui.services.navigation.ContentSelector.Mode.OPEN,
                function (result, dataDescriptor) {
                    // S1422616: When the ContentSelector is closed, restart the NavChecker (which ultimately ends up restarting the Router) so
                    // we can safely navigate to views and properly receive NavChecker discard dialogs
                    var viewToUse = controllerToUse.getView();
                    var navCheckerCallback = $.proxy(viewToUse.destroy, viewToUse);
                    controllerToUse.navChecker.start(navCheckerCallback);

                    // User clicked 'Open'
                    if ( result.detail === sas.hc.ui.services.navigation.ContentSelector.CloseDetail.COMMIT ) {
                        // get the ID for the selected exploration
                        var id = dataDescriptor.getId(result.selectedItem);

                        // Set view to busy state
                        viewToUse.setBusy(true);

                        // Release the current cube resources before opening a saved Exploration
                        $.when (
                            DataSourceCacheManager.releaseAllLoadedDataSources()
                        ).done(function() {
                            // Open the selected Exploration
                            var promise = sas.analytics.riskhp.utils.ViewManager.loadSavedExploration(id);
                            promise.done(sas.analytics.riskhp.utils.HprUtil.createSuccessHandler(controllerToUse.getView(),null));

                            // This method is ran in the failure handler when it completes
                            var openExplorationFailureHandlerCompletionFunction = function() {
                                // Check to see if we are in ExplorationView, and if so route to WelcomeView when there is a failure
                                var explorationView = sas.analytics.riskhp.utils.ViewManager.getExplorationView();
                                if (explorationView) {
                                    var explorationViewController = explorationView.getController();
                                    explorationViewController.navChecker.setDirty(false); // Have to set navChecker dirty state to 'false' so router will navigate correctly
                                    sas.analytics.riskhp.utils.ViewManager.loadHprWelcomeView();
                                }
                            };
                            promise.fail(sas.analytics.riskhp.utils.HprUtil.createFailureHandler(controllerToUse.getView(),true,openExplorationFailureHandlerCompletionFunction));
                        }).fail(function() {
                            viewToUse.setBusy(false);
                            sas.analytics.riskhp.utils.HprUtil.displayGenericErrorDialog();
                        });
                    }
                }
            );
        };

        // Would love to use "sas.risk.common.utils.NavChecker" here like in other places of the UI, but it only supports navigating
        // to another view using the router.  In this case, we need to stay on the current view and reset.  Just use the MessageDialog
        // which the NavChecker uses behind the scenes.
        if (_.isFunction(mainController.isExplorationDirty)) {
            // If the Exploration is dirty, put up a 'confirm' prompt to see if the user wants to navigate away
            if (mainController.isExplorationDirty()) {
                sas.analytics.riskhp.utils.HprUtil._displayDirtyExplorationMessageDialog(onOpenExploration);
            } else {
                onOpenExploration(mainController);
            }
        } else {
            // Need to use HprWelcome.controller in this case
            var welcomeViewController = sas.analytics.riskhp.utils.ViewManager.getWelcomeView().getController(); // Controller class for HprWelcome.view
            onOpenExploration(welcomeViewController); // This method is called by numerous controller's, handle the case where 'isExplorationDirty' is not defined
        }
    },

    /**
     * Displays the content selector with the desired configuration parameters.
     *
     * @param id {String} the id for the dialog
     * @param mode {sas.hc.ui.services.navigation.ContentSelector.Mode} the mode of the Content Selector
     * @param callback {Function} callback function to handle the response from the Content Selector
     * @param name {String} default name of the exploration to save (optional)
     */
    showContentSelector: function(id, mode, callback, name) {
        var ContentSelector = sas.hc.ui.services.navigation.ContentSelector,
            FilteredItemsConfig = sas.hc.ui.services.navigation.config.FilteredItemsConfig,
            FilterItem = sas.hc.ui.services.navigation.config.FilterItem,
            ModelConfig = sas.hc.ui.services.navigation.config.ModelConfig,
            ObjectTypeConstants = sas.hc.ui.services.navigation.types.ObjectTypeConstants,
            ServiceConfig = sas.hc.ui.services.navigation.config.ServiceConfig,
            NavigatorToolBar = sas.hc.ui.services.navigation.classes.NavigatorToolBar;

        var sasFoldersSvcConfig = new ServiceConfig(ServiceConfig.Types.SAS_FOLDERS,
            new FilteredItemsConfig(
                [new FilterItem("RiskExploration")],
                false,
                [ObjectTypeConstants.TYPE_SYSTEM_FOLDER_NAME, ObjectTypeConstants.TYPE_JCR_FOLDER_NAME, ObjectTypeConstants.TYPE_REPOSITORY_ROOT_FOLDER_NAME]
            )
        );

        // default configuration for all use cases
        var config = {
            id : id,
            modelConfig : new ModelConfig([sasFoldersSvcConfig]),
            toolbarActionsToRemove: [NavigatorToolBar.DELETE_ITEM]
        };

        // if a name is specified, add it to the config
        if (name) {
            config.nameFieldChoices = [name];
        }

        // allow search for open
        if (mode === ContentSelector.Mode.OPEN) {
            config.searchEnabled = true;
        }

        // show the content selector
        ContentSelector.show(
            mode,
            config,
            function closeHandler(result){
                callback(result, sasFoldersSvcConfig.getDataDescriptor());
            }
        );
    },

    /**
     * Create and open a warning MessageDialog instance that gives the user an option to discard their unsaved changes.  Modeled after the
     * Risk Commons MessageDialog used in the 'sas.risk.common.utils.NavChecker' class.
     * @param callbackFn Function to callback when the user decides to Discard changes
     * @private
     */
    _displayDirtyExplorationMessageDialog: function(callbackFn) {
        var discardText = sas.analytics.riskhp.utils.HprUtil.getBundle().getText('hprisk.discardButton.txt');
        var oMessageDialog = new sas.hc.m.MessageDialog({
            type: sas.hc.m.MessageDialogType.Warning,
            title: null, // passing null defaults to the application name
            text: sas.analytics.riskhp.utils.HprUtil.getBundle().getText('ExplorationView.unsaved.changes.txt'),
            contentWidth: "28rem",
            buttonActions: [
                discardText,
                {
                    text: sas.hc.m.MessageDialogAction.Cancel,
                    initialFocus: true
                }
            ],
            buttonPress: function (e) {
                var button = e.getParameter("origin");
                if (button.getText() === discardText) {
                    // To truly discard changes for the Exploration, we need to go back out to the Welcome screen and then call the callbackFn
                    var explorationViewController = sas.analytics.riskhp.utils.ViewManager.getExplorationView().getController();
                    explorationViewController.navChecker.setDirty(false); // Have to set navChecker dirty state to 'false' so router will navigate correctly
                    sas.analytics.riskhp.utils.ViewManager.loadHprWelcomeView();

                    // Since we navigated back out to Welcome screen, we need that controller and view so we can set busy state properly for the callbackFn's
                    var welcomeViewController = sas.analytics.riskhp.utils.ViewManager.getWelcomeView().getController(); // Controller class for HprWelcome.view

                    // Call the callbackFn with a slight pause because we are changing views and it can be disorienting
                    setTimeout(function() {
                        callbackFn(welcomeViewController);
                    }, 0);
                }
            }
        });
        oMessageDialog.open();
    },

    //*****************************************************************************************************
    //*****************************************************************************************************
    // Page functions
    //*****************************************************************************************************
    //*****************************************************************************************************
    /**
     * Return a unique count (starting at 1) for the number of Pages.
     * @returns {int} Count of Pages
     */
    getUniquePageCount: function() {
        var type = "page";
        var currentCount = 0;
        if(sas.analytics.riskhp.utils.HprUtil._pageCountMap.hasOwnProperty(type)) {
            currentCount = sas.analytics.riskhp.utils.HprUtil._pageCountMap[type];
        }
        sas.analytics.riskhp.utils.HprUtil._pageCountMap[type] = currentCount + 1;
        return sas.analytics.riskhp.utils.HprUtil._pageCountMap[type];
    },

    /**
     * Reset the count of Pages to 0.
     */
    resetUniquePageCount: function() {
        sas.analytics.riskhp.utils.HprUtil._pageCountMap["page"] = 0;
    },

    //*****************************************************************************************************
    //*****************************************************************************************************
    // Data Item functions
    //*****************************************************************************************************
    //*****************************************************************************************************
    /**
     * Given one of the dataitem type constants returns the title for that type.  For now this is a plural displayable
     * string
     * @param dataType
     * @returns {string}
     */
    dataTypeTitle: function(dataType) {
        switch(dataType) {
            case sas.analytics.riskhp.models.dataitems.RiskCube.RISKCUBE:
                return sas.analytics.riskhp.utils.HprUtil.getBundle().getText("DataPane.node.RiskCubes.title.txt");
            case sas.analytics.riskhp.models.dataitems.Classvar.CLASSVAR:
                return sas.analytics.riskhp.utils.HprUtil.getBundle().getText("DataPane.node.ClassVars.title.txt");
            case sas.analytics.riskhp.models.dataitems.Hierarchy.HIERARCHY:
                return sas.analytics.riskhp.utils.HprUtil.getBundle().getText("DataPane.node.Hierarchies.title.txt");
            case sas.analytics.riskhp.models.dataitems.Horizon.HORIZON:
                return sas.analytics.riskhp.utils.HprUtil.getBundle().getText("DataPane.node.Horizons.title.txt");
            case sas.analytics.riskhp.models.dataitems.RelativeHorizon.RELATIVEHORIZON:
                // Use Horizons label in this case, as Relative Horizons should have same label
                return sas.analytics.riskhp.utils.HprUtil.getBundle().getText("DataPane.node.Horizons.title.txt");
            case sas.analytics.riskhp.models.dataitems.AbsoluteHorizon.ABSOLUTEHORIZON:
                // Use Horizons label in this case, as Absolute Horizons should have same label
                return sas.analytics.riskhp.utils.HprUtil.getBundle().getText("DataPane.node.Horizons.title.txt");
            case sas.analytics.riskhp.models.dataitems.Outvar.OUTVAR:
                return sas.analytics.riskhp.utils.HprUtil.getBundle().getText("DataPane.node.OutVars.title.txt");
            case sas.analytics.riskhp.models.dataitems.Compvar.COMPVAR:
                return sas.analytics.riskhp.utils.HprUtil.getBundle().getText("DataPane.node.CompVars.title.txt");
            case sas.analytics.riskhp.models.dataitems.Benchvar.BENCHVAR:
                return sas.analytics.riskhp.utils.HprUtil.getBundle().getText("DataPane.node.BenchmarkVars.title.txt");
            case sas.analytics.riskhp.models.dataitems.Statistic.STATISTIC:
                return sas.analytics.riskhp.utils.HprUtil.getBundle().getText("DataPane.node.Statistics.title.txt");
            case sas.analytics.riskhp.models.dataitems.ResultSet.RESULTSET:
                return sas.analytics.riskhp.utils.HprUtil.getBundle().getText("DataPane.node.ResultSets.title.txt");
            case sas.analytics.riskhp.models.dataitems.Scenario.SCENARIO:
                return sas.analytics.riskhp.utils.HprUtil.getBundle().getText("DataPane.node.Scenarios.title.txt");
        }
        return sas.analytics.riskhp.utils.HprUtil.getBundle().getText("DataPane.node.unknown.title.txt");
    },

    //*****************************************************************************************************
    //*****************************************************************************************************
    // AJAX helper functions
    //*****************************************************************************************************
    //*****************************************************************************************************
    /**
     * Displays a generic warning message dialog with default warning message and default title.  A more specific warning message
     * can be provided if is so desired.
     * @param errorMessage If specified, will be used in warning dialog.  Otherwise, a generic warning message will be used.
     */
    displayGenericWarningDialog: function(errorMessage) {
        var errorText = errorMessage ? errorMessage : sas.analytics.riskhp.utils.HprUtil.getBundle().getText('warning.message.txt');
        sas.hc.m.MessageDialog.warning({
            text: errorText,
            title: sas.analytics.riskhp.utils.HprUtil.getBundle().getText("warning.window.title")
        });
    },

    /**
     * Displays a generic error message dialog with default error message and default title.  A more specific error message
     * can be provided if is so desired.
     * @param errorMessage If specified, will be used in error dialog.  Otherwise, a generic error message will be used.
     */
    displayGenericErrorDialog: function(errorMessage) {
        var errorText = errorMessage ? errorMessage : sas.analytics.riskhp.utils.HprUtil.getBundle().getText('error.message.txt');
        sas.hc.m.MessageDialog.error({
            text: errorText,
            title: sas.analytics.riskhp.utils.HprUtil.getBundle().getText("error.window.title")
        });
    },

    /**
     * Displays a detailed error message dialog.
     * @param jqxhr If specified, response information will be used in error dialog.
     *      Otherwise, a generic error message will be used.
     */
    displayDetailedErrorDialog: function(jqxhr) {
        if (jqxhr) {
            if (jqxhr.responseJSON || jqxhr.getParameter("responseText")) {
                var responseJSON = jqxhr.responseJSON;
                if (!responseJSON) {
                    try {
                        responseJSON = JSON.parse(jqxhr.getParameter("responseText"));
                    }
                    catch(error) {
                        sas.analytics.riskhp.utils.HprUtil.displayGenericErrorDialog();
                        return;
                    }
                }

                // if (responseJSON) {
                    var mainErrorMsg = undefined;

                    // Create the message text to display
                    var errorResourceKey = "HPRisk.ErrorCode." + responseJSON.errorCode + ".txt";
                    var hasErrorMsg = sas.analytics.riskhp.utils.HprUtil.getBundle().hasText(errorResourceKey);
                    if (hasErrorMsg) {
                        mainErrorMsg = sas.analytics.riskhp.utils.HprUtil.getBundle().getText(errorResourceKey);
                    }

                    if (!hasErrorMsg || mainErrorMsg === undefined || mainErrorMsg.length === 0) {
                        mainErrorMsg = sas.analytics.riskhp.utils.HprUtil.getBundle().getText("HPRisk.ErrorCode.1000.txt");
                    }

                    // Format the details into a proper string.
                    var detailString = "";
                    _.each(responseJSON.details, function (str) {
                        detailString += (str + "\n");
                    });

                    // build detail text for the dialog
                    var detailText = sas.analytics.riskhp.utils.HprUtil.getBundle().getText("HPRisk.detailed.error.dialog.error.message.fmt", [responseJSON.message]) +
                        "\n" +
                        sas.analytics.riskhp.utils.HprUtil.getBundle().getText("HPRisk.detailed.error.dialog.error.code.fmt", [responseJSON.errorCode]) +
                        "\n" +
                        sas.analytics.riskhp.utils.HprUtil.getBundle().getText("HPRisk.detailed.error.dialog.status.code.fmt", [responseJSON.httpStatusCode]);

                    // if there were details in the response, add them to the detail text
                    if (detailString) {
                        detailText += ("\n\n" + detailString);
                    }

                    // Show the error dialog
                    sas.hc.m.MessageDialog.error({
                        text: mainErrorMsg,
                        title: sas.analytics.riskhp.utils.HprUtil.getBundle().getText("error.window.title"),
                        detailText: detailText
                    });
                    return;
                }
            // }
        }
        sas.analytics.riskhp.utils.HprUtil.displayGenericErrorDialog();
    },

    /**
     * Creates a common failure handler function that when called will turn off the view's busy indicator
     * and popup a dialog with the message (if the showMessage flag is true).  Will call the completionHandler
     * if supplied as well
     * @param view
     * @param {boolean} showMessage
     * @param completionHandler optional
     * @returns {failureFunction}
     */
    createFailureHandler: function(view, showMessage, completionHandler) {
        return function(jqxhr, textStatus, error) {
            if (showMessage) {
               sas.analytics.riskhp.utils.HprUtil.displayDetailedErrorDialog(jqxhr);
            }
            // Clear the parent view busy indicator after a slight delay; This prevents a
            // flicker for the parent view when busy indicator disappears before new view renders
            setTimeout(function() {
                if (completionHandler) {
                    completionHandler(jqxhr, textStatus, error);
                }
                view.setBusy(false);
            }, 1000);
        };
    },

    /**
     * Creates a common success handler function that when called will turn off the view's busy indicator.
     * Will call the completionHandler if supplied as well
     * @param view
     * @param completionHandler optional
     * @returns {successFunction}
     */
    createSuccessHandler: function(view, completionHandler) {
        return function() {
            // Clear the parent view busy indicator after a slight delay; This prevents a
            // flicker for the parent view when busy indicator disappears before new view renders
            setTimeout(function() {
                if (completionHandler) {
                    completionHandler();
                }
                view.setBusy(false);
            }, 1000);
        };
    },

    //*****************************************************************************************************
    //*****************************************************************************************************
    // Selected Panels View helper functions
    //*****************************************************************************************************
    //*****************************************************************************************************
    /**
     * Returns the title of the selected Page in the Exploration
     * @returns {String} The title of the selected Page
     */
    getSelectedPageTitle: function() {
        var explorationView = sas.analytics.riskhp.utils.ViewManager.getExplorationView();
        var controller = explorationView.getController();
        var page = controller.getSelectedPage();
        return page.getText();
    },

    /**
     * Returns the selected PanelsView. May return null.
     * @returns {*} The selected PanelsView or null.
     */
    getSelectedRiskPanelsView: function() {
        var explorationView = sas.analytics.riskhp.utils.ViewManager.getExplorationView();
        var controller = explorationView.getController();
        var page = controller.getSelectedPage();
        return page.getContent()[0];
    },

    /**
     * Returns the id of the selected container in the PanelsView for the currently selected page in the exploration
     * view.  May return null
     * @returns {*} Id of the selected container or null if none is selected.
     */
    getSelectedPanelId: function() {
        // Get the delegate object
        var selectedRiskPanelsView = sas.analytics.riskhp.utils.HprUtil.getSelectedRiskPanelsView();
        if (selectedRiskPanelsView) {
            var delegate = selectedRiskPanelsView.getDelegate();
            return delegate.getSelectedContainerId();
        }
        return null;
    },

    /**
     * Given the id of a component contained in a RiskPanelsView this method returns the owning RiskPanelsView object.
     * First, we try to find the component using jQuery selectors.  If that returns 'null', which usually means the
     * tab/page the component is on is not active or selected, we then use OpenUI5 code.  The 'delegateComponentPropertyString'
     * parameter is used in the 2nd case.
     *
     * In the end, the method may return null.
     *
     * @param id id of a contained component
     * @param delegateComponentPropertyString String representation of the list of id's to look for
     * @returns {Element|Object|sap.ui.core.Element|null}
     */
    getPanelsViewForComponent: function(id, delegateComponentPropertyString) {
        // First we try to use the method in RiskPanelsView which uses JQuery selectors; That method will
        // return null if the PanelsView is not active or currently displayed because it is not in the DOM
        var panelsView = sas.risk.common.panelsview.RiskPanelsView.prototype.findPanelsView(id);
        if (panelsView === null && (delegateComponentPropertyString && delegateComponentPropertyString.length > 0)) {
            // If it is null, try to use OpenUI5 code to locate the correct RiskPanelsView
            var explorationView = sas.analytics.riskhp.utils.ViewManager.getExplorationView();
            var mainTabbar = explorationView.getMainTabbar();
            // Loop through the tabs/pages and inspect each RiskPanelsView and it's Delegate for the component
            $.each(mainTabbar.getItems(), function(index, page) {
                var potentialRiskPanelsView = page.getContent()[0]; // This is a RiskPanelsView control
                var delegate = potentialRiskPanelsView.getDelegate();
                // The Delegate holds the ids of the controls it contains, search through them for a match
                if (_.contains(delegate[delegateComponentPropertyString], id)) {
                    // If we find a match, set the appropriate RiskPanelsView and stop iterating
                    panelsView = potentialRiskPanelsView;
                    return false;
                }
            });
        }

        return panelsView;
    },

    // TODO: Add a getSelectedPanel or getSelectedPanelContent method so callers don't have to get the panel id, get the
    // panel and then get the content/control off of that.

    //*****************************************************************************************************
    //*****************************************************************************************************
    // Event helper functions
    //*****************************************************************************************************
    //*****************************************************************************************************
    /**
     * Update all the right-side panels (i.e. Options, Filters, Hierarchies) for a selected Page.  Currently, this
     * displays Options for the selected Page and displays an Zero State for the Filters and Hierarchies
     * Panes since they are not valid for Page objects.
     * @param selectedPage {sas.hc.m.IconTabFilter} The Page which is selected in the Exploration View
     */
    updateRightPanelsForSelectedPage: function(selectedPage) {
        // Publish an event that will update the Options pane based on a selected Page
        sas.eventBus.publish(HprConstants.CHANNEL_OPTIONS_PANE, HprConstants.EVENT_OPTIONS_PANE_CHANGED, {
            optionsType: HprConstants.HPR_OPTIONS_PANE_PAGE_TYPE,
            pageId: selectedPage.getKey() // Key of selected IconTabFilter is the id we want
        });

        // Publish an event that will update the Filters pane based on a Zero State
        sas.eventBus.publish(HprConstants.CHANNEL_FILTERS_PANE, HprConstants.EVENT_FILTERS_PANE_CHANGED, {
            filtersType: HprConstants.HPR_FILTERS_PANE_ZERO_STATE_TYPE
        });

        // Publish an event that will update the Hierarchies pane based on a Zero State
        sas.eventBus.publish(HprConstants.CHANNEL_HIERARCHIES_PANE, HprConstants.EVENT_HIERARCHIES_PANE_CHANGED, {
            hierarchiesType: HprConstants.HPR_HIERARCHIES_PANE_ZERO_STATE_TYPE
        });

        // Publish an event that will update the Display Rules pane based on a Zero State
        sas.eventBus.publish(HprConstants.CHANNEL_DISPLAY_RULES_PANE, HprConstants.EVENT_DISPLAY_RULES_PANE_CHANGED, {
            displayRulesType: HprConstants.HPR_DISPLAY_RULES_PANE_ZERO_STATE_TYPE
        });
    },

    /**
     * Update all the right-side panels (i.e. Options, Filters, Hierarchies) for a selected Crosstab.
     * @param selectedCrosstabId {String} The id of the selected Crosstab
     */
    updateRightPanelsForSelectedCrosstab: function(selectedCrosstabId) {
        // Publish an event that will update the Options pane based on the selected Crosstab
        sas.eventBus.publish(HprConstants.CHANNEL_OPTIONS_PANE, HprConstants.EVENT_OPTIONS_PANE_CHANGED, {
            optionsType: HprConstants.HPR_OPTIONS_PANE_CROSSTAB_TYPE,
            crosstabId: selectedCrosstabId
        });

        // Publish an event that will update the Filters pane based on the selected Crosstab
        sas.eventBus.publish(HprConstants.CHANNEL_FILTERS_PANE, HprConstants.EVENT_FILTERS_PANE_CHANGED, {
            filtersType: HprConstants.HPR_FILTERS_PANE_CROSSTAB_TYPE,
            crosstabId: selectedCrosstabId
        });

        // Publish an event that will update the Hierarchies pane based on the selected Crosstab
        sas.eventBus.publish(HprConstants.CHANNEL_HIERARCHIES_PANE, HprConstants.EVENT_HIERARCHIES_PANE_CHANGED, {
            hierarchiesType: HprConstants.HPR_HIERARCHIES_PANE_CROSSTAB_TYPE,
            crosstabId: selectedCrosstabId
        });

        // Publish an event that will update the Display Rules pane based on the selected Crosstab
        sas.eventBus.publish(HprConstants.CHANNEL_DISPLAY_RULES_PANE, HprConstants.EVENT_DISPLAY_RULES_PANE_CHANGED, {
            displayRulesType: HprConstants.HPR_DISPLAY_RULES_PANE_CROSSTAB_TYPE,
            crosstabId: selectedCrosstabId
        });
    },

    /**
     * Update all the right-side panels (i.e. Options, Filters, Hierarchies) for a selected generic
     * TableViewer (i.e. Simulated Values Table).
     * @param selectedTableViewerId
     */
    updateRightPanelsForSelectedTableViewer: function(selectedTableViewerId) {
        // Publish an event that will update the Options pane based on the selected TableViewer
        sas.eventBus.publish(HprConstants.CHANNEL_OPTIONS_PANE, HprConstants.EVENT_OPTIONS_PANE_CHANGED, {
            optionsType: HprConstants.HPR_OPTIONS_PANE_TABLE_VIEWER_TYPE,
            tableViewerId: selectedTableViewerId
        });

        // Publish an event that will update the Filters pane to a Zero State (always Zero State)
        sas.eventBus.publish(HprConstants.CHANNEL_FILTERS_PANE, HprConstants.EVENT_FILTERS_PANE_CHANGED, {
            filtersType: HprConstants.HPR_FILTERS_PANE_ZERO_STATE_TYPE
        });

        // Publish an event that will update the Hierarchies pane based on a TableViewer (always Zero State)
        sas.eventBus.publish(HprConstants.CHANNEL_HIERARCHIES_PANE, HprConstants.EVENT_HIERARCHIES_PANE_CHANGED, {
            hierarchiesType: HprConstants.HPR_HIERARCHIES_PANE_ZERO_STATE_TYPE
        });

        // Publish an event that will update the Display Rules pane based on a TableViewer (always Zero State)
        sas.eventBus.publish(HprConstants.CHANNEL_DISPLAY_RULES_PANE, HprConstants.EVENT_DISPLAY_RULES_PANE_CHANGED, {
            displayRulesType: HprConstants.HPR_DISPLAY_RULES_PANE_ZERO_STATE_TYPE
        });
    },

    /**
     * Update all the right-side panels (i.e. Options, Filters, Hierarchies) for a selected Crosstab which is in ZeroState.
     * @param selectedCrosstabId {String} The id of the selected Crosstab
     */
    updateRightPanelsForZeroStateCrosstab: function(selectedCrosstabId) {
        // Publish an event that will update the Options pane based on the selected Crosstab
        sas.eventBus.publish(HprConstants.CHANNEL_OPTIONS_PANE, HprConstants.EVENT_OPTIONS_PANE_CHANGED, {
            optionsType: HprConstants.HPR_OPTIONS_PANE_ZERO_STATE_TYPE,
            crosstabId: selectedCrosstabId
        });

        // Quick check to see if the selected Crosstab has Filters; If so, we need to keep Filters pane active (see S1374991)
        // so user has a chance to clear the offending filter and get Crosstab back to a good state
        var selectedCrosstab = sap.ui.getCore().byId(selectedCrosstabId);
        var filtersType = HprConstants.HPR_FILTERS_PANE_ZERO_STATE_TYPE; // Default for this case is ZeroState Filters pane
        if (selectedCrosstab) {
            var filters = selectedCrosstab.context.getFilters();
            if (filters && filters.length > 0) {
                // If the Crosstab had filters, make sure the Filters pane reflects the filters for the Crosstab
                filtersType = HprConstants.HPR_FILTERS_PANE_CROSSTAB_TYPE;
            }
        }

        // Publish an event that will update the Filters pane based on the selected Crosstab
        sas.eventBus.publish(HprConstants.CHANNEL_FILTERS_PANE, HprConstants.EVENT_FILTERS_PANE_CHANGED, {
            filtersType: filtersType,
            crosstabId: selectedCrosstabId
        });

        // Publish an event that will update the Hierarchies pane based on the selected Crosstab
        sas.eventBus.publish(HprConstants.CHANNEL_HIERARCHIES_PANE, HprConstants.EVENT_HIERARCHIES_PANE_CHANGED, {
            hierarchiesType: HprConstants.HPR_HIERARCHIES_PANE_ZERO_STATE_TYPE,
            crosstabId: selectedCrosstabId
        });

        // Publish an event that will update the Display Rules pane based on the selected Crosstab
        sas.eventBus.publish(HprConstants.CHANNEL_DISPLAY_RULES_PANE, HprConstants.EVENT_DISPLAY_RULES_PANE_CHANGED, {
            displayRulesType: HprConstants.HPR_DISPLAY_RULES_PANE_ZERO_STATE_TYPE,
            crosstabId: selectedCrosstabId
        });
    },

    /**
     * Update all the right-side panels (i.e. Options, Filters, Hierarchies) for a selected Plot.  The Hierarchies Pane
     * is set to zero state because it is not valid for a Plot.
     * @param selectedPlotId {String} The id of the selected Plot
     * @param optionsType {String} The string that identifies which type of Options pane to display
     * @param filtersType {String} The string that identifies which type of Filters pane to display
     */
    updateRightPanelsForSelectedPlot: function(selectedPlotId, optionsType, filtersType) {
        // Get the selected Plot
        var selectedPlot = sap.ui.getCore().byId(selectedPlotId);
        var forceUpdate = false;
        if (selectedPlot) {
            forceUpdate = selectedPlot.getChartModel().shouldRefreshRightPanel;
            selectedPlot.getChartModel().shouldRefreshRightPanel = false;
        }

        // Publish an event that will update the Options pane based on a Plot
        sas.eventBus.publish(HprConstants.CHANNEL_OPTIONS_PANE, HprConstants.EVENT_OPTIONS_PANE_CHANGED, {
            optionsType: optionsType,
            plotId: selectedPlotId,
            forceUpdate: forceUpdate
        });

        // Publish an event that will update the Filters pane based on a Plot
        sas.eventBus.publish(HprConstants.CHANNEL_FILTERS_PANE, HprConstants.EVENT_FILTERS_PANE_CHANGED, {
            filtersType: filtersType,
            plotId: selectedPlotId,
            forceUpdate: forceUpdate
        });

        // Publish an event that will update the Hierarchies pane based on a Plot (always Zero State for Plots)
        sas.eventBus.publish(HprConstants.CHANNEL_HIERARCHIES_PANE, HprConstants.EVENT_HIERARCHIES_PANE_CHANGED, {
            hierarchiesType: HprConstants.HPR_HIERARCHIES_PANE_ZERO_STATE_TYPE
        });

        // Publish an event that will update the Display Rules pane based on the selected Plot
        sas.eventBus.publish(HprConstants.CHANNEL_DISPLAY_RULES_PANE, HprConstants.EVENT_DISPLAY_RULES_PANE_CHANGED, {
            displayRulesType: HprConstants.HPR_DISPLAY_RULES_PANE_ZERO_STATE_TYPE
        });
    },

    /**
     * Update all the right-side panels (i.e. Options, Filters, Hierarchies) for a selected Portfolio Details.  The Hierarchies Pane
     * is set to zero state because it is not valid for a Portfolio Details.
     * @param selectedPortfolioDetailsId {String} The id of the selected Portfolio Details
     * @param optionsType {String} The string that identifies which type of Options pane to display
     * @param filtersType {String} The string that identifies which type of Filters pane to display
     */
    updateRightPanelsForSelectedPortfolioDetails: function(selectedPortfolioDetailsId, optionsType, filtersType) {
        // Publish an event that will update the Options pane based on a Portfolio Details
        sas.eventBus.publish(HprConstants.CHANNEL_OPTIONS_PANE, HprConstants.EVENT_OPTIONS_PANE_CHANGED, {
            optionsType: optionsType,
            portfolioDetailsId: selectedPortfolioDetailsId
        });

        // Publish an event that will update the Filters pane based on a Portfolio Details
        sas.eventBus.publish(HprConstants.CHANNEL_FILTERS_PANE, HprConstants.EVENT_FILTERS_PANE_CHANGED, {
            filtersType: filtersType,
            portfolioDetailsId: selectedPortfolioDetailsId
        });

        // Publish an event that will update the Hierarchies pane based on a Portfolio Details (always Zero State for Portfolio Details)
        sas.eventBus.publish(HprConstants.CHANNEL_HIERARCHIES_PANE, HprConstants.EVENT_HIERARCHIES_PANE_CHANGED, {
            hierarchiesType: HprConstants.HPR_HIERARCHIES_PANE_ZERO_STATE_TYPE
        });

        // Publish an event that will update the Display Rules pane based on the selected Portfolio Details
        sas.eventBus.publish(HprConstants.CHANNEL_DISPLAY_RULES_PANE, HprConstants.EVENT_DISPLAY_RULES_PANE_CHANGED, {
            displayRulesType: HprConstants.HPR_DISPLAY_RULES_PANE_ZERO_STATE_TYPE
        });
    },

    /**
     * Update all the right-side panels (i.e. Options, Filters, Hierarchies) for a selected Risk Factor Sensitivities
     * table.  The Hierarchies Pane is set to zero state because it is not valid for a Risk Factor Sensitivities table.
     * @param selectedRfsId {String} The id of the selected control
     * @param optionsType {String} The string that identifies which type of Options pane to display
     * @param filtersType {String} The string that identifies which type of Filters pane to display
     */
    updateRightPanelsForSelectedRiskFactorSensitivities: function(selectedRfsId, optionsType, filtersType) {
        // Publish an event that will update the Options pane based on a RFS
        sas.eventBus.publish(HprConstants.CHANNEL_OPTIONS_PANE, HprConstants.EVENT_OPTIONS_PANE_CHANGED, {
            optionsType: optionsType,
            rfsId: selectedRfsId
        });

        // Publish an event that will update the Filters pane based on a RFS
        sas.eventBus.publish(HprConstants.CHANNEL_FILTERS_PANE, HprConstants.EVENT_FILTERS_PANE_CHANGED, {
            filtersType: filtersType,
            rfsId: selectedRfsId
        });

        // Publish an event that will update the Hierarchies pane based on a RFS (always Zero State for RFS)
        sas.eventBus.publish(HprConstants.CHANNEL_HIERARCHIES_PANE, HprConstants.EVENT_HIERARCHIES_PANE_CHANGED, {
            hierarchiesType: HprConstants.HPR_HIERARCHIES_PANE_ZERO_STATE_TYPE
        });

        // Publish an event that will update the Display Rules pane based on the selected RFS (always Zero State)
        sas.eventBus.publish(HprConstants.CHANNEL_DISPLAY_RULES_PANE, HprConstants.EVENT_DISPLAY_RULES_PANE_CHANGED, {
            displayRulesType: HprConstants.HPR_DISPLAY_RULES_PANE_ZERO_STATE_TYPE
        });
    },

    /**
     * Update all the right-side panels (i.e. Options, Filters, Hierarchies) to a Zero State.  Based on the
     * "bForceOptionsToPageType" parameter, determine if we need to display Page options instead of the Zero State.
     * @param bForceOptionsToPageType Boolean that causes the code to run through some logic to determine
     * if we need to display Page options instead of Zero State
     */
    updateRightPanelsForZeroState: function(bForceOptionsToPageType) {
        // This method is called by numerous parts of the UI to set the right-side panels to Zero State;
        // In some cases we need to check to make sure that displaying Page Options is a valid choice;
        // Make this choice based on the "bForceOptionsToPageType" parameter
        if (!bForceOptionsToPageType) {
            bForceOptionsToPageType = undefined;
        }

        // Default to Zero State view for Options pane unless we need to update for Page options
        var optionsType = HprConstants.HPR_OPTIONS_PANE_ZERO_STATE_TYPE;
        var pageId = null;
        if (bForceOptionsToPageType) {
            // Going to use Page options now, need to find a valid Page options to display
            optionsType = HprConstants.HPR_OPTIONS_PANE_PAGE_TYPE;
            var explorationView = sas.analytics.riskhp.utils.ViewManager.getExplorationView();
            var currentPageOptionsModel = explorationView.getOptionsPane().getOptionsPaneView().getOptionsPanelModel(HprConstants.HPR_OPTIONS_PANE_PAGE_TYPE);
            var currentSelectedTabId = explorationView.getMainTabbar().getSelectedItem().getKey();
            if (currentPageOptionsModel && currentPageOptionsModel.getData() && currentPageOptionsModel.getData().options) {
                var currentPageOptionsId = currentPageOptionsModel.getData().options.id;

                // Since this method is called by numerous parts of the UI, try to make a best guess as
                // to which Page options to display based on the current Page options model and which tab
                // is currently selected; This really comes into play when adding a new Tab/Page and when
                // trying to edit Page options for another tab that is currently not selected
                if ((currentPageOptionsId !== currentSelectedTabId)) {
                    pageId = currentSelectedTabId;
                } else {
                    pageId = currentPageOptionsId;
                }
            } else {
                // Fall back to currently selected tab/page
                pageId = currentSelectedTabId;
            }
        }

        // Publish an event that will update the Options pane to a Zero State
        sas.eventBus.publish(HprConstants.CHANNEL_OPTIONS_PANE, HprConstants.EVENT_OPTIONS_PANE_CHANGED, {
            optionsType: optionsType,
            pageId: pageId ? pageId : null
        });

        // Publish an event that will update the Filters pane to a Zero State
        sas.eventBus.publish(HprConstants.CHANNEL_FILTERS_PANE, HprConstants.EVENT_FILTERS_PANE_CHANGED, {
            filtersType: HprConstants.HPR_FILTERS_PANE_ZERO_STATE_TYPE
        });

        // Publish an event that will update the Hierarchies pane to a Zero State
        sas.eventBus.publish(HprConstants.CHANNEL_HIERARCHIES_PANE, HprConstants.EVENT_HIERARCHIES_PANE_CHANGED, {
            hierarchiesType: HprConstants.HPR_HIERARCHIES_PANE_ZERO_STATE_TYPE
        });

        // Publish an event that will update the Display Rules pane based on the selected Crosstab
        sas.eventBus.publish(HprConstants.CHANNEL_DISPLAY_RULES_PANE, HprConstants.EVENT_DISPLAY_RULES_PANE_CHANGED, {
            displayRulesType: HprConstants.HPR_DISPLAY_RULES_PANE_ZERO_STATE_TYPE
        });
    },

    //*****************************************************************************************************
    //*****************************************************************************************************
    // Data Source/Cube functions
    //*****************************************************************************************************
    //*****************************************************************************************************
    /**
     * Get the displayable label of the DataSource.  This will usually be the 'label' property from the mid-tier (if it exists) concatenated
     * with the 'type' property.  If the 'label' property does not exist, then the 'name' property is used.  The resulting displayable label
     * will be something like "AutoPrice (price)".
     * @returns String The displayable label of this DataSource
     */
    getCubeDisplayLabel: function(dataSourceLabel, dataSourceName, dataSourceType) {
        // Default these variables to the ones passed in
        var displayLabel = dataSourceLabel;
        var originalLabel = dataSourceLabel;
        var originalName = dataSourceName;
        var originalType = dataSourceType;

        var formatStringsToUse = []; // Array to hold strings to pass to formatter

        // Method which determines which resource string to use based on length of 'aFormatStringsToUse'
        var _getCubeDisplayLabelHelper = function(aFormatStringsToUse) {
            if (aFormatStringsToUse.length === 1) {
                // No 'type' present, so don't use format string for it
                return sas.analytics.riskhp.utils.HprUtil.getBundle().getText("HPRisk.data.source.label.without.type.fmt", aFormatStringsToUse);
            } else {
                // The 'type' of DataSource is present, so use format string for it
                return sas.analytics.riskhp.utils.HprUtil.getBundle().getText("HPRisk.data.source.label.with.type.fmt", aFormatStringsToUse);
            }
        };

        // Retrieve the resource string for 'type' if it exists
        if (originalType && originalType.length > 0) {
            var desiredTypeKey = "HPRisk.data.source." + originalType + ".type.lc.txt";
            var hasText = sas.analytics.riskhp.utils.HprUtil.getBundle().hasText(desiredTypeKey);
            if (hasText) {
                // Pull the string from the Resource Bundle representing the DataSource type
                var displayType = sas.analytics.riskhp.utils.HprUtil.getBundle().getText(desiredTypeKey);
                formatStringsToUse.push(displayType);
            }
        }

        // See if we have to use the 'label' or 'name' of the DataSource to build a correct displayable label
        if (originalLabel && originalLabel.length > 0) {
            var parenIndex = originalLabel.indexOf(" (");
            // If the 'label' does not contain " (", see if we can concatenate the 'type' to it to create the displayable label;
            // If it does contain " (", just go ahead and use it as-is
            if (parenIndex === -1) {
                formatStringsToUse.unshift(originalLabel); // Add the label to the front of the array
                displayLabel = _getCubeDisplayLabelHelper(formatStringsToUse);
            }
        } else {
            // If we have to use the 'name', first determine if we have to strip off the ".rskcdesc" string.
            var dotIndex = originalName.indexOf(".rskcdesc");
            if (dotIndex !== -1) {
                formatStringsToUse.unshift(originalName.substring(0, dotIndex)); // Add the name without '.rskcdesc' to the front of the array
                displayLabel = _getCubeDisplayLabelHelper(formatStringsToUse);
            } else {
                formatStringsToUse.unshift(originalName); // Add the name to the front of the array
                displayLabel = _getCubeDisplayLabelHelper(formatStringsToUse);
            }

        }

        return displayLabel;
    },

    /**
     * Get the displayable path of the DataSource.  This will usually be the DataSource 'workGroup' folder path appended with the 'relPath' folder path.  If
     * the 'relPath' does not exist it is just the 'workGroup' folder path.  NOTE: The application folder is not contained in this cube path as it was deemed
     * not useful.
     * @returns {String}
     */
    getCubePath: function(workGroup, relPath) {
        var path = workGroup + '/';
        if (relPath) {
            path += relPath + '/';
        }
        return path;
    },

    /**
     * Given a data source id and a list of candidate substitute data source ids, return the first data source id in the
     * list that is of the same type as the passed in data source id.  If the list has only one data source id in it then
     * that id is returned.
     * @param {string} currentDataSourceId
     * @param {[string]} candidateDataSourceIds
     * @returns The found id or undefined
     */
    getCompatibleDataSourceId: function(currentDataSourceId, candidateDataSourceIds) {
        if (candidateDataSourceIds.length === 1) {
            return candidateDataSourceIds[0];
        }
        var currentDataSourceType = currentDataSourceId.split('+')[0];
        return _.find(candidateDataSourceIds, function(newDSId) {
            return newDSId.split('+')[0] === currentDataSourceType;
        });
    },

    /**
     * Parse value items to group the values into a single object instead of
     * multiple objects containing one value each.
     * @param filterValuesItems
     * @returns {*}
     */
    parseFilterValuesData: function(filterValuesItems) {
        var modelData = filterValuesItems[0];

        // When applying a filter for selected items, the number of items returned by the
        // midtier matches the number of items selected in the filter list. This is because
        // the filter is constructed as a filter for each selected value. The midtier returns
        // an item per filter (each containing one value), so  we need to consolidate the
        // values for each item into a single array and update the count on the returned
        // modelData.
        if (filterValuesItems.length > 1) {
            for (var i = 1; i < filterValuesItems.length; i++) {
                if (!modelData.values) {
                    modelData.values = [];
                }
                _.each(filterValuesItems[i].values, function(value) {
                    modelData.values.push(value);
                });
            }
            modelData.count = modelData.values.length;
        }

        // translate the data for model consumption
        return modelData;
    },

    /**
     * Validate the filters passed in against the dataSource and return the valid filters.
     * @param filters Filters we are trying to see if we can use in 'dataSource'
     * @param dataSource The new dataSource
     * @param columnDescriptorMap Portfolio column ColumnDescriptor objects
     * @returns {*} Valid filters for the dataSource
     */
    validateFilters: function(filters, dataSource, columnDescriptorMap) {
        var myDeferred = $.Deferred();
        var HprConstants = sas.analytics.riskhp.utils.HprConstants;

        // compare filters (do this last because we have to make a midtier call to verify values)
        var remainingFilters = _.filter(filters, function(filter) {
            if (filter.type === HprConstants.CLASS_VARIABLE_DATA_ITEM_TYPE) {
                return !!dataSource.getDataItemById(filter.id);
            }
            else {  // portfolioColumn
                return !!columnDescriptorMap[filter.id];
            }
        });

        if (remainingFilters && remainingFilters.length > 0) {
            var filterItems = {};
            var filterValuesQueries = [];

            // build a values query for each remaining filter
            _.each(remainingFilters, function(remainingFilter) {
                var filterValuesQueryObjectTemplate = {
                    version: 1,
                    type: remainingFilter.type,
                    id: remainingFilter.id,
                    countOnly: false,
                    start: 0,
                    limit: 2500
                };

                // capture the data item represented by the filter
                filterItems[remainingFilter.id] = remainingFilter.type === HprConstants.CLASS_VARIABLE_DATA_ITEM_TYPE ? dataSource.getDataItemById(remainingFilter.id) : columnDescriptorMap[remainingFilter.id];

                // if this is a class variable or a string based instrument variable (portfolio column), build the search array to return only the selected values
                if (remainingFilter.type === HprConstants.CLASS_VARIABLE_DATA_ITEM_TYPE || filterItems[remainingFilter.id].getDataType() === HprConstants.HPR_DATA_TYPE_STRING) {
                    _.each(remainingFilter.values, function (value) {
                        filterValuesQueries.push($.extend(true, {search: [{
                            value: filterItems[remainingFilter.id].getDataType() === HprConstants.HPR_DATA_TYPE_STRING ? value : Number(value),
                            type: filterItems[remainingFilter.id].getDataType() === HprConstants.HPR_DATA_TYPE_STRING ? "exact" : "equals"
                        }]}, filterValuesQueryObjectTemplate));
                    });
                }
                else {
                    // for all other instrument variable (portfolio column) filters, return the template
                    filterValuesQueries.push(filterValuesQueryObjectTemplate);
                }
            });

            // process each query and update the filter values (if the filter resolves to no values, remove it)
            sas.analytics.riskhp.utils.HprAPI.updateCubeValues(dataSource.getId(), filterValuesQueries).done(
                function(data) {
                    // group the values data by ID
                    var valuesData = _.groupBy(data.items, function(item) {
                        return item.id;
                    });

                    // map the filters by ID
                    var remainingFiltersMap = _.indexBy(remainingFilters, "id");

                    // update each filter with the values returned
                    _.each(valuesData, function(valuesDataItems) {
                        var parsedValuesData = sas.analytics.riskhp.utils.HprUtil.parseFilterValuesData(valuesDataItems);
                        var filter = remainingFiltersMap[parsedValuesData.id];

                        if (parsedValuesData.type === HprConstants.CLASS_VARIABLE_DATA_ITEM_TYPE || filterItems[parsedValuesData.id].getDataType() === HprConstants.HPR_DATA_TYPE_STRING) {
                            filter.values = _.filter(filter.values, function(filterValue) {
                                if (filterItems[parsedValuesData.id].getDataType() !== HprConstants.HPR_DATA_TYPE_STRING) {
                                    filterValue = Number(filterValue);
                                }
                                return _.findIndex(parsedValuesData.values, function(valueObject) {return valueObject.value === filterValue;}) !== -1;
                            });
                        }
                        else {
                            if (filterItems[parsedValuesData.id].minValue < parsedValuesData.min) {
                                filterItems[parsedValuesData.id].minValue = parsedValuesData.min;
                            }
                            if (filterItems[parsedValuesData.id].maxValue > parsedValuesData.max) {
                                filterItems[parsedValuesData.id].maxValue = parsedValuesData.max;
                            }
                        }
                    });

                    // remove filters with no remaining values or for which values were not returned
                    remainingFilters = _.filter(remainingFilters, function (remainingFilter) {
                        var valuesForFilter = !!valuesData[remainingFilter.id];
                        return !!valuesData[remainingFilter.id] &&
                            (remainingFilter.values && remainingFilter.values.length > 0 ||
                            remainingFilter.hasOwnProperty("minValue") && remainingFilter.hasOwnProperty("maxValue"));
                    });

                    myDeferred.resolve(remainingFilters);
                }
            ).fail(
                // if the value query failed, remove all filters
                function(jqxhr, textStatus, error) {
                    myDeferred.resolve([]);
                }
            );
        }
        else {
            myDeferred.resolve([]);
        }

        return myDeferred;
    },

    /**
     * Given a list of dataitems, retrieves the last one in the array with the given type.  May
     * return null if the type is not found.
     * @param dataItemsArray
     * @param type
     * @returns {null}
     */
    findLastDataItemByType: function(dataItemsArray, type) {
        var items = _.filter(dataItemsArray, function(dataItem) {
            return dataItem.type === type;
        });
        if (items.length > 0) {
            return _.last(items);
        }
        return null;
    },

    /**
     * Based on the state of the current DataSource and the application settings return the horizon type that should be
     * shown
     * @returns {string} One of the HprConstants.DISPLAY_HORIZONS_ constants (there are only two)
     * @private
     */
    getDesiredHorizonDisplayType: function(dataSource) {
        // If data source is side by side need to show relative
        var desiredHorizonDisplayType = sas.analytics.riskhp.utils.HprConstants.DISPLAY_HORIZONS_RELATIVE_VALUE;

        // If data source is not a side by side then get the display type from the app settings
        if (!dataSource.isSideBySide()) {
            desiredHorizonDisplayType = sas.analytics.riskhp.settings.ApplicationSettings.getDisplayHorizons();
        }
        return desiredHorizonDisplayType;
    },

    /**
     * Searches the url parameters for the given parameter name and returns it if found otherwize returns undefined.
     * @param paramName
     * @returns {*}
     */
    getURLParameter: function(paramName) {
        var query = window.location.search.substring(1);  // returns everything after the ?
        var vars = query.split("&");
        for (var i = 0; i < vars.length; i++) {
            var pair = vars[i].split("=");
            if (pair[0] === paramName) {
                return pair[1];
            }
        }
        return undefined;
    },

    copyTextToClipboard: function (text) {
        var textArea = document.createElement("textarea");

        //
        // *** This styling is an extra step which is likely not required. ***
        //
        // Why is it here? To ensure:
        // 1. the element is able to have focus and selection.
        // 2. if element was to flash render it has minimal visual impact.
        // 3. less flakyness with selection and copying which **might** occur if
        //    the textarea element is not visible.
        //
        // The likelihood is the element won't even render, not even a flash,
        // so some of these are just precautions. However in IE the element
        // is visible whilst the popup box asking the user for permission for
        // the web page to copy to the clipboard.
        //

        // Place in top-left corner of screen regardless of scroll position.
        textArea.style.position = 'fixed';
        textArea.style.top = 0;
        textArea.style.left = 0;

        // Ensure it has a small width and height. Setting to 1px / 1em
        // doesn't work as this gives a negative w/h on some browsers.
        textArea.style.width = '2em';
        textArea.style.height = '2em';

        // We don't need padding, reducing the size if it does flash render.
        textArea.style.padding = 0;

        // Clean up any borders.
        textArea.style.border = 'none';
        textArea.style.outline = 'none';
        textArea.style.boxShadow = 'none';

        // Avoid flash of white box if rendered for any reason.
        textArea.style.background = 'transparent';


        textArea.value = text;

        document.body.appendChild(textArea);
        textArea.focus();
        textArea.select();

        try {
            var successful = document.execCommand('copy');
            var msg = successful ? 'successful' : 'unsuccessful';
            // console.log('Copying text command was ' + msg);
        } catch (err) {
            // console.log('Oops, unable to copy');
        }

        document.body.removeChild(textArea);
    }
};