// (c) 2015, SAS Institute Inc.

jQuery.sap.declare("sas.analytics.riskhp.utils.ViewManager");
jQuery.sap.require("sas.analytics.riskhp.utils.HprUtil");
jQuery.sap.require("sas.analytics.riskhp.utils.HprConstants");
jQuery.sap.require("sas.analytics.riskhp.models.HPRiskModelManager");
jQuery.sap.require("sas.analytics.riskhp.models.ComputedVariablesManager");
jQuery.sap.require("sas.analytics.riskhp.models.DisplayRulesManager");
jQuery.sap.require("sap.ui.core.routing.Router");

sas.analytics.riskhp.utils.ViewManager = (function() {
    return {
        getRouter: function(router) {
            return sap.ui.core.routing.Router.getRouter(sas.analytics.riskhp.utils.HprConstants.HPR_ROUTER_ID);
        },

        /**
         * Returns the active (currently rendered) view.
         * @returns {sap.ui.core.Element|Element}
         */
        getActiveView: function () {
            //var wv = this.getWelcomeView();
            //var ev = this.getExplorationView();
            //return this.getWelcomeView() ? this.getWelcomeView() : this.getExplorationView();

            var router = this.getRouter();
            if (router) {
                return sap.ui.getCore().byId(router.getActiveViewName());
            }
            return null;
        },

        /**
         * Isolates the code to get the one and only Welcome view. Expected to return
         * null if there is no Welcome view yet.
         * @returns {sap.ui.core.Element|Element}
         */
        getWelcomeView: function () {
            return sap.ui.getCore().byId(sas.analytics.riskhp.utils.HprConstants.HPR_WELCOME_VIEW_ID);
        },

        /**
         * Isolates the code to get the one and only Exploration view. Expected to return
         * null if there is no Exploration View yet.
         * @returns {sap.ui.core.Element|Element}
         */
        getExplorationView: function () {
            return sap.ui.getCore().byId(sas.analytics.riskhp.utils.HprConstants.HPR_EXPLORATION_VIEW_ID);
        },

        /**
         * Load the Welcome view.
         */
        loadHprWelcomeView: function () {
            this.getRouter().navTo(sas.analytics.riskhp.utils.HprConstants.HPR_WELCOME_VIEW_ID, null, true);
        },

        /**
         * Load a new exploration.  The specified name of the new exploration and the data from the data source will be used to populate
         * the core model so the UI can access it.  Once the exploration is loaded, navigate to the Exploration view.
         *
         * @param newExplorationName The name of the new exploration
         * @param dataSourceDetails The dataSourceDetails object (which consists of a DataSourceParams object + multiple types of supported dataSources) needed to load the correct DataSource
         * @param selectedDataSourceType The selected type of DataSource to open
         * @returns {*} - a restricted promise
         */
        loadNewExploration: function (newExplorationName, dataSourceDetails, selectedDataSourceType) {
            var self = this;

            DataSourceCacheManager.reset(); // Reset the Cache Manager
            ComputedVariablesManager.reset();
            DisplayRulesManager.reset();

            var promise = sas.analytics.riskhp.utils.HprAPI.loadNewExploration(dataSourceDetails, selectedDataSourceType);
            promise.done(function () {
                // Create an object to simulate an Exploration
                var exploration = sas.analytics.riskhp.models.HPRiskModelManager.createNewExplorationObject(newExplorationName);

                // Reset the ModelManager for the new content
                sas.analytics.riskhp.models.HPRiskModelManager.reset();

                // Populate the model
                sas.analytics.riskhp.models.HPRiskModelManager.setExplorationModel(exploration);

                // If there is already an Exploration View then we just need to reset it's
                // controller otherwise we need to navigate to it.
                var explorationView = self.getExplorationView();
                if (explorationView) {
                    explorationView.getController().reset(true);
                }
                else {
                    self.getRouter().navTo(sas.analytics.riskhp.utils.HprConstants.HPR_EXPLORATION_VIEW_ID, null, true);
                }
            });
            //TODO handle error?

            return promise;
        },

        /**
         * Load the saved exploration.  Appropriate data about the exploration and cube will be loaded into the core model so the UI can access it.
         * Once the exploration is loaded, navigate to the Exploration view.
         *
         * @param explorationId The exploration to load
         * @returns {*} - a restricted promise
         */
        loadSavedExploration: function (explorationId) {
            var self = this;

            DataSourceCacheManager.reset(); // Reset the Cache Manager
            ComputedVariablesManager.reset();
            DisplayRulesManager.reset();

            var promise = sas.analytics.riskhp.utils.HprAPI.loadSavedExploration(explorationId);
            promise.done(function (exploration) {

                // TODO: dacagl - Start of Compatibility Check code
                // We have the exploration and cube data here (exploration & cube)
                // 1.  Initial validation of Exploration (TBD)
                // 2.  Apply changed labels to cube data so other components bound to it will pick up the new labels
                // See if there are any saved objects on the Exploration representing a changed label for a Data Item; If so, apply the new label
                // to the correct Data Item stored in the 'cube' data

                // Populate the model
                sas.analytics.riskhp.models.HPRiskModelManager.setExplorationModel(exploration);

                // Reset the ModelManager for the new content
                sas.analytics.riskhp.models.HPRiskModelManager.reset();

                // If there is already an Exploration View then we just need to reset it's
                // controller otherwise we need to navigate to it.
                var explorationView = self.getExplorationView();
                if (explorationView) {
                    // During load of Saved Exploration, don't allow events to fire during reset of UI
                    explorationView.getController().setResetExplorationEventsAllowed(false);
                    explorationView.getController().reset(true);
                    setTimeout(function () {
                        // After a brief period (UI has finished re-rendering after loading saved Exploration), turn events back on; This was a
                        // problem when loading a saved Exploration that had a different selected property pane from the one currently displayed
                        // in the UI; The splitter resize methods fire multiple times in this case
                        explorationView.getController().setResetExplorationEventsAllowed(true);
                    }, 1000);
                }
                else {
                    self.getRouter().navTo(sas.analytics.riskhp.utils.HprConstants.HPR_EXPLORATION_VIEW_ID, null, true);
                }
            });
            return promise;
        },

        /**
         * Open a data source without going through the DataSourceCacheManager.  Primarily used to retrieve the data from the data source so
         * it can be compared to another data source.
         *
         * @param dataSourceDetails The dataSourceDetails object (which consists of a DataSourceParams object + multiple types of supported dataSources) needed to load the new DataSource
         * @param selectedDataSourceType The selected type of the new DataSource
         * @returns {*} - a restricted promise
         */
        previewDataSource: function (dataSourceDetails, selectedDataSourceType) {
            // No promise.done or promise.fail functions yet, not sure if they are needed
            var promise = sas.analytics.riskhp.utils.HprAPI.previewDataSource(dataSourceDetails, selectedDataSourceType);
            return promise;
        }
    }
}());