/*
 * '(c) 2017, SAS Institute Inc.'
 */

jQuery.sap.declare("sas.analytics.riskhp.views.datapane.DataPaneModelManager");

jQuery.sap.declare("sap.ui.model.json.JSONModel");

jQuery.sap.require("sas.analytics.riskhp.utils.HprUtil");
jQuery.sap.require("sas.analytics.riskhp.utils.HprConstants");
jQuery.sap.require("sas.analytics.riskhp.models.DataSourceParams");
jQuery.sap.require("sas.analytics.riskhp.models.HPRiskModelManager");

/**
 * This is an odd class.  It's intent is to provide support for the DataPane when it first stands up and when it needs
 * be saved.  Additionally it provides useful helper methods for the DataPane when it is in use.  I have tried to document
 * the methods (at least the public ones) and how they are intended to be used.  It also keeps some state while the DataPane
 * is in use such as the horizons' cache, and the current DataSource.  In the end this is a combination of the DataPane's
 * view model (at least start and end) some utility methods from the DataPaneUtils class, and some state.
 * @name sas.analytics.riskhp.views.datapane.DataPaneModelManager
 */
sas.analytics.riskhp.views.datapane.DataPaneModelManager = (function () {

    var nodeOrderMap = {};
    var expandedNodeIds = []; // Keep track of the ids for opened tree nodes
    var horizonsOrderCache = {};
    var dataSource = null;
    var selectedDataSourceId = null;
    var treeTableModel = null;
    //var isTreeTableModelHydrated = false;
    var DELETED_DATAITEM = "DELETED_DATAITEM";

    /**
     * Helper method to encapsulate the check as to whether a root node should be added.
     * @param type
     * @param items
     * @returns {boolean}
     * @private
     */
    function _shouldAddRootNode(type, items) {
        if (items && items.length > 0) {
            return true;
        }
        // TODO: Will need to remove scenarios category when the cube doesn't support them in is't capabilities. See David A.
        return (type === sas.analytics.riskhp.models.dataitems.Scenario.SCENARIO && dataSource.isStressCube());
    }

    /**
     * For use when creating a new exploration to build the default tree node model for the data pane
     * @returns {{}} The tree for the data pane as a single node
     * @private
     */
    function _buildTreeNodesModel() {
        // Create a Root Node with a placeholder for children
        var rootNode = {children: []};

        // Build the nodes using the model created for a new Exploration and add the nodes to the Root Node
        _buildNodes(rootNode, sas.analytics.riskhp.models.dataitems.Classvar.CLASSVAR, HprConstants.HPR_DATA_PANE_CLASS_VARIABLES_NODE_ID);
        _buildNodes(rootNode, sas.analytics.riskhp.models.dataitems.Hierarchy.HIERARCHY, HprConstants.HPR_DATA_PANE_HIERARCHIES_NODE_ID);
        _buildNodes(rootNode, sas.analytics.riskhp.models.dataitems.RiskCube.RISKCUBE, HprConstants.HPR_DATA_PANE_RISK_CUBES_NODE_ID);
        if (!dataSource.isStressCube()) {
            _buildNodes(rootNode, sas.analytics.riskhp.models.dataitems.Horizon.HORIZON, HprConstants.HPR_DATA_PANE_HORIZONS_NODE_ID);
        }
        _buildNodes(rootNode, sas.analytics.riskhp.models.dataitems.Outvar.OUTVAR, HprConstants.HPR_DATA_PANE_OUTPUT_VARIABLES_NODE_ID);
        _buildNodes(rootNode, sas.analytics.riskhp.models.dataitems.Compvar.COMPVAR, HprConstants.HPR_DATA_PANE_COMPUTED_VARIABLES_NODE_ID);
        _buildNodes(rootNode, sas.analytics.riskhp.models.dataitems.Benchvar.BENCHVAR, HprConstants.HPR_DATA_PANE_BENCHMARK_VARIABLES_NODE_ID);
        _buildNodes(rootNode, sas.analytics.riskhp.models.dataitems.Statistic.STATISTIC, HprConstants.HPR_DATA_PANE_STATISTICS_NODE_ID);
        if (dataSource.isStressCube()) {
            _buildNodesForScenario(rootNode);
        }

        return rootNode;
    }

    /**
     * Helper method to build a branch for the tree table model for the given dataType
     * @param parentNode
     * @param dataType
     * @param dataTypeParentNodeId
     * @private
     */
    function _buildNodes(parentNode, dataType, dataTypeParentNodeId) {
        // Get the appropriate data times for the given data type
        var dataTypeItems = dataSource.getDataItemsForType(dataType);

        // If the data items exist, build a parent node and populate it with children nodes
        if (_shouldAddRootNode(dataType, dataTypeItems)) {
            // Parent node contains some of the same properties as the children node to help with binding in the HprTreeTableRow class
            var dataTypeParentNode = {
                label: sas.analytics.riskhp.utils.HprUtil.dataTypeTitle(dataType),
                nodeType: HprConstants.HPR_TREETABLE_LEVEL_TOP,
                folderType: dataType,
                id: dataTypeParentNodeId,
                isHprMoveable: false,
                isHprDraggable: false,
                isHprSelectable: false,
                isHprEditable: false
            };

            // Add a lineage property which will help with search filtering on nested folders
            dataTypeParentNode.lineage = dataTypeParentNode.label;

            // Horizons require some special processing because they can be displayed to the user in different ways (Absolute vs. Relative)
            // based on the type of cube and the Display Horizons Preference setting; The Horizon Data Item objects we read in when the cube
            // is opened are considered the Absolute Horizons.  So if we don't need Relative Horizons, just use the original Horizon Data Item
            // objects as they are
            if (dataType === sas.analytics.riskhp.models.dataitems.Horizon.HORIZON) {
                var horizonDisplayType = sas.analytics.riskhp.utils.HprUtil.getDesiredHorizonDisplayType(dataSource);
                if (horizonDisplayType === HprConstants.DISPLAY_HORIZONS_RELATIVE_VALUE) {
                    // Update the folder node's folderType property with the correct type of Horizon
                    dataTypeParentNode.folderType = sas.analytics.riskhp.models.dataitems.RelativeHorizon.RELATIVEHORIZON;

                    // Construct new Relative Horizon Data Items and apply the label changes to them
                    dataTypeItems = sas.analytics.riskhp.models.DataSourceUtil.generateRelativeHorizonDataItems(dataSource.getHorizons(), true, dataSource.getId());
                }

                // Construct the children nodes for this type of folder node
                _buildHorizonChildNodes(dataTypeParentNode, dataTypeItems, horizonDisplayType);
            }
            else {
                // Construct the children nodes for this type of folder node
                _buildChildNodes(dataTypeParentNode, dataTypeItems);
            }

            if (!parentNode.children) {
                parentNode.children = [];
            }
            parentNode.children.push(dataTypeParentNode);
        }
    }

    function _buildNodesForScenario(parentNode) {
        var resultSets = dataSource.getDataItemsForType(sas.analytics.riskhp.models.dataitems.ResultSet.RESULTSET);

        // Parent node contains some of the same properties as the children node to help with binding in the HprTreeTableRow class
        var scenarioParentNode = {
            label: sas.analytics.riskhp.utils.HprUtil.dataTypeTitle(sas.analytics.riskhp.models.dataitems.Scenario.SCENARIO),
            nodeType: sas.analytics.riskhp.utils.HprConstants.HPR_TREETABLE_LEVEL_TOP,
            folderType: sas.analytics.riskhp.models.dataitems.Scenario.SCENARIO,
            id: sas.analytics.riskhp.utils.HprConstants.HPR_DATA_PANE_SCENARIOS_NODE_ID,
            isHprMoveable: false,
            isHprDraggable: false,
            isHprSelectable: false,
            isHprEditable: false
        };

        // Add a lineage property which will help with search filtering on nested folders
        scenarioParentNode.lineage = scenarioParentNode.label;

        // Construct the children nodes for the scenario folder node
        _buildScenarioChildNodes(scenarioParentNode, resultSets);

        if (!parentNode.children) {
            parentNode.children = [];
        }
        parentNode.children.push(scenarioParentNode);
    }

    /**
     * Construct Data Pane children Tree nodes by looping through the 'dataTypeItems'
     * @param folderNode
     * @param dataTypeItems
     */
    function _buildChildNodes(folderNode, dataTypeItems) {
        // create a temporary child array and get the child order map
        var folderNodeChildren = [],
            folderNodeChildrenOrderMap = nodeOrderMap[folderNode.id] ? nodeOrderMap[folderNode.id].children : {};

        // Loop through the items and create children nodes
        _.each(dataTypeItems, function (dataTypeItem) {
            // Children node contains DataItem object and some of the same properties as the parent node
            // to help with binding in the HprTreeTableRow class
            var node = {
                id: dataTypeItem.id,
                dataItem: dataTypeItem,
                nodeType: dataTypeItem.type,
                isHprMoveable: true,
                isHprDraggable: true,
                isHprSelectable: true,
                isHprEditable: true,
                lineage: folderNode.lineage
            };
            if (dataTypeItem.type === sas.analytics.riskhp.models.dataitems.Hierarchy.HIERARCHY) {
                node.id = node.dataItem.getId();
                node.lineage = folderNode.lineage + " " + node.dataItem.label; // Must update because we are adding sub-nodes

                // create a temporary child array and get the child order map
                var hierarchyChildren = [],
                    hierarchyChildrenOrderMap = folderNodeChildrenOrderMap[node.id] ? folderNodeChildrenOrderMap[node.id].children : {};

                _.each(dataTypeItem.classVariables, function(classVariable){
                    var classVariableNode = {
                        id: classVariable.id,
                        dataItem: classVariable,
                        nodeType: classVariable.type,
                        dataItemChild: true,
                        isHprMoveable: false,
                        isHprDraggable: false,
                        isHprSelectable: false,
                        isHprEditable: false,
                        lineage: node.lineage
                    };
                    hierarchyChildren.push(classVariableNode);
                });
                node.children = _.sortBy(hierarchyChildren, function(childNode) {
                    if (hierarchyChildrenOrderMap && hierarchyChildrenOrderMap[childNode.id]) {
                        return hierarchyChildrenOrderMap[childNode.id].index;
                    }
                    return sas.risk.common.utils.CommonConstants.MAX_INTEGER;
                });
            }
            folderNodeChildren.push(node);
        });

        // sort the nodes
        folderNode.children = _.sortBy(folderNodeChildren, function(childNode) {
            if (folderNodeChildrenOrderMap && folderNodeChildrenOrderMap[childNode.id]) {
                return folderNodeChildrenOrderMap[childNode.id].index;
            }
            return sas.risk.common.utils.CommonConstants.MAX_INTEGER;
        });
    }

    /**
     * Construct Data Pane children Tree nodes by looping through the 'dataTypeItems'
     * @param folderNode
     * @param dataTypeItems
     * @param horizonDisplayType
     */
    function _buildHorizonChildNodes(folderNode, dataTypeItems, horizonDisplayType) {
        // create a temporary child array and get the child order map
        var folderNodeChildren = [],
            folderNodeChildrenOrderMap = nodeOrderMap[folderNode.id] ? nodeOrderMap[folderNode.id].children : {};

        // Loop through the items and create children nodes
        _.each(dataTypeItems, function (dataTypeItem) {
            // Children node contains DataItem object and some of the same properties as the parent node
            // to help with binding in the HprTreeTableRow class
            var node = {
                id: dataTypeItem.id,
                dataItem: dataTypeItem,
                nodeType: dataTypeItem.type,
                isHprMoveable: true,
                isHprDraggable: true,
                isHprSelectable: true,
                isHprEditable: true,
                lineage: folderNode.lineage
            };
            folderNodeChildren.push(node);
        });

        // If we have cached for this display type,
        if (sas.analytics.riskhp.views.datapane.DataPaneModelManager.hasCachedHorizons(horizonDisplayType)) {
            var horizonsOrderMap = horizonsOrderCache[horizonDisplayType];

            // Keep the cache clean
            delete horizonsOrderCache[horizonDisplayType];

            // sort the horizons using the cache nodes
            folderNode.children = _.sortBy(folderNodeChildren, function (childNode) {
                if (horizonsOrderMap && horizonsOrderMap[childNode.id]) {
                    return horizonsOrderMap[childNode.id].index;
                }
                return sas.risk.common.utils.CommonConstants.MAX_INTEGER;
            });
        }
        else {
            // sort the nodes
            folderNode.children = _.sortBy(folderNodeChildren, function (childNode) {
                if (folderNodeChildrenOrderMap && folderNodeChildrenOrderMap[childNode.id]) {
                    return folderNodeChildrenOrderMap[childNode.id].index;
                }
                return sas.risk.common.utils.CommonConstants.MAX_INTEGER;
            });
        }
    }

    /**
     * Construct Data Pane children tree nodes for the scenarios folder.
     * @param scenarioParentNode
     * @param resultSets
     */
    function _buildScenarioChildNodes(scenarioParentNode, resultSets) {
        var HprConstants = sas.analytics.riskhp.utils.HprConstants;
        var HprUtil = sas.analytics.riskhp.utils.HprUtil;
        var DataSourceUtil = sas.analytics.riskhp.models.DataSourceUtil;
        var DataSourceCacheManager = sas.analytics.riskhp.models.DataSourceCacheManager;
        var RelativeHorizon = sas.analytics.riskhp.models.dataitems.RelativeHorizon;

        DataSourceCacheManager.getDataSource(DataSourceCacheManager.getSelectedDataSourceId()).done(
            function(dataSource) {
                // determine the horizon display setting
                var horizonDisplayType = HprUtil.getDesiredHorizonDisplayType(dataSource);

                // create a temporary child array and get the child order map
                var scenarioParentNodeChildren = [],
                    scenarioParentNodeChildrenOrderMap = nodeOrderMap[scenarioParentNode.id] ? nodeOrderMap[scenarioParentNode.id].children : {};

                // Loop through the result sets and create children nodes
                _.each(resultSets, function (resultSet) {
                    // Children node contains DataItem object and some of the same properties as the parent node
                    // to help with binding in the HprTreeTableRow class
                    var resultSetNode = {
                        id: resultSet.id,
                        dataItem: resultSet,
                        nodeType: resultSet.type,
                        isHprMoveable: true,
                        isHprDraggable: true,
                        isHprSelectable: true,
                        isHprEditable: true,
                        lineage: scenarioParentNode.lineage + " " + resultSet.label,
                    };

                    // create a temporary child array and get the child order map
                    var resultSetNodeChildren = [],
                        resultSetNodeChildrenOrderMap = scenarioParentNodeChildrenOrderMap[resultSetNode.id] ? scenarioParentNodeChildrenOrderMap[resultSetNode.id].children : {};

                    // add the base case horizon
                    if (resultSet.getBasecaseHorizon()) {
                        var basecaseHorizonId = resultSet.getBasecaseHorizon().id,
                            horizonDataItem = dataSource.getDataItemById(basecaseHorizonId),
                            basecaseHorizon = horizonDisplayType === HprConstants.DISPLAY_HORIZONS_RELATIVE_VALUE ? new RelativeHorizon(horizonDataItem) : horizonDataItem;

                        resultSetNodeChildren.push({
                            id: basecaseHorizon.id,
                            dataItem: basecaseHorizon,
                            nodeType: basecaseHorizon.type,
                            isHprMoveable: false,
                            isHprDraggable: true,
                            isHprSelectable: true,
                            isHprEditable: false,
                            lineage: resultSetNode.lineage
                        });
                    }

                    // add scenarios
                    _.each(resultSet.getScenarios(), function(scenario){
                        var scenarioNode = {
                            id: scenario.id,
                            dataItem: scenario,
                            nodeType: scenario.type,
                            dataItemChild: true,
                            isHprMoveable: true,
                            isHprDraggable: true,
                            isHprSelectable: true,
                            isHprEditable: false,
                            lineage: resultSetNode.lineage
                        };
                        resultSetNodeChildren.push(scenarioNode);
                    });

                    resultSetNode.children = _.sortBy(resultSetNodeChildren, function(childNode) {
                        if (resultSetNodeChildrenOrderMap && resultSetNodeChildrenOrderMap[childNode.id]) {
                            return resultSetNodeChildrenOrderMap[childNode.id].index;
                        }
                        return sas.risk.common.utils.CommonConstants.MAX_INTEGER;
                    });

                    scenarioParentNodeChildren.push(resultSetNode);
                });

                // sort the nodes
                scenarioParentNode.children = _.sortBy(scenarioParentNodeChildren, function(childNode) {
                    if (scenarioParentNodeChildrenOrderMap && scenarioParentNodeChildrenOrderMap[childNode.id]) {
                        return scenarioParentNodeChildrenOrderMap[childNode.id].index;
                    }
                    return sas.risk.common.utils.CommonConstants.MAX_INTEGER;
                });
            }
        );
    }

    /**
     * Given the root object of the DataPane's treetable and an id for the toplevel node that is desired returns the
     * corresponding object for that branch.  May return undefined if the branch is not in the list.
     * @param treeTableRootNode
     * @param nodeId
     * @private
     */
    function _getBranchByNodeId(treeTableRootNode, nodeId) {
        return _.find(treeTableRootNode.children, function(item) {
            return (item.nodeType === HprConstants.HPR_TREETABLE_LEVEL_TOP && item.id === nodeId);
        });
    }

    /**
     * Given the current treeTableRootNode inspects the horizons branch of ti to determine if the horizons it contains
     * are RelativeHorizons.  Returns true if they are, false otherwise
     * @param treeTableRootNode
     * @returns {boolean}
     * @private
     */
    function _displayingRelativeHorizons(treeTableRootNode) {

        // Find the top level node that is the horizons list
        var horizonsBranch = _getBranchByNodeId(treeTableRootNode,HprConstants.HPR_DATA_PANE_HORIZONS_NODE_ID);

        // If the horizons branch was not found or it had no children return false
        if (!horizonsBranch || !horizonsBranch.children || horizonsBranch.children.length === 0) {
            return false;
        }

        // Only need to check first child as we ASSUME all the children are of the same type
        var firstChild = horizonsBranch.children[0];
        return (firstChild && firstChild.nodeType === sas.analytics.riskhp.models.dataitems.RelativeHorizon.RELATIVEHORIZON);
    }

    /**
     * Takes the current state of the data pane's tree table model and compares the horizons that are being shown to what
     * is stored in the cache.  Since that is the most correct ordering of horizons of that type, if there is an old
     * list for that type in the horizons cache remove it
     * @param treeTableModelData
     * @private
     */
    function _updateHorizonsCacheBeforeConvertingToJSON(treeTableModelData) {
        if (dataSource && dataSource.isStressCube() || Object.keys(horizonsOrderCache).length === 0) {
            return;     // stress cubes do not have a root horizon node, so this is a no-op
        }

        var isShowingRelativeHorizons = _displayingRelativeHorizons(treeTableModelData);
        if (isShowingRelativeHorizons) {
            delete horizonsOrderCache[HprConstants.DISPLAY_HORIZONS_RELATIVE_VALUE];
        }
        else {
            delete horizonsOrderCache[HprConstants.DISPLAY_HORIZONS_ABSOLUTE_VALUE];
        }
    }

    /**
     * Returns a map of sorted child nodes keyed by data item ID. The sorted child nodes
     * are keyed by the parent node ID.
     * @param treeTableModelData
     * @private
     */
    function _buildChildNodeOrderMap(nodes, areRootNodes) {
        var nodeMap = {};

        for (var i = 0; i < nodes.length; i++) {
            if (areRootNodes === true) {
                nodeMap[nodes[i].id] = {children: _buildChildNodeOrderMap(nodes[i].children)};
            }
            else {
                nodeMap[nodes[i].id] = {index: i};

                if (nodes[i].children && nodes[i].children.length > 0) {
                    nodeMap[nodes[i].id].children = _buildChildNodeOrderMap(nodes[i].children);
                }
            }
        }

        return nodeMap;
    }


    /**
     * Helper function to take an object that "might" be a JSON model and get the data off of it, otherwise the object
     * is assumed to be the data from a JSON model and is returned.
     * @param model
     * @returns {*}
     * @private
     */
    function _getDataFromModel(model) {
        if (model instanceof sap.ui.model.json.JSONModel) {
            return model.getData();
        }
        return model;
    }

    return {

        /**
         * Create a model for the Data Pane fragment.  It is populated with information from the currently loaded
         * exploration in the HPRiskModelManager object.
         * @param isOpeningExploration {Boolean} true if this is being called while opening an exploration
         */
        initDataPaneModel: function (isOpeningExploration) {
            // Init the data pane model from the currently loaded exploration model
            var explorationModel = sas.analytics.riskhp.models.HPRiskModelManager.getExplorationModel();

            // Set the selected data source id
            selectedDataSourceId = DataSourceCacheManager.getSelectedDataSourceId();

            // model represents a saved exploration if there are pages in the original analysis layout
            var explorationPages = explorationModel.getProperty("/exploration/analysis/layout/pages");

            // Tree table is reconstructed every time. Expanded nodes is not persisted.
            treeTableModel = null;
            expandedNodeIds = [];

            if (isOpeningExploration === true && explorationPages && explorationPages.length > 0) {
                // Know it is a saved exploration so source the attributes for the data pane model from what is stored in the exploration
                var dataPaneModelData = explorationModel.getProperty("/exploration/analysis/layout/dataPane");
                nodeOrderMap = jQuery.extend(true, {}, dataPaneModelData.nodeOrderMap);
                horizonsOrderCache = jQuery.extend(true, {}, dataPaneModelData.horizonNodeChildrenCache);
            }
            else {
                // See if the selected DataSourceParam object holds a copy of the DataPane TreeTable model, if it does use it; This
                // allows us to keep the order of the nodes in the TreeTable model if the user has dragged them around and keeps the
                // updated DataItem labels;
                var selectedDataSourceParam = DataSourceCacheManager.getSelectedDataSourceParam();
                var selectedDataSourceTreeTableModel = selectedDataSourceParam ? selectedDataSourceParam.getDataSourceTreeTableModelObject(selectedDataSourceId) : undefined;
                if (selectedDataSourceTreeTableModel) {
                    nodeOrderMap = jQuery.extend(true, {}, selectedDataSourceTreeTableModel.nodeOrderMap);
                    horizonsOrderCache = jQuery.extend(true, {}, selectedDataSourceTreeTableModel.horizonNodeChildrenCache);
                } else {
                    nodeOrderMap = {};
                    horizonsOrderCache = {};
                }
            }
        },

        /**
         * Update the 'dataPaneExpandedNodeIds' array by adding a selectedRowId if it was expanded or deleting it if it was closed.
         *
         * @param selectedRowId
         * @param expanded
         */
        updateExpandedNodeIds: function (selectedRowId, expanded) {

            // Retrieve the list of expanded node ids
            var newExpandedNodeIdsArray = DataPaneModelManager.getExpandedNodeIds();

            // If the selected row id is expanded add it to the list IFF it isn't already in there (we are using the list as a Set)
            if (expanded) {

                // Only add the id if it is not in the array already
                if ($.inArray(selectedRowId, newExpandedNodeIdsArray) === -1) {
                    newExpandedNodeIdsArray.push(selectedRowId);
                }
            }
            else {
                // Row is not expanded so remove it from the list by looping through the array backwards (since indices
                // and length will change as items are removed) and remove the id if it's found.
                for (var i = newExpandedNodeIdsArray.length; i--;) {
                    if (newExpandedNodeIdsArray[i] === selectedRowId) {
                        newExpandedNodeIdsArray.splice(i, 1);
                    }
                }
            }

            expandedNodeIds = newExpandedNodeIdsArray;
        },

        /**
         * Returns the expanded tree node ids for the Data Pane
         *
         * @returns {Array} dataPaneExpandedNodeIds
         */
        getExpandedNodeIds: function () {
            return $.extend(true, [], expandedNodeIds); // only return a copy to be safe.
        },

        /**
         * Reset the array to empty for the expanded tree nodes on the Data Pane
         */
        resetExpandedNodeIds: function () {
            expandedNodeIds = [];
        },

        /**
         * Given the current treeTableModel generates a dehydrated version of the state of the DataPane suitable for
         * storing in a saved exploration.
         * @returns {{dataSourceTreeTableModel: {}, horizonNodeChildrenCache: {}}}
         */
        toJSON: function () {
            var treeTableModelData = undefined;

            if (dataSource) {
                // We need a dataSource to continue
                treeTableModelData = _getDataFromModel(this.getTreeTableModel());
            }

            // return previous values if no tree model
            if (!treeTableModelData) {
                return {
                    horizonNodeChildrenCache: horizonsOrderCache,
                    nodeOrderMap: nodeOrderMap
                };
            }

            // Clean up the horizons cache with the results of the latest treetable model
            _updateHorizonsCacheBeforeConvertingToJSON(treeTableModelData);

            // Build and return the JSON representation to be stored.
            return {
                horizonNodeChildrenCache: horizonsOrderCache,
                nodeOrderMap: _buildChildNodeOrderMap(treeTableModelData.children, true)
            };
        },

        /**
         * Retrieves the id of the data source this model manager is pulling data items from
         * @returns {*}
         */
        getSelectedDataSourceId: function () {
            return selectedDataSourceId;
        },

        /**
         * Set the data source this model manager should use
         * @param loadedDataSource
         */
        setSelectedDataSource: function (loadedDataSource) {
            dataSource = loadedDataSource;
        },

        /**
         * Retrieves the selected data source this model manager is using
         * @returns {*}
         */
        getSelectedDataSource: function() {
            return dataSource;
        },

        /**
         * Returns a valid tree table model to be encapsulated by a sap JSONModel and placed on the  DataPane's tree table
         * This model is only valid right after a call to initDataPaneModel() as the model is sourced from the currently
         * loaded exploration.  In a reset scenario one would set the new exploration in HPRiskModelManager then as
         * a part of the reset of the DataPane the initDataPaneModel() method should be called and the DataSource should
         * be set.  Once that is done this method can be called to get the model for the DataPane's tree table.  The
         * returned model will be hydrated.
         * @returns {*}
         */
        getTreeTableModel: function () {
            // We only want to build the treeTableModel once
            if (!treeTableModel) {
                treeTableModel = _buildTreeNodesModel();
            }
            return treeTableModel;
        },

        /**
         * Construct Data Pane children Tree nodes by looping through the 'dataTypeItems'
         * @param folderNode
         * @param dataTypeItems
         */
        dataPaneTreeConstructChildrenNodes: function(folderNode, dataTypeItems, horizonType) {
            if (horizonType) {
                _buildHorizonChildNodes(folderNode, dataTypeItems, horizonType);
            }
            else {
                _buildChildNodes(folderNode, dataTypeItems);
            }
        },

        /**
         * Construct Data Pane children tree nodes for the scenarios folder.
         * @param scenarioParentNode
         * @param resultSets
         */
        dataPaneTreeConstructScenarioChildrenNodes: function(scenarioParentNode, resultSets) {
            _buildScenarioChildNodes(scenarioParentNode, resultSets);
        },

        /**
         * Return the branch from the passed in treeTableModelData that represents the ClassVars. May return undefined
         * if that branch doesn't exist
         * @param treeTableModelData This is a result of calling treetable.getModel().oData
         * @returns {*}
         */
        getClassVariableBranch: function(treeTableModelData) {
            return _getBranchByNodeId(treeTableModelData,HprConstants.HPR_DATA_PANE_CLASS_VARIABLES_NODE_ID)
        },

        /**
         * Do any of the ClassVariables in the treeTable need to have their counts updated.
         * @param treeTableModelData
         * @returns {boolean}
         */
        shouldClassVariableCountsBeUpdated: function (treeTableModelData) {
            // Get the class variable branch
            var classVariableBranch = DataPaneModelManager.getClassVariableBranch(treeTableModelData);

            // If there isn't one the answer is false
            if (!classVariableBranch) {
                return false;
            }

            // walk the children of the class var branch to see if any of the children have a dataitem with a true updateCount value
            var childNodes = classVariableBranch.children;
            var childThatNeedsCountUpdated = _.find(childNodes, function(child){
                var dataItem = child.dataItem;
                return (dataItem && dataItem.getUpdateCount());
            });

            // if a child was found the answer is true otherwise it is false
            return !(_.isNull(childThatNeedsCountUpdated) || _.isUndefined(childThatNeedsCountUpdated));
        },

        /**
         * Set the passed in list as the list for the given horizon type to be cached.
         * @param nodesArray a list of hydrated horizon nodes
         * @param horizonType one of the HprConstants.DISPLAY_HORIZONS_ constants
         */
        setHorizonsCache: function(nodesArray, horizonType) {
            horizonsOrderCache[horizonType] = _buildChildNodeOrderMap(nodesArray);
        },

        /**
         * Does the horizon cache hold a list for the passed in horizonType
         * @param horizonType one of the HprConstants.DISPLAY_HORIZONS_ constants
         */
        hasCachedHorizons: function(horizonType) {
            return _.has(horizonsOrderCache,horizonType);
        },

        updateStressCubeHorizonsForDisplayType: function(treeTableRootNode) {
            var desiredHorizonDisplayType = sas.analytics.riskhp.utils.HprUtil.getDesiredHorizonDisplayType(dataSource);

            // get the set of result set horizons
            var horizonDataItems = _.filter(dataSource.masterDataItemArray, function(dataItem) { return dataItem.type === sas.analytics.riskhp.models.dataitems.Horizon.HORIZON && dataItem.resultSetId; });

            var horizonsMappedByResultSet = {};
            _.each(horizonDataItems, function(horizonDataItem) {
                horizonsMappedByResultSet[horizonDataItem.resultSetId] = horizonDataItem;
            });

            var parentBranchNode = _getBranchByNodeId(treeTableRootNode, HprConstants.HPR_DATA_PANE_SCENARIOS_NODE_ID),
                DataPaneUtil = sas.analytics.riskhp.views.datapane.DataPaneUtil,
                newHorizon;
            _.each(parentBranchNode.children, function(rsChildNode) {
                _.each(rsChildNode.children, function(childNode) {
                    if (desiredHorizonDisplayType === HprConstants.DISPLAY_HORIZONS_RELATIVE_VALUE) {
                        if(childNode.nodeType === sas.analytics.riskhp.models.dataitems.Horizon.HORIZON) {
                            newHorizon = new sas.analytics.riskhp.models.dataitems.RelativeHorizon(dataSource.getDataItemById(DataPaneUtil.getNodeDataItemId(childNode)));
                        }
                    }
                    else {
                        if (childNode.nodeType === sas.analytics.riskhp.models.dataitems.RelativeHorizon.RELATIVEHORIZON) {
                            newHorizon = horizonsMappedByResultSet[DataPaneUtil.getNodeDataItemId(rsChildNode)];
                        }
                    }

                    if (newHorizon) {
                        childNode.nodeType = newHorizon.type;

                        // dehydrated might have both, but the data item is just an object and will be replaced at hydration time
                        if (childNode.dataItemId) {
                            childNode.dataItemId = newHorizon.id;   // dehydrated, so replace the data item
                        }
                        else {
                            childNode.dataItem = newHorizon;        // hydrated, so update the data item id
                        }

                        // Reset the new Horizon whenever we use it since we are in an each loop
                        newHorizon = null;
                    }
                });
            });
        }
    }
}());

DataPaneModelManager = sas.analytics.riskhp.views.datapane.DataPaneModelManager;