// (c) 2016, SAS Institute Inc.

jQuery.sap.require("sas.analytics.riskhp.models.HPRiskModelManager");
jQuery.sap.require('sas.analytics.riskhp.settings.ApplicationSettings');
jQuery.sap.require("sas.analytics.riskhp.utils.HprUtil");
jQuery.sap.require("sas.analytics.riskhp.views.datapane.DraggedRowData");
jQuery.sap.require("sas.analytics.riskhp.views.datapane.DataPaneUtil");

sap.ui.define([
    'jquery.sap.global',
    'sap/ui/base/Object',
    'sas/hc/ui/core/delegate/Draggable',
    'sas/analytics/riskhp/views/datapane/DraggedRowData'],
    function (jQuery, BaseObject, Draggable, DraggedRowData) {
        "use strict";

        var $ = jQuery;
        var DataPaneUtil = sas.analytics.riskhp.views.datapane.DataPaneUtil;
        var HprUtil = sas.analytics.riskhp.utils.HprUtil;
        var HPRiskModelManager = sas.analytics.riskhp.models.HPRiskModelManager;

        /**
         * Creates a Draggable delegate that can be attached to Controls requiring
         * capabilities dragging.
         * @class Delegate for dragging Table rows.
         * @author SAS
         * @extends sas.hc.ui.core.delegate.Draggable
         * @version 4.0
         * @constructor
         * @name sas.analytics.riskhp.views.datapane.DraggableRowDelegate
         * @public
         */
        var RowDraggable = Draggable.extend("sas.analytics.riskhp.views.datapane.DraggableRowDelegate", {

            /**
             * The constructor.
             *
             * @param {sap.ui.core.Control} oControl the Control of which this Draggable is the delegate
             * @param {object} oDragOptions options to utilize while dragging. Based on jQuery UI draggable options.
             * @param {boolean} bEnableTouch set to false to disable touch dragging.
             */
            constructor: function (oControl, oDragOptions, bEnableTouch) {
                // Use defaults and allow overrides by consumer
                var oConfig = $.extend({}, {
                    appendTo: "body",
                    cursor: "move",
                    revert: "invalid",
                    opacity: 1.0,
                    helper: this._fnCreateDragHelper,
                    cursorAt: {top: 5, left: 5},
                    distance: 20,
                    scroll: false,
                    stop: function(event, ui) {
                        // Clean up draggables we collected so we are back at a clean state
                        sas.analytics.riskhp.models.HPRiskModelManager.setDraggableObjects([]);

                        // Clear the dragged (but not officially selected) row
                        var model = DataPaneUtil.getDataPaneTreeTableSelectionModel();
                        model.clearDraggedRowPath();
                    }
                }, oDragOptions);

                Draggable.call(this, oControl, oConfig, bEnableTouch);
            },

            /**
             * Create an element that is dragged instead of the original element.
             *
             * @param event
             * @returns {*}
             * @private
             */
            _fnCreateDragHelper: function (event) {
                var $draggable = $(this), // This will be the draggable element
                    oTable,
                    $helper,
                    helperLabel,
                    rowSelectionModel = DataPaneUtil.getDataPaneTreeTableSelectionModel();

                // S1432372: Guard against something not being initialized correctly in the creation of the Helper DIV
                var $errorHelper = $('<div class="rowHoverHelper">').text(HprUtil.getBundle().getText("DataPane.error.creating.drag.helper.txt"));
                if ($draggable && $draggable.data("sasTableRow") && $draggable.data("sasTableRow").getParent()) {
                    oTable = $draggable.data("sasTableRow").getParent();
                } else {
                    return $errorHelper;
                }

                var draggedFromId = this.id;

                // $draggable.on('mouseup',function(){
                //     $(this).draggable('destroy');
                //     //in case you like to use if after
                //     $(this).hide();
                //     //as in your code
                //     //$(this).remove();
                // });

                /// Need the rowPath of the draggedFromId
                var draggedRowPath = DataPaneUtil.getRowPathForRowId(draggedFromId);
                // Check to see if the current dragged id is in the list of selected items
                // if so make sure the dragged row path is empty, otherwise add it to the list of dragged items
                if (rowSelectionModel.isRowPathSelected(draggedRowPath)) {
                    rowSelectionModel.setDraggedRowPath(null);
                }
                else {
                    rowSelectionModel.setDraggedRowPath(draggedRowPath);
                    rowSelectionModel.selectRow(draggedFromId,true);
                }

                // Store the table rows (actually id's) which are being dragged
                var draggedRows = [];

                // Get the model for the table
                var tableModel = oTable.getModel();

                // Get a list of the row paths that are selected in the table
                var selectedRowPaths = rowSelectionModel.getSelectedRows();

                // Need the currently selected dataSourceId as that is the data source all the dragged items are assocaited with
                var dataSourceId = DataSourceCacheManager.getSelectedDataSourceId();

                // Get the list of Objects the table knows are selected
                var selectedTableRowObjects = DataPaneUtil.getSelectedTableRowObjects();

                // Convert these data items to DraggedRowData objects and add them to the draggedRows array.
                _.each(selectedTableRowObjects, function(dataItemSuperObject){
                    if (dataItemSuperObject.dataItem) {
                        // This code retrieves the table path for the row so we can add that to the DraggedRowData object we will build
                        var matchingRowPath = _.find(selectedRowPaths, function(selectedRowPath) {
                            var tempDataItemSuperObject = tableModel.getProperty(selectedRowPath);
                            if (tempDataItemSuperObject && tempDataItemSuperObject.dataItem) {
                                return tempDataItemSuperObject.dataItem.getId() === dataItemSuperObject.dataItem.getId();
                            }
                            return false;
                        });

                        // Have everything to create a DraggedRowData object so create it and add it to our list
                        draggedRows.push( new DraggedRowData(dataItemSuperObject, dataSourceId, matchingRowPath) );
                    }
                });

                // Construct an appropriate label for the helper div
                if (draggedRows) {
                    switch (draggedRows.length) {
                        case 0:
                            helperLabel = "";
                            break;
                        case 1:
                            // When there is a single DataItem being dragged, use the ApplicationSettings 'showLabels'
                            // property to build the helperLabel so it will match the DataPane TreeTable node label
                            var showLabels = sas.analytics.riskhp.settings.ApplicationSettings.getShowLabels();
                            helperLabel = draggedRows[0].getDisplayLabel(showLabels);
                            break;
                        default:
                            // Use a generic label for more than 1 data items being dragged
                            helperLabel = HprUtil.getBundle().getText('DataPane.multipleDataItems.fmt', [draggedRows.length]);
                    }
                }

                // Create a new outer DIV element which will serve as the draggable DIV, and set the new label on it
               if (helperLabel.length === 0) {
                   // If the text for the helper is empty this means the user is trying to drag a top
                   // level folder which isn't allowed. However to make jQuery happy with this design
                   // we have to return a helper div (or it will crash) So we construct one but make it
                   // not visible.  Yes a HACK but it works.
                   $helper = $('<div class="rowHoverHelper" style="display: none;">').text(helperLabel);
               }
               else {
                    $helper = $('<div class="rowHoverHelper">').text(helperLabel);
               }

               // Attempt to remove the helper when an exception occurs - first try.
                $helper.on('mouseup',function(){
                    //$helper.draggable('destroy');
                    //in case you like to use if after
                    //$(this).hide();
                    //as in your code
                    $helper.remove();
                });

                // Construct an array of objects representing the dragged items
                // Store the array of dragged objects in the HPRiskModelManager so the drop handlers will have access
                // to them when they consume the drop
                if (draggedRows && draggedRows.length > 0) {
                    HPRiskModelManager.setDraggableObjects(draggedRows);
                }

                return $helper;
            },

            /**
             * Must be set to draggable in onAfterRendering function.
             */
            onAfterRendering: function () {
                var oTable = this._oControl;
                var aRows = oTable.getRows(),
                    oDragOptions = this._oDragOptions,
                    oConfig,
                    $domRefs;

                function getDomRefs(oRow) {
                    var $refs = oRow.getDomRef("fixed");
                    $refs = $($refs).add(oRow.getDomRef());
                    return $refs;
                }

                function assignDraggableToRow(oRow) {
                    $domRefs = getDomRefs(oRow);
                    oConfig = $.extend({}, oDragOptions, { });
                    $domRefs.draggable(oConfig);
                    $domRefs.data("sasTableRow", oRow); // Store the oRow object onto the dom ref at the "sasTableRow" key !!!! Whaaaatttt?!
                }

                // For each row, configure the draggable options for it
                aRows.forEach(function (oRow) {
                    assignDraggableToRow(oRow);
                });

                // Instead of listening as each row's context is updated, we'll listen to the table for '_rowsUpdated' events
                oTable.attachEvent("_rowsUpdated", function (oEvent) {
                    // Loop through all rows to enable/disable drag state
                    // Add the draggable options again for each row
                    // NOTE: This solves a problem with hidden rows in TreeTable being able to be dragged
                    var myRows = oTable.getRows();
                    myRows.forEach(function (oRow) {
                        assignDraggableToRow(oRow);
                    });
                });
            }
        });
        return RowDraggable;
    },
    true
);