// (c) 2017, SAS Institute Inc.

jQuery.sap.require("sas.analytics.riskhp.views.datapane.DataPaneUtil");
jQuery.sap.require("sas.analytics.riskhp.models.HPRiskModelManager");

sap.ui.define(['jquery.sap.global', 'sap/ui/base/Object', 'sas/hc/ui/core/delegate/Droppable'],
    function(jQuery, BaseObject, Droppable) {
        "use strict";

        var $ = jQuery;
        var DataPaneUtil = sas.analytics.riskhp.views.datapane.DataPaneUtil;
        var HPRiskModelManager = sas.analytics.riskhp.models.HPRiskModelManager;

        /**
         * Creates a Droppable delegate for the TreeTable in the DataPane.  This is a heavily customized version of the
         * "sas.hc.ui.core.delegate.Droppable" version.  However, it is almost just like the "sas.hc.ui.table.DroppableTableDelegate"
         * class with some changes to support our needs.
         * @class Delegate for dropping (i.e. moving) rows in the DataPane.
         * @author maasti
         * @extends sas.hc.ui.core.delegate.Droppable
         * @constructor
         * @name sas.analytics.riskhp.views.datapane.DroppableRowDelegate
         * @public
         */
        var RowDroppable = Droppable.extend("sas.analytics.riskhp.views.datapane.DroppableRowDelegate", {

            /**
             * The constructor.
             *
             * @param {sap.ui.core.Control} oControl the Control of which this Droppable is the delegate
             * @param {object} oDropOptions options to utilize while dropping. Based on jQuery UI droppable options.
             * @param {boolean} bEnableTouch set to false to disable touch dropping.
             * @param {boolean} bAllowRowReorder set to false to disable row re-ordering.
             */
            constructor: function (oControl, oDropOptions, bEnableTouch, bAllowRowReorder) {
                // Use defaults and allow overrides by consumer
                var oConfig = $.extend({}, {}, oDropOptions);
                if (bAllowRowReorder === undefined) {
                    bAllowRowReorder = oConfig.allowRowReorder; // Flexible way to assign options somewhat consistent with UI5
                    // This could *also* be undefined, but that is fine
                }
                this._bAllowRowReorder = bAllowRowReorder;

                // Have to keep track of DOM elements that have been made droppable so that they can be cleaned up
                // and avoid a memory leak
                this._droppableDomRefs = [];

                Droppable.call(this, oControl, oConfig, bEnableTouch);
            },

            _exit: function () {
                for (var i=0; i<this._droppableDomRefs.length; i++) {
                    if (this._droppableDomRefs[i]) {
                        var $droppable = this._droppableDomRefs[i].data('ui-droppable');
                        if ($droppable) {
                            $droppable.destroy();
                        }
                    }
                }
                delete this._droppableDomRefs;

                if (Droppable.prototype.exit) {
                    Droppable.prototype.exit.apply(this, arguments);
                }
            },

            /**
             * Helper method that determines if the draggable is currently over the upper-half or bottom-half of the droppable.
             * @param ui
             * @param $droppable
             * @returns {boolean}
             */
            _isInUpperHalf: function(ui, $droppable) {
                var $draggable = ui.draggable || ui.helper;
                return (ui.offset.top + $draggable.outerHeight() / 2) <= ($droppable.offset().top + $droppable.outerHeight() / 2);
            },

            /**
             * Controls which draggable elements are accepted by the droppable.
             * 
             * @param draggable The draggable elements
             * @returns {*}
             * @private
             */
            _canAcceptDrop: function(draggable) {
                if (draggable.length > 0) {
                    var model = DataPaneUtil.getDataPaneTreeTableSelectionModel();
                    // 'this.id' resolves to the table row id (thanks to the binding with the drop options on the table rows in onAfterRendering method) which in this case is the droppable id
                    return model.validDropArea(draggable[0].id, this.id);
                }
                return false;
            },

            // Must clean up existing droppables if any
            onBeforeRendering: function () {
                for (var i=0; i<this._droppableDomRefs.length; i++) {
                    if (this._droppableDomRefs[i]) {
                        var $droppable = this._droppableDomRefs[i].data('ui-droppable');
                        if ($droppable) {
                            $droppable.destroy();
                        }
                    }
                }

                this._droppableDomRefs = [];
            },

            /**
             * Must be set to droppable in onAfterRendering function.
             */
            onAfterRendering: function() {
                var self = this;
                this._hasRendered = true;

                var oTable = this._oControl,
                    aRows = oTable.getRows(),
                    oDropOptions = this._oDropOptions,
                    oDelegate = this,
                    $domRefs,
                    oConfig,
                    oCopyOfConfig;

                function getDomRefs(oRow) {
                    var $refs = oRow.getDomRef("fixed");
                    $refs = $($refs).add(oRow.getDomRef());
                    return $refs;
                }

                function oConfigDropOptions(oRow) {
                    return {
                        disabled: (!oRow.getBindingContext()),
                        tolerance: "touch",
                        accept: self._canAcceptDrop,
                        over: function(event,ui) {
                            ui.droppable = $(this);
                            ui.draggable.data("current-droppable", $(this)); // Keep some custom data about the droppable for later use

                            // Remove the hover styles from everything
                            $(".dataPaneTopRowDropZoneHovered").removeClass("dataPaneTopRowDropZoneHovered");
                            $(".dataPaneBottomRowDropZoneHovered").removeClass("dataPaneBottomRowDropZoneHovered");

                            // Figure out if we have a valid drop area and style it if so.
                            var droppableId = ui.droppable[0].id;
                            var draggableId = ui.draggable[0].id;
                            if (self._isInUpperHalf(ui, ui.droppable)) {
                                var model = DataPaneUtil.getDataPaneTreeTableSelectionModel();
                                if (!model.validDropAreaForFolder(draggableId, droppableId)) {
                                    $(".hprrow_bottom_dropzone",$("#"+droppableId)).removeClass("dataPaneBottomRowDropZoneHovered");
                                    $(".hprrow_top_dropzone",$("#"+droppableId)).addClass("dataPaneTopRowDropZoneHovered");
                                }
                            } else {
                                $(".hprrow_top_dropzone",$("#"+droppableId)).removeClass("dataPaneTopRowDropZoneHovered");
                                if (!DataPaneUtil.isTreeNodeExpanded(droppableId)) {
                                    $(".hprrow_bottom_dropzone",$("#"+droppableId)).addClass("dataPaneBottomRowDropZoneHovered");
                                }
                            }
                        },
                        out: function(event,ui) {
                            ui.droppable = $(this);

                            var droppableId = ui.droppable[0].id;
                            if (DataPaneUtil.isRowIdFirstChildOfFolder(droppableId)) {
                                if (self._isInUpperHalf(ui, ui.droppable)) {
                                    $(".hprrow_top_dropzone",$("#"+droppableId)).addClass("dataPaneTopRowDropZoneHovered");
                                    $(".hprrow_bottom_dropzone",$("#"+droppableId)).removeClass("dataPaneBottomRowDropZoneHovered");
                                } else {
                                    $(".hprrow_top_dropzone",$("#"+droppableId)).removeClass("dataPaneTopRowDropZoneHovered");
                                }
                            } else if (DataPaneUtil.isRowIdLastChildOfFolder(droppableId)) {
                                if (self._isInUpperHalf(ui, ui.droppable)) {
                                    $(".hprrow_bottom_dropzone",$("#"+droppableId)).removeClass("dataPaneBottomRowDropZoneHovered");
                                } else {
                                    $(".hprrow_top_dropzone",$("#"+droppableId)).removeClass("dataPaneTopRowDropZoneHovered");

                                    if ((!DataPaneUtil.isTreeNodeExpanded(droppableId) && !DataPaneUtil.isSubItemAndParentFolderExpanded(droppableId)) ||
                                        (DataPaneUtil.isLastSubItemForFolder(droppableId))) {
                                        $(".hprrow_bottom_dropzone",$("#"+droppableId)).addClass("dataPaneBottomRowDropZoneHovered");
                                    }
                                }
                            } else {
                                $(".hprrow_top_dropzone",$("#"+droppableId)).removeClass("dataPaneTopRowDropZoneHovered");
                                $(".hprrow_bottom_dropzone",$("#"+droppableId)).removeClass("dataPaneBottomRowDropZoneHovered");
                            }

                            // Remove the custom data
                            var oDrop = ui.draggable.data("current-droppable");
                            if (oDrop && oDrop.length > 0 && oDrop[0] === this) {
                                ui.draggable.removeData("current-droppable");
                            }
                        },
                        deactivate: function(event,ui) {
                            // Remove the hover styles from everything
                            $(".dataPaneTopRowDropZoneHovered").removeClass("dataPaneTopRowDropZoneHovered");
                            $(".dataPaneBottomRowDropZoneHovered").removeClass("dataPaneBottomRowDropZoneHovered");
                        },
                        drop: function(event,ui) {
                            ui.droppable = $(this);

                            // Sometimes (not all the time) the drop event is fired twice; This wreaks havoc on dropping sub-folders and keeping
                            // track of indexes; Here we rely on some custom data to alert us if the droppable has already been dropped; If so, bail
                            var oDrop = ui.draggable.data("current-droppable");
                            if (oDrop && oDrop.length > 0 && oDrop[0] === this) {
                                ui.draggable.removeData("current-droppable");
                            } else {
                                return false; // Nothing to drop or already dropped one time
                            }

                            // Remove the hover styles from everything
                            $(".dataPaneTopRowDropZoneHovered").removeClass("dataPaneTopRowDropZoneHovered");
                            $(".dataPaneBottomRowDropZoneHovered").removeClass("dataPaneBottomRowDropZoneHovered");

                            // TODO: When we add folders this algorithm needs to be re-looked at to make sure it can handle
                            // TODO: moving items between folders and up and down the tree.
                            // TODO: Look at cell control

                            // pull out the target (which is the droppable) row's id
                            var targetRowId = ui.droppable[0].id;

                            // Get the rowData for the row being dropped on and it's parent path (RowData is a generic
                            // object that contains two attributes, the row's path (under the path attribute) and the
                            // row's model (under the model attribute))
                            var rowData = DataPaneUtil.getRowData(targetRowId);

                            // The targetRowPath is the path in the model where the item(s) are being dropped
                            var targetRowPath = rowData.path;

                            // Grab the table model and the data for the target row from the table's model
                            var tableModel = self._oControl.getModel();
                            var targetRowTreeNodeObject = tableModel.getProperty(targetRowPath);

                            // We are dragging a sub-folder and trying to drop below a sub-item of a different type in a sibling sub-folder; In other words, we
                            // are dragging a sub-folder and trying to drop it under a child of a sibling sub-folder; In this case we need to re-calculate the targetRowPath
                            // based on their parent's path
                            if (DataPaneUtil.isRowIdLastChildOfFolder(targetRowId) && !DataPaneUtil.areTreeNodesSameType(ui.draggable[0].id, targetRowId) && targetRowTreeNodeObject.dataItemChild) {
                                // Working with sub-folders and sub-items here; Need to manipulate targetRowPath because we need to move droppable inside parent node
                                targetRowPath = DataPaneUtil.getParentsPathFromPath(targetRowPath);
                                targetRowTreeNodeObject = tableModel.getProperty(targetRowPath); // Get new tree node object
                            }

                            // Use the targetRowPath to get the path of its parent's children
                            var parentsChildrenPathOfTargetRowPath = DataPaneUtil.getParentsChildrenPathFromPath(targetRowPath);

                            // March through the children and save their expansion state so we can re-apply after nodes have been moved to give the
                            // illusion that their state has been maintained after dragging/dropping
                            var expansionStateObject = DataPaneUtil.createChildrenPathExpansionStateObject(parentsChildrenPathOfTargetRowPath);

                            // itemsToMove will hold an array of the objects from the table model that need to be moved
                            // these items will be selected based on the list of draggable objects held in the HPRiskModelManager
                            var itemsToMove = [];

                            // We need a sorted list of the row paths that are being dragged, so we do that here
                            var draggedRows = HPRiskModelManager.getDraggableObjects();
                            var selectedRowPaths = _.map(draggedRows, function(draggedRow){
                                return draggedRow.rowPath;
                            });
                            var sortedSelectedRowPaths = DataPaneUtil.getSortedSelectedRowPaths(selectedRowPaths);

                            // Accumulate the objects from the TreeTable's model that will be moved in their sorted order.
                            _.each(sortedSelectedRowPaths, function(selectedItemPath){
                                var item = tableModel.getProperty(selectedItemPath); // May need to make a copy of this
                                if (item !== null){
                                    itemsToMove.push(item);
                                }
                            });

                            // Get the current state of the list the target row is sitting in.  The idea is we'll get
                            // this array correct and put it back into the model (after it is fixed)
                            var parentsChildrenOfTargetRowArray = tableModel.getProperty(parentsChildrenPathOfTargetRowPath);

                            // Find the index of the row to move ahead of (may be one of the rows being moved)
                            var moveToIndex = _.findIndex(parentsChildrenOfTargetRowArray, function(item){
                                return item.dataItem.name === targetRowTreeNodeObject.dataItem.name;
                            });

                            // Re-calculate the moveIndex based on whether we are in upper-half or bottom-half of droppable; This essentially
                            // drops the draggable before or after the desired row and matches up with where the hover styles are displayed
                            var indexAddition = self._isInUpperHalf(ui, ui.droppable) ? 0 : 1;
                            moveToIndex = moveToIndex + indexAddition;

                            // If we actually have a valid index, start the process of moving the dragged rows
                            if (moveToIndex !== -1) {
                                // Remove the rows we are going to move, this will leave null "placeholders" in the model
                                _.each(sortedSelectedRowPaths, function(selectedItemPath){
                                    tableModel.setProperty(selectedItemPath,null);
                                });

                                // Build three arrays, head is the first part up to the index we want to put the
                                // items we are moving at, tail is everything after and including the move index.
                                var head = parentsChildrenOfTargetRowArray.slice(0,moveToIndex);
                                var tail = parentsChildrenOfTargetRowArray.slice(moveToIndex,parentsChildrenOfTargetRowArray.length);

                                // Merge everything back to one array
                                var newArray = head.concat(itemsToMove,tail);

                                // Clean up any nulls in the resultant array by filtering them out
                                newArray = _.filter(newArray, function(item){
                                    return item !== null;
                                });

                                // Finally push the updated and cleaned up array back into the tableModel which will cause
                                // a refresh on the tree table.
                                tableModel.setProperty(parentsChildrenPathOfTargetRowPath,newArray);

                                // Re-apply the expansion state so sub-folder will be expanded or collapsed like before the move
                                DataPaneUtil.applyExpansionStateObject(parentsChildrenPathOfTargetRowPath, expansionStateObject);

                                // Clear out the set of dragged rows (to be safe).
                                HPRiskModelManager.setDraggableObjects([]);

                                // Clear the table's selections.  If this is not desired then we will need to run another
                                // process to figure out what rows are now selected.  How we will set that, I have no
                                // idea since the TreeTable doesn't seem to support that.  Maybe we could do it view jQuery?
                                DataPaneUtil.clearSelections(false); // Don't scroll the TreeTable to the top after a drop

                                // Publish an event that will update the dirtyState of the Exploration view
                                sas.eventBus.publish(HprConstants.CHANNEL_EXPLORATION, HprConstants.EVENT_UPDATE_DIRTY_STATE, {
                                    dirty: true
                                });
                            }
                        }
                    }
                }

                aRows.forEach(function(oRow) {
                    $domRefs = getDomRefs(oRow);

                    // Create a copy of the original drop options object
                    oConfig = $.extend({}, oDropOptions, oConfigDropOptions(oRow));

                    // init the droppable on the dom refs for this row
                    $domRefs.droppable(oConfig);

                    // save which UI5 row
                    $domRefs.data("sasTableRow", oRow);
                    this._droppableDomRefs.push($domRefs);
                }, this);

                // Instead of listening as each row's context is updated, we'll listen to the table for '_rowsUpdated' events; This event is ran
                // after scrolling the TreeTable, expanding/collapsing nodes, etc.
                oTable.attachEvent("_rowsUpdated", function (oEvent) {
                    // Loop through all rows to enable/disable drag state
                    var myRows = oTable.getRows();
                    myRows.forEach(function (oRow) {
                        // Add the draggable options again for each row
                        var $refs = getDomRefs(oRow);

                        // Create a copy of the original drop options object and re-apply to the rows
                        oCopyOfConfig = $.extend({}, oDropOptions, oConfigDropOptions(oRow));
                        $refs.droppable(oCopyOfConfig);
                    });
                });
            },

            /**
             * Get the 'allowRowReorder' property which determines whether the associated Table is a drop target for its
             * own rows.
             * @returns {boolean} true if rows are allowed to be reordered within a table
             */
            getAllowRowReorder: function() {
                return this._bAllowRowReorder;
            },

            /**
             * Set the 'allowRowReorder' property which determines whether the associated Table is a drop target for its
             * own rows.
             */
            setAllowRowReorder: function(bValue) {
                this._bAllowRowReorder = bValue;
            }
        });
        return RowDroppable;
    },
    true
);