// (c) 2017, SAS Institute Inc.

jQuery.sap.declare('sas.analytics.riskhp.views.datapane.HprDataPanelTreeTableSelectionModel');

jQuery.sap.require("sas.analytics.riskhp.views.datapane.DataPaneUtil");

/**
 * This is the model that "runs" the show for the DataPanel selection behaviour.
 *
 * @name sas.analytics.riskhp.views.datapane.HprDataPanelTreeTableSelectionModel
 */
sap.ui.base.ManagedObject.extend("sas.analytics.riskhp.views.datapane.HprDataPanelTreeTableSelectionModel", {
    metadata: {
        properties: {
            "_draggedRowPath" : {type: "string", defaultValue: null, visibility: "hidden"},
            "selecting" : {type : "boolean", defaultValue: true},
            "selectedRows" : {type: "any", defaultValue: []}
        },

        aggregations: {}
    },

    logger: sas.log.getLogger("sas.analytics.riskhp.views.datapane.HprDataPanelTreeTableSelectionModel"),

    /**
     * Updates the model state for the change in selectness.  It is up to the caller to call rowData.model.refresh(true)
     * when it knows it is finished updating the model
     * @param rowData The data object for a row
     * @param bSelected Whether the row should be selected or unselected
     * @private
     */
    _selectRow: function(rowData, bSelected) {
        // Retrieve the RowData object for the current row
        if (rowData === null || rowData === undefined) {
            return false;
        }

        var isSelectable = rowData.model.getProperty(rowData.path+"/isHprSelectable");
        if (!isSelectable) {
            this.logger.debug("Not selecting row because it is not selectable");
            return false;
        }
        // For now we don't support selecting a toplevel row
        if (rowData.model.getProperty(rowData.path+"/nodeType") === sas.analytics.riskhp.utils.HprConstants.HPR_TREETABLE_LEVEL_TOP) {
            return false;
        }

        // if (rowData.model.getProperty(rowData.path+"/dataItemChild")) {
        //     return false;
        // }

        // Get the array of rows that are currently selected so we can update it
        var selectedRows = this.getSelectedRows();

        if (bSelected) {
            // Add the row's path to the list of rows that are selected
            var index = _.indexOf(selectedRows,rowData.path);
            if (index === -1) {
                selectedRows.push(rowData.path);
            }
        }
        else {
            // Remove the row's path from the list of rows that are selected, since _.without returns a copy
            // we have to set the list of selected rows to the new copy
            var newSelectedRows = _.without(selectedRows,rowData.path);
            this.setSelectedRows(newSelectedRows);
        }

        this.logger.debug("Selected rows are: "+this.getSelectedRows());

        return true;
    },

    /**
     * Moves the selected row on a table up or down based on the bMoveDown flag
     * @param table
     * @param bMoveDown true to move the selection down, false to move it up
     * @param model an instance of the HPRDataPanelTreeTableModel
     * @private
     */
    _moveSelection: function(table, bMoveDown, model){
        // Get the currently selected index and the rows from the table
        var selectedIndex = table.getSelectedIndex();
        var rows = table.getRows();

        // Move the selection on the table based on the currently selected index (within the allowed range)
        if (bMoveDown) {
            if (selectedIndex < rows.length-1) {
                table.setSelectedIndex(selectedIndex+1);
            }
        }
        else {
            if (selectedIndex > 0) {
                table.setSelectedIndex(selectedIndex-1);
            }
        }

        // Now that the table has been updated we need to change the selected row in the model
        selectedIndex = table.getSelectedIndex();
        var selectedRow = rows[selectedIndex];
        if (selectedRow !== null && selectedRow !== undefined) {
            model.selectRow(selectedRow.getId(),true);
        }
        else {
            this.logger.debug("could not get row at index:" + selectedIndex);
        }

        // Refocus on the table.  This may not be needed
        table.focus();
    },

    /**
     * Method to determine if the draggable can be dropped on the droppable.  This method will check the selection model
     * to see if it is currently being filtered, it will also check the object associated with the draggableId AND all
     * the rows that are currently selected (held in the selectedRows property) against the object associated with the
     * droppableId.  To be acceptable, the selection model can not be currently filtered and the backing dataitem
     * objects must have the same nodeType attribute.  If either parameter is null then false is returned.
     * @param draggableId
     * @param droppableId
     * @returns {boolean}
     */
    validDropArea: function(draggableId, droppableId) {
        // Drop area is not valid if selection model is filtered
        if (this.isFilteredModel()) {
            return false;
        }

        // Make sure we have two values to check
        if (!draggableId || !droppableId) {
            return false;
        }

        // Get the rowData for both ids and make sure we have them both
        var draggableRowData = sas.analytics.riskhp.views.datapane.DataPaneUtil.getRowData(draggableId);
        var droppableRowData = sas.analytics.riskhp.views.datapane.DataPaneUtil.getRowData(droppableId);
        if (!draggableRowData || ! droppableRowData) {
            return false;
        }

        if (!draggableRowData.model.getProperty(draggableRowData.path+"/isHprDraggable")) {
            return false;
        }

        // Grab the nodeType for both objects and make sure they are the same
        var draggableType = draggableRowData.model.getProperty(draggableRowData.path+"/nodeType");
        var droppableType = droppableRowData.model.getProperty(droppableRowData.path+"/nodeType");
        if (draggableType === droppableType) {
            // Additionally walk through the list of selected rows to make sure they also have the same nodeType
            // if any one of them doesn't match then return false
            var selectedRows = this.getSelectedRows();
            var model = droppableRowData.model;
            var result = true;
            _.each(selectedRows, function(selectedRow){
                // get the data item from the model
                var selectedRowType = model.getProperty(selectedRow+"/nodeType");
                // check to see if it is of the right type, if not this is not a valid drop area.
                if (selectedRowType !== droppableType) {
                    result = false;
                }

                // Check to see if we are dragging Scenarios or Horizons (Both Absolute and Relative types); We can not move a Scenario or Horizon from one Result Set to another Result Set
                if (selectedRowType === sas.analytics.riskhp.models.dataitems.Scenario.SCENARIO || selectedRowType === sas.analytics.riskhp.models.dataitems.Horizon.HORIZON ||
                    selectedRowType === sas.analytics.riskhp.models.dataitems.RelativeHorizon.RELATIVEHORIZON) {
                    // Compare the parent row path's for the selectedRow against the droppable
                    var parentPathOfSelectedRow = sas.analytics.riskhp.views.datapane.DataPaneUtil.getParentsPathFromPath(selectedRow);
                    var parentPathOfDroppableRowPath = sas.analytics.riskhp.views.datapane.DataPaneUtil.getParentsPathFromPath(droppableRowData.path);
                    if (parentPathOfSelectedRow !== parentPathOfDroppableRowPath) {
                        // Can't drag a Scenario from on Result Set parent to another
                        result = false;
                    }
                }
            });

            return result;
        } else {
            // There are some cases where the nodeTypes can be different (think sub-folders) and we need to determine if we have a valid drop area
            return this._validDropAreaForFolderHelper(draggableType, droppableType, draggableRowData, droppableRowData, droppableId);
        }
    },

    /**
     * Method to determine if the droppable is a sub-folder and a valid drop area.
     * @param draggableId
     * @param droppableId
     * @returns {boolean}
     */
    validDropAreaForFolder: function(draggableId, droppableId) {
        // Drop area is not valid if selection model is filtered
        if (this.isFilteredModel()) {
            return false;
        }

        // Make sure we have two values to check
        if (!draggableId || !droppableId) {
            return false;
        }

        // Get the rowData for both ids and make sure we have them both
        var draggableRowData = sas.analytics.riskhp.views.datapane.DataPaneUtil.getRowData(draggableId);
        var droppableRowData = sas.analytics.riskhp.views.datapane.DataPaneUtil.getRowData(droppableId);
        if (!draggableRowData || ! droppableRowData) {
            return false;
        }

        if (!draggableRowData.model.getProperty(draggableRowData.path+"/isHprDraggable")) {
            return false;
        }

        // Grab the nodeType for both objects and make sure they are the same
        var draggableType = draggableRowData.model.getProperty(draggableRowData.path+"/nodeType");
        var droppableType = droppableRowData.model.getProperty(droppableRowData.path+"/nodeType");

        return this._validDropAreaForFolderHelper(draggableType, droppableType, draggableRowData, droppableRowData, droppableId);
    },

    /**
     * Method that does the work to determine if sub-folders and data items within those sub-folders are a valid drop area.
     * @param draggableType
     * @param droppableType
     * @param draggableRowData
     * @param droppableRowData
     * @param droppableId
     * @returns {boolean}
     * @private
     */
    _validDropAreaForFolderHelper: function(draggableType, droppableType, draggableRowData, droppableRowData, droppableId) {
        var result = false;

        // Since we have cases of sub-folders (ResultSets hold Scenarios & Hierarchies holds Class Variables), we need to determine if they are valid drop areas
        if ((draggableType === sas.analytics.riskhp.models.dataitems.ResultSet.RESULTSET && droppableType === sas.analytics.riskhp.models.dataitems.Scenario.SCENARIO) ||
            (draggableType === sas.analytics.riskhp.models.dataitems.Hierarchy.HIERARCHY && droppableType === sas.analytics.riskhp.models.dataitems.Classvar.CLASSVAR)) {

            var DataPaneUtil = sas.analytics.riskhp.views.datapane.DataPaneUtil;
            var parentsPathOfDroppableRowPath = DataPaneUtil.getParentsPathFromPath(droppableRowData.path);

            // If the draggable path is not the same as the droppable's parent path which means we are dragging a sub-folder around
            if (draggableRowData.path !== parentsPathOfDroppableRowPath) {
                // Determine if the droppable is the last child/treenode for the sub-folder
                if (DataPaneUtil.isRowIdLastChildOfFolder(droppableId)) {
                    // If it is the last child, we may have to declare this a valid drop area; Use the top-level folder path to make sure we are still inside the same top-level folder
                    if (DataPaneUtil.getTopLevelFolderPath(draggableRowData.path) === DataPaneUtil.getTopLevelFolderPath(parentsPathOfDroppableRowPath)) {
                        result = true;
                    }
                }
            }
        }

        return result;
    },

    /**
     * Clears the list of rows this model deems as being selected.
     */
    clearSelectedRows: function() {
        this.logger.debug("Cleared selected rows");
        this.setSelectedRows([]);
    },

    _removeRowPathFromSelectedRows: function(rowPath) {
        var selectedRows = this.getSelectedRows();
        var newSelectedRows = _.reject(selectedRows, function(selectedRowPath) {
            return selectedRowPath === rowPath;
        });
        this.setSelectedRows(newSelectedRows);
    },

    hasDraggedRowPath: function() {
        var path = this.getProperty("_draggedRowPath");
        return (path !== null && path !== undefined && path !== "" );
    },

    setDraggedRowPath: function(path) {
        this.setProperty("_draggedRowPath",path);
    },

    clearDraggedRowPath: function() {
        var draggedRowPath = this.getDraggedRowPath();
        if (draggedRowPath) {
            this._removeRowPathFromSelectedRows(draggedRowPath);
            this.setProperty("_draggedRowPath",null);
        }
    },

    getDraggedRowPath: function() {
        if (this.hasDraggedRowPath()) {
            return this.getProperty("_draggedRowPath");
        }
        return null;
    },

    /**
     * Updates the model based on the selected state passed in for the given rowId
     * @param rowId The id of the row or child of the row to be selected or unselected (and therefore included in the overall selection)
     * @param bSelected Flag to indicate whether the row should be selected or unselected
     */
    selectRow: function(rowId, bSelected) {
        // Get the array of rows that are currently selected so we can update it
        var selectedRows = this.getSelectedRows();

        // Retrieve the RowData object for the current row
        var rowData = sas.analytics.riskhp.views.datapane.DataPaneUtil.getRowData(rowId);
        if (!rowData) {
            return;
        }
        this._selectRow(rowData,bSelected);
        rowData.model.refresh(true);
    },

//    /**
//     * Updates the model based on the selected state passed in for the given rowPath.  The rowData for the passed in rowPath is manually constructed
//     * in this method.  This is because there is not an easy way using the Tree/Table APIs to find the corresponding 'id' of the row associated
//     * with a given path for a row.  This is especially true if the parent node is collapsed in the Tree/Table component and the children nodes
//     * are not visible.
//     * @param rowPath The path of the row or child of the row to be selected or unselected (and therefore included in the overall selection)
//     * @param table The Table component where the rowPath is stored in the table's model
//     * @param bSelected Flag to indicate whether the row should be selected or unselected
//     */
//    selectRowUsingPath: function(rowPath, table, bSelected) {
////        this.logger.debug("selectRowUsingPath: "+rowPath);
//        // Construct a mock RowData object for the current path
//        var rowData = {
//            path: rowPath,
//            model: table.getModel()
//        };
//        this._selectRow(rowData,bSelected);
// //       rowData.model.refresh(true);
//    },

    /**
     * Selects the nodes/rows with the specified indexes.
     * @param indices
     * @return an array of the indices that were successfully selected
     */
    selectRowsByIndex: function(indices) {
        var self = this;

        var selectedRowIndices = [];
        _.each(self._convertIndicesToRows(indices), function(rowData) {
            if (self._selectRow(rowData, true)) {
                selectedRowIndices.push(rowData.index);
            }
        });
        return selectedRowIndices;
    },

    /**
     * Deselects the nodes/rows with the specified indexes.
     * @param indices
     * @return an array of the indices that were successfully deselected
     */
    deselectRowsByIndex: function(indices) {
        var self = this;

        var deselectedRowIndices = [];
        _.each(self._convertIndicesToRows(indices), function(rowData) {
            if (self._selectRow(rowData, false)) {
                deselectedRowIndices.push(rowData.index);
            }
        });
        return deselectedRowIndices;
    },

    isRowPathSelected: function(path) {
        if (path === null || path === undefined) {
            return false;
        }
        var selectedRows = this.getSelectedRows();

        return (_.indexOf(selectedRows, path) !== -1);

    },

    isRowSelectable: function(model,path) {
        if (path === null || path === undefined) {
            return false;
        }
        return model.getProperty(path+"/isHprSelectable");
    },

    isSelectableRowType: function(type) {
        return (type !== null && type !== undefined && type !== sas.analytics.riskhp.utils.HprConstants.HPR_TREETABLE_LEVEL_TOP);
    },

    /**
     * Reset the 'filteredModel' property to "null"
     */
    resetFilteredModel: function() {
        this.filteredModel = null;
    },

    /**
     * Returns whether the selection model has been filtered
     * @returns {boolean} true | false Returns a boolean that signifies that the selection model has been filtered
     */
    isFilteredModel: function() {
        return !!this.filteredModel;
    },

    /**
     * Applies the specified search value to the data pane tree model and produces a filtered model containing
     * row objects that map back to the tree model with indices corresponding to their location in the filtered
     * tree model.
     *
     * @param searchValue
     * @param showLabels
     */
    applyFilters: function(searchValue, showLabels) {
        var self = this;

        // clear the row selections
        self.clearSelectedRows();

        // if no search value, remove the filtered model
        if (!searchValue) {
            this.filteredModel = null;
            return;
        }

        var unfilteredModel = sas.analytics.riskhp.views.datapane.DataPaneModelManager.getTreeTableModel();
        var rowDataModel = new sap.ui.model.json.JSONModel(unfilteredModel);

        var __filterTreeNode = function(node, path) {
            var nodeMatch = false;

            // if the node has a data item, check the search value against label or name
            if (node.dataItem) {
                if (showLabels === true) {
                    nodeMatch = node.dataItem.label && node.dataItem.label.toUpperCase().includes(searchValue.toUpperCase());
                }
                else {
                    nodeMatch = node.dataItem.name && node.dataItem.name.toUpperCase().includes(searchValue.toUpperCase());
                }
            }

            // if we still don't have a match, check the search value against the lineage
            if (nodeMatch === false) {
                nodeMatch = node.lineage && node.lineage.toUpperCase().includes(searchValue.toUpperCase());
            }

            // filter the children of the node (if they exist)
            if (node.children && node.children.length > 0) {
                var unfilteredChildNodes = node.children,
                    filteredChildNode;

                node.children = [];

                for (var i = 0; i < unfilteredChildNodes.length; i++) {
                    filteredChildNode = __filterTreeNode($.extend(true, {}, unfilteredChildNodes[i]), path + "/children/" + i);
                    if (filteredChildNode) {
                        node.children.push(filteredChildNode);
                    }
                }
            }

            // if this node matched, or its descendants did, return it
            if (nodeMatch === true || node.children && node.children.length > 0) {
                return {
                    id: node.id,
                    model: rowDataModel,
                    path: path,
                    children: node.children
                };
            }
            else {
                return null;
            }
        };

        // otherwise, apply the filters to create a filtered model
        var filteredModel = {children: []};
        var filteredNode;
        for (var i = 0; i < unfilteredModel.children.length; i++) {
            filteredNode = __filterTreeNode($.extend(true, {}, unfilteredModel.children[i]), "/children/" + i);
            if (filteredNode) {
                filteredModel.children.push(filteredNode);
            }
        }
        self.filteredModel = filteredModel;
    },

    /**
     * Converts indexes to tree node objects containing information about the
     * node, including ID and path(?).
     */
    _convertIndicesToRows: function(indices) {
        var flatRowDataList = [];

        var isFiltered = !!this.filteredModel;
        var treeTableModel = isFiltered ? this.filteredModel : sas.analytics.riskhp.views.datapane.DataPaneModelManager.getTreeTableModel();
        var expandedNodeIds = sas.analytics.riskhp.views.datapane.DataPaneModelManager.getExpandedNodeIds();

        // this is needed for HprDataPaneTreeTableSelectionModel._selectRow
        var rowDataModel = new sap.ui.model.json.JSONModel(treeTableModel);

        var __isNodeExpanded = function(nodeId) {
            return _.indexOf(expandedNodeIds, nodeId) !== -1;
        };

        var __flattenTreeNodes = function(treeNodes, parentPath) {
            for (var i = 0; i < treeNodes.length; i++) {
                // add a row data object for each node (filtered model nodes already have path and model--see applyFilters)
                flatRowDataList.push({
                    model: isFiltered ? treeNodes[i].model : rowDataModel,
                    path: isFiltered ? treeNodes[i].path : (parentPath + "/" + i)
                });

                // if this node is expanded, process it's children
                if (__isNodeExpanded(treeNodes[i].id)) {
                    if (treeNodes[i].children && treeNodes[i].children.length > 0) {
                        __flattenTreeNodes(treeNodes[i].children, parentPath + "/" + i + "/children");
                    }
                }
            }
        };

        __flattenTreeNodes(treeTableModel.children, "/children");

        var rowObjects = [];
        _.each(indices, function(index) {
            flatRowDataList[index].index = index;
            rowObjects.push(flatRowDataList[index]);
        });
        return rowObjects;
    },

    /**
     * Handles the keyboard navigation, by updating the model based on what key was selected.
     * @param table
     * @param event
     */
    handleKeyEvent: function(table, event) {
        if (event.keyCode === 40 ||  // arrow down
            event.keyCode === 38 ||  // arrow up
            event.keyCode === 32) { // space key
            var model = sas.analytics.riskhp.views.datapane.DataPaneUtil.getDataPaneTreeTableSelectionModel();

            switch (event.keyCode){
                case 40: // down arrow
                    this._moveSelection(table, true, model);
                    break;
                case 38: // up arrow
                    this._moveSelection(table, false, model);
                    break;
                case 32: // space key
                    // Retrieve the row from the table that is selected
                    var selectedIndex = table.getSelectedIndex();
                    var rows = table.getRows();
                    var selectedRow = rows[selectedIndex];
                    if (!selectedRow) {
                        return;
                    }
                    // Get the checkbox on the selected row, if it has one and toggle it's state
                    var rowPath = sas.analytics.riskhp.views.datapane.DataPaneUtil.getRowPathForRowId(selectedRow.getId());
                    var newSelectionState = !this.isRowPathSelected(rowPath);
                    this.selectRow(selectedRow.getId(), newSelectionState);
                    // var checkbox = this._findCheckbox(selectedRow.getId());
                    // if (checkbox) {
                    //     var newSelectionState = !checkbox.getSelected();
                    //     this.checkRow(selectedRow.getId(), newSelectionState);
                    //     checkbox.setSelected(newSelectionState);
                    // }
                    table.focus();
                break;
            default:
                break;
            }
        }
    }
});