jQuery.sap.declare("sas.risk.common.widgets.PersistedSearch");

jQuery.sap.require("sas.hc.m.SearchField");
jQuery.sap.require("sas.hc.m.SearchFieldRenderer");
jQuery.sap.require("sas.hc.ui.table.Table");
jQuery.sap.require("sas.hc.ui.table.TreeTable");

sas.hc.m.SearchField.extend("sas.risk.common.widgets.PersistedSearch", {
    metadata: {
        properties: {
            /**
             * Unique id that will be used to identify the search control for which filterString will be saved and applied
             */
            persistenceId: {type: "string"},
            /**
             * Table that will be filtered using the persisted filter string
             */
            searchTable: {type: "any"},
            /**
             * Function to call for table filter passing the filter string
             */
            filterFunction: {type: "any"}
        }
    },
    renderer: sas.hc.m.SearchFieldRenderer
});

sas.risk.common.widgets.PersistedSearch.prototype.init = function(){
    sas.hc.m.SearchField.prototype.init.apply(this, arguments);
    //set up the beforeRendering listener to apply the saved search string as value of the search field
    this.addEventDelegate({
        "onBeforeRendering": $.proxy(this.applyFilterString, this)
    });
    //On live change of the search field keep updating the persisted filter string
    this.attachLiveChange($.proxy(this.persistFilterString, this));
};

sas.risk.common.widgets.PersistedSearch.prototype.onBeforeRendering = function(){
    if(_.isEmpty(this.getPersistenceId()) || !this.getSearchTable() || !this.getFilterFunction()){
        console.error("Missing attributes: Since persistenceId, searchTable and filterFunction attributed are not provided therefore 'PersistedSearch' will not work as expected");
    }
    sas.hc.m.SearchField.prototype.onBeforeRendering.apply(this, arguments);
};

sas.risk.common.widgets.PersistedSearch.prototype.setSearchTable = function(table){
    var that = this;
    if(!(table instanceof sas.hc.ui.table.Table || table instanceof sas.hc.ui.table.TreeTable)){
        console.error("You must pass a valid table to the 'setSearchTable' function for PersistedSearch to function properly");
    }else{
        this.setProperty("searchTable", table);

        var afterRenderingEventDelegate = {
            "onAfterRendering": function(e){
                var filterString = that.getValue();
                var trimmedFilterStr = filterString.trim();
                //If table is not ready yet with it's binding being updated then skip filtering at this point.
                //Without this check, due to timing and the sequence in which onAfterRendering and _updateBindingContexts are executed,
                // by the time table starts to work on it's rows, it might be already filtered
                //I wish there was a good way to know when the table has all the data ready and bound irrespective of the sequence in which view adds the table and table model is set and populated
                var tableBinding = table.getBinding("rows");
                var tableContexts = tableBinding ? tableBinding.getContexts() : null;
                if(_.isEmpty(tableContexts) && !_.isEmpty(trimmedFilterStr)){ //S1439170
                    return;
                }

                that.setValue(trimmedFilterStr);
                //If the trimmed version of the filter string is empty, then remove this filterString from session storage
                if(trimmedFilterStr.length === 0){
                    sessionStorage.removeItem(that.getPersistenceKey());
                }
                //If the trimmed version of the string is different from the original string that means user had trailing or leading spaces.
                // While applying the filter string back, don't keep the leading or trailing spaces. Otherwise that will confuse users with the default cursor position being at the beginning.
                if(filterString !== trimmedFilterStr){
                    sessionStorage.setItem(that.getPersistenceKey(), trimmedFilterStr);
                }
                var filterFunc = that.getFilterFunction();
                if(filterFunc){
                    filterFunc(trimmedFilterStr);
                }

                // After the first pass during initial load when contexts are ready, remove this event delegate so that this code is not executed on next OnAfterRendering calls
                table.removeEventDelegate(afterRenderingEventDelegate, that);

                //When table data updates for example due to setData, then put the event delegate back again to be executed next on next after rendering
                var rowBinding = table.getBindingInfo("rows");
				// defensive check added with 3.2
                if (rowBinding && tableBinding) {				 
					var tableModel = table.getModel(rowBinding.model);
					var rowsBindingPath = tableBinding.getPath();
					var tableDataBinding = new sap.ui.model.json.JSONPropertyBinding(tableModel, rowsBindingPath, tableModel.getContext(rowsBindingPath));
					tableDataBinding.attachChange(function(e){
						table.addEventDelegate(afterRenderingEventDelegate, that);
					});
				}
            }		 
        };
        table.addEventDelegate(afterRenderingEventDelegate);
    }
};

sas.risk.common.widgets.PersistedSearch.prototype.setFilterFunction = function(filterFunc){
    if(!_.isFunction(filterFunc)){
        console.error("You must provide a valid filter function in the 'setSearchTable' function for the PersistedSearch to function properly");
    }else{
        this.setProperty("filterFunction", filterFunc);
    }
};

sas.risk.common.widgets.PersistedSearch.prototype.getPersistenceKey = function(){
    return this.getPersistenceId();
};

sas.risk.common.widgets.PersistedSearch.prototype.persistFilterString = function(){
    var filterString = this.getValue();
    if(!!filterString){
        sessionStorage.setItem(this.getPersistenceKey(), filterString);
    }else{
        sessionStorage.removeItem(this.getPersistenceKey());
    }
};

sas.risk.common.widgets.PersistedSearch.prototype.applyFilterString = function(){
    var persistedStr = sessionStorage.getItem(this.getPersistenceKey());
    this.setValue(persistedStr);
};