
/************************************************************************
 *  Copyright (C), 2015-2017, SAS Institute Inc., Cary, NC, USA. All Rights Reserved.
 *
 *  NAME:       create_single_scenario.sas
 *
 *  PURPOSE:    Macro to create the scenanal data set that constains 
 *              the scenarios perturbed according to the sensitivty
 *              run specifications.
 *
 *  NOTE:       This macro does no validation since it is being called 
 *              by an API.
 *
 *              SCENARIO_DS 
 *              SENSITIVITY_OUTPUT_DS
 *              PERTURB_BASESCENARIO
 *              SHOCKS
 *                  The SHOCKS variable is a macro-quoted string of JSON
 *                  that contains an array of shock objects.  These
 *                  objects consist of a scenario name, a perturbation
 *                  name, a variable type, a shock type, a value,
 *                  and optionally, a fade in, fade out, floor, and cap.
 *
 *                  This variable is parsed with the JSON libname engine
 *                  and results in a data set in work that contains one
 *                  row for each shock.
 * 
 ************************************************************************/

%macro hpr_create_sensitivity_pkg(SCENARIO_DS=,
                                  SENSITIVITY_OUTPUT_DS=,
                                  PERTURB_BASESCENARIO=,
                                  SHOCKS=);

    %local dsid dateBased usesAliasType usesAliasName usesAliasValue horizonBased MaxTime
numofScenarios shocknum timeInd time rc;

    /*  Get Indexing and Alias resolutions */
    %let dsid           = %sysfunc(open(&SCENARIO_DS));
    %let dateBased      = %eval((%sysfunc(varnum(&dsid,Date)) > 0));
    %let usesAliasType  = %eval((%sysfunc(varnum(&dsid,change_type)) > 0));
    %let usesAliasName  = %eval((%sysfunc(varnum(&dsid,variable_name)) > 0));
    %let usesAliasValue = %eval((%sysfunc(varnum(&dsid,change_value)) > 0));
    %let horizonBased   = %eval((%sysfunc(varnum(&dsid,horizon)) > 0));
    %let rc             = %sysfunc(close(&dsid));

    %if (&horizonBased and &dateBased) %then %do;
        proc sql;
            select min(horizon) into: minHorizon
            from &scenario_ds;
        quit;
        %if (&minHorizon>0) %then %let dateBased = 0;
    %end;

    /******************************************************************
     *  Convert the SHOCKS JSON string to a data set WORK.SHOCKS
     *  that contains one row for each shock
     ******************************************************************/

    /*  Write the JSON to a temporary file (the libname engine cannot
        read a string directly, only a file or URL */

    filename myjson temp;
    data _null_;
        file myjson;
        put "&SHOCKS";
    run;

    /*  Now use the JSON libname engine to parse the data and
        create data set(s) from it.  The only data set we care
        about is ITEMS (which contains the individual shocks).
        We do some column renaming and dropping for convenience,
        and end up with a data set WORK.SHOCKS */

    libname myjson json;
    data work.shocks;
        set myjson.items;
        rename  scenarioName=scenName
                sourceName=name
                sourceType=varType
                perturbType=type
                shockValue=value
                fadeIn=fi
                fadeOut=fo
                floor=flr
                cap=lowercasecap
                ;
        drop ordinal_root ordinal_items;
    run;

    /*  Done with the JSON libname engine. Clear the libname
        and filename */

    libname myjson clear;
    filename myjson clear;

    /******************************************************************
     *  Done with stuff that was done to unpack the JSON in the
     *  SHOCKS parameter
     ******************************************************************/

    /*  Determine whether the baseline scenario data set is date based
        or horizon based */

    %if (&dateBased) %then %do;
        proc sql;
            create table dates as
            select distinct date format 8.0 as date from &SCENARIO_DS. where scenario_name="&PERTURB_BASESCENARIO.";
        quit;
        proc sort data=dates; by date; run;
        proc sql;
            select date into :timeInd separated by ' '
            from dates;
        quit;
        data _NULL_;
            set &SCENARIO_DS.(obs=1);
            CALL symput("dateFormat",vformat(DATE));
        run;
    %end;
    %else %do;
        proc sql outobs=1 noprint;
            select interval into :baseInterval
            from &SCENARIO_DS. where scenario_name="&PERTURB_BASESCENARIO.";
        quit;
        proc sql;
            create table Horizon as
            select distinct Horizon as Horizon from &SCENARIO_DS. where scenario_name="&PERTURB_BASESCENARIO.";
        quit;
        proc sort data=horizon; by horizon; run;
        proc sql noprint;
            select Horizon into :timeInd separated by ' '
            from horizon;
        quit;
    %end;
    %let maxTime=&sqlobs;
    %put MaxTime=&maxTime.;

    %if (&syserr eq 0) %then %do;

        data &SENSITIVITY_OUTPUT_DS.;
            length SCENARIO_NAME $32. BASELINE $32. LABEL $255.;
            drop scenName name varType value fi fo flr lowercasecap type;
            %if (&dateBased) %then %do;
                format DATE &dateFormat;
            %end;
            %else %do;
                length INTERVAL $10.;
            %end;
            %if (&usesAliasName) %then %do;
                length VARIABLE_NAME $32.;
            %end;
            %else %do;
                length _NAME_ $32.;
            %end;
            %if (&usesAliasType) %then %do;
                length CHANGE_TYPE $5.;
            %end;
            %else %do;
                length _type_ $5.;
            %end;
            set shocks;

            if (fi=0) then fi=.;
            if (fo=0 or fo>=&maxtime.) then fo=.;

            if (varType=301) then do;
                    %if (&usesAliasName) %then %do;
                        variable_name = name;
                    %end;
                    %else %do;
                        _NAME_        = name;
                    %end;
                    %if (&usesAliasValue) %then %do;
                        CHANGE_VALUE  = value;
                    %end;
                    %else %do;
                        _VALUE_ = value;
                    %end;
                    %if (&usesAliasType) %then %do;
                        CHANGE_TYPE = type;
                    %end;
                    %else %do;
                        _TYPE_ = type;
                    %end;
                    SCENARIO_NAME = scenName;
                    BASELINE      = "&PERTURB_BASESCENARIO.";
                    FLOOR         = flr;
                    CAP           = lowercasecap;
                    %if (&dateBased=0) %then %do;
                        HORIZON   = .;
                        INTERVAL  = '';
                    %end;
                    output;
            end;
            else do;
                %do time=1 %to &maxTime;
                        %if (&usesAliasName) %then %do;
                            variable_name = name;
                        %end;
                        %else %do;
                            _NAME_        = name;
                        %end;
                        SCENARIO_NAME = scenName;
                        BASELINE      = "&PERTURB_BASESCENARIO.";
                        %if (&dateBased) %then %do;
                            %let dt=%qscan(&timeInd, &time, %str( ));
                            Date="&dt.";
                        %end;
                        %else %do;
                            INTERVAL      = "&baseInterval.";
                            HORIZON       = &time;
                        %end;
                        %if (&usesAliasType) %then %do;
                            CHANGE_TYPE = type;
                        %end;
                        %else %do;
                            _TYPE_      = type;
                        %end;
                        %if (&usesAliasValue) %then %do;
                            CHANGE_VALUE  = min(1, &time/fi)*min(1,(&maxTime-&time)/(&maxTime-fo))*value;
                        %end;
                        %else %do; 
                            _VALUE_       = min(1, &time/fi)*min(1,(&maxTime-&time)/(&maxTime-fo))*value;
                        %end;
                        FLOOR         = flr;
                        CAP           = lowercasecap;
                    output;
                %end;
            end;

        run;

        %if (&syserr eq 0) %then %do;
            proc append 
                data=&SCENARIO_DS.(where=(scenario_name eq "&PERTURB_BASESCENARIO.")) 
                base=&SENSITIVITY_OUTPUT_DS. 
              force nowarn;
            run;
            %if (&dateBased) %then %do;
                proc sort data=&SENSITIVITY_OUTPUT_DS.;
                   by scenario_name date _NAME_;
                run;
            %end;
            %else %do;
                proc sort data=&SENSITIVITY_OUTPUT_DS.;
                   by scenario_name horizon _NAME_;
                run;            
            %end;
        %end;
    %end;

    /* Ignore WARNINGS */
    %if ((&syserr eq 0) | (&syserr eq 4)) %then %do;
    %end;
    %else %do;
        %let rc=&syserr;
    %end;

    %let dsid           = %sysfunc(open(shocks));
    %let shockNum       = %sysfunc(attrn(&dsid,NLOBS));
    %let rc             = %sysfunc(close(&dsid));

    data msarc;
        format rc 8.0 msg $256.;
        rc=&rc;
        if (&rc eq 0) then do;
            %let numOfScenarios=%eval(&shockNum.+1);
            msg="&numOfScenarios. scenarios written to &SENSITIVITY_OUTPUT_DS.";
        end;
        else do;
            if (&rc eq -1) then msg='Error: There is a misalignment in input scenarios';
            else msg='Error creating scenarios';
        end;
    run;


%mend hpr_create_sensitivity_pkg;
